
/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifdef ON_DEMAND
#include "cm.h"

typedef enum CM_conn_state_cli {
    CM_CONN_STATE_C_IDLE,
    CM_CONN_STATE_C_REQUESTING,
} CM_conn_state_cli;

typedef enum CM_conn_state_srv {
    CM_CONN_STATE_S_IDLE,
    CM_CONN_STATE_S_REQUESTED,
} CM_conn_state_srv;

#define CM_MSG_TYPE_REQ     0
#define CM_MSG_TYPE_REP     1
#define CM_MSG_TYPE_RTU     2

#define CM_MSG_FLAG       201

typedef struct cm_msg {
    uint8_t cm_flag;
    uint32_t req_id;
    uint32_t server_rank;
    uint32_t client_rank;
    uint8_t msg_type;
    uint16_t lid;
    uint32_t qpn;
    MPICM_conn_type conn_type;
} cm_msg;

#define DEFAULT_CM_MSG_RECV_BUFFER_SIZE   1024
#define DEFAULT_CM_SEND_DEPTH             10
#define DEFAULT_CM_MAX_SPIN_COUNT         5000   
#define DEFAULT_CM_THREAD_STACKSIZE   (1024*1024)

/*In microseconds*/
#define CM_DEFAULT_TIMEOUT      500000
#define CM_MIN_TIMEOUT           20000

#define CM_UD_DEFAULT_PSN   0

#define CM_UD_SEND_WR_ID  11
#define CM_UD_RECV_WR_ID  13

static int cm_send_depth;
static int cm_recv_buffer_size;
static int cm_ud_psn;
static int cm_req_id_global;
static int cm_max_spin_count;
static pthread_t cm_comp_thread, cm_timer_thread;
static pthread_mutex_t cm_conn_state_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cm_cond_new_pending = PTHREAD_COND_INITIALIZER;
struct timespec cm_timeout;
long cm_timeout_usec;
size_t cm_thread_stacksize;

MPICM_ib_context cm_ib_context;

int *cm_listening;              /*listening type to all peers */
MPICM_conn_type *cm_conn_type;  /*conntion type to all peers */
CM_conn_state_cli *cm_state_cli;        /*connecting states as client to all peers */
CM_conn_state_srv *cm_state_srv;        /*connecting states as server to all peers */

struct ibv_comp_channel *cm_ud_comp_ch;
struct ibv_qp *cm_ud_qp;
struct ibv_cq *cm_ud_recv_cq;
struct ibv_cq *cm_ud_send_cq;
struct ibv_mr *cm_ud_mr;
struct ibv_ah **cm_ah;          /*Array of address handles of peers */
struct ibv_qp **cm_rc_qp;       /*Array of qp handles of peers */
struct ibv_mr **cm_rc_mr;       /*Array of rc qp temp mr */
uint32_t *cm_ud_qpn;            /*Array of ud pqn of peers */
uint16_t *cm_lid;               /*Array of lid of all procs */
void **cm_rc_buf;               /*Array of rc buf */
void *cm_ud_buf;
void *cm_ud_send_buf;           /*length is set to 1 */
void *cm_ud_recv_buf;
int cm_ud_recv_buf_index;
int page_size;

int (*cm_listen_callback) (int, struct ibv_qp *, void *);
void *cm_listen_callback_arg;

/*array of cm_connect_callback funcions*/
int (**cm_connect_callback) (int, struct ibv_qp *, void *);

/*array of cm_connect_callback args*/
void **cm_connect_callback_arg;

int cm_post_ud_recv(void *buf, int size)
{
    struct ibv_sge list;
    struct ibv_recv_wr wr;
    struct ibv_recv_wr *bad_wr;

    memset(&list, 0, sizeof(struct ibv_sge));
    list.addr = (uintptr_t) buf;
    list.length = size + 40;
    list.lkey = cm_ud_mr->lkey;
    memset(&wr, 0, sizeof(struct ibv_recv_wr));
    wr.next = NULL;
    wr.wr_id = CM_UD_RECV_WR_ID;
    wr.sg_list = &list;
    wr.num_sge = 1;

    return ibv_post_recv(cm_ud_qp, &wr, &bad_wr);
}

struct ibv_qp *cm_create_rc_qp()
{
    struct ibv_qp *qp;
    struct ibv_qp_init_attr init_attr;
    struct ibv_qp_attr attr;

    memcpy(&init_attr, &(cm_ib_context.rc_qp_init_attr),
           sizeof(struct ibv_qp_init_attr));

    qp = ibv_create_qp(cm_ib_context.pd, &init_attr);
    if (!qp) {
        CM_ERR("Couldn't create RC QP");
        return NULL;
    }

    memcpy(&attr, &(cm_ib_context.rc_qp_attr_to_init),
           sizeof(struct ibv_qp_attr));

    if (ibv_modify_qp(qp, &attr, cm_ib_context.rc_qp_mask_to_init)) {
        CM_ERR("Failed to modify RC QP to INIT");
        return NULL;
    }

    return qp;
}

int cm_send_rc_msg(cm_msg * msg)
{
    int peer = msg->server_rank;
    struct ibv_sge list;
    struct ibv_send_wr wr;
    struct ibv_send_wr *bad_wr;

    memcpy(cm_rc_buf[peer], msg, sizeof(cm_msg));

    memset(&list, 0, sizeof(struct ibv_sge));
    list.addr = (uintptr_t) cm_rc_buf[peer];
    list.length = sizeof(cm_msg);
    list.lkey = cm_rc_mr[peer]->lkey;

    memset(&wr, 0, sizeof(struct ibv_send_wr));
    wr.wr_id = cm_ib_context.confim_send_wr_id;
    wr.sg_list = &list;
    wr.num_sge = 1;
    wr.opcode = IBV_WR_SEND;
    wr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_SOLICITED;

    return ibv_post_send(cm_rc_qp[peer], &wr, &bad_wr);
}

int cm_post_ud_packet(cm_msg * msg)
{
    int peer;
    int ret;
    struct ibv_sge list;
    struct ibv_send_wr wr;
    struct ibv_send_wr *bad_wr;

    struct ibv_wc wc;
    int ne;

    CM_DBG("cm_post_ud_packet Enter");
    
    if (msg->msg_type == CM_MSG_TYPE_REP) {
        peer = msg->client_rank;
    } else {
        peer = msg->server_rank;
    }

    memcpy(cm_ud_send_buf + 40, msg, sizeof(cm_msg));
    memset(&list, 0, sizeof(struct ibv_sge));
    list.addr = (uintptr_t) cm_ud_send_buf + 40;
    list.length = sizeof(cm_msg);
    list.lkey = cm_ud_mr->lkey;

    memset(&wr, 0, sizeof(struct ibv_send_wr));
    wr.wr_id = CM_UD_SEND_WR_ID;
    wr.sg_list = &list;
    wr.num_sge = 1;
    wr.opcode = IBV_WR_SEND;
    wr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_SOLICITED;
    wr.wr.ud.ah = cm_ah[peer];
    wr.wr.ud.remote_qpn = cm_ud_qpn[peer];
    wr.wr.ud.remote_qkey = 0;

    ret = ibv_post_send(cm_ud_qp, &wr, &bad_wr);
    if (ret) {
        CM_ERR("ibv_post_send to ud qp failed");
    }

    /*poll for completion */
    CM_DBG("cm_post_ud_packet Poll");
    
    while (1) {
        ne = ibv_poll_cq(cm_ud_send_cq, 1, &wc);
        if (ne < 0) {
            CM_ERR("poll CQ failed %d", ne);
            return -1;
        } else if (ne == 0)
            continue;

        if (wc.status != IBV_WC_SUCCESS) {
            CM_ERR("Failed status %d for wr_id %d",
                    wc.status, (int) wc.wr_id);
            return -1;
        }

        if (wc.wr_id == CM_UD_SEND_WR_ID) {
            break;
        } else {
            CM_ERR("unexpected completion, wr_id: %d",
                    (int) wc.wr_id);
            return -1;
        }
    }

    CM_DBG("cm_post_ud_packet Exit");

    return 0;
}

/*move qp to rts*/

int cm_enable_qp_init_to_rts(int peer, struct ibv_qp *qp,
                             uint16_t dlid, int dqpn)
{
    struct ibv_qp_attr attr;

    CM_DBG("cm_enable_qp_init_to_rts Enter");
    
    memcpy(&attr, &(cm_ib_context.rc_qp_attr_to_rtr),
           sizeof(struct ibv_qp_attr));
    attr.dest_qp_num = dqpn;
    attr.ah_attr.dlid = dlid;

    if (ibv_modify_qp(qp, &attr, cm_ib_context.rc_qp_mask_to_rtr)) {
        CM_ERR("Failed to modify QP to RTR");
        return -1;
    }

    memcpy(&attr, &(cm_ib_context.rc_qp_attr_to_rts),
           sizeof(struct ibv_qp_attr));

    if (ibv_modify_qp(qp, &attr, cm_ib_context.rc_qp_mask_to_rts)) {
        CM_ERR("Failed to modify QP to RTS");
        return -1;
    }

    /*Register memory for send */

    cm_rc_buf[peer] = memalign(page_size, sizeof(cm_msg));
    if (!cm_rc_buf[peer]) {
        CM_ERR("Couldn't allocate working buf");
        return -1;
    }

    memset(cm_rc_buf[peer], 0, sizeof(cm_msg));

    cm_rc_mr[peer] = ibv_reg_mr(cm_ib_context.pd, cm_rc_buf[peer],
                                sizeof(cm_msg), IBV_ACCESS_LOCAL_WRITE);

    if (!cm_rc_mr[peer]) {
        CM_ERR("Couldn't allocate MR");
        return -1;
    }
    CM_DBG("cm_enable_qp_init_to_rts Exit");
    
    return 0;
}

/*move qp to rtr*/

int cm_enable_qp_init_to_rtr(int peer, struct ibv_qp *qp,
                             uint16_t dlid, int dqpn)
{
    struct ibv_qp_attr attr;

    CM_DBG("cm_enable_qp_init_to_rtr Enter");
        
    memcpy(&attr, &(cm_ib_context.rc_qp_attr_to_rtr),
           sizeof(struct ibv_qp_attr));

    attr.dest_qp_num = dqpn;
    attr.ah_attr.dlid = dlid;

    if (ibv_modify_qp(qp, &attr, cm_ib_context.rc_qp_mask_to_rtr)) {
        CM_ERR("Failed to modify QP to RTR");
        return -1;
    }

    /*Register memory for recv */
    if (cm_ib_context.rc_qp_init_attr.srq != NULL) {
        CM_DBG("cm_enable_qp_init_to_rtr Exit");
        return 0;
    }

    cm_rc_buf[peer] = memalign(page_size, sizeof(cm_msg));
    if (!cm_rc_buf[peer]) {
        CM_ERR("Couldn't allocate working buf");
        return -1;
    }

    memset(cm_rc_buf[peer], 0, sizeof(cm_msg));

    cm_rc_mr[peer] = ibv_reg_mr(cm_ib_context.pd, cm_rc_buf[peer],
                                sizeof(cm_msg), IBV_ACCESS_LOCAL_WRITE);
    if (!cm_rc_mr[peer]) {
        CM_ERR("Couldn't allocate MR");
        return -1;
    }

    /*Post recv */
    {
        struct ibv_sge list;
        struct ibv_recv_wr wr;
        struct ibv_recv_wr *bad_wr;
        list.addr = (uintptr_t) cm_rc_buf[peer];
        list.length = sizeof(cm_msg);
        list.lkey = cm_rc_mr[peer]->lkey;
        memset(&wr, 0, sizeof(wr));
        wr.wr_id = cm_ib_context.confim_recv_wr_id;
        wr.sg_list = &list;
        wr.num_sge = 1;
        if (ibv_post_recv(qp, &wr, &bad_wr)) {
            CM_ERR("Couldn't post rc recv");
            return -1;
        }
    }
    
    CM_DBG("cm_enable_qp_init_to_rtr Exit");
        
    return 0;
}

/*Just modify to rts, not register anything*/

int cm_enable_qp_rtr_to_rts(struct ibv_qp *qp)
{
    struct ibv_qp_attr attr;
    CM_DBG("cm_enable_qp_rtr_to_rts Enter");
    
    memcpy(&attr, &(cm_ib_context.rc_qp_attr_to_rts),
           sizeof(struct ibv_qp_attr));

    if (ibv_modify_qp(qp, &attr, cm_ib_context.rc_qp_mask_to_rts)) {
        CM_ERR("Failed to modify QP to RTS");
        return -1;
    }
    CM_DBG("cm_enable_qp_rtr_to_rts Exit");
    
    return 0;
}

typedef struct cm_packet {
    struct timeval timestamp;        /*the time when timer begins */
    cm_msg payload;
} cm_packet;

typedef struct cm_pending {
    int cli_or_srv;             /*pending as a client or server */
    int peer;                   /*if peer = self rank, it's head node */
    cm_packet *packet;
    struct cm_pending *next;
    struct cm_pending *prev;
} cm_pending;

int cm_pending_num;

#define CM_PENDING_SERVER   0
#define CM_PENDING_CLIENT   1

cm_pending *cm_pending_head = NULL;

cm_pending *cm_pending_create()
{
    CM_DBG("cm_pending_create Enter");
    cm_pending *temp = (cm_pending *) malloc(sizeof(cm_pending));
    memset(temp, 0, sizeof(cm_pending));
    CM_DBG("cm_pending_create Exit");
    return temp;
}

int cm_pending_init(cm_pending * pending, cm_msg * msg)
{
    CM_DBG("cm_pending_init Enter");
    
    if (msg->msg_type == CM_MSG_TYPE_REQ) {
        pending->cli_or_srv = CM_PENDING_CLIENT;
        pending->peer = msg->server_rank;
    } else if (msg->msg_type == CM_MSG_TYPE_REP) {
        pending->cli_or_srv = CM_PENDING_SERVER;
        pending->peer = msg->client_rank;
    } else {
        CM_ERR("error message type");
        return -1;
    }
    pending->packet = (cm_packet *) malloc(sizeof(cm_packet));
    memcpy(&(pending->packet->payload), msg, sizeof(cm_msg));
    CM_DBG("cm_pending_init Exit");
    
    return 0;
}

cm_pending *cm_pending_search_peer(int peer, int cli_or_srv)
{
    cm_pending *pending = cm_pending_head;
    CM_DBG("cm_pending_search_peer Enter");
    
    while (pending->next != cm_pending_head) {
        pending = pending->next;
        if (pending->cli_or_srv == cli_or_srv && pending->peer == peer) {
            return pending;
        }
    }
    CM_DBG("cm_pending_search_peer Exit");
     
    return NULL;
}

int cm_pending_append(cm_pending * node)
{
    CM_DBG("cm_pending_append Enter");
    
    cm_pending *last = cm_pending_head->prev;
    last->next = node;
    node->next = cm_pending_head;
    cm_pending_head->prev = node;
    node->prev = last;
    cm_pending_num++;
    CM_DBG("cm_pending_append Exit");
    
    return 0;
}

int cm_pending_remove_and_destroy(cm_pending * node)
{
    CM_DBG("cm_pending_remove_and_destroy Enter");
        
    free(node->packet);
    node->next->prev = node->prev;
    node->prev->next = node->next;
    free(node);
    cm_pending_num--;
    CM_DBG("cm_pending_remove_and_destroy Exit");
    
    return 0;
}

int cm_pending_list_init()
{
    cm_pending_num = 0;
    cm_pending_head = cm_pending_create();
    cm_pending_head->peer = cm_ib_context.rank;
    cm_pending_head->prev = cm_pending_head;
    cm_pending_head->next = cm_pending_head;
    return 0;
}

int cm_pending_list_finalize()
{
    while (cm_pending_head->next != cm_pending_head) {
        cm_pending_remove_and_destroy(cm_pending_head->next);
    }
    assert(cm_pending_num==0);
    free(cm_pending_head);
    cm_pending_head = NULL;
    return 0;
}

/*functions for cm protocol*/
int cm_send_ud_msg(cm_msg * msg)
{
    int ret;
    cm_pending *pending;
    struct timeval now;

    CM_DBG("cm_send_ud_msg Enter");
    pending = cm_pending_create();
    if (cm_pending_init(pending, msg)) {
        CM_ERR("cm_pending_init failed");
        return -1;
    }
    cm_pending_append(pending);

    gettimeofday(&now, NULL);
    pending->packet->timestamp = now;
    ret = cm_post_ud_packet(&(pending->packet->payload));
    if (ret) {
        CM_ERR("cm_post_ud_packet failed %d", ret);
        return -1;
    }
    if (cm_pending_num == 1) {
        pthread_cond_signal(&cm_cond_new_pending);
    }
    CM_DBG("cm_send_ud_msg Exit");
    
    return 0;
}

int cm_accept_and_cancel(cm_msg * msg)
{
    cm_msg msg_send;

    /*Prepare QP */
    CM_DBG("cm_accept_and_cancel Enter");
    
    if (cm_enable_qp_init_to_rtr(msg->client_rank,
                                 cm_rc_qp[msg->client_rank], msg->lid,
                                 msg->qpn)) {
        CM_ERR("cm_enable_qp_init_to_rtr failed");
        return -1;
    }

    /*call callback function. This must be non-blocking, No send descriptor can be post */
    if (cm_listen_callback) {
        (*cm_listen_callback) (msg->client_rank,
                               cm_rc_qp[msg->client_rank],
                               cm_listen_callback_arg);
    }


    /*Prepare rep msg */
    memcpy(&msg_send, msg, sizeof(cm_msg));
    msg_send.msg_type = CM_MSG_TYPE_REP;
    msg_send.lid = cm_lid[cm_ib_context.rank];
    msg_send.qpn = (cm_rc_qp[msg->client_rank])->qp_num;

    /*Send rep msg */


    if (cm_send_ud_msg(&msg_send)) {
        CM_ERR("cm_send_ud_msg failed");
        return -1;
    }

    cm_state_srv[msg->client_rank] = CM_CONN_STATE_S_REQUESTED;

    CM_DBG("cm_accept_and_cancel Cancel");
    /*Cancel client role */
    {
        cm_pending *pending;
        pending =
            cm_pending_search_peer(msg->client_rank, CM_PENDING_CLIENT);
        if (NULL == pending) {
            CM_ERR("Can't find pending entry");
            return -1;
        }

        cm_pending_remove_and_destroy(pending);
        if (NULL == pending)
            cm_state_cli[msg->client_rank] = CM_CONN_STATE_C_IDLE;
    }
    CM_DBG("cm_accept_and_cancel Exit");
    
    return 0;
}


int cm_accept(cm_msg * msg)
{
    cm_msg msg_send;
    /*Prepare QP */
    CM_DBG("cm_accpet Enter");
    
    cm_rc_qp[msg->client_rank] = cm_create_rc_qp();

    if (cm_enable_qp_init_to_rtr(msg->client_rank,
                                 cm_rc_qp[msg->client_rank], msg->lid,
                                 msg->qpn)) {
        CM_ERR("cm_enable_qp_init_to_rtr failed");
        return -1;
    }

    /*call callback function. This must be non-blocking, No send descriptor can be post */
    if (cm_listen_callback) {
        (*cm_listen_callback) (msg->client_rank,
                               cm_rc_qp[msg->client_rank],
                               cm_listen_callback_arg);
    }

    /*Prepare rep msg */
    memcpy(&msg_send, msg, sizeof(cm_msg));
    msg_send.msg_type = CM_MSG_TYPE_REP;
    msg_send.lid = cm_lid[cm_ib_context.rank];
    msg_send.qpn = (cm_rc_qp[msg->client_rank])->qp_num;

    /*Send rep msg */
    if (cm_send_ud_msg(&msg_send)) {
        CM_ERR("cm_send_ud_msg failed");
        return -1;
    }

    cm_state_srv[msg->client_rank] = CM_CONN_STATE_S_REQUESTED;
    CM_DBG("cm_accpet Exit");
    return 0;
}

int cm_enable(cm_msg * msg)
{
    cm_msg msg_to_send;
    CM_DBG("cm_enable Enter");
    if (cm_enable_qp_init_to_rts(msg->server_rank,
                                 cm_rc_qp[msg->server_rank],
                                 msg->lid, msg->qpn)) {
        CM_ERR("cm_enable_qp_init_to_rtr failed");
        return -1;
    }

    /* call callback function. 
     * This must be non-blocking, 
     * No send descriptor can be post*/
    if (cm_connect_callback[msg->server_rank]) {
        (*cm_connect_callback[msg->server_rank])
            (msg->server_rank,
             cm_rc_qp[msg->server_rank],
             cm_connect_callback_arg[msg->server_rank]);
    }

    /*Send RTU */
    memcpy(&msg_to_send, msg, sizeof(cm_msg));
    msg_to_send.qpn = -1;
    msg_to_send.lid = -1;
    msg_to_send.msg_type = CM_MSG_TYPE_RTU;
    if (cm_send_rc_msg(&msg_to_send)) {
        CM_ERR("cm_send_rc_msg failed");
        return -1;
    }
    /*Mark QP */
    cm_state_cli[msg->server_rank] = CM_CONN_STATE_C_IDLE;
    cm_conn_type[msg->server_rank] = msg->conn_type;
    CM_DBG("cm_enable Exit");
    return 0;
}

int cm_handle_msg(cm_msg * msg)
{
    CM_DBG("Handle cm_msg: msg_type: %d, client_rank %d, server_rank %d",
           msg->msg_type, msg->client_rank, msg->server_rank);
    switch (msg->msg_type) {
    case CM_MSG_TYPE_REQ:
        {
            pthread_mutex_lock(&cm_conn_state_lock);
            if (cm_conn_type[msg->client_rank] == msg->conn_type) {
                /*already existing */
                pthread_mutex_unlock(&cm_conn_state_lock);
                return 0;
            } else if (cm_state_srv[msg->client_rank] !=
                       CM_CONN_STATE_S_IDLE) {
                /*already a pending request from that peer */
                pthread_mutex_unlock(&cm_conn_state_lock);
                return 0;
            } else if (cm_state_cli[msg->client_rank] !=
                       CM_CONN_STATE_C_IDLE) {
                /*already initiated a request to that peer */
                if (msg->client_rank > cm_ib_context.rank) {
                    /*that peer should be server */
                    pthread_mutex_unlock(&cm_conn_state_lock);
                    return 0;
                } else {
                    /*myself should be server */
                    cm_accept_and_cancel(msg);
                }
            } else if (!(cm_listening[msg->client_rank] & msg->conn_type)) {
                /*no such service enabled */
                pthread_mutex_unlock(&cm_conn_state_lock);
                return 0;
            } else {
                cm_accept(msg);
            }

            pthread_mutex_unlock(&cm_conn_state_lock);
        }

        break;

    case CM_MSG_TYPE_REP:
        {
            cm_pending *pending;
            pthread_mutex_lock(&cm_conn_state_lock);

            if (cm_state_cli[msg->server_rank] !=
                CM_CONN_STATE_C_REQUESTING) {
                /*not waiting for any reply */
                pthread_mutex_unlock(&cm_conn_state_lock);
                return 0;
            }

            pending =
                cm_pending_search_peer(msg->server_rank,
                                       CM_PENDING_CLIENT);
            if (NULL == pending) {
                CM_ERR("Can't find pending entry");
                return -1;
            }
            cm_pending_remove_and_destroy(pending);
            cm_enable(msg);
            pthread_mutex_unlock(&cm_conn_state_lock);
        }
        break;
    case CM_MSG_TYPE_RTU:
        {
            cm_pending *pending;
            pthread_mutex_lock(&cm_conn_state_lock);

            if (cm_state_srv[msg->client_rank] !=
                CM_CONN_STATE_S_REQUESTED) {
                /*not waiting for comfirm */
                pthread_mutex_unlock(&cm_conn_state_lock);
                return 0;
            }

            pending =
                cm_pending_search_peer(msg->client_rank,
                                       CM_PENDING_SERVER);
            if (NULL == pending) {
                CM_ERR("Can't find pending entry");
                return -1;
            }
            cm_pending_remove_and_destroy(pending);
            if (cm_enable_qp_rtr_to_rts(cm_rc_qp[msg->client_rank])) {
                CM_ERR("cm_enable_qp_rtr_to_rts failed");
                return -1;
            }
            cm_state_srv[msg->client_rank] = CM_CONN_STATE_S_IDLE;
            cm_conn_type[msg->client_rank] = msg->conn_type;
            pthread_mutex_unlock(&cm_conn_state_lock);
        }
        break;
    default:
        CM_ERR("Unknown msg type: %d", msg->msg_type);
        return -1;
    }
    CM_DBG("cm_handle_msg Exit");
    return 0;
}

void *cm_timeout_handler(void *arg)
{
    struct timeval now;
    int delay, ret;
    cm_pending *p;
    struct timespec remain;
    while (1) {
        pthread_mutex_lock(&cm_conn_state_lock);
        while (cm_pending_num == 0) {
            pthread_cond_wait(&cm_cond_new_pending, &cm_conn_state_lock);
        }
        while (1) {
            pthread_mutex_unlock(&cm_conn_state_lock);
            nanosleep(&cm_timeout,&remain);/*Not handle the EINTR*/
            pthread_mutex_lock(&cm_conn_state_lock);
            if (cm_pending_num == 0) {
                break;
            }
            CM_DBG("Time out");
            p = cm_pending_head;
            if (NULL == p) {
                CM_ERR("cm_pending_head corrupted");
            }
            gettimeofday(&now, NULL);
            while (p->next != cm_pending_head) {
                p = p->next;
                delay = (now.tv_sec - p->packet->timestamp.tv_sec) * 1000000
                    + (now.tv_usec - p->packet->timestamp.tv_usec);
                if (delay > cm_timeout_usec) {       /*Timer expired */
                    CM_DBG("Resend");
                    p->packet->timestamp = now;
                    ret = cm_post_ud_packet(&(p->packet->payload));
                    if (ret) {
                        CM_ERR("cm_post_ud_packet failed %d", ret);
                    }
                    gettimeofday(&now,NULL);
                }
            }
            CM_DBG("Time out exit");
        }
        pthread_mutex_unlock(&cm_conn_state_lock);
    }
    return NULL;
}

void *cm_completion_handler(void *arg)
{
    while (1) {
        struct ibv_cq *ev_cq;
        void *ev_ctx;
        struct ibv_wc wc;
        int ne;
        int spin_count;
        int ret;

        CM_DBG("Waiting for cm message");

        do {
            ret = ibv_get_cq_event(cm_ud_comp_ch, &ev_cq, &ev_ctx);
            if (ret && errno != EINTR ) {
                CM_ERR("Failed to get cq_event: %d", ret);
                return NULL;
            }
        } while(ret && errno == EINTR);

        ibv_ack_cq_events(ev_cq, 1);

        if (ev_cq != cm_ud_recv_cq) {
            CM_ERR("CQ event for unknown CQ %p", ev_cq);
            return NULL;
        }

        CM_DBG("Processing cm message");
          
        spin_count = 0;
        do {
            ne = ibv_poll_cq(cm_ud_recv_cq, 1, &wc);
            if (ne < 0) {
                CM_ERR("poll CQ failed %d", ne);
                return NULL;
            } else if (ne == 0) {
                spin_count++;
                continue;
            }

            spin_count = 0;

            if (wc.status != IBV_WC_SUCCESS) {
                CM_ERR("Failed status %d for wr_id %d",
                        wc.status, (int) wc.wr_id);
                return NULL;
            }

            if (wc.wr_id == CM_UD_RECV_WR_ID) {
                void *buf =
                    cm_ud_recv_buf +
                    cm_ud_recv_buf_index * (sizeof(cm_msg) + 40) + 40;
                cm_msg *msg = (cm_msg *) buf;
                cm_handle_msg(msg);
                CM_DBG("Post recv");
                cm_post_ud_recv(buf - 40, sizeof(cm_msg));
                cm_ud_recv_buf_index =
                    (cm_ud_recv_buf_index + 1) % cm_recv_buffer_size;
            }
        }while (spin_count < cm_max_spin_count);

        CM_DBG("notify_cq");
        if (ibv_req_notify_cq(cm_ud_recv_cq, 1)) {
            CM_ERR("Couldn't request CQ notification");
            return NULL;
        }
    }
    return NULL;
}

int MPICM_Init_UD(MPICM_ib_context * context, uint32_t * ud_qpn)
{
    int i, ret;
    memcpy(&cm_ib_context, context, sizeof(MPICM_ib_context));
    char *value;

    /*Initialization */
    cm_listening = malloc(cm_ib_context.size * sizeof(int));
    cm_conn_type = malloc(cm_ib_context.size * sizeof(MPICM_conn_type));
    cm_state_cli = malloc(cm_ib_context.size * sizeof(CM_conn_state_cli));
    cm_state_srv = malloc(cm_ib_context.size * sizeof(CM_conn_state_srv));
    cm_ah = malloc(cm_ib_context.size * sizeof(struct ibv_ah *));
    cm_rc_qp = malloc(cm_ib_context.size * sizeof(struct ibv_qp *));
    cm_rc_mr = malloc(cm_ib_context.size * sizeof(struct ibv_mr *));
    cm_ud_qpn = malloc(cm_ib_context.size * sizeof(uint32_t));
    cm_lid = malloc(cm_ib_context.size * sizeof(uint16_t));
    cm_rc_buf = malloc(cm_ib_context.size * sizeof(void *));

    cm_connect_callback =
        malloc(cm_ib_context.size *
               sizeof(int (*)(int, struct ibv_qp *, void *)));
    cm_connect_callback_arg = malloc(cm_ib_context.size * sizeof(void *));

    cm_req_id_global = 0;

    page_size = sysconf(_SC_PAGESIZE);

    if ((value = getenv("VIADEV_CM_SEND_DEPTH")) != NULL) {
        cm_send_depth = atoi(value);
    } else {
        cm_send_depth = DEFAULT_CM_SEND_DEPTH;
    }

    if ((value = getenv("VIADEV_CM_RECV_BUFFERS")) != NULL) {
        cm_recv_buffer_size = atoi(value);
    } else {
        cm_recv_buffer_size = DEFAULT_CM_MSG_RECV_BUFFER_SIZE;
    }

    if ((value = getenv("VIADEV_CM_UD_PSN")) != NULL) {
        cm_ud_psn = atoi(value);
    } else {
        cm_ud_psn = CM_UD_DEFAULT_PSN;
    }

    if ((value = getenv("VIADEV_CM_MAX_SPIN_COUNT")) != NULL) {
        cm_max_spin_count = atoi(value);
    } else {
        cm_max_spin_count = DEFAULT_CM_MAX_SPIN_COUNT;
    }
    
    if ((value = getenv("VIADEV_CM_THREAD_STACKSIZE")) != NULL) {
        cm_thread_stacksize = atoi(value);
    } else {
        cm_thread_stacksize = DEFAULT_CM_THREAD_STACKSIZE;
    }
   
    if ((value = getenv("VIADEV_CM_TIMEOUT")) != NULL) {
        cm_timeout_usec = atoi(value)*1000;
    } else { 
        cm_timeout_usec = CM_DEFAULT_TIMEOUT;
    }
    if (cm_timeout_usec < CM_MIN_TIMEOUT) {
        cm_timeout_usec = CM_MIN_TIMEOUT;
    }

    cm_timeout.tv_sec = cm_timeout_usec/1000000;
    cm_timeout.tv_nsec = (cm_timeout_usec-cm_timeout.tv_sec*1000000)*1000;

    cm_ud_buf =
        memalign(page_size,
                 (sizeof(cm_msg) + 40) * (cm_recv_buffer_size + 1));
    if (!cm_ud_buf) {
        CM_ERR("Couldn't allocate work buf");
        return -1;
    }
    
    memset(cm_ud_buf, 0,
           (sizeof(cm_msg) + 40) * (cm_recv_buffer_size + 1));
    cm_ud_send_buf = cm_ud_buf;
    cm_ud_recv_buf = cm_ud_buf + sizeof(cm_msg) + 40;

    cm_ud_comp_ch = ibv_create_comp_channel(context->context);
    if (!cm_ud_comp_ch) {
        CM_ERR("Couldn't create completion channel");
        return -1;
    }

    cm_ud_mr = ibv_reg_mr(cm_ib_context.pd, cm_ud_buf,
                          (sizeof(cm_msg) +
                           40) * (cm_recv_buffer_size + 1),
                          IBV_ACCESS_LOCAL_WRITE);
    if (!cm_ud_mr) {
        CM_ERR("Couldn't allocate MR");
        return -1;
    }

    cm_ud_recv_cq =
        ibv_create_cq(cm_ib_context.context, cm_recv_buffer_size, NULL,
                      cm_ud_comp_ch, 0);
    if (!cm_ud_recv_cq) {
        CM_ERR("Couldn't create CQ");
        return -1;
    }

    cm_ud_send_cq =
        ibv_create_cq(cm_ib_context.context, cm_send_depth, NULL, NULL, 0);
    if (!cm_ud_send_cq) {
        CM_ERR("Couldn't create CQ");
        return -1;
    }

    {
        struct ibv_qp_init_attr attr;
        memset(&attr, 0, sizeof(struct ibv_qp_init_attr));
        attr.send_cq = cm_ud_send_cq;
        attr.recv_cq = cm_ud_recv_cq;
        attr.cap.max_send_wr = cm_send_depth;
        attr.cap.max_recv_wr = cm_recv_buffer_size;
        attr.cap.max_send_sge = 1;
        attr.cap.max_recv_sge = 1;
        attr.qp_type = IBV_QPT_UD;

        cm_ud_qp = ibv_create_qp(cm_ib_context.pd, &attr);
        if (!cm_ud_qp) {
            CM_ERR("Couldn't create UD QP");
            return -1;
        }
    }

    *ud_qpn = cm_ud_qp->qp_num;
    {
        struct ibv_qp_attr attr;
        memset(&attr, 0, sizeof(struct ibv_qp_attr));

        attr.qp_state = IBV_QPS_INIT;
        attr.pkey_index = 0;
        attr.port_num = cm_ib_context.hca_port;
        attr.qkey = 0;

        if ((ret = ibv_modify_qp(cm_ud_qp, &attr,
                                 IBV_QP_STATE |
                                 IBV_QP_PKEY_INDEX |
                                 IBV_QP_PORT | IBV_QP_QKEY))) {
            CM_ERR("Failed to modify QP to INIT, ret = %d, errno=%d",
                    ret, errno);
            return -1;
        }
    }
    {
        struct ibv_qp_attr attr;
        memset(&attr, 0, sizeof(struct ibv_qp_attr));

        attr.qp_state = IBV_QPS_RTR;
        if (ibv_modify_qp(cm_ud_qp, &attr, IBV_QP_STATE)) {
            CM_ERR("Failed to modify QP to RTR");
            return -1;
        }
    }

    {
        struct ibv_qp_attr attr;
        memset(&attr, 0, sizeof(struct ibv_qp_attr));

        attr.qp_state = IBV_QPS_RTS;
        attr.sq_psn = cm_ud_psn;
        if (ibv_modify_qp(cm_ud_qp, &attr, IBV_QP_STATE | IBV_QP_SQ_PSN)) {
            CM_ERR("Failed to modify QP to RTS");
            return -1;
        }
    }

    for (i = 0; i < cm_recv_buffer_size; i++) {
        /*  cm_msg * msg = (cm_msg *)(cm_ud_recv_buf+(sizeof(cm_msg)+40)*i+40); */
        if (cm_post_ud_recv
            (cm_ud_recv_buf + (sizeof(cm_msg) + 40) * i, sizeof(cm_msg))) {
            CM_ERR("cm_post_ud_recv failed");
            return -1;
        }
    }
    cm_ud_recv_buf_index = 0;

    if (ibv_req_notify_cq(cm_ud_recv_cq, 1)) {
        CM_ERR("Couldn't request CQ notification");
        return -1;
    }

    for (i = 0; i < cm_ib_context.size; i++) {
        cm_conn_type[i] = MPICM_IB_NONE;
        cm_rc_qp[i] = NULL;
        cm_state_cli[i] = CM_CONN_STATE_C_IDLE;
        cm_state_srv[i] = CM_CONN_STATE_S_IDLE;
        cm_listening[i] = 0;
    }

    cm_pending_list_init();
    return 0;
}

int MPICM_Connect_UD(uint32_t * qpns, uint16_t * lids)
{
    int i, ret;
    pthread_attr_t attr;
    if (pthread_attr_init(&attr))
    {
        CM_ERR("pthread_attr_init failed\n");
        return -1;
    }
    
    /*Copy qpns and lids */
    memcpy(cm_ud_qpn, qpns, cm_ib_context.size * sizeof(uint32_t));
    memcpy(cm_lid, lids, cm_ib_context.size * sizeof(uint16_t));

    /*Create address handles */
    for (i = 0; i < cm_ib_context.size; i++) {
        struct ibv_ah_attr ah_attr;
        memset(&ah_attr, 0, sizeof(ah_attr));
        ah_attr.is_global = 0;
        ah_attr.dlid = cm_lid[i];
        ah_attr.sl = 0;
        ah_attr.src_path_bits = 0;
        ah_attr.port_num = cm_ib_context.hca_port;
        cm_ah[i] = ibv_create_ah(cm_ib_context.pd, &ah_attr);
        if (!cm_ah[i]) {
            CM_ERR("Failed to create AH");
            return -1;
        }
    }

    /*Spawn cm thread */
    ret = pthread_attr_setstacksize(&attr,cm_thread_stacksize);
    if (ret && ret != EINVAL) 
    {
        CM_ERR("pthread_attr_setstacksize failed\n");
        return -1;
    }
    pthread_create(&cm_comp_thread, &attr, cm_completion_handler, NULL);
    pthread_create(&cm_timer_thread, &attr, cm_timeout_handler, NULL);
    return 0;
}

int MPICM_Finalize_UD()
{
    int i;

    CM_DBG("In MPICM_Finalize_UD");

    cm_pending_list_finalize();
    /*Cancel cm thread */
    pthread_cancel(cm_comp_thread);
    pthread_cancel(cm_timer_thread);
    
    /*Clean up */
    for (i = 0; i < cm_ib_context.size; i++)
        ibv_destroy_ah(cm_ah[i]);
    ibv_destroy_qp(cm_ud_qp);

    ibv_destroy_comp_channel(cm_ud_comp_ch);
    ibv_destroy_cq(cm_ud_recv_cq);
    ibv_destroy_cq(cm_ud_send_cq);
    ibv_dereg_mr(cm_ud_mr);

    free(cm_ud_buf);
    free(cm_listening);
    free(cm_conn_type);
    free(cm_state_cli);
    free(cm_state_srv);
    free(cm_ah);
    free(cm_rc_qp);
    free(cm_rc_mr);
    free(cm_ud_qpn);
    free(cm_lid);
    free(cm_connect_callback);
    free(cm_connect_callback_arg);
    free(cm_rc_buf);

    CM_DBG("MPICM_Finalize_UD done");
    return 0;
}

int MPICM_Listen(int available_conn_type_mask,
                 int (*callback) (int, struct ibv_qp *, void *),
                 void *callback_arg)
{
    int i;

    pthread_mutex_lock(&cm_conn_state_lock);

    for (i = 0; i < cm_ib_context.size; i++) {
        if (i == cm_ib_context.rank) {
            continue;
        }
        cm_listening[i] = available_conn_type_mask;
    }

    cm_listen_callback = callback;
    cm_listen_callback_arg = callback_arg;
    pthread_mutex_unlock(&cm_conn_state_lock);
    return 0;
}

int MPICM_Connect_req(int peer_rank, MPICM_conn_type conn_type_req,
                      int (*callback) (int, struct ibv_qp *, void *),
                      void *callback_arg)
{
    cm_msg req;
    pthread_mutex_lock(&cm_conn_state_lock);

    if (cm_ib_context.rank == peer_rank) {
        pthread_mutex_unlock(&cm_conn_state_lock);
        return 1;
    }
    if (cm_conn_type[peer_rank] == conn_type_req) {
        pthread_mutex_unlock(&cm_conn_state_lock);
        return 2;
    }
    if (cm_state_cli[peer_rank] == CM_CONN_STATE_C_REQUESTING) {
        pthread_mutex_unlock(&cm_conn_state_lock);
        return 3;
    }
    if (cm_state_srv[peer_rank] != CM_CONN_STATE_S_IDLE) {
        /*handling the request from peer */
        pthread_mutex_unlock(&cm_conn_state_lock);
        return 0;
    }

    CM_DBG("Sending Req to rank %d", peer_rank);
    cm_state_cli[peer_rank] = CM_CONN_STATE_C_REQUESTING;
    cm_rc_qp[peer_rank] = cm_create_rc_qp();
    cm_connect_callback[peer_rank] = callback;
    cm_connect_callback_arg[peer_rank] = callback_arg;

    req.cm_flag = CM_MSG_FLAG;
    req.msg_type = CM_MSG_TYPE_REQ;
    req.client_rank = cm_ib_context.rank;
    req.server_rank = peer_rank;
    req.req_id = ++cm_req_id_global;
    req.conn_type = conn_type_req;
    req.lid = cm_lid[cm_ib_context.rank];
    req.qpn = (cm_rc_qp[req.server_rank])->qp_num;
    if (cm_send_ud_msg(&req)) {
        CM_ERR("cm_send_ud_msg failed");
        return -1;
    }
    pthread_mutex_unlock(&cm_conn_state_lock);
    return 0;
}

MPICM_conn_type MPICM_Get_conn_type(int peer_rank)
{
    pthread_mutex_lock(&cm_conn_state_lock);
    MPICM_conn_type conn = cm_conn_type[peer_rank];
    pthread_mutex_unlock(&cm_conn_state_lock);
    return conn;
}

int MPICM_Lookup_peer_rank(uint32_t qpn, int *peer_rank)
{
    int i;
    for (i = 0; i < cm_ib_context.size; i++) {
        if (cm_rc_qp[i] != NULL && cm_rc_qp[i]->qp_num == qpn) {
            *peer_rank = i;
            return 0;
        }
    }
    return -1;
}

int MPICM_Server_connection_establish(int peer_rank, void *msg)
{
    if (NULL == msg) {
        cm_handle_msg((cm_msg *) cm_rc_buf[peer_rank]);
        ibv_dereg_mr(cm_rc_mr[peer_rank]);
        free(cm_rc_buf[peer_rank]);
    } else {
        cm_handle_msg((cm_msg *) msg);
    }
    CM_DBG("MPICM_Server_connection_establish, peer_rank %d", peer_rank);
    return 0;
}

int MPICM_Client_connection_establish(int peer_rank)
{
    ibv_dereg_mr(cm_rc_mr[peer_rank]);
    free(cm_rc_buf[peer_rank]);
    CM_DBG("MPICM_Client_connection_establish, peer_rank %d", peer_rank);
    return 0;
}

#endif
