
/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef MPID_CM_H
#define MPID_CM_H

#ifdef ON_DEMAND
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <malloc.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <time.h>
#include <signal.h>
#include <infiniband/verbs.h>
#include <errno.h>
#include <assert.h>

typedef enum MPICM_conn_type {
    MPICM_IB_NONE = 1,
    MPICM_IB_RC_PT2PT = 1 << 1,
}MPICM_conn_type;

typedef struct MPICM_ib_context{
    int size;
    int rank;
    struct ibv_context *context;
    int  hca_port;
    struct ibv_pd *pd;

    uint64_t confim_recv_wr_id;
    uint64_t confim_send_wr_id;
    struct ibv_qp_init_attr rc_qp_init_attr;
    struct ibv_qp_attr rc_qp_attr_to_init;
    enum ibv_qp_attr_mask rc_qp_mask_to_init;
    struct ibv_qp_attr rc_qp_attr_to_rtr;
    enum ibv_qp_attr_mask rc_qp_mask_to_rtr;
    struct ibv_qp_attr rc_qp_attr_to_rts;
    enum ibv_qp_attr_mask rc_qp_mask_to_rts;
}MPICM_ib_context;

/*
MPICM_Init_UD
Initialize MPICM based on UD connection, need to be called before calling 
any of following interfaces, it will return a qpn for the established ud qp, 
mpi lib needs to exchange the ud qpn with other procs
*/
int MPICM_Init_UD(MPICM_ib_context * context, uint32_t *ud_qpn);

/*
MPICM_Connect_UD
Provide connect information to UD
*/
int MPICM_Connect_UD(uint32_t *qpns, uint16_t *lids);

/*
MPICM_Listen
Listen to a set of connection services, prepare to accept connection requests, 
need to be called in initialization phase before any other process requesting 
connection, if callback is not NULL,  the callback function will be called
before the it sends reply message to client.
first parameter to callback is the client rank
*/
int MPICM_Listen(int available_conn_type_mask, 
        int (*callback)(int, struct ibv_qp *, void *), 
        void *callback_arg);

/*
MPICM_Connect_req
Request to establish a type of connection, it guarantees the request being 
sent, use needs to call MPICM_Get_conn_type to check established connection.
*/
int MPICM_Connect_req(int peer_rank, MPICM_conn_type conn_type_req, 
        int (*callback)(int, struct ibv_qp *, void *), void *callback_arg);

/*
MPICM_Get_conn_type
Get the current available connection type to another process.
*/
MPICM_conn_type MPICM_Get_conn_type(int peer_rank);

/*
MPICM_Finalize_UD
Finalize cm, clean supporting thread
*/
int MPICM_Finalize_UD();

/*
MPICM_Lookup_peer_rank
If there is some pending request corresponding to the qpn, peer rank is set, and return
0, otherwise, -1 is returned
*/
int MPICM_Lookup_peer_rank(uint32_t qpn, int *peer_rank);

/*
MPICM_Server_connection_done
When polls a recv completion with the wr_id == confim_recv_wr_id, call this function
to mark the connection established. Look up peer rank by qpn first. 
In SRQ case, the message will not be in the buffer managed by cmlib, so user can 
specify the address of the message.
*/
int MPICM_Server_connection_establish(int peer_rank, void *msg);

/*
MPICM_Server_connection_done
When polls a send completion with the wr_id == confim_send_wr_id, call this function
to dreg mr and free send buf. Look up peer rank by qpn first.
*/
int MPICM_Client_connection_establish(int peer_rank);

#define CM_ERR(args...)  \
    fprintf(stderr, "[Rank %d][%s: line %d]", cm_ib_context.rank ,__FILE__, __LINE__); \
    fprintf(stderr, args); \
    fprintf(stderr, "\n");

#ifdef CM_DEBUG
#define CM_DBG(args...)  \
    fprintf(stderr, "[Rank %d][%s: line %d]", cm_ib_context.rank ,__FILE__, __LINE__); \
    fprintf(stderr, args); \
    fprintf(stderr, "\n");
#define CM_DBG_MPI(args...) \
    fprintf(stderr, "[Rank %d][%s: line %d]", viadev.me ,__FILE__, __LINE__); \
    fprintf(stderr, args); \
    fprintf(stderr, "\n");
#else
#define CM_DBG(args...)
#define CM_DBG_MPI(args...)
#endif

#endif  /* ON_DEMAND */
#endif  /* MPID_CM_H */
