/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "collutils.h"
#include "viaparam.h"
#include "vbuf.h"
#include "viapacket.h"
#include "ibverbs_const.h"

/*
 * ==============================================================
 * Initialize global parameter variables to default values
 * ==============================================================
 */

/* 
 *  Which HCA device is used?
 *  (1) If a user specifies "VIADEV_DEVICE" environmental variable,
 *      the specified HCA device will be used.
 *  (2) Otherwise, MVAPICH first gets a list of the device IDs 
 *      of the available HCA devices on the physical node.
 *  (3) Then, it chooses the first device by default.
 */
#define DEFAULT_HCA_ID      (0)

/*
 * The user definition of the cluster size. This is used to define the
 * memory usage scheme. For larger clusters, less and smaller
 * buffers are used. This will effect performance but will ensure
 * memory scalability for larger clusters.
 * 0  - Small cluster
 * 1  - Medium size cluster
 * 2  - Large cluster
 * -1 - Auto-size (1-64: small, 65-256: Medium, 257+: Large)
 */
int viadev_cluster_size = -1;

/*  Limit the size of the registration cache. This is the amount of
 *  pages registered but not used.
 *  Note that for overlapping pages the pages can be counted more
 *  than once. To fix that one can change the caching algorithm so that
 *  each page is only registered once.
 *  VIADEV_REG_CACHE_HAVE_LIMIT flag enables the cache limit feature.
 *  VIADEV_DREG_CACHE_LIMIT = 0 indicates no cache limit.
 */
unsigned long int viadev_dreg_cache_limit = 0;

/* Max number of entries on the Send Q of QPs per connection.
 * Should be about (prepost_depth + extra).
 * Must be within NIC MaxQpEntries limit.
 * Size will be adjusted below.
 */
unsigned long viadev_sq_size = 200;

/* Max number of entries on the RecvQ of QPs per connection.
 * computed to be:
 * prepost_depth + viadev_prepost_rendezvous_extra + viadev_prepost_noop_extra
 * Must be within NIC MaxQpEntries limit.
 */
unsigned long viadev_rq_size;

/* Max number of entries on the completion queue at a time.
 * Needed when constructing the completion queue, but can
 * be grown dynamically if needed.
 * Should be about (viadev_sq_size + viadev_rq_size)*num_vi.
 * Must be within NIC MaxCQEntries limit.
 * Size will be adjusted below.
 */
unsigned long viadev_cq_size = 40000;

/* number of vbufs used in RDMA_FAST_PATH on each VI */
int viadev_num_rdma_buffer = VIADEV_NUM_RDMA_BUFFER;

#ifdef ADAPTIVE_RDMA_FAST_PATH

/* Adaptive RDMA Eager flow
 * Added two configuration parameters:
 *
 * > viadev_rdma_eager_limit
 *       Limits the number of connections that can be upgraded to use
 *       the RDMA buffers for Eager flow.
 *       If this parameter is not set, there will be no limit. 
 *       -1 : The limit is 2 * log2(number of processes)
 *       Any other value will be the actual limit.
 *
 * > viadev_rdma_eager_threshold
 *       Define the number of Eager packets the receiver will get
 *       before it decide to alloacte RDMA buffers for this connection
 *       -1 : allocate buffers on initialization.
 */
int viadev_rdma_eager_limit = -1;
int viadev_rdma_eager_threshold = 10;

#endif

#ifdef SRQ

uint32_t viadev_srq_size = 512;
uint32_t viadev_srq_limit = 30;
uint32_t viadev_max_r3_oust_send = 32;

#endif


/* number of vbufs to pre-post as receives on each VI */
int viadev_prepost_depth = VIADEV_PREPOST_DEPTH;
int viadev_prepost_threshold = 10;
int viadev_initial_prepost_depth = 12;
int viadev_initial_credits;

/* number of additional initial vbufs to allocate per VI
 * after initial allocation is exhausted.
 */
static int viadev_vbuf_extra = 10;


/* The total size of each vbuf. Used to be the eager threshold, but
 * it can be smaller, so that each eager message will span over few
 * vbufs
 */
int viadev_vbuf_total_size = VBUF_TOTAL_SIZE;

/* In the case of small numbers of VIs, allocate at least
 * this number of VBUFs whenever a new region is allocated.
 */

/* number of vbufs to allocate initially.
 * This will be re-defined after reading the parameters below
 * to scale to the number of VIs and other factors.
 */
int viadev_vbuf_pool_size = 512;

/* number of vbufs to allocate in a secondary region if we should
 * run out of the initial allocation.  This is re-computed (below)
 * once other parameters are known.
 */
int viadev_vbuf_secondary_pool_size = 128;

/* max (total) number of vbufs to allocate, after which process 
 * terminates with a fatal error.
 * -1 means no limit.
 */
int viadev_vbuf_max = -1;

/* The number of "extra" vbufs that will be posted as receives
 * on a connection in anticipation of an R3 rendezvous message.
 * The TOTAL number of VBUFs posted on a receive queue at any
 * time is viadev_prepost_depth + viadev_prepost_rendezvous_extra
 * regardless of the number of outstanding R3 sends active on
 * a connection.
 */
int viadev_prepost_rendezvous_extra = VIADEV_PREPOST_RENDEZVOUS_EXTRA;

/* The message size (in bytes) at which we switch from an EAGER
 * protocol to the three-way RENDEZVOUS protocol.
 * In reality, smaller messages might be sent using R3
 * rather than EAGER if remote credits are low.
 */
int viadev_rendezvous_threshold = DEFAULT_RENDEZVOUS_THRESHOLD;

/* the number of times the MPID_deviceCheck will poll VipCQDone
 * before it reverts to blocking on VipCQWait.
 * This value could influence latency and should probably be
 * adjusted dynamically.
 */
int viadev_spin_count = 100;

/* allow some extra buffers for non-credited packets (eg. NOOP) */
#ifdef MEMORY_RELIABLE
int viadev_prepost_noop_extra = 16;
#else
int viadev_prepost_noop_extra = 8;
#endif



/* dont allow EAGER sends or R3 data packets to consume all
 * remote credits on a connection.  This parameter records
 * the number of credits per connection that will be preserved
 * for non-data, control packets.
 */
#ifdef SRQ
int viadev_credit_preserve = 100;
#else
int viadev_credit_preserve = 10;
#endif

/* As receive descriptors (VBUFs) are consumed and reaped on
 * a connection, the receive side posts more receive VBUFs.
 * Each time a message is sent to the remote end of a connection,
 * a credit update value, representing the number of new receive
 * VBUFs posted since the last message was sent, is piggybacked
 * on the message.  The remote end can then update its remote
 * credit value, possibly unblocking stalled R3 data sends.
 * If the receiving side has no reason to send a message, the
 * credit update accumulate and do not get communicated to the
 * remote end.  If the credit update value exceeds this
 * viadev_credit_notify_threshold value, a NoOp message is
 * sent to the remote end with the sole purpose of communicating
 * the credit update and thereby preventing deadlock.
 *
 * There are two places where NoOp operations may be sent.
 * The first is after all the entries on the CQ are processed,
 * indicating that all incoming and outgoing messages have
 * completed.  If there are more than viadev_credit_notify_threshold
 * local credits we send a NoOp.
 */
int viadev_credit_notify_threshold = 5;

/* The second place is while we are processing receive descriptors.
 * If the local credit value gets reasonably large, we are not
 * sending messages to the remote side.  The remote end of this
 * connection may be stalled waiting for us to process all the
 * recieves so lets send a NoOp immediately so that they can
 * continue work.
 * This viadev_dynamic_credit_threshold should be a fraction
 * of the maximum receive prepost depth, but substantially
 * greater than the viadev_credit_notify_threshold.
 * It is re-computed below.
 */
int viadev_dynamic_credit_threshold = 10;

/* The RDMA collective register the entire user buffers to provide
 * optimal performance. However, for very large system sizes,
 * it might be difficult to register the entire buffer due to
 * physical memory limitations.
 *
 * Always overridable by use of env. var VIADEV_PT2PT_FAILOVER
 * which is set in MB. By default, it is 256 MB, assuming nodes
 * have atleast 512 MB.
 */

unsigned long viadev_pt2pt_failover = 268435456;

/* Environment variables to control Alltoall
 * algorithm being used. By default the values
 * used are the same as default MPICH
 */

int mpir_alltoall_short_msg = 256;
int mpir_alltoall_medium_msg = 32768;

#ifdef BLOCKING_SUPPORT
unsigned long long viadev_max_spin_count = VIADEV_MAX_SPIN_COUNT;
#endif

#ifdef _SMP_
#ifdef _AFFINITY_
/* Affinity is enabled by default */
unsigned int viadev_enable_affinity=1;
#endif 
#endif

/* 
 * Initialization parameters.
 */
unsigned int      viadev_ndreg_entries = 1000;
int               viadev_default_mtu = VIADEV_DEFAULT_MTU;
unsigned long     viadev_max_rdma_size = VIADEV_MAX_RDMA_SIZE;
uint8_t           viadev_default_qp_ous_rd_atom = VIADEV_DEFAULT_QP_OUS_RD_ATOM;
uint32_t          viadev_default_psn = VIADEV_DEFAULT_PSN;
uint16_t          viadev_default_pkey_ix = VIADEV_DEFAULT_PKEY_IX;
uint8_t           viadev_default_min_rnr_timer = VIADEV_DEFAULT_MIN_RNR_TIMER;
uint8_t           viadev_default_service_level = VIADEV_DEFAULT_SERVICE_LEVEL;
uint8_t           viadev_default_time_out = VIADEV_DEFAULT_TIME_OUT;
uint8_t           viadev_default_static_rate = VIADEV_DEFAULT_STATIC_RATE;
uint8_t           viadev_default_src_path_bits = VIADEV_DEFAULT_SRC_PATH_BITS;
int               viadev_max_fast_eager_size = VIADEV_MAX_FAST_EAGER_SIZE;
uint8_t           viadev_default_retry_count = VIADEV_DEFAULT_RETRY_COUNT;
uint8_t           viadev_default_rnr_retry = VIADEV_DEFAULT_RNR_RETRY;
uint32_t          viadev_default_max_sg_list = VIADEV_DEFAULT_MAX_SG_LIST;
uint8_t           viadev_default_max_rdma_dst_ops = VIADEV_DEFAULT_MAX_RDMA_DST_OPS;

int               viadev_default_port = -1;
int               viadev_max_ports = 2;
uint32_t          viadev_max_inline_size = 128;


void viadev_init_parameters(int num_proc, int me)
{
    char *value;

    /* ===============================================================
     * Examine the environment variables to see in any of these
     * were re-defined on the command line.
     * ===============================================================
     */

    /* Get the appropriate IB device */
    strncpy(viadev.device_name, VIADEV_INVALID_DEVICE, 32);

    if ((value = getenv("VIADEV_DEVICE")) != NULL) {
        strncpy(viadev.device_name, value, 32);
    }    

    if ((value = getenv("VIADEV_DEFAULT_PORT")) != NULL) {
        viadev_default_port = atoi(value);
        if(viadev_default_port < 0) {
            error_abort_all(GEN_ASSERT_ERR,
                    "User supplied port (%d) invalid\n",
                    viadev_default_port);
        }
    }

    if ((value = getenv("VIADEV_MAX_PORTS")) != NULL) {
        viadev_max_ports = atoi(value);
        if(viadev_max_ports < 0) {
            error_abort_all(GEN_ASSERT_ERR,
                    "User supplied max. ports (%d) invalid\n",
                    viadev_max_ports);
        }
    }

    if ((value = getenv("VIADEV_MAX_INLINE_SIZE")) != NULL) {
        viadev_max_inline_size = atoi(value);
        if(viadev_max_inline_size < 0) {
            error_abort_all(GEN_ASSERT_ERR,
                    "User supplied max. inline size (%d) invalid\n",
                    viadev_max_inline_size);
        }
    }

    if ((value = getenv("VIADEV_CLUSTER_SIZE")) != NULL) {
        if (strcmp(value,"SMALL")==0)
            viadev_cluster_size = 0;
        else if (strcmp(value,"MEDIUM")==0)
            viadev_cluster_size = 1;
        else if (strcmp(value,"LARGE")==0)
            viadev_cluster_size = 2;
        else if (strcmp(value,"AUTO")==0)
            viadev_cluster_size = -1;
        else
            viadev_cluster_size = 0;
        /* Now that I know the cluster size, I can set the default values
         * for some of the parameters that depand on the cluter size.
         * The command line parameters MUST be parsed below this point
         * to make sure we enable the user to overide this parameters!
         */
        VBUF_SIZE_CALC(num_proc);
        RENDEZVOUS_THRESHOLD_CALC(num_proc);
    }

    if ((value = getenv("VIADEV_DREG_CACHE_LIMIT")) != NULL) {
        viadev_dreg_cache_limit = atol(value);
    }

#ifdef BLOCKING_SUPPORT
    if ((value = getenv("VIADEV_MAX_SPIN_COUNT")) != NULL) {
        viadev_max_spin_count = atoll(value);
    }
#endif
    if ((value = getenv("VIADEV_MAX_RDMA_SIZE")) != NULL) {
        viadev_max_rdma_size = (unsigned long)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MTU")) != NULL) {
        if (strcmp(value,"MTU256")==0)
            viadev_default_mtu = IBV_MTU_256;
        else if (strcmp(value,"MTU512")==0)
            viadev_default_mtu = IBV_MTU_512;
        else if (strcmp(value,"MTU1024")==0)
            viadev_default_mtu = IBV_MTU_1024;
        else if (strcmp(value,"MTU2048")==0)
            viadev_default_mtu = IBV_MTU_2048;
        else if (strcmp(value,"MTU4096")==0)
            viadev_default_mtu = IBV_MTU_4096;
        else
            viadev_default_mtu = IBV_MTU_1024;
    }

    if ((value = getenv("VIADEV_DEFAULT_QP_OUS_RD_ATOM")) != NULL) {
        viadev_default_qp_ous_rd_atom = (uint8_t)atoi(value);
    }

    if ((value = getenv("VIADEV_DEFAULT_PSN")) != NULL) {
        viadev_default_psn = (uint32_t)atol(value);
    }

    if ((value = getenv("VIADEV_DEFAULT_PKEY_IX")) != NULL) {
        viadev_default_pkey_ix = (uint16_t)atol(value);
    }

    if ((value = getenv("VIADEV_DEFAULT_MIN_RNR_TIMER")) != NULL) {
        viadev_default_min_rnr_timer = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_SERVICE_LEVEL")) != NULL) {
        viadev_default_service_level = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_TIME_OUT")) != NULL) {
        viadev_default_time_out = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_STATIC_RATE")) != NULL) {
        viadev_default_static_rate = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_SRC_PATH_BITS")) != NULL) {
        viadev_default_src_path_bits = (uint8_t)atoi(value);
    }
    if ((value = getenv("VIADEV_MAX_FAST_EAGER_SIZE")) != NULL) {
        viadev_max_fast_eager_size = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_RETRY_COUNT")) != NULL) {
        viadev_default_retry_count = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_RNR_RETRY")) != NULL) {
        viadev_default_rnr_retry = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_SG_LIST")) != NULL) {
        viadev_default_max_sg_list = (uint32_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_RDMA_DST_OPS")) != NULL) {
        viadev_default_max_rdma_dst_ops = (uint8_t)atol(value);
    }
    if ((value = getenv("VIADEV_NDREG_ENTRIES")) != NULL) {
        viadev_ndreg_entries = (unsigned int)atoi(value);
    }
    if ((value = getenv("VIADEV_SQ_SIZE")) != NULL) {
        viadev_sq_size = atol(value);
    }
    if ((value = getenv("VIADEV_CQ_SIZE")) != NULL) {
        viadev_cq_size = atol(value);
    }
    if ((value = getenv("VIADEV_NUM_RDMA_BUFFER")) != NULL) {
        viadev_num_rdma_buffer = atoi(value);
        if (viadev_num_rdma_buffer < 0)
            viadev_num_rdma_buffer = 0;
        if (viadev_num_rdma_buffer == 1)
            viadev_num_rdma_buffer = 2; /* need 1 extra */
    }

#ifdef ADAPTIVE_RDMA_FAST_PATH
    if ((value = getenv("VIADEV_ADAPTIVE_RDMA_THRESHOLD")) != NULL) {
        viadev_rdma_eager_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_ADAPTIVE_RDMA_LIMIT")) != NULL) {
        viadev_rdma_eager_limit = atoi(value);
        if (viadev_rdma_eager_limit == -1)
            viadev_rdma_eager_limit = log_2(viadev.np);
    } else {
        viadev_rdma_eager_limit = viadev.np;
    }
#endif

#ifdef SRQ

    if ((value = getenv("VIADEV_SRQ_SIZE")) != NULL) {
        viadev_srq_size = (uint32_t) atoi(value);
    }

    if ((value = getenv("VIADEV_SRQ_LIMIT")) != NULL) {
        viadev_srq_limit = (uint32_t) atoi(value);

        if(viadev_srq_limit > viadev_srq_size) {
            error_abort_all(GEN_ASSERT_ERR,
                    "SRQ limit shouldn't be greater than SRQ size\n");
        }
    }


    if ((value = getenv("VIADEV_MAX_R3_OUST_SEND")) != NULL) {
        viadev_max_r3_oust_send = (uint32_t) atoi(value);
    }

#endif
    
    if ((value = getenv("VIADEV_PREPOST_DEPTH")) != NULL) {
        viadev_prepost_depth = atoi(value);
        if(viadev_prepost_depth <= 0) {
            viadev_prepost_depth = 1;
        }
    }
    if ((value = getenv("VIADEV_PREPOST_THRESHOLD")) != NULL) {
        viadev_prepost_threshold = atoi(value);
        if(viadev_prepost_threshold <= 0) {
            viadev_prepost_threshold = 1;
        }
    }
    if ((value = getenv("VIADEV_VBUF_POOL_SIZE")) != NULL) {
        viadev_vbuf_pool_size = atoi(value);
        if(viadev_vbuf_pool_size <= 0) {
            viadev_vbuf_pool_size = 1;
        }
    }
    if ((value = getenv("VIADEV_VBUF_TOTAL_SIZE")) != NULL) {
        viadev_vbuf_total_size= atoi(value);
        if(viadev_vbuf_total_size <= 0) {
            viadev_vbuf_total_size = 1;
        }
    }

    if ((value = getenv("VIADEV_VBUF_SECONDARY_POOL_SIZE")) != NULL) {
        viadev_vbuf_secondary_pool_size = atoi(value);
        if ( viadev_vbuf_secondary_pool_size <= 0 ) {
            viadev_vbuf_secondary_pool_size = 1;
        }
    }
    if ((value = getenv("VIADEV_VBUF_EXTRA")) != NULL) {
        viadev_vbuf_extra = atoi(value);
        if( viadev_vbuf_extra <= 0) {
            viadev_vbuf_extra = 1;
        }
    }
    if ((value = getenv("VIADEV_VBUF_MAX")) != NULL) {
        viadev_vbuf_max = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_RENDEZVOUS_EXTRA")) != NULL) {
        viadev_prepost_rendezvous_extra = atoi(value);
        if(viadev_prepost_rendezvous_extra <= 0) {
            viadev_prepost_rendezvous_extra = 1;
        }
    }
    if ((value = getenv("VIADEV_RENDEZVOUS_THRESHOLD")) != NULL) {
        viadev_rendezvous_threshold = atoi(value);
        if(viadev_rendezvous_threshold <= 0) {
            viadev_rendezvous_threshold = 1;
        }
    }
    if ((value = getenv("VIADEV_SPIN_COUNT")) != NULL) {
        viadev_spin_count = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_NOOP_EXTRA")) != NULL) {
        viadev_prepost_noop_extra = atoi(value);
        if(viadev_prepost_noop_extra <= 0) {
            viadev_prepost_noop_extra = 1;
        }
    }
    if ((value = getenv("VIADEV_INITIAL_PREPOST_DEPTH")) != NULL) {
        viadev_initial_prepost_depth = atoi(value);
        if(viadev_initial_prepost_depth <= 0) {
            viadev_initial_prepost_depth = 1;
        }
    }
    if (viadev_initial_prepost_depth >
        viadev_prepost_depth + viadev_prepost_noop_extra)
        viadev_initial_prepost_depth =
            viadev_prepost_depth + viadev_prepost_noop_extra;
    if (viadev_initial_prepost_depth <= 0)
        viadev_initial_prepost_depth = 1;
    if (viadev_initial_prepost_depth <= viadev_prepost_noop_extra) {
        viadev_initial_credits = viadev_initial_prepost_depth;
    } else {
        viadev_initial_credits =
            viadev_initial_prepost_depth - viadev_prepost_noop_extra;
    }
    if ((value = getenv("VIADEV_CREDIT_PRESERVE")) != NULL) {
        viadev_credit_preserve = atoi(value);
        if(viadev_credit_preserve <= 0) {
            viadev_credit_preserve = 1;
        }
    }

#ifdef SRQ
    /* In case of SRQ, initial credits should always be
     * greater than credit preserve, since sender doesn't
     * need to bother about window size */
    viadev_initial_credits = viadev_credit_preserve + 100;
#endif
    
    if ((value = getenv("VIADEV_CREDIT_NOTIFY_THRESHOLD")) != NULL) {
        viadev_credit_notify_threshold = atoi(value);
        if(viadev_credit_notify_threshold <= 0) {
            viadev_credit_notify_threshold = 1;
        }
    }
    if ((value = getenv("VIADEV_DYNAMIC_CREDIT_THRESHOLD")) != NULL) {
        viadev_dynamic_credit_threshold = atoi(value);
        if(viadev_dynamic_credit_threshold <= 0) {
            viadev_dynamic_credit_threshold = 1;
        }
    }
    if ((viadev_prepost_depth - viadev_prepost_threshold) <=
        MAX(viadev_credit_notify_threshold,
            viadev_dynamic_credit_threshold)) {

        if(viadev_prepost_depth < 1 + MAX(viadev_credit_notify_threshold,
                    viadev_dynamic_credit_threshold)) {
            error_abort_all(GEN_EXIT_ERR,
                "Error: VIADEV_PREPOST_DEPTH should atleast exceed %d"
                " based on your current parameter set\n", 1 + 
                MAX(viadev_credit_notify_threshold, viadev_dynamic_credit_threshold));
        } else {
            viadev_prepost_threshold =
                viadev_prepost_depth - 1 - MAX(viadev_credit_notify_threshold,
                        viadev_dynamic_credit_threshold);
        }
    }
    if (viadev_prepost_threshold < 0)
        viadev_prepost_threshold = 0;

#ifdef _SMP_
    /* Should set in KBytes */
    if ( (value = getenv("SMP_EAGERSIZE")) != NULL ) {
        smp_eagersize = atoi(value);
    }
    /* Should set in MBytes */
    if ( (value = getenv("SMPI_LENGTH_QUEUE")) != NULL ) {
        smpi_length_queue = atoi(value);
    }
#ifdef _AFFINITY_
    if ((value = getenv("VIADEV_ENABLE_AFFINITY")) != NULL) {
	    viadev_enable_affinity = atoi(value);
    }
#endif
#endif

    /* compute Receive Q size */
    viadev_rq_size =
        viadev_prepost_depth + viadev_prepost_rendezvous_extra +
        viadev_prepost_noop_extra;

     if ((value = getenv("VIADEV_PT2PT_FAILOVER")) != NULL) {
         viadev_pt2pt_failover = atol(value) * (1024 * 1024);
     }

     if ((value = getenv("MPIR_ALLTOALL_SHORT_MSG")) != NULL) {
         mpir_alltoall_short_msg = atoi(value);
     }

     if ((value = getenv("MPIR_ALLTOALL_MEDIUM_MSG")) != NULL) {
         mpir_alltoall_medium_msg = atoi(value);
     }
}

void dump_param_values(FILE * fd, int me)
{
    char *int_fmt;
    char *str_fmt;

    int_fmt = (char *) "%3d: %32s = %d\n";
    str_fmt = (char *) "%3d: %32s = %s\n";

    /*
       char *hex_fmt = "%3d: %32s = %#x\n";
     */

    fprintf(fd, int_fmt, me, "VIADEV_MAX_RDMA_SIZE", (int)viadev_max_rdma_size);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MTU", (int)viadev_default_mtu);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_QP_OUS_RD_ATOM", (int)viadev_default_qp_ous_rd_atom);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_PSN", (int)viadev_default_psn);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_PKEY_IX", (int)viadev_default_pkey_ix);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MIN_RNR_TIMER", (int)viadev_default_min_rnr_timer);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_SERVICE_LEVEL", (int)viadev_default_service_level);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_TIME_OUT", (int)viadev_default_time_out);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_STATIC_RATE", (int)viadev_default_static_rate);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_SRC_PATH_BITS", (int)viadev_default_src_path_bits);
    fprintf(fd, int_fmt, me, "VIADEV_MAX_FAST_EAGER_SIZE", (int)viadev_max_fast_eager_size);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_RETRY_COUNT", (int)viadev_default_retry_count);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_RNR_RETRY", (int)viadev_default_rnr_retry);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MAX_SG_LIST", (int)viadev_default_max_sg_list);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MAX_RDMA_DST_OPS", (int)viadev_default_max_rdma_dst_ops);
    fprintf(fd, int_fmt, me, "VIADEV_NDREG_ENTRIES", viadev_ndreg_entries);
    fprintf(fd, int_fmt, me, "VIADEV_SQ_SIZE", (int) viadev_sq_size);
    fprintf(fd, int_fmt, me, "VIADEV_CQ_SIZE", (int) viadev_cq_size);
    fprintf(fd, int_fmt, me, "VIADEV_NUM_RDMA_BUFFER",
            viadev_num_rdma_buffer);
    fprintf(fd, int_fmt, me, "VIADEV_PREPOST_DEPTH", viadev_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_THRESHOLD", viadev_prepost_threshold);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_POOL_SIZE", viadev_vbuf_pool_size);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_TOTAL_SIZE", viadev_vbuf_total_size);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_SECONDARY_POOL_SIZE",
            viadev_vbuf_secondary_pool_size);
    fprintf(fd, int_fmt, me, "VIADEV_VBUF_MAX", viadev_vbuf_max);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_RENDEZVOUS_EXTRA",
            viadev_prepost_rendezvous_extra);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_NOOP_EXTRA", viadev_prepost_noop_extra);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_PRESERVE",
            viadev_credit_preserve);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_NOTIFY_THRESHOLD",
            viadev_credit_notify_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_DYNAMIC_CREDIT_THRESHOLD",
            viadev_dynamic_credit_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_RENDEZVOUS_THRESHOLD",
            viadev_rendezvous_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_SPIN_COUNT", viadev_spin_count);
}


/* ================================================================
 * Adjust global parameters to conform with NIC attributes
 * ================================================================ */

/* This section only applicatble to VIA */
