
/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "vbuf.h"

/* head of list of allocated vbuf regions */
static vbuf_region *vbuf_region_head = NULL;

/* free_vbuf_head is the head of the free list */
static vbuf *free_vbuf_head = NULL;

static int vbuf_n_allocated = 0;
static long num_free_vbuf = 0;
static long num_vbuf_get = 0;
static long num_vbuf_free = 0;

static void allocate_vbuf_region(int nvbufs)
{
    int i, j;
    struct vbuf_region *reg;
    void *mem;
    int alignment = VBUF_TOTAL_SIZE;
    vbuf *cur;

    if(free_vbuf_head != NULL) {
        error_abort_all(GEN_ASSERT_ERR, "free_vbuf_head = NULL");
    }

    if (MAX_VBUFS > 0) {
        nvbufs = MIN(nvbufs, MAX_VBUFS - vbuf_n_allocated);
        if (nvbufs <= 0) {
            error_abort_all(GEN_EXIT_ERR,
                    "VBUF alloc failure, limit exceeded");
        }
    }

    reg = (struct vbuf_region *) malloc(sizeof(struct vbuf_region));
    if (NULL == reg) {
        error_abort_all(GEN_EXIT_ERR,
                "Unable to malloc a new struct vbuf_region");
    }

    mem = (void *) malloc(nvbufs * sizeof(vbuf) + (alignment - 1));
    if (NULL == mem) {
        error_abort_all(GEN_EXIT_ERR, "unable to malloc vbufs");
    }

    memset(mem, 0, nvbufs * sizeof(vbuf) + (alignment - 1));

    vbuf_n_allocated += nvbufs;
    num_free_vbuf += nvbufs;
    reg->malloc_start = mem;
    reg->malloc_end = (void *) ((char *) mem + nvbufs * sizeof(vbuf) +
            alignment - 1);
    reg->count = nvbufs;
    free_vbuf_head = (vbuf *) (((aint_t) mem + (aint_t) (alignment - 1)) &
            ~((aint_t) alignment - 1));

    /* Register the vbuf region */
    for(j = 0; j < num_hcas; j++)
        reg->mem_handle[j] = register_memory(free_vbuf_head,
                nvbufs * sizeof(vbuf), j);

    /* init the free list */
    for (i = 0; i < nvbufs - 1; i++) {
        cur = free_vbuf_head + i;
        cur->desc.next = free_vbuf_head + i + 1;
        cur->region = reg;
    }
    /* last one needs to be set to NULL */
    cur = free_vbuf_head + nvbufs - 1;
    cur->desc.next = NULL;
    cur->region = reg;

    /* thread region list */
    reg->next = vbuf_region_head;
    vbuf_region_head = reg;

}

void allocate_vbufs(void)
{
    allocate_vbuf_region(VBUF_POOL_SIZE);
}

void deallocate_vbufs(void)
{
    int j;
    vbuf_region *r = vbuf_region_head;
    while (r) {
        for (j = 0; j < num_hcas; j++) {
            if (NULL != r->mem_handle[j]) {
                if(deregister_memory(r->mem_handle[j])) {
                    error_abort_all(IBV_RETURN_ERR, 
                            "could not deregister vbuf region\n");
                }
            }
        }
        r = r->next;
    }
}

/* Get a vbuf off a free list */

vbuf* get_vbuf(void)
{
    vbuf *v;

    if (NULL == free_vbuf_head) {
        allocate_vbuf_region(VBUF_SECONDARY_POOL_SIZE);
        if (NULL == free_vbuf_head) {
            error_abort_all(GEN_EXIT_ERR, "No free vbufs\n");
        }
    }

    v = free_vbuf_head;
    num_free_vbuf--;
    num_vbuf_get++;

    /* this correctly handles removing from single entry free list */

    free_vbuf_head = free_vbuf_head->desc.next;
#if defined(RDMA_FAST_PATH)
    v->padding = NORMAL_VBUF_FLAG;
#endif

    v->shandle = NULL;

    return (v);
}

void vbuf_init_recv(vbuf * v, 
				unsigned long len, int subchannel)
{
    v->desc.u.rr.next = NULL;
    v->desc.u.rr.wr_id = (aint_t) v;
    v->desc.u.rr.num_sge = 1;
    v->desc.u.rr.sg_list = &(v->desc.sg_entry);
	assert(subchannel >= 0);

    v->desc.sg_entry.addr = (uintptr_t) v->buffer;
    v->desc.sg_entry.length = len;
    v->desc.sg_entry.lkey = v->region->mem_handle[subchannel/num_ports]->lkey;
	v->subchannel = subchannel;
	v->desc.subchannel = subchannel;
#ifdef RDMA_FAST_PATH
    v->padding = NORMAL_VBUF_FLAG;
#endif
}

void vbuf_init_send(vbuf * v, unsigned long len, int subchannel)
{
	assert(subchannel >= 0);
	v->desc.u.sr.send_flags = IBV_SEND_SIGNALED;
    v->desc.u.sr.opcode = IBV_WR_SEND;
    v->desc.u.sr.wr_id = (aint_t) v;
    v->desc.u.sr.num_sge = 1;
    v->desc.u.sr.sg_list = &(v->desc.sg_entry);
	v->desc.subchannel = subchannel;
    v->desc.sg_entry.addr = (uintptr_t) v->buffer;
    v->desc.sg_entry.length = len;
    v->desc.sg_entry.lkey = v->region->mem_handle[subchannel/num_ports]->lkey;
#ifdef RDMA_FAST_PATH
    v->padding = NORMAL_VBUF_FLAG;
#endif
	v->subchannel = subchannel;
}

void vbuf_init_rput(vbuf * v, void *local_address,
                    uint32_t lkey, void *remote_address,
                    uint32_t rkey, int len)
{
    v->desc.u.sr.send_flags = IBV_SEND_SIGNALED;
    v->desc.u.sr.opcode = IBV_WR_RDMA_WRITE;
    v->desc.u.sr.wr_id = (aint_t) v;

    v->desc.u.sr.num_sge = 1;
    v->desc.u.sr.sg_list = &(v->desc.sg_entry);

    v->desc.sg_entry.length = len;
    v->desc.sg_entry.lkey = lkey;
    v->desc.sg_entry.addr = (uintptr_t) local_address;
    v->desc.u.sr.wr.rdma.remote_addr = (uintptr_t) remote_address;
    v->desc.u.sr.wr.rdma.rkey = rkey;

#ifdef RDMA_FAST_PATH
    v->padding = RPUT_VBUF_FLAG;
#endif

    D_PRINT("RDMA write\n");
}

/*
 * Put a vbuf back on the free list 
 */

void release_vbuf(vbuf * v)
{
    /* note this correctly handles appending to empty free list */

    assert(v != free_vbuf_head);

    v->desc.next = free_vbuf_head;

#if defined(RDMA_FAST_PATH)
    if ((v->padding != NORMAL_VBUF_FLAG)
        && (v->padding != RPUT_VBUF_FLAG)) {
        error_abort_all(GEN_EXIT_ERR, "vbuf %p not correct!!!\n", v);
    }
#endif

    free_vbuf_head = v;
    num_free_vbuf++;
    num_vbuf_free++;
}
