/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "mpid_bind.h"
#include "ibverbs_header.h"
#include "viutil.h"
#include "viapacket.h"
#include "queue.h"
#include "viaparam.h"
#include "mpid_smpi.h"

#define NIC_0 0
#define SUBCHANNEL_0 0

/* 
 * The flowlist is a list of connections that have received new flow credits
 * and may be able to make progress on pending rendezvous sends.  The
 * flowlist is a singly-linked list of connections, linked by the "nextflow"
 * field in the connection. In order to prevent loops, there is also a
 * connection field "inflow" which is set to one if it is in the flowlist
 * and zero otherwise. It must be initialized to 0 and it only changed by
 * the flow control logic in MPID_CheckDevice
 * 
 * We add a connection to the flowlist when either 1) it gets new credits
 * and there is a rendezvous in progress 2) a rendezvous ack is received
 * 
 * Note: flowlist fields are touched only routines in this file. 
 */

static ibv_connection_t *flowlist;
#define PUSH_FLOWLIST(c) {                                          \
    if (0 == c->inflow) {                                           \
        c->inflow = 1;                                              \
        c->nextflow = flowlist;                                     \
        flowlist = c;                                               \
    }                                                               \
}

#define POP_FLOWLIST() {                                            \
    if (flowlist != NULL) {                                         \
        ibv_connection_t *_c;                                       \
        _c = flowlist;                                              \
        flowlist = _c->nextflow;                                    \
        _c->inflow = 0;                                             \
        _c->nextflow = NULL;                                        \
    }                                                               \
}

/*
 * Attached to each connection is a list of send handles that
 * represent rendezvous sends that have been started and acked but not
 * finished. When the ack is received, the send is placed on the list;
 * when the send is complete, it is removed from the list.  The list
 * is an "in progress sends" queue for this connection.  We need it to
 * remember what sends are pending and to remember the order sends
 * were acked so that we complete them in that order. This
 * prevents a situation where we receive an ack for message 1, block
 * because of flow control, receive an ack for message 2, and are able
 * to complete message 2 based on the new piggybacked credits.
 * 
 * The list head and tail are given by the shandle_head and
 * shandle_tail entries on viadev_connection_t and the list is linked
 * through the nexthandle entry on a send handle.  
 *
 * The queue is FIFO because we must preserve order, so we maintain
 * both a head and a tail. 
 *
 */

#define RENDEZVOUS_IN_PROGRESS(c, s) {                              \
    if (NULL == c->shandle_tail) {                                  \
        c->shandle_head = s;                                        \
    } else {                                                        \
        c->shandle_tail->nexthandle = s;                            \
    }                                                               \
    c->shandle_tail = s;                                            \
    s->nexthandle = NULL;                                           \
}

#define RENDEZVOUS_DONE(c) {                                        \
    c->shandle_head = c->shandle_head->nexthandle;                  \
        if (NULL == c->shandle_head) {                              \
            c->shandle_tail = NULL;                                 \
        }                                                           \
}


static void viadev_rendezvous_push(MPIR_SHANDLE * s)
{
    vbuf *v;
    viadev_packet_r3_data *h;
    int subchannel;
    char *databuf;
    int datamax = VBUF_DATA_SIZE(viadev_packet_r3_data);
    static int count = 0;
    ibv_connection_t *c = (ibv_connection_t *) s->connection;
    count++;

    char *start_addr, *remote;
    int i;

    if (s->protocol == VIADEV_PROTOCOL_RPUT) {
        int nbytes;
        char *stripe_addr[MAX_SUBCHANNELS];
        int stripe_size[MAX_SUBCHANNELS];

        if (s->bytes_sent != 0) {
            error_abort_all(GEN_ASSERT_ERR,
                                "s->bytes_sent != 0 Rendezvous Push");
        }

        s->total_stripe_finished = 0;
        for (i = 0; i < num_subchannels; i++)
            s->rput_flag[i] = 0;

        if (s->bytes_total > 0) {
            assert(s->dreg_entry != NULL);
            assert(s->remote_address != NULL);
        }
        if (lm_scheduling == ADAPTIVE_STRIPING) {
            for (i = 0; i < num_subchannels; i++) {
                s->initial_weight[i] = c->s_weight[i];
                s->stripe_finish_time[i] = 0;
            }
        }
        while (s->bytes_sent < s->bytes_total) {
            nbytes = s->bytes_total - s->bytes_sent;
            if (nbytes > ibv_dev.maxtransfersize) {
                nbytes = ibv_dev.maxtransfersize;
            }

        /* schedule this rdma transfer */
            start_addr = (char *)(s->local_address)+s->bytes_sent;
            remote = (char *)(s->remote_address)+s->bytes_sent;
            schedule_rdma(c, start_addr, nbytes, 
                        stripe_addr, stripe_size, s->do_stripe);

            for (i = 0; i < num_subchannels; i++) {
                if (stripe_size[i] == 0)
                    continue;
                v = get_vbuf();
                v->shandle = s;
                assert(v != NULL);
                s->rput_flag[i] = 1;
                v->subchannel = i;
                v->desc.subchannel = i;

                viadev_rput(c, v,
                            stripe_addr[i],
                            ((dreg_entry *) s->dreg_entry)->memhandle[i/num_ports]->lkey,
                            remote,
                            s->remote_memhandle[i/num_ports].rkey, stripe_size[i]);
                remote += stripe_size[i];
            }
            s->bytes_sent += nbytes;
        }
        assert(s->bytes_sent == s->bytes_total);

        viadev_rput_finish(s);
        s->nearly_complete = 1;

        return;
    }

    /* if we get here, then protocol is R3 */

    subchannel = choose_lm_subchannel(c);
   	assert(subchannel >= 0);
   
    while (c->remote_credit[subchannel] >= CREDIT_PRESERVE &&
           s->nearly_complete == 0) {
        v = get_vbuf();

        /* figure out where header and data are */
        h = (viadev_packet_r3_data *) VBUF_BUFFER_START(v);
        databuf = ((char *) h) + sizeof(viadev_packet_r3_data);

        /* set up the packet */
        PACKET_SET_HEADER(h, c, VIADEV_PACKET_R3_DATA);
        h->rreq = s->receive_id;

        /* figure out how much data to send in this packet */
        h->bytes_in_this_packet = s->bytes_total - s->bytes_sent;
        if (datamax < h->bytes_in_this_packet) {
            h->bytes_in_this_packet = datamax;
        }

        /* copy next chunk of data in to vbuf */
        memcpy(databuf, ((char *) s->local_address) +
               s->bytes_sent, h->bytes_in_this_packet);

        s->bytes_sent += h->bytes_in_this_packet;

        /* finally send it off */
        vbuf_init_send(v, sizeof(viadev_packet_r3_data) +
                       h->bytes_in_this_packet, subchannel);
        /* If this is the last packet, mark the vbuf so that the MPI
         * send is marked complete when the VIA send completes. 
         * Do not remove from the shandle list on the connection. Whoever
         * called us will do that. 
         */
        if (s->bytes_sent == s->bytes_total) {
            /* mark MPI send complete when VIA send completes */
            v->shandle = s;
            s->nearly_complete = 1;
        }
        viadev_post_send(c, v, subchannel);

    }
    if (s->nearly_complete) {
        D_PRINT("VI %3d R3 PUSH COMPLETE, total %d",
                c->global_rank, s->bytes_total);
    } else {
        D_PRINT("VI %3d R3 PUSH STALLED, total %d, sent %d",
                c->global_rank, s->bytes_total, s->bytes_sent);
    }
}


static void process_flowlist(void)
{
    MPIR_SHANDLE *s;
    while (flowlist) {

        /* Push on the the first ongoing receive with 
         * viadev_rendezvous_push. If the receive
         * finishes, it will advance the shandle_head
         * pointer on the connection. 
         * 
         * xxx the side effect of viadev_rendezvous_push is
         * bad practice. Find a way to do this so the logic
         * is obvious. 
         */

        s = flowlist->shandle_head;
        while (s != NULL) {
            viadev_rendezvous_push(s);
            if (!s->nearly_complete) {
                break;
            }
            RENDEZVOUS_DONE(flowlist);
            s = flowlist->shandle_head;
        }
        /* now move on to the next connection */
        POP_FLOWLIST();
    }
}

#ifdef RDMA_FAST_PATH
static int poll_rdma_buffer(vbuf **vbuf_addr, int *out_of_order)
{
    int i, j;
    vbuf *v;
    viadev_packet_header *h;
    VBUF_FLAG_TYPE size;
    volatile VBUF_FLAG_TYPE *tail;
    volatile VBUF_FLAG_TYPE *head;
    *vbuf_addr = NULL;
    *out_of_order = 0;

    if (0 == viadev_num_rdma_buffer)
        return 1;
    /* currently poll all connections */
    for (i = 0; i < ibv_dev.np; i++) {
        ibv_connection_t *c = &ibv_dev.connections[i];

        if (i == ibv_dev.me)
            continue;
        if(c->remote_address_received < num_subchannels) {
            continue;
        }
        for(j = 0; j < num_subchannels ; j++) {
            /* get the current polling buffer */
            v = &(c->RDMA_recv_buf[j][c->p_RDMA_recv[j]]);
            head = &(v->head_flag);
            /* check the flag */
            if (*head) {
                size = (*head & FAST_RDMA_SIZE_MASK);
                tail = (VBUF_FLAG_TYPE *) ((aint_t) (v->buffer) + size);

            /* check tail */

            /* If tail has not come go ahead and poll other
             * connections. Busy waiting for a message being
             * DMA'd to user memory causes unnecessary loss of
             * CPU cycles */
                if (*tail != *head)
                    continue;
                /* out of order message? */
                readbar();
                h = (viadev_packet_header *) (v->buffer);
                if (h->type != VIADEV_PACKET_NOOP
                    && h->id != c->next_packet_expected) {
                    continue;
                }

                *vbuf_addr = v;
                *out_of_order = 0;

                /* clear head flag */
                *head = 0;

                /* advance receive pointer */
                if (++(c->p_RDMA_recv[j]) >= viadev_num_rdma_buffer)
                    c->p_RDMA_recv[j] = 0;
                return 0;

            } else
                continue;
        }
    }
        return 1;
}
#endif

static void viadev_process_send(void* vbuf_addr)
{
    vbuf *v = NULL;
    viadev_packet_header *header;
    ibv_connection_t *c = NULL;


    MPIR_SHANDLE *s;

    /* process a completed descriptor on the send queue.  
     * dispatch to an appropriate routine based on the type 
     * in the header field of the vbuf. xxx Later (with RDMA) we will 
     * have to handle packets that have immediate data 
     * but no packet header in the vbuf.  */

    v = (vbuf *) (vbuf_addr);

	assert(vbuf_addr != NULL);
    /* since we have a completion, one more WQE is available
     * first process any existing WQEs on extended send q
     * if we implement send coallesing for RDMA fast Path (MAX_NO_COMPLETION)
     * we will need to handle >1 additions to send_wqes_avail
     */
	v->subchannel = v->desc.subchannel;
    c = &(ibv_dev.connections[v->grank]);
    c->send_wqes_avail[v->subchannel]++;
    if (c->ext_sendq_head[v->subchannel]) {
        viadev_ext_sendq_send(c, v->subchannel);
    }

#if defined(RDMA_FAST_PATH)
    if (v->padding == RPUT_VBUF_FLAG) {

        if (ADAPTIVE_STRIPING == lm_scheduling) {

            s = (MPIR_SHANDLE *)(v->shandle);
            s->stripe_finish_time[v->subchannel] = 
                            MPI_Wtime();
        }
        
        release_vbuf(v);
        return;
    }
#else
    if (v->desc.u.sr.opcode == IBV_WR_RDMA_WRITE) {

        if (ADAPTIVE_STRIPING == lm_scheduling) {

            s = (MPIR_SHANDLE *)(v->shandle);
            s->stripe_finish_time[v->subchannel] =
                            MPI_Wtime();
        }

		release_vbuf(v);
        return;
    }
#endif


    header = (viadev_packet_header *) VBUF_BUFFER_START(v);
	D_PRINT("got send completion, subchannel[%d], id[%d]\n", v->subchannel, header->id);
			
    switch (header->type) {
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    case FAST_EAGER_CACHED:
#endif
#endif
    case VIADEV_PACKET_EAGER_START:
    case VIADEV_PACKET_EAGER_NEXT:
    case VIADEV_PACKET_R3_DATA:
        s = (MPIR_SHANDLE *) (v->shandle);
        if (s != NULL)
            SEND_COMPLETE(s);
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;

    case VIADEV_PACKET_RPUT_FINISH:
        s = (MPIR_SHANDLE *) (v->shandle);
        if (s == NULL) {
            error_abort_all(GEN_ASSERT_ERR,
                                "s == NULL, s is the send "
                                "handler of the rput finish");
        }
        
        s->total_stripe_finished += ((viadev_packet_rput_finish * )header )->striping_weight;
        if (s->total_stripe_finished == num_subchannels){
             if ((lm_scheduling == ADAPTIVE_STRIPING) && 
                     (((viadev_packet_rput_finish*)header)->striping_weight <  num_subchannels)){
                 adjust_weights(c, s->stripe_start_time, s->stripe_finish_time, s->initial_weight);
             }
            SEND_COMPLETE(s);
        }

#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;

#if defined(RDMA_FAST_PATH)
    case VIADEV_RDMA_ADDRESS:
#endif
    case VIADEV_PACKET_RENDEZVOUS_START:
    case VIADEV_PACKET_RENDEZVOUS_REPLY:
    case VIADEV_PACKET_NOOP:
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_BARRIER:
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;
    default:
        error_abort_all(IBV_STATUS_ERR,
                            "Unknown packet type %d in "
                            "viadev_process_send", header->type);
    }
}

static void viadev_incoming_eager_start(vbuf * v, 
        ibv_connection_t * c,
        viadev_packet_eager_start * header)
{
    MPIR_RHANDLE *rhandle;
    int found;


    MPID_Msg_arrived(header->envelope.src_lrank,
                     header->envelope.tag,
                     header->envelope.context, &rhandle, &found);

#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
#define MAX_FAST_EAGER_SIZE 255

    /* caching header */
    if ((header->header.type == VIADEV_PACKET_EAGER_START) &&
        (header->envelope.data_length < 256) &&
        (v->padding != NORMAL_VBUF_FLAG)) {
        memcpy(&(c->cached_incoming[v->desc.subchannel]), header,
               sizeof(viadev_packet_eager_start));
    }

#endif
#endif

    /* This information needs to be filled in whether a preposted 
     * receive was found or not. 
     */

    rhandle->connection = c;
    rhandle->s.count = header->envelope.data_length;

    rhandle->protocol = VIADEV_PROTOCOL_EAGER;
    rhandle->vbufs_received = 1;
    rhandle->subchannel = v->desc.subchannel;
    if (found) {

        int truncated = 0;
        int copy_bytes;
#if !defined(DISABLE_HEADER_CACHING) && defined (RDMA_FAST_PATH)
        char *vbuf_data;
#else
        char *vbuf_data = ((char *) header) +
            sizeof(viadev_packet_eager_start);
#endif
        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            truncated = 1;
        } else {
            rhandle->len = header->envelope.data_length;
        }

#ifdef  RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
        if (header->header.type == FAST_EAGER_CACHED) {
            vbuf_data =
                ((char *) (VBUF_BUFFER_START(v))) + FAST_EAGER_HEADER_SIZE;
        header->bytes_in_this_packet = header->header.fast_eager_size;
        copy_bytes = rhandle->len;
    } else {
            vbuf_data = ((char *) header) +
                sizeof(viadev_packet_eager_start);
            copy_bytes = MIN(rhandle->len, 
                    header->bytes_in_this_packet);
        }
#else
        copy_bytes = MIN(rhandle->len, 
                header->bytes_in_this_packet);
#endif

#else
        copy_bytes = MIN(rhandle->len, 
                header->bytes_in_this_packet);
#endif

        /* matching packet is here, can no longer cancel recv */
        rhandle->can_cancel = 0;

        rhandle->vbufs_expected =
            viadev_calculate_vbufs_expected(rhandle->len,
                                            rhandle->protocol);

         if(copy_bytes > 0) {
             memcpy((char *) (rhandle->buf), vbuf_data, copy_bytes);
         }

        rhandle->bytes_copied_to_user = copy_bytes;
        if (rhandle->vbufs_expected == 1) {
            RECV_COMPLETE(rhandle);

            if (truncated) {
                rhandle->s.MPI_ERROR = MPI_ERR_TRUNCATE;
            }

        } else {
            c->rhandle = rhandle;
            rhandle->s.MPI_ERROR =
                truncated ? MPI_ERR_TRUNCATE : MPI_SUCCESS;
        }
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG)
            release_vbuf(v);
        else
            release_recv_rdma(c, v);
#else
        release_vbuf(v);
#endif
    } else {
        /* Recv not yet posted, set size of incoming data in
         * unexpected queue rhandle
         */
        rhandle->len = header->envelope.data_length;
        rhandle->vbuf_head = v;
        rhandle->vbuf_tail = v;
        rhandle->bytes_copied_to_user = 0;
        v->desc.next = NULL;
        c->rhandle = rhandle;
    }
}

static void viadev_incoming_eager_next(vbuf * v, ibv_connection_t * c,
                                viadev_packet_eager_next * header)
{
    MPIR_RHANDLE *rhandle = c->rhandle;
    int copy_bytes = 0;

    if (rhandle == NULL) {
        error_abort_all(GEN_ASSERT_ERR, "rhandle is NULL");
    }

    rhandle->vbufs_received++;

    if (rhandle->bytes_copied_to_user > 0) {
        /* this is a user receive handle, and has already been matched */
        char *vbuf_data = ((char *) header) +
            sizeof(viadev_packet_eager_next);

        copy_bytes = MIN(header->bytes_in_this_packet,
                (rhandle->len - rhandle->bytes_copied_to_user));

    if (copy_bytes) {
            memcpy((char *) (rhandle->buf) +
                   rhandle->bytes_copied_to_user,
                   vbuf_data, copy_bytes);
            rhandle->bytes_copied_to_user += copy_bytes;
        }

        if (rhandle->vbufs_expected == rhandle->vbufs_received) {
            int err = rhandle->s.MPI_ERROR;
            RECV_COMPLETE(rhandle);
            rhandle->s.MPI_ERROR = err;
        }
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG)
            release_vbuf(v);
        else
            release_recv_rdma(c, v);
#else
         release_vbuf(v);
#endif
    } else {
        /* haven't posted the message yet. accumulate vbufs on a chain */
        ((vbuf *) rhandle->vbuf_tail)->desc.next = v;
        v->desc.next = NULL;
        rhandle->vbuf_tail = v;
    }
}

static void viadev_incoming_rendezvous_start(vbuf * v, 
        ibv_connection_t * c,
        viadev_packet_rendezvous_start *header)
{

    MPIR_RHANDLE *rhandle;
    int found;


    MPID_Msg_arrived(header->envelope.src_lrank,
                     header->envelope.tag,
                     header->envelope.context, &rhandle, &found);

    /* This information needs to be filled in whether a preposted 
     * receive was found or not.  */
    rhandle->connection = c;
    rhandle->send_id = header->sreq;
    rhandle->s.count = header->envelope.data_length;
    rhandle->vbufs_received = 0;
    rhandle->vbuf_head = NULL;
    rhandle->vbuf_tail = NULL;
    rhandle->replied = 0;
    rhandle->protocol = header->protocol;
    rhandle->subchannel = v->desc.subchannel;
    rhandle->total_stripe_received = 0;

	assert(rhandle->subchannel >= 0);
    if (!found) {
        /* recv not yet posted, set size of incoming data in
         * unexpected queue rhandle */
        rhandle->len = header->envelope.data_length;
    } else {
        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            error_abort_all(IBV_STATUS_ERR,
                                "message truncated. ask %d got %d",
                                rhandle->len,
                                header->envelope.data_length);
        }

        rhandle->len = header->envelope.data_length;

        /* packet now arrived, can no-longer cancel recv */
        rhandle->can_cancel = 0;

        /* can we match sender's protocol? */
        switch (header->protocol) {
        case VIADEV_PROTOCOL_R3:
            viadev_recv_r3(rhandle);
            break;
        case VIADEV_PROTOCOL_RGET:
            viadev_recv_rget(rhandle);
            break;
        case VIADEV_PROTOCOL_RPUT:
            viadev_recv_rput(rhandle);
            break;
        case VIADEV_PROTOCOL_RENDEZVOUS_UNSPECIFIED:
            error_abort_all(IBV_STATUS_ERR,
                                "VIRS: unspecified protocol %d\n",
                                rhandle->protocol);
            break;
        default:
            error_abort_all(IBV_STATUS_ERR,
                                "VIRS: invalid protocol %d\n",
                                rhandle->protocol);
        }
    }
}


/* 
 * We previously initiated a rendezvous send. 
 * Receiver has posted a matching receive and is 
 * telling us about it. 
 * The fact that they are telling us about it means
 * it is not an rget. It is either an r3 or an rput
 * depending on whether we are sending data directly 
 * into the user buffer on the other side
 * For the first implementation, only r3 is supported, 
 * copied through vbufs on both sides. 
 * 
 * Since messages can be very large and vbuf resources
 * are limited, we have to handle the case where there
 * are not enough preposted vbufs on the other side
 * to handle our data (this is true even for rput, 
 * which still consumes vbufs). 
 * 
 * For a non-pathological receiver (philosophical
 * debates about the MPI progress rule being of little
 * relevance to the real world -- and for the record we 
 * state that in the initial single-threaded imlementation 
 * we do not obey the strongest version of the rule)
 * we will receive credit renewals that allow us
 * to proceed. However, to get those credits, we have
 * to process some incoming packets. To handle this, we
 * put send handles for started-but-not-completed
 * rendezvous transfers on a queue of send
 * handles attached to the connection. MPID_DeviceCheck
 * makes progress on these through "viadev_rendezvous_push_transfer"
 * when it has run out of packets to process. 
 * 
 * All we do in this routine is set up the send handle and 
 * defer actually pushing data until all packets are processed. 
 *
 */

void viadev_incoming_rendezvous_reply(vbuf * v, 
        ibv_connection_t * c, 
        viadev_packet_rendezvous_reply * header)
{
    int i;
    MPIR_SHANDLE *s;
    /* first make sure we have a match */
    s = (MPIR_SHANDLE *) (ID_TO_REQ(header->sreq));
    if (NULL == s) {
        error_abort_all(GEN_ASSERT_ERR,
                            "s == NULL, send handler at "
                            "viadev_incoming_rendezvous_reply");
    }

    /* fill in some fields */
    s->bytes_sent = 0;
    s->receive_id = header->rreq;

    /* we are the sender and initially set a protocol.  The
     * receiver may not be able to support our protocol and
     * may have reverted to R3.  If so, we have to revert as
     * well, and unregister any memory */

    switch (s->protocol) {
    case VIADEV_PROTOCOL_RPUT:
        if (header->protocol == VIADEV_PROTOCOL_R3) {
            if (s->dreg_entry != NULL) {
                dreg_unregister(s->dreg_entry);
                s->dreg_entry = NULL;
            }
            s->remote_address = NULL;
            s->protocol = VIADEV_PROTOCOL_R3;
        } else {
            assert(header->protocol == VIADEV_PROTOCOL_RPUT);

            s->remote_address = header->buffer_address;
            for (i = 0; i < num_hcas; i++)
                s->remote_memhandle[i] = header->memhandle[i];

        }
        break;
    case VIADEV_PROTOCOL_RGET:
        if (header->protocol == VIADEV_PROTOCOL_R3) {
            s->protocol = VIADEV_PROTOCOL_R3;
        } else {
            /* should never have gotten here if both sides
             * agreed on RGET */
            error_abort_all(IBV_STATUS_ERR,
                                "VIRR: RGET protocol failure");
        }
        break;

    case VIADEV_PROTOCOL_R3:
        assert(header->protocol == VIADEV_PROTOCOL_R3);
        break;

    default:
        error_abort_all(IBV_STATUS_ERR,
                            "VIRR: invalid sender protocol [%d]",
                            s->protocol);
    }

    /* 
     * even if we have an rput, which we know we could finish, 
     * put it on the list so that it is processed in order. 
     * Need to satisfy MPI ordering requirements. 
     */

    RENDEZVOUS_IN_PROGRESS(c, s);

    /*
     * this is where all rendezvous transfers are started,
     * so it is the only place we need to set this kludgy
     * field
     */

    s->nearly_complete = 0;

    PUSH_FLOWLIST(c);

}

static void viadev_incoming_r3_data(vbuf *v, ibv_connection_t *c,
        viadev_packet_r3_data *h)
{
    MPIR_RHANDLE *rhandle;
    void *databuf;
    int bytes_to_copy;


    /* find the rhandle for this data */
    rhandle = (MPIR_RHANDLE *) ID_TO_REQ(h->rreq);

    databuf = ((char *) h) + sizeof(viadev_packet_r3_data);
    bytes_to_copy = h->bytes_in_this_packet;

    rhandle->vbufs_received++;

    memcpy(((char *) rhandle->buf) +
           rhandle->bytes_copied_to_user, databuf, bytes_to_copy);

    rhandle->bytes_copied_to_user += bytes_to_copy;

    c->rendezvous_packets_expected--;

    /* sanity check */
    assert(rhandle->bytes_copied_to_user <= rhandle->len);

    if (rhandle->bytes_copied_to_user == rhandle->len) {
        assert(rhandle->vbufs_received == rhandle->vbufs_expected);
        RECV_COMPLETE(rhandle);
        D_PRINT("VI %3d R3 RECV COMPLETE total %d",
                c->global_rank, rhandle->len);
    }
}

#ifdef RDMA_FAST_PATH
static void process_remote_rdma_address(ibv_connection_t *c,
        uint32_t remote_hndl, struct vbuf* remote_address, int subchannel)
{
    int i;

	assert(c != NULL);
    c->remote_address_received++;
    c->remote_RDMA_buf_hndl[subchannel] = remote_hndl;
    c->remote_RDMA_buf[subchannel] = remote_address;

	assert(subchannel >=  0);
    for (i = 0; i < viadev_num_rdma_buffer; i++) {

        /* set global_rank */
        c->RDMA_send_buf[subchannel][i].grank = c->global_rank;
        c->RDMA_recv_buf[subchannel][i].grank = c->global_rank;

        c->RDMA_recv_buf[subchannel][i].desc.next = NULL;
		c->RDMA_recv_buf[subchannel][i].subchannel = subchannel;
		c->RDMA_send_buf[subchannel][i].subchannel = subchannel;
		c->RDMA_send_buf[subchannel][i].desc.subchannel = subchannel;
		c->RDMA_recv_buf[subchannel][i].desc.subchannel = subchannel;	
        /* fill out the descriptors */
        c->RDMA_send_buf[subchannel][i].padding = FREE_FLAG;

        /* Mark initialization state of RDMA buffers to be busy */
        c->RDMA_recv_buf[subchannel][i].padding = BUSY_FLAG;

        /* this field to be filled out later */
        c->RDMA_send_buf[subchannel][i].desc.sg_entry.length = 0;

        c->RDMA_send_buf[subchannel][i].desc.sg_entry.addr = (uintptr_t)
            (&((c->RDMA_send_buf[subchannel] + i)->head_flag));

        c->RDMA_send_buf[subchannel][i].desc.sg_entry.lkey =
            c->RDMA_send_buf_hndl[subchannel]->lkey;

        c->RDMA_send_buf[subchannel][i].desc.u.sr.sg_list =
            &(c->RDMA_send_buf[subchannel][i].desc.sg_entry);

        c->RDMA_send_buf[subchannel][i].desc.u.sr.send_flags = 
            IBV_SEND_SIGNALED;
        c->RDMA_send_buf[subchannel][i].desc.u.sr.next = NULL;

        c->RDMA_send_buf[subchannel][i].desc.u.sr.opcode = IBV_WR_RDMA_WRITE;
        c->RDMA_send_buf[subchannel][i].desc.u.sr.wr_id = (aint_t) & (c->RDMA_send_buf[subchannel][i]);
        c->RDMA_send_buf[subchannel][i].desc.u.sr.wr.rdma.rkey = c->remote_RDMA_buf_hndl[subchannel];
        c->RDMA_send_buf[subchannel][i].desc.u.sr.wr.rdma.remote_addr = (uintptr_t)
            (&(c->remote_RDMA_buf[subchannel][i]));
        c->RDMA_send_buf[subchannel][i].desc.u.sr.num_sge = 1;

    }
}

static void viadev_incoming_rdma_address(vbuf *v, ibv_connection_t *c,
        viadev_packet_rdma_address *header)
{
    int subchannel;
    subchannel = header->subchannel;
    /* address also received */
    if (c->remote_address_received >= num_subchannels) {
        error_abort_all(GEN_EXIT_ERR,
                "global rank for this connection: %d",
                c->global_rank);
    }
    process_remote_rdma_address(c, header->RDMA_hndl,
            header->RDMA_address, subchannel);
}

#endif

static void viadev_process_recv(void* vbuf_addr)
{
    vbuf *v;
    viadev_packet_header *header;
    ibv_connection_t *c;
    int needed;
    int barrier_id;
    int subchannel;

    /* *All* packets come IN through this routine and go OUT
     * through viadev_post_send. This is asymmetrical, so 
     * it doesn't feel right. But perhaps it's ok?
     */

    /* Process a completed descriptor on the recv queue.  dispatch to
     * an appropriate routine based on the type in the header field of
     * the vbuf. Flow control credits are always accumulated before
     * the packet is processed, and we check the sequence number.
     * When done, we release the vbuf and repost if necessary. 
     */

    v = (vbuf *) vbuf_addr;
    c = &(ibv_dev.connections[v->grank]);

    subchannel = v->desc.subchannel;
    header = (viadev_packet_header *) VBUF_BUFFER_START(v);

	assert(subchannel >= 0);
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    if (header->type == FAST_EAGER_CACHED) {
        c->cached_incoming[subchannel].header.id = header->id;
        c->cached_incoming[subchannel].header.type = FAST_EAGER_CACHED;
        c->cached_incoming[subchannel].envelope.data_length = header->fast_eager_size;
        c->cached_incoming[subchannel].header.fast_eager_size = 
            header->fast_eager_size;
        header = &(c->cached_incoming[subchannel].header);
    }
#endif
#endif
#if defined(RDMA_FAST_PATH)
    /* handle out of order messages for send/receive channel */
    /* RDMA channel don't return out of order message !!! */
    if (v->padding == NORMAL_VBUF_FLAG) {
#if 0
        while(header->type != VIADEV_PACKET_NOOP &&
               header->id != c->next_packet_expected) {
            /* switch to the rdma channel */
#ifdef _SMP_
            if(SMP_INIT) {
                smpi_net_lookup();
            }
#endif
            r1 = poll_rdma_buffer((void *) &v1, &ofo);
            if (0 == r1) {
                viadev_process_recv(v1);
            }
        }
#endif
    }
#endif
    if (header->type != VIADEV_PACKET_NOOP) {
        if (header->id != c->next_packet_expected) {
            error_abort_all(GEN_ASSERT_ERR,
                                "header->id != c->next_packet_expected");
        }
        c->next_packet_expected++;
    }

    /* just consumed a vbuf from a VI receive queue, decrement prepost
     * count and check if we should re-post.  Want to do this here
     * since we may be sending a packet in response to this receive
     * and want it to include the credit update asap.
     */
#if defined(RDMA_FAST_PATH)
    if (v->padding == NORMAL_VBUF_FLAG)
        c->preposts[subchannel]--;
#else
    c->preposts[subchannel]--;
#endif
    needed = PREPOST_DEPTH + PREPOST_NOOP_EXTRA
        + MIN(PREPOST_RENDEZVOUS_EXTRA,
              c->rendezvous_packets_expected);
    if (c->initialized[subchannel]) {
#if defined(RDMA_FAST_PATH)
        if (header->type == VIADEV_PACKET_NOOP
            && v->padding == NORMAL_VBUF_FLAG) {
#else
        if (header->type == VIADEV_PACKET_NOOP) {
#endif
            PREPOST_VBUF_RECV(c, subchannel);
            /* noops don't count for credits */
            c->local_credit[subchannel]--;
        } else if (c->preposts[subchannel] < DEFAULT_MAX_RECV_WQE &&
                   c->preposts[subchannel] + PREPOST_THRESHOLD < needed) {
            do {
                PREPOST_VBUF_RECV(c, subchannel);
            } while (c->preposts[subchannel] < DEFAULT_MAX_RECV_WQE 
                    && c->preposts[subchannel] < needed);
        }
    } else {
        int i;
        for (i = 0;
             i <
             PREPOST_NOOP_EXTRA - (INITIAL_PREPOST_DEPTH -
                                          INITIAL_CREDITS); ++i) {
            PREPOST_VBUF_RECV(c, subchannel);
            c->local_credit[subchannel]--;
        }

        /* fill out buffers to make up for small initial preposts */
        while (c->preposts[subchannel] < DEFAULT_MAX_RECV_WQE && c->preposts[subchannel] < needed) {
            PREPOST_VBUF_RECV(c, subchannel);
        }
        c->initialized[subchannel] = 1;
    }                           /* c->initialized */

    if (header->type == VIADEV_PACKET_R3_DATA) {
        D_PRINT("[%d]header->type=%d is R3 DATA\n", ibv_dev.me,
                header->type);
        /* got an R3 Data packet, do we need to pre-post additional
         * receives?  This next call will ensure that we keep the
         * the number of additional pre-posts for R3 data sends
         * up to the maximum allowed.
         */
        if (c->rendezvous_packets_expected > 0) {
            viadev_prepost_for_rendezvous(c, 0, subchannel);
        }
    }

    /* update our knowledge of remote sides credit count */
    c->remote_cc[subchannel] = header->remote_credit;

    /* accumulate vbuf credits and remember */
    if (header->vbuf_credit > 0) {
        c->remote_credit[subchannel] += header->vbuf_credit;
        if (c->backlog[subchannel].len > 0) {
            /* We have pending messages on the backlog queue.
             * Send them out immediately.
             */
            viadev_backlog_send(c, subchannel);
        }
        /* if any credits remain, schedule rendezvous progress */
        if ((c->remote_credit[subchannel] > 0) && (c->shandle_head != NULL)) {
            PUSH_FLOWLIST(c);
        }
    }
#if defined(RDMA_FAST_PATH)
    /* process rdma credit */
    if (header->rdma_credit > 0) {
        c->ptail_RDMA_send[subchannel] += header->rdma_credit;
        if (c->ptail_RDMA_send[subchannel] >= viadev_num_rdma_buffer)
            c->ptail_RDMA_send[subchannel] -= viadev_num_rdma_buffer;
    }
#endif
    switch (header->type) {
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    case FAST_EAGER_CACHED:
#endif
#endif
    case VIADEV_PACKET_EAGER_START:
        viadev_incoming_eager_start(v, c,
                                    (viadev_packet_eager_start *) header);
        /* dont release vbuf may have been unexpected. 
         * eager packet processing
         * will release vbufs when possible.  */
        break;
    case VIADEV_PACKET_EAGER_NEXT:
        viadev_incoming_eager_next(v, c,
                                   (viadev_packet_eager_next *) header);
        /* dont release vbuf may have been unexpected. 
         * eager packet processing
         * will release vbufs when possible. 
         */
        break;
    case VIADEV_PACKET_RENDEZVOUS_START:
		V_PRINT("Received rendezvous_start\n");
        viadev_incoming_rendezvous_start(v, c,
                                         (viadev_packet_rendezvous_start *)
                                         header);
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_RENDEZVOUS_REPLY:
		V_PRINT("Received rendezvous_reply\n");

        viadev_incoming_rendezvous_reply(v, c,
                                         (viadev_packet_rendezvous_reply *)
                                         header);
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_R3_DATA:
        viadev_incoming_r3_data(v, c, (viadev_packet_r3_data *) header);
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;

#if defined(RDMA_FAST_PATH)
    case VIADEV_RDMA_ADDRESS:

        viadev_incoming_rdma_address(v, c, (viadev_packet_rdma_address *)
                                     header);

        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
        break;
#endif

    case VIADEV_PACKET_RPUT_FINISH:
		V_PRINT("Received rendezvous_finish\n");

        viadev_incoming_rput_finish(v, c,
                                    (viadev_packet_rput_finish *) header);
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_NOOP:
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_BARRIER:
        barrier_id = ((viadev_packet_barrier *) header)->barrier_id;
        D_PRINT("[%d] got barrier [%d] from [%d]",
                ibv_dev.me, barrier_id, c->global_rank);
        c->barrier_id = barrier_id;
#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    default:
        error_abort_all(IBV_STATUS_ERR,
                            "Unknown packet type %d in "
                            "viadev_process_recv", header->type);
    }

    viadev_send_noop_ifneeded(c, subchannel);
}



static void spin_and_wait(struct ibv_wc *sc)
{
    int ret = 1, ne = -1;
    int i;
#ifdef RDMA_FAST_PATH
    vbuf *v;
    int ofo;
#endif

    sc->wr_id = 0;

    while(1) {

#ifdef _SMP_
        if (SMP_INIT) {
            if (MPID_SMP_Check_incoming() == MPI_SUCCESS) {
                sc->wr_id = 0;
                ret = 0;
                break;
            }
        }
#endif

#if defined(RDMA_FAST_PATH)
        ret = poll_rdma_buffer(&v, &ofo);
        if (ret == 0) {
            sc->wr_id = (aint_t) v;
            sc->opcode = IBV_WC_RECV;
            break;
        }
#endif
        for(i = 0; i < num_hcas; i++) {
            ne = ibv_poll_cq(ibv_dev.cq_hndl[i], 1, sc);
            if (ne > 0)
                break;
            } 
            if (ne > 0) {
				readbar();	
                if (sc->status != IBV_WC_SUCCESS) {
                    error_abort_all(IBV_RETURN_ERR, 
                        "[%s:%d] Got completion with error code %d\n",
                        ibv_dev.my_name, ibv_dev.me, sc->status);
                }

            ret = 0;

            break;
        }

        if (ret == 1)
            continue;
        /* error when comes here */

        error_abort_all(GEN_ASSERT_ERR,
                "[%s:%d] Error in (%s:%d): ret=%d\n",
                ibv_dev.my_name, ibv_dev.me, __FILE__, __LINE__, ret);
        break;
    }
}

int MPID_DeviceCheck(MPID_BLOCKING_TYPE blocking)
{
    int ret1, ret2;
    vbuf *vbuf_addr;
    int gotone = 0, ne = -1;
    struct ibv_wc wc;

    int i;
#ifdef RDMA_FAST_PATH
    int ofo;
#endif

checkstart:
    flowlist = NULL;

    /* process all waiting packets */
    ret1 = ret2 = 0;

    /* Check the fastest channel */
#ifdef _SMP_
    if (SMP_INIT) {
        if (MPI_SUCCESS == MPID_SMP_Check_incoming())
            gotone = 1;
    }
#endif /* _SMP_ */

    while(ret1 == 0) {

#ifdef RDMA_FAST_PATH
        ret1 = poll_rdma_buffer(&vbuf_addr, &ofo);
        if (ret1 == 0) {
            gotone = 1;
            viadev_process_recv(vbuf_addr);
        }
#endif
        for (i = 0; i < num_hcas; i++){
            ne = ibv_poll_cq(ibv_dev.cq_hndl[i], 1, &wc);
            if (ne > 0)
                break;
        }
        if(ne < 0) {
            error_abort_all(IBV_RETURN_ERR, "Error polling CQ\n");
        } else if(ne > 1) {
            error_abort_all(IBV_RETURN_ERR, "Asked only one completion, "
                    "got more!\n");
        } else if (ne == 1) {
            /* Valid completion */
            gotone = 1;
            vbuf_addr = (vbuf *)(uintptr_t *) wc.wr_id;
            ret1 = 0;

            if(IBV_WC_SUCCESS != wc.status) {
                error_abort_all(IBV_RETURN_ERR, 
                        " [%s:%d] Error code %d in polled desc!\n",
                        ibv_dev.my_name, ibv_dev.me, wc.status);
            }

            if( (IBV_WC_SEND == wc.opcode) ||
                    (IBV_WC_RDMA_WRITE == wc.opcode) ||
                    (IBV_WC_RDMA_READ == wc.opcode)) {
                viadev_process_send(vbuf_addr);
            } else {
                viadev_packet_header *h;
         
                ibv_connection_t *c;
                h = (viadev_packet_header *)
                        (VBUF_BUFFER_START(vbuf_addr));
				assert(vbuf_addr->grank >= 0);
                c = &(ibv_dev.connections[vbuf_addr->grank]);
                if (h->id == c->next_packet_expected  || 
                        h->type == VIADEV_PACKET_NOOP) {
                   viadev_process_recv(vbuf_addr);
                } else {
                    ret1 = 1;
                    enqueue_ofo_queue(vbuf_addr, c);
                }
            }
        } else {
            ret1 = 1;
        }
		do{
        	ret1 = check_ofo_queues(&vbuf_addr);
	        if (ret1 == 0) {
    	        viadev_process_recv(vbuf_addr);
        	}
		}while(0 == ret1);

    }

    if(flowlist) {
        process_flowlist();
        /* In case some new credits have come in while processing
         * go check again. Little overhead to do this.
         */
        goto checkstart;
    }

    if (gotone || blocking == MPID_NOTBLOCKING) {
        return gotone ? MPI_SUCCESS : -1;       /* nonblock must return -1 */
    }

    spin_and_wait(&wc);

    if(wc.wr_id) {

        vbuf_addr = (void *) ((aint_t) wc.wr_id);

        if( (IBV_WC_SEND == wc.opcode) ||
                (IBV_WC_RDMA_WRITE == wc.opcode) ||
                (IBV_WC_RDMA_READ == wc.opcode)) {
            viadev_process_send(vbuf_addr);
        } else {
            viadev_packet_header *h;
            ibv_connection_t *c;
            h = (viadev_packet_header *)VBUF_BUFFER_START(vbuf_addr);
            c = &(ibv_dev.connections[vbuf_addr->grank]);
            if (h->id == c->next_packet_expected  
                || h->type == VIADEV_PACKET_NOOP) {
                viadev_process_recv(vbuf_addr);
            } else {
                ret1 = 1;
                enqueue_ofo_queue(vbuf_addr, c);
            }
        }
    }
	
	do{
	    ret1 = check_ofo_queues(&vbuf_addr);
    	if (ret1 == 0) {
        	viadev_process_recv(vbuf_addr);
    	}
	} while(0 == ret1);

    if(flowlist) {
        process_flowlist();
    }
    
    return MPI_SUCCESS;
}
