/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */


/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#ifdef _SMP_
#ifdef _AFFINITY_
#include <sched.h>
#endif /*_AFFINITY_*/
#endif

#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "viadev.h"
#include "viaparam.h"
#include "viapriv.h"
#include "process/pmgr_client.h"
#include "mpid_smpi.h"
#include "rdma_fast_path.h"

#define NIC_0 0
#define SUBCHANNEL_0 0

extern int viadev_default_mtu;
extern int viadev_default_psn;
extern int viadev_default_min_rnr_timer;
extern int viadev_default_service_level;
extern int viadev_default_src_path_bits;
extern int viadev_default_time_out;
extern int viadev_default_retry_count;
extern int viadev_default_rnr_retry;

#ifdef _SMP_
#ifdef _AFFINITY_
extern unsigned int viadev_enable_affinity;
#endif
#endif

static void set_malloc_options(void)
{
    /* Set glibc/stdlib malloc options to prevent handing
     * memory back to the system (brk) upon free.
     * Also, dont allow MMAP memory for large allocations.
     * This because we dont unregister memory when refcounts
     * go to zero.
     * Only need these options when RDMA operations are being
     * used.  Otherwise, MVAPICH layer owns all registered memory
     * and we will only free it if it is de-registered anyway.
     */

#ifdef LAZY_MEM_UNREGISTER
    mallopt(M_TRIM_THRESHOLD, -1);
    mallopt(M_MMAP_MAX, 0);
#endif /* LAZY_MEM_UNREGISTER */
}

static void open_hca(void)
{
    struct ibv_device **dev_list;
    int i = 0;

    dev_list = ibv_get_device_list(NULL);

    for(i = 0; i < num_hcas; i++) { 
        ibv_dev.nic[i] = dev_list[i];

        ibv_dev.context[i] = ibv_open_device(dev_list[i]);

        if(!ibv_dev.context[i]) {
            error_abort_all(GEN_EXIT_ERR, 
						"Error getting HCA context\n");
        }

        ibv_dev.ptag[i] = ibv_alloc_pd(ibv_dev.context[i]);

        if(!ibv_dev.ptag[i]) {
            error_abort_all(GEN_EXIT_ERR, 
						"Error getting protection domain\n");
        }
    }
}

static void get_lid(void)
{
    struct ibv_port_attr port_attr;
    int i, j, count = 0;

    uint16_t calc_lid[MAX_SUBCHANNELS];
    /* num_phy_ports refers to the number of physical ports
     * to be used */

    for(j = 0; j < num_hcas; j++) {
        count = 0;
        for(i = viadev_default_port - 1; i < num_phy_ports + 
                viadev_default_port - 1; i++){
            if(ibv_query_port(ibv_dev.context[j], i % MAX_PORTS + 1, &port_attr)) {
                error_abort_all(GEN_EXIT_ERR, "Error getting LID\n");
            }
            calc_lid[count] = port_attr.lid;
            count++;
	    	D_PRINT("subchannel[%d], LID %d\n", i - 1, port_attr.lid);
        }
        /* num ports refers to the number of subchannels */

	    for(i = 0; i < num_ports; i++)
		    ibv_dev.my_hca_lid[j * num_ports + i] = 
                calc_lid[i % num_phy_ports];	
    }
}

/* Create Completion Queue */
static void create_cq(void)
{
    int i;
    for(i = 0; i < num_hcas; i++){
#ifdef GEN2_OLD_CQ_VERB
        ibv_dev.cq_hndl[i] = ibv_create_cq(ibv_dev.context[i], 
            DEFAULT_MAX_CQ_SIZE, NULL);
#else
        ibv_dev.cq_hndl[i] = ibv_create_cq(ibv_dev.context[i],
                            DEFAULT_MAX_CQ_SIZE, NULL, NULL, 0);
#endif

        if(!ibv_dev.cq_hndl[i]) {
            error_abort_all(GEN_EXIT_ERR, "Error creating CQ\n");
        }
    }
}
static void create_qps(void)
{
    int i, j;

    struct ibv_qp_attr qp_attr;

    for(i = 0; i < ibv_dev.np; i++) {

        if (i == ibv_dev.me){
            continue;
        }
        for(j = 0; j < num_subchannels; j++) { 
            struct ibv_qp_init_attr attr;
            memset(&attr, 0, sizeof(attr));
            attr.send_cq = ibv_dev.cq_hndl[j / num_ports];
            attr.recv_cq = ibv_dev.cq_hndl[j / num_ports];
            attr.cap.max_send_wr  = DEFAULT_MAX_SEND_WQE;
            attr.cap.max_recv_wr  = DEFAULT_MAX_RECV_WQE;
            attr.cap.max_send_sge = DEFAULT_MAX_SEND_SGE;
            attr.cap.max_recv_sge = DEFAULT_MAX_RECV_SGE;
            attr.cap.max_inline_data = MAX_INLINE_SIZE;
            attr.qp_type = IBV_QPT_RC;

            ibv_dev.qp_hndl[i * num_subchannels + j] = 
                ibv_create_qp(ibv_dev.ptag[j/num_ports], &attr);
            if(!ibv_dev.qp_hndl[i * num_subchannels + j]) {
                error_abort_all(GEN_EXIT_ERR, "Error creating QP\n");
            }
        }
    }

    inline_size = MAX_INLINE_SIZE;
	D_PRINT("num subchannels = %d\n", num_subchannels);
    for(i = 0; i < ibv_dev.np; i++) {
        
        if (i == ibv_dev.me){
            continue;
        }
        for(j = 0; j < num_subchannels; j++) {
            qp_attr.qp_state = IBV_QPS_INIT;
            qp_attr.pkey_index = 0;
            qp_attr.port_num = (viadev_default_port + j % num_phy_ports > MAX_PORTS ? 1 : viadev_default_port + j % num_phy_ports) ;
            qp_attr.qp_access_flags = IBV_ACCESS_LOCAL_WRITE | 
									IBV_ACCESS_REMOTE_WRITE;

            if(ibv_modify_qp(ibv_dev.qp_hndl[i * num_subchannels + j], 
                        &qp_attr,
                        IBV_QP_STATE              |
                        IBV_QP_PKEY_INDEX         |
                        IBV_QP_PORT               |
                        IBV_QP_ACCESS_FLAGS)) {
                error_abort_all(GEN_EXIT_ERR, "Couldn't modify QP to init\n");
            }
        }
    }
}

static void ib_init(void)
{
    open_hca();
    get_lid();
    create_cq();
    create_qps();
    return;
}

#ifdef _SMP_
static int get_host_id(char *myhostname, int hostname_len)
{
    int host_id = 0;
    struct hostent *hostent;

    hostent = gethostbyname(myhostname);
    if (hostent == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                "Null value returned by gethostbyname");
    }
    host_id = ((struct in_addr *) hostent->h_addr_list[0])->s_addr;
    return host_id;
}
#endif /* _SMP_ */

static void ib_rank_lid_table(int *alladdrs, int local_addr_len)
{
    int i, j;

#ifdef _SMP_
    int *hostnames_j = NULL;
    int *smpi_ptr = NULL;

#ifdef _AFFINITY_
    long N_CPUs_online;
    unsigned long affinity_mask=1;
    unsigned long affinity_mask_len=sizeof(affinity_mask);

    /*Get number of CPU on machine*/
    if ( (N_CPUs_online=sysconf (_SC_NPROCESSORS_ONLN)) < 1 ) {
            perror("sysconf");
    }
#endif /*_AFFINITY_*/

    hostnames_j = (int *) malloc(ibv_dev.np * sizeof(int));
    smpi.local_nodes = (unsigned int *) malloc(ibv_dev.np * sizeof(int));
    smpi_ptr = (unsigned int *) malloc(ibv_dev.np * sizeof(int));

    if (hostnames_j == NULL || smpi.local_nodes == NULL || smpi_ptr == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                "malloc failure: in ib_rank_lid_table for SMP");
    }

#endif /* _SMP_ */

	for(i = 0; i < 2 * ibv_dev.np * num_subchannels; i++)
			D_PRINT(" alladdrs[i] = %d\n", alladdrs[i]);
    for (i = 0; i < ibv_dev.np; i++) {
        for(j = 0; j < num_subchannels; j++) {
            ibv_dev.lid_table[i * num_subchannels + j] = 
                alladdrs[i*num_subchannels + j];
            ibv_dev.qp_table[i * num_subchannels + j] = 
                alladdrs[i*num_subchannels + j  + ibv_dev.np * num_subchannels];
			D_PRINT("remote rank[%d], subchannel[%d], remote LID[%d], remote Qp[%d] \n", 
							i, j, ibv_dev.lid_table[i * num_subchannels + j],
						  ibv_dev.qp_table[i * num_subchannels + j]);	
#ifdef _SMP_
            hostnames_j[i] = alladdrs[2 * ibv_dev.np * num_subchannels + i];
#endif /* _SMP_ */
        }
    }

#ifdef _SMP_
    smpi.only_one_device = 1;
    smpi.num_local_nodes = 0;
    for (j = 0; j < ibv_dev.np; j++) {
        if (hostnames_j[ibv_dev.me] == hostnames_j[j]) {
             if (j == ibv_dev.me) {
                 smpi.my_local_id = smpi.num_local_nodes;
#ifdef _AFFINITY_
                if ( viadev_enable_affinity > 0 ){
                        affinity_mask=1<<(smpi.my_local_id % N_CPUs_online);
                        if ( sched_setaffinity(0,affinity_mask_len,&affinity_mask)<0 ) {
                                perror("sched_setaffinity");
                        }
                }
#endif /* _AFFINITY_ */
             }
             smpi.local_nodes[j] = smpi.num_local_nodes;
             smpi_ptr[smpi.num_local_nodes] = j;
             smpi.num_local_nodes++;
        } else {
            smpi.only_one_device = 0;
            smpi.local_nodes[j] = -1;
        }
    }
    smpi.l2g_rank = (unsigned int *)malloc(smpi.num_local_nodes * sizeof(int));
    if (smpi.l2g_rank == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                            "malloc: in ib_rank_lid_table for SMP");
    }
    for (j = 0; j < smpi.num_local_nodes; j++) {
        smpi.l2g_rank[j] = smpi_ptr[j];
    }
    free(smpi_ptr);
    free(hostnames_j);
#endif /* _SMP_ */
}

static void ib_qp_enable(void)
{
    int i, j;
    struct ibv_qp_attr attr;

    attr.qp_state               = IBV_QPS_RTR;
    attr.path_mtu               = DEFAULT_MTU; 
    attr.rq_psn                 = DEFAULT_PSN;
    attr.max_dest_rd_atomic     = DEFAULT_DEST_RD_ATOMIC;
    attr.min_rnr_timer          = DEFAULT_MIN_RNR_TIMER;
    attr.ah_attr.is_global      = 0;
    attr.ah_attr.sl             = DEFAULT_SERVICE_LEVEL;
    attr.ah_attr.src_path_bits  = DEFAULT_SRC_PATH_BITS;
    attr.ah_attr.static_rate    = 0;
    for(i = 0; i < ibv_dev.np; i++) {
        if(i == ibv_dev.me) {
            continue;
        }
        for(j = 0; j < num_subchannels; j++) {
            attr.dest_qp_num = ibv_dev.qp_table[i * num_subchannels + j];
            attr.ah_attr.dlid = ibv_dev.lid_table[i * num_subchannels + j];
            /* Using number of physical ports for the loop around in case
             * the number of queue pairs per port is greater 1 */
            attr.ah_attr.port_num = (viadev_default_port  - 1 + j) % num_phy_ports + 1;
			V_PRINT("remote rank[%d], subchannel[%d], remote qp[%d], remote LID[%d]\n", 
							i, j, attr.dest_qp_num, attr.ah_attr.dlid);
            if (ibv_modify_qp(ibv_dev.qp_hndl[i * num_subchannels + j], &attr,
                        IBV_QP_STATE              |
                        IBV_QP_AV                 |
                        IBV_QP_PATH_MTU           |
                        IBV_QP_DEST_QPN           |
                        IBV_QP_RQ_PSN             |
                        IBV_QP_MAX_DEST_RD_ATOMIC |
                        IBV_QP_MIN_RNR_TIMER)) {
                error_abort_all(GEN_EXIT_ERR, "Failed to modify QP to RTR\n");
            }
        }
    }

    attr.qp_state       = IBV_QPS_RTS;
    attr.timeout        = DEFAULT_TIMEOUT;
    attr.retry_cnt      = DEFAULT_RETRY_COUNT;    
    attr.rnr_retry      = DEFAULT_RNR_RETRY;    
    attr.sq_psn         = DEFAULT_PSN;
    attr.max_rd_atomic  = DEFAULT_DEST_RD_ATOMIC;

    for(i = 0; i < ibv_dev.np; i++) {
        if(i == ibv_dev.me) {
            continue;
        }
        for(j = 0; j < num_subchannels; j++) 
        if (ibv_modify_qp(ibv_dev.qp_hndl[i * num_subchannels + j], &attr,
                    IBV_QP_STATE              |
                    IBV_QP_TIMEOUT            |
                    IBV_QP_RETRY_CNT          |
                    IBV_QP_RNR_RETRY          |
                    IBV_QP_SQ_PSN             |
                    IBV_QP_MAX_QP_RD_ATOMIC)) {
            error_abort_all(GEN_EXIT_ERR, "Failed to modify QP to RTS\n");
        }
    }
}

static void ib_finalize(void)
{
    int i, j;

    /* unregister all user buffer registration */
    while (dreg_evict());

    /* Destroy all RC Qps */
    for(i = 0; i < ibv_dev.np; i++) {
        if(i == ibv_dev.me) {
            continue;
        }
        for(j = 0; j < num_subchannels; j++) {
            if(ibv_destroy_qp(ibv_dev.qp_hndl[i * num_subchannels + j])) {
                error_abort_all(GEN_EXIT_ERR,"Couldn't destroy QP\n");
            }
        }
    }

	D_PRINT("Executing Finalize\n");
    finalize_rdma_fast_path();

    deallocate_vbufs();

    for(i = 0; i < num_hcas; i++){
        if(ibv_destroy_cq(ibv_dev.cq_hndl[i])) {
            error_abort_all(GEN_EXIT_ERR,"Couldn't destroy CQ\n");
        }
    
        if(ibv_dealloc_pd(ibv_dev.ptag[i])) {
            error_abort_all(GEN_EXIT_ERR,"Couldn't deallocate PD\n");
        }

        if(ibv_close_device(ibv_dev.context[i])) {
            error_abort_all(GEN_EXIT_ERR,"Couldn't close HCA\n");
        }
    }
    free(ibv_dev.qp_hndl);
}

static void init_mpi_connections(void)
{
    int i, j, k;
    ibv_connection_t *c;

    /* Allocate all the connection structures */
    ibv_dev.connections = (ibv_connection_t *) 
        malloc(sizeof(ibv_connection_t) * ibv_dev.np);
    if (NULL == ibv_dev.connections) {
        error_abort_all(GEN_EXIT_ERR,
                "Unable to malloc connection structures");
    }

    allocate_vbufs();
    for(i = 0; i < ibv_dev.np; i++) {
        c = &ibv_dev.connections[i];
        c->global_rank = i;
        c->next_packet_expected = 1;
        c->next_packet_to_send = 1;
        c->shandle_head = NULL; 
        c->shandle_tail = NULL; 
        c->rhandle = NULL; 
        c->nextflow = NULL; 
        c->inflow = 0;
        c->rendezvous_packets_expected = 0;
        viadev_init_backlog_queue(c);
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
        /* init cached packet header */
        for(j = 0; j < num_subchannels; j++) {
            memset(&(c->cached_outgoing[j]), 0,
                sizeof(viadev_packet_eager_start));
            memset(&(c->cached_incoming[j]), 0,
                sizeof(viadev_packet_eager_start));
            c->cached_hit = c->cached_miss = 0;
            c->cached_incoming[j].header.type = c->cached_outgoing[j].header.type
                = FAST_EAGER_CACHED;
        }
#endif
#endif

        for(j = 0; j < num_subchannels; j++) { 
	        if (i != ibv_dev.me) {
    	        c->qp_hndl[j] = (struct ibv_qp*) 
							ibv_dev.qp_hndl[i * num_subchannels + j];

        	    /* prepost receives */
            	for (k = 0; k < INITIAL_PREPOST_DEPTH; k++) {
	                PREPOST_VBUF_RECV(c, j);
    	        }

	            c->preposts[j] = 0;
    	        c->send_wqes_avail[j] = DEFAULT_MAX_SEND_WQE;
        	    c->ext_sendq_head[j] = NULL;
            	c->ext_sendq_tail[j] = NULL;

	            c->remote_credit[j] = INITIAL_CREDITS;
    	        c->remote_cc[j] = INITIAL_CREDITS;
        	    c->local_credit[j] = 0;
            	c->preposts[j] = INITIAL_PREPOST_DEPTH;
	            c->initialized[j] = (INITIAL_PREPOST_DEPTH
                    ==
                    PREPOST_DEPTH +
                    PREPOST_NOOP_EXTRA);
            /* Initialize RDMA fast path */
                
            } else {
            	/* these should be ignored for
	             * self-communication */
    	        c->qp_hndl[j] = NULL;
        	    c->remote_credit[j] = 0;
            	c->remote_cc[j] = 0;
	            c->local_credit[j] = 0;
#ifdef RDMA_FAST_PATH
    	        c->phead_RDMA_send[j] = 0;
        	    c->ptail_RDMA_send[j] = 0;
            	c->p_RDMA_recv[j] = 0;
	            c->p_RDMA_recv_tail[j] = 0;
#endif /* RDMA_FAST_PATH */
    	    }
        }

		if (i != ibv_dev.me) {
        	init_rdma_fast_path(c);
		}
    }

    /* call barrier to make sure all preposting has been done */
    i = pmgr_barrier();

    /* Initialize the out of order queues for all processes */
	init_ofo_queues();

    /* Initialize the weights for each subchannel */
    for (i = 0; i < ibv_dev.np; i++) {
        ibv_connection_t *c = &ibv_dev.connections[i];
        for (j = 0; j < num_subchannels; j++) {
                c->s_weight[j] = 4;
        }
    }

    /* Initialize the weight for all subchannels */
    if (lm_scheduling == ADAPTIVE_STRIPING) {
        for (i = 0; i < ibv_dev.np; i++) {
            ibv_connection_t *c = &ibv_dev.connections[i];
            for (j = 0; j < num_subchannels - 1; j++) {
                c->s_weight[j] = DYNAMIC_TOTAL_WEIGHT / num_subchannels;
            }
        c->s_weight[num_subchannels - 1] = DYNAMIC_TOTAL_WEIGHT -
            DYNAMIC_TOTAL_WEIGHT / num_subchannels * (num_subchannels - 1);
        }
    }

    /* RDMA buffer address exchange */
    rdma_buffer_address_exchange();
}

/* Fist function called from the ADI Layer */
int MPID_VIA_Init(int *argc, char ***argv, int *size, int *rank)
{
    int *alladdrs = NULL;
    int *local_addr = NULL;
    int local_addr_len, mypid, i, j;

    set_malloc_options();

    /* Init the barrier counter */
    ibv_dev.barrier_id = 0;

    /* Find out from the process manager who I am  */
    if (!pmgr_client_init(argc, argv, &ibv_dev.np, &ibv_dev.me,
                &ibv_dev.global_id)) {
        error_abort_all(GEN_EXIT_ERR, "pmgr_client_init failed");
    }

    /* Report my size & rank to upper layer */
    *size = ibv_dev.np;
    *rank = ibv_dev.me;

    /* Make this dynamic soon!!! */
    strncpy (ibv_dev.execname, ((const char **) (*argv))[0], 256);

    ibv_dev.my_name = (char *) malloc(sizeof(char) * HOSTNAME_LEN);
    if (NULL == ibv_dev.my_name) {
        error_abort_all(GEN_EXIT_ERR, "malloc failed");
    }

    if (gethostname(ibv_dev.my_name, sizeof(char) * HOSTNAME_LEN) < 0) {
        error_abort_all(GEN_EXIT_ERR, "gethostname failed");
    }

    /* For both SMP and no-SMP, we have the same address length */
    /* Using MAX_SUBCHANNELS during the allocation, becuase the actual
     * number of subchannels will be decided at the run time */

    alladdrs = (int *) malloc(2 * MAX_SUBCHANNELS * ibv_dev.np * sizeof(int) + ibv_dev.np * sizeof(int));
    local_addr = (int *) malloc(2 * MAX_SUBCHANNELS * ibv_dev.np * sizeof(int));
    ibv_dev.lid_table = (uint16_t *) malloc(ibv_dev.np * MAX_SUBCHANNELS * sizeof(uint16_t));
    ibv_dev.qp_table = (uint32_t *) malloc(ibv_dev.np * MAX_SUBCHANNELS * sizeof(uint32_t));
    ibv_dev.qp_hndl = (struct ibv_qp **) malloc(ibv_dev.np * MAX_SUBCHANNELS *  sizeof(struct ibv_qp *));
    ibv_dev.pids     = (uint32_t *) malloc(ibv_dev.np * sizeof(uint32_t));

    ibv_dev.pids[ibv_dev.me] = mypid =  (int) getpid();

    ibv_dev.maxtransfersize = DEFAULT_MAX_RDMA_SIZE;

    if (alladdrs == NULL || local_addr == NULL ||
            ibv_dev.lid_table == NULL || ibv_dev.qp_table == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                "malloc for alladdrs/local_addr/lid_table");
    }

	viadev_init_parameters(ibv_dev.np, ibv_dev.me);

    /* Initialize InfiniBand Resources and Connections */
    ib_init();

    for(i = 0; i < ibv_dev.np; i++) {
        for(j = 0; j < num_subchannels; j++){
            if(ibv_dev.me == i) {
                /* Put a junk value, which will not be used for the
                 * communication */
                local_addr[i * num_subchannels + j] = ibv_dev.my_hca_lid[j];
            } else {
                local_addr[i * num_subchannels + j] = ibv_dev.qp_hndl
												[i * num_subchannels + j]->qp_num;
            }
        }
    }

    /* Without the hostname this is the length of the data transfer */
    local_addr_len = num_subchannels * 	ibv_dev.np * sizeof(int);

#ifdef _SMP_
    local_addr[ibv_dev.np * num_subchannels] = get_host_id
						(ibv_dev.my_name, HOSTNAME_LEN);
#else /* Useless integer for non-SMP code */
    local_addr[ibv_dev.np * num_subchannels] = ibv_dev.me;
#endif /* _SMP_ */
	
    /* Adding the length by an integer for SMP */
    local_addr_len += sizeof(int);
	
    pmgr_exchange_addresses(local_addr, local_addr_len, 
            alladdrs);

    ib_rank_lid_table(alladdrs, local_addr_len);
    ib_qp_enable();

    init_mpi_connections();

    dreg_init();
    free(local_addr);
    free(alladdrs);
    return MPI_SUCCESS;
}


void MPID_VIA_Finalize(void)
{
    ib_finalize();

	M_PRINT("done with ib_finalize\n");
    if (ibv_dev.connections) {
        free(ibv_dev.connections);
    }

    pmgr_finalize();

    if (ibv_dev.lid_table) {
        free(ibv_dev.lid_table);
    }
    if (ibv_dev.qp_table) {
        free(ibv_dev.qp_table);
    }
    if (ibv_dev.pids) {
        free(ibv_dev.pids);
    }
    if(ibv_dev.processes) {
        free(ibv_dev.processes);
    }

#ifdef _SMP_
    if (smpi.local_nodes) {
        free(smpi.local_nodes);
    }
    if (smpi.l2g_rank) {
        free(smpi.l2g_rank);
    }
#endif

}

void MPID_Wtime(double *t)
{
    struct timeval tv;
    static int initialized = 0;
    static int sec_base;
    gettimeofday(&tv, NULL);
    if (!initialized) {
        sec_base = tv.tv_sec;
        initialized = 1;
    }
    *t = (double) (tv.tv_sec - sec_base) + (double) tv.tv_usec * 1.0e-6;
}

void MPID_Wtick(double *t)
{
    *t = 0.000010;
}
