/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 * 
 * Developed at Berkeley Lab as part of MVICH.
 * 
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *           Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 * 
 */

#include "mpid.h"
#include "ibverbs_header.h"
#include "viapriv.h"
#include "mpid_smpi.h"

int num_ports = 1;
int num_hcas = 1;
int num_subchannels = 2;
int num_phy_ports = 1;
int num_qp_per_port = 2;

int backlog_counter = 0;
int link_speed[MAX_SUBCHANNELS];

int striping_threshold = VBUF_BUFFER_SIZE;

int rdma_scheduling = RDMA_EVEN_STRIPING;
int sr_scheduling = SR_USE_FIRST;
int sr_rr_window = DEFAULT_RR_WINDOW;

int	sm_scheduling = ROUND_ROBIN;
int lm_scheduling = STRIPE_BLOCKING;

int progress = PARTIAL;
/* Global structure containing information about ADI2 device */
ibv_info_t ibv_dev;

char nullsbuffer, nullrbuffer;

/* 
 * post a descriptor to the send queue
 * all outgoing packets go through this routine. 
 * takes a connection rather than a vi because we need to 
 * update flow control information on the connection. Also
 * it turns out it is always called in the context of a connection, 
 * i.e. it would be post_send(c->vi) otherwise. 
 */

#define BACKLOG_ENQUEUE(q,v) {                                      \
    v->desc.next = NULL;                                            \
    if (q->vbuf_tail == NULL) {                                     \
        q->vbuf_head = v;                                           \
    } else {                                                        \
        q->vbuf_tail->desc.next = v;                                \
    }                                                               \
    q->vbuf_tail = v;                                               \
    q->len++;                                                       \
}


#define BACKLOG_DEQUEUE(q,v)  {                                     \
    v = q->vbuf_head;                                               \
    q->vbuf_head = v->desc.next;                                    \
    if (v == q->vbuf_tail) {                                        \
        q->vbuf_tail = NULL;                                        \
    }                                                               \
    q->len--;                                                       \
    v->desc.next = NULL;                                            \
}


struct ibv_mr* register_memory(void* buf, int len, int hca)
{
    /* Register with a particular HCA */
        return (ibv_reg_mr(ibv_dev.ptag[hca], buf, len,
            IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE));
}

int deregister_memory(struct ibv_mr* mr)
{
    return (ibv_dereg_mr(mr));
}

/* init the backlog queue */
void viadev_init_backlog_queue(ibv_connection_t * c)
{
    int i = 0;
    for(i = 0; i < num_subchannels; i++) {
        c->backlog[i].len = 0;
        c->backlog[i].vbuf_head = NULL;
        c->backlog[i].vbuf_tail = NULL;
    }
}

/* to handle Send Q overflow, we maintain an extended send queue
 * above the HCA.  This permits use to have a virtually unlimited send Q
 * depth (limited by number of vbufs available for send)
 */
void viadev_ext_sendq_queue(ibv_connection_t * c, vbuf * v,
        int subchannel)
{
    v->desc.next = NULL;
    if (c->ext_sendq_head[subchannel] == NULL) {
        c->ext_sendq_head[subchannel] = v;
    } else {
        c->ext_sendq_tail[subchannel]->desc.next = v;
    }
    c->ext_sendq_tail[subchannel] = v;
}

/* dequeue and send as many as we can from the extended send queue
 * this is called in each function which may post send prior to it attempting
 * its send, hence ordering of sends is maintained
 */
void viadev_ext_sendq_send(ibv_connection_t * c,
        int subchannel)
{
    vbuf *v;
    struct ibv_send_wr *bad_wr;

    while (c->send_wqes_avail[subchannel] && c->ext_sendq_head[subchannel]) {
        /* dequeue one vbuf */
        v = c->ext_sendq_head[subchannel];
        c->ext_sendq_head[subchannel] = v->desc.next;
        if (v == c->ext_sendq_tail[subchannel]) {
            c->ext_sendq_tail[subchannel] = NULL; 
        }
        v->desc.next = NULL; 
        /* put the vbuf on the HCA's send Q */
        c->send_wqes_avail[subchannel]--;

        if(ibv_post_send(c->qp_hndl[subchannel], &(v->desc.u.sr), &bad_wr)) {
            error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
        }
    }
}

/* Just got a credit update, send off as many backlogged
 * messages as possible
 */
void viadev_backlog_send(ibv_connection_t * c, int subchannel)
{
    struct ibv_send_wr *bad_wr;

    ibv_backlog_queue_t *q = &(c->backlog[subchannel]);

    while ((q->len > 0) && (c->remote_credit[subchannel] > 0)) {
        vbuf *v = NULL;
        viadev_packet_header *p;
        assert(q->vbuf_head != NULL);
        BACKLOG_DEQUEUE(q, v);

        /* Assumes packet header is at beginning of packet structure */
        p = (viadev_packet_header *) VBUF_BUFFER_START(v);
        
        PACKET_SET_CREDITS(p, c, subchannel);
        c->remote_credit[subchannel]--;

        v->grank = c->global_rank;
        v->subchannel = subchannel;

        if (!c->send_wqes_avail[subchannel]) {
            viadev_ext_sendq_queue(c, v, subchannel);
            continue;
        }
        c->send_wqes_avail[subchannel]--;

        if(ibv_post_send(c->qp_hndl[subchannel], &(v->desc.u.sr), &bad_wr)) {
            error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
        }

    }
}


/* post a descriptor to the receive queue */
void viadev_post_recv(ibv_connection_t * c, vbuf * v,
        int subchannel)
{
    struct ibv_recv_wr *bad_wr;
	v->desc.subchannel = subchannel;
	v->subchannel = subchannel;
    v->grank = c->global_rank;
    if(ibv_post_recv(c->qp_hndl[subchannel], &(v->desc.u.rr), &bad_wr)) {
        error_abort_all(IBV_RETURN_ERR,
                "Error posting recv\n");
    }
}

void viadev_post_send(ibv_connection_t * c, vbuf * v, 
        int subchannel)
{
    struct ibv_send_wr *bad_wr;

    viadev_packet_header *p =
        (viadev_packet_header *) VBUF_BUFFER_START(v);

    if (c->remote_credit[subchannel] > 0 || p->type == VIADEV_PACKET_NOOP) {
        /* if we got here, the backlog queue better be  empty */
        assert(c->backlog[subchannel].len == 0 || p->type == VIADEV_PACKET_NOOP);
        PACKET_SET_CREDITS(p, c, subchannel);
        if (p->type != VIADEV_PACKET_NOOP)
            c->remote_credit[subchannel]--;
        v->grank = c->global_rank;

        if (!c->send_wqes_avail[subchannel]) {
            viadev_ext_sendq_queue(c, v, subchannel);
            return;
        }
        c->send_wqes_avail[subchannel]--;

        if(v->desc.sg_entry.length < inline_size) {
            v->desc.u.sr.send_flags =
                IBV_SEND_SIGNALED;
        } else {
            v->desc.u.sr.send_flags =
                IBV_SEND_SIGNALED;
        }

        if(ibv_post_send(c->qp_hndl[subchannel], 
                    &(v->desc.u.sr), &bad_wr)) {
            error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
        }

        if (PREPOST_THRESHOLD && v->shandle != NULL
                && c->initialized[subchannel]) {
            int needed =
                PREPOST_DEPTH + PREPOST_NOOP_EXTRA +
                MIN(PREPOST_RENDEZVOUS_EXTRA,
                        c->rendezvous_packets_expected);
            assert(needed >= 0);
            while (c->preposts[subchannel] < DEFAULT_MAX_RECV_WQE && c->preposts[subchannel] < needed) {
                PREPOST_VBUF_RECV(c, subchannel);
            }
            viadev_send_noop_ifneeded(c, subchannel);
        }
    } else {
        /* must delay the send until more credits arrive.
         * put on end of queue */
        backlog_counter++;
        ibv_backlog_queue_t *q = &c->backlog[subchannel];
        BACKLOG_ENQUEUE(q, v);
    }
}
/*
 * calculate_vbufs_expected is called whenever we get
 * a match between a posted receive and an incoming packet. 
 * This happens either when a receive is posted and matches
 * an entry on the unexpected queue or when a packet
 * comes in that matches a pre-posted receive. 
 * 
 * In general, it is not until this match that we have information
 * from both sides that allows us to figure out how many packets
 * will be sent to complete the transfer. 
 * 
 * When the transfer is through a rendezvous protocol, the
 * specific type of rendezvous will have been determined before
 * this routine is called. 
 *
 */


int viadev_calculate_vbufs_expected(int nbytes, viadev_protocol_t protocol)
{
    int calc, first_packet, next_packet;
    switch (protocol) {
    case VIADEV_PROTOCOL_EAGER:
        {
            first_packet = VBUF_BUFFER_SIZE -
                sizeof(viadev_packet_eager_start);
            next_packet = VBUF_BUFFER_SIZE -
                sizeof(viadev_packet_eager_next);
            break;
        }
    case VIADEV_PROTOCOL_R3:
        {
            first_packet = VBUF_BUFFER_SIZE -
                sizeof(viadev_packet_r3_data);
            next_packet = first_packet;
            break;
        }
    case VIADEV_PROTOCOL_RPUT:
        {
            /* this is number of rputs by sender.
             * receiver gets only one packet */
            first_packet = ibv_dev.maxtransfersize;
            next_packet = ibv_dev.maxtransfersize;
            break;
        }
    default:
        {
            /* return to MPI */
            first_packet = 1;
            next_packet = 1;
            error_abort_all(GEN_EXIT_ERR,
                                "calculate_vbufs_expected: unexpected protocol %d",
                                protocol);
        }
    }
    nbytes -= first_packet;
    if (nbytes > 0) {
        calc = (2 + (nbytes - 1) / next_packet);
    } else {
        calc = (1);
    }
    return calc;
}

#ifdef RDMA_FAST_PATH
void release_recv_rdma(ibv_connection_t * c, vbuf * v)
{
    vbuf *next_free;
    int next;
    int i;
    int subchannel = v->desc.subchannel;
    assert(v->desc.subchannel == v->subchannel);
    next = (c->p_RDMA_recv_tail[subchannel] + 1) % viadev_num_rdma_buffer;
    next_free = &(c->RDMA_recv_buf[subchannel][next]);

    v->padding = FREE_FLAG;
    if (v != next_free) {
        return;
    }
    for (i = next; i != c->p_RDMA_recv[subchannel]; 
        i = (i + 1) % viadev_num_rdma_buffer) {
        if (c->RDMA_recv_buf[subchannel][i].padding == FREE_FLAG) {
            c->rdma_credit[subchannel]++;
            c->p_RDMA_recv_tail[subchannel] = 
            (c->p_RDMA_recv_tail[subchannel] + 1) % viadev_num_rdma_buffer;
            c->RDMA_recv_buf[subchannel][i].padding = BUSY_FLAG;
         } else
            break;
         }

#if 0
    /* search all free buffers */
    for (i = next; i != c->p_RDMA_recv;) {

        if (c->RDMA_recv_buf[i].padding == FREE_FLAG) {
            c->rdma_credit++;
            if (++(c->p_RDMA_recv_tail) >= viadev_num_rdma_buffer)
                c->p_RDMA_recv_tail = 0;
            c->RDMA_recv_buf[i].padding = BUSY_FLAG;
        } else break;
        if (++i >= viadev_num_rdma_buffer)
            i = 0;
    }
#endif
}
#endif

void viadev_send_noop_ifneeded(ibv_connection_t *c,
        int subchannel)
{
    /* Dont want remote end to stall too long waiting for a credit
     * update.  If the local credits have built up above the dynamic
     * threshold, send a no-op immediately
     * c->remote_cc should accurately reflect the remote sides credit situation.
     * If its low and we have built-up local credits, send a noop
     * for the purposes of a credit refresh and to prevent a deadlock.
     */
    if (c->local_credit[subchannel] >= DYNAMIC_CREDIT_THRESHOLD
        || (c->remote_cc[subchannel] <= CREDIT_PRESERVE &&
            c->local_credit[subchannel] >= CREDIT_NOTIFY_THRESHOLD)
        ) {
        viadev_send_noop(c, subchannel);
    }
}

void viadev_send_noop(ibv_connection_t * c, int subchannel)
{
    /* always send a noop when it is needed even if there is a backlog.
     * noops do not consume credits.
     * this is necessary to avoid credit deadlock.
     * RNR NAK will protect us if receiver is low on buffers.
     * by doing this we can force a noop ahead of any other queued packets.
     */
    int i = subchannel;
	assert(subchannel >= 0 && i >= 0);
	
#if defined(RDMA_FAST_PATH)
    vbuf *v;
    viadev_packet_noop *p;

    if (fast_rdma_ok(c, sizeof(viadev_packet_noop), 0, i)) {
        v = &(c->RDMA_send_buf[i][c->phead_RDMA_send[i]]);
        v->shandle = NULL;
        p = (viadev_packet_noop *) v->buffer;
        PACKET_SET_HEADER(p, c, VIADEV_PACKET_NOOP);
        post_fast_rdma_with_completion(c, sizeof(viadev_packet_noop), i);
    } else {
        v = get_vbuf();
        p = (viadev_packet_noop *) VBUF_BUFFER_START(v);
        PACKET_SET_HEADER(p, c, VIADEV_PACKET_NOOP);
        vbuf_init_send(v, sizeof(viadev_packet_header), i);
        viadev_post_send(c, v, i);
    }

#else
    vbuf *v = get_vbuf();
    viadev_packet_noop *p = (viadev_packet_noop *) VBUF_BUFFER_START(v);
    D_PRINT("VI %3d SEND NOOP with %d credits", c->global_rank,
            c->local_credit[i]);
    PACKET_SET_HEADER(p, c, VIADEV_PACKET_NOOP);
    vbuf_init_send(v, sizeof(viadev_packet_header), i);
    viadev_post_send(c, v, i);
#endif
}

#ifdef RDMA_FAST_PATH
int fast_rdma_ok(ibv_connection_t * c, int data_size,
        int is_eager_packet, int subchannel)
{
    /* not OK if buffers are full */
    if (c->phead_RDMA_send[subchannel] == 
            c->ptail_RDMA_send[subchannel]) {
        return 0;
    }

    /* not OK if data size is too large */
    if (data_size > VBUF_BUFFER_SIZE) {
        return 0;
    }

    /* not OK if there are sends waiting in the backlog ??? */
    if (c->backlog[subchannel].len > 0) {
        return 0;
    }

    /* See if head buffer is marked free */
    if(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]].padding == BUSY_FLAG) {
        return 0;
    }

    /* otherwise OK */
    return 1;
}
#endif

#ifdef RDMA_FAST_PATH
void post_fast_rdma_with_completion(ibv_connection_t * c, int len
        , int subchannel)
{
    int n = c->phead_RDMA_send[subchannel];
    VBUF_FLAG_TYPE flag;
    struct ibv_send_wr *bad_wr;

    viadev_packet_header *p = (viadev_packet_header *)
        (c->RDMA_send_buf[subchannel][n].buffer);

    /* Mark RDMA buffer busy */
    c->RDMA_send_buf[subchannel][n].padding = BUSY_FLAG;

    assert(subchannel >= 0);
    /* set credits info */
#ifndef DISABLE_HEADER_CACHING
    if ((p->type != FAST_EAGER_CACHED) &&
        (p->type != VIADEV_PACKET_EAGER_START))
        PACKET_SET_CREDITS(p, c, subchannel);
	if ((p->type == VIADEV_PACKET_EAGER_START) &&
        (len >= 256))
        PACKET_SET_CREDITS(p, c, subchannel);
#else
    PACKET_SET_CREDITS(p, c, subchannel);
#endif

    if (++(c->phead_RDMA_send[subchannel]) >= viadev_num_rdma_buffer)
        c->phead_RDMA_send[subchannel] = 0;

    /* set flags */
    if (*(VBUF_FLAG_TYPE *) (c->RDMA_send_buf[subchannel][n].buffer + len)
        == len) {
        flag = (VBUF_FLAG_TYPE) (len + FAST_RDMA_ALT_TAG);
    } else {
        flag = (VBUF_FLAG_TYPE) len;
    }

    /* set head flag */
    c->RDMA_send_buf[subchannel][n].head_flag = flag;
    /* set tail flag */
    *(VBUF_FLAG_TYPE *) (c->RDMA_send_buf[subchannel][n].buffer + len)
        = flag;
    c->RDMA_send_buf[subchannel][n].desc.u.sr.wr_id = 
        (aint_t) &(c->RDMA_send_buf[subchannel][n]);

	c->RDMA_send_buf[subchannel][n].desc.subchannel = subchannel;
	c->RDMA_send_buf[subchannel][n].subchannel = subchannel;
    c->RDMA_send_buf[subchannel][n].desc.sg_entry.length =
        len + VBUF_FAST_RDMA_EXTRA_BYTES;

    if (!c->send_wqes_avail[subchannel]) {
        viadev_ext_sendq_queue(c, &c->RDMA_send_buf[subchannel][n],
                subchannel);
        return;
    }
    c->send_wqes_avail[subchannel]--;
#ifdef _IBM_EHCA_
        c->RDMA_send_buf[subchannel][n].desc.u.sr.send_flags =
            IBV_SEND_SIGNALED;
#else
    if(c->RDMA_send_buf[subchannel][n].desc.sg_entry.length < MAX_INLINE_SIZE) {
        c->RDMA_send_buf[subchannel][n].desc.u.sr.send_flags =
            IBV_SEND_SIGNALED | IBV_SEND_INLINE;
    } else {
        c->RDMA_send_buf[subchannel][n].desc.u.sr.send_flags =
            IBV_SEND_SIGNALED;
    }
#endif
    if(ibv_post_send(c->qp_hndl[subchannel], 
               &c->RDMA_send_buf[subchannel][n].desc.u.sr, &bad_wr)) {
        error_abort_all(IBV_RETURN_ERR,"Error posting send"
                " post size : %d, inline size : %d\n", 
                c->RDMA_send_buf[subchannel][n].desc.sg_entry.length,
                inline_size);
    }
}
#ifndef DISABLE_HEADER_CACHING

int search_header_cache(ibv_connection_t * c,
                        viadev_packet_eager_start * h, int subchannel)
{
    viadev_packet_eager_start *cached = &(c->cached_outgoing[subchannel]);

    if ((cached->header.vbuf_credit == h->header.vbuf_credit) &&
        (cached->header.remote_credit == h->header.remote_credit) &&
        (cached->header.rdma_credit == h->header.rdma_credit) &&
        (cached->envelope.context == h->envelope.context) &&
        (cached->envelope.tag == h->envelope.tag) &&
        (cached->envelope.src_lrank == h->envelope.src_lrank)) {
        c->cached_hit++;
        return 1;
    }

    c->cached_miss++;
    memcpy(cached, h, sizeof(viadev_packet_eager_start));
    return 0;
}

#endif
#endif

void viadev_rput(ibv_connection_t * c, vbuf * v, void *local_address,
                 uint32_t lkey, void *remote_address,
                 uint32_t rkey, int nbytes)
{
    D_PRINT("viadev_rput: RDMA write\n");

    vbuf_init_rput(v, local_address, lkey,
                   remote_address, rkey, nbytes);

    viadev_post_rdmawrite(c, v, v->subchannel);
}

void viadev_rput_finish(MPIR_SHANDLE *s)
{

#if defined(RDMA_FAST_PATH)
    int rdma_ok;
    vbuf *v;
    void *buffer;
    viadev_packet_rput_finish *packet;
#else
    vbuf *v;
    void *buffer;
    viadev_packet_rput_finish *packet;
#endif
    ibv_connection_t *c = (ibv_connection_t *) (s->connection);

    int subchannel = 0;
    int weight[MAX_SUBCHANNELS];
    int total_weight, last_channel;

    last_channel = 0; total_weight = num_subchannels;
    for (subchannel = 0; subchannel < num_subchannels; subchannel++) {
        if (s->rput_flag[subchannel] == 0)
            weight[subchannel] = 0;
        else {
            weight[subchannel] = 1;
            total_weight--;
            last_channel = subchannel;
        }
    }

    weight[last_channel] += total_weight;

    for (subchannel = 0; subchannel < num_subchannels; subchannel++) {

    if (s->rput_flag[subchannel] == 0)
        continue;

#if defined(RDMA_FAST_PATH)

    rdma_ok = fast_rdma_ok(c,
        sizeof(viadev_packet_rput_finish), 0, subchannel);

    if (rdma_ok) {
        v = &(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]]);
        v->shandle = NULL;
        packet = (viadev_packet_rput_finish*) v->buffer;
    }
    else {
        v = get_vbuf();
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    }
#else
        v = get_vbuf();
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;

#endif

    /* fill in the packet */
    PACKET_SET_HEADER(packet, c, VIADEV_PACKET_RPUT_FINISH);
    packet->rreq      =  REQ_TO_ID(s->receive_id);

    /* mark MPI send complete when VIA send completes */
    v->shandle = s;
    packet->striping_weight =  weight[subchannel];

	assert(subchannel >= 0);
#if defined(RDMA_FAST_PATH)
    if (rdma_ok)
        post_fast_rdma_with_completion(c,
            sizeof(viadev_packet_rput_finish), subchannel);
    else {
        vbuf_init_send(v, sizeof(viadev_packet_rput_finish), subchannel);
        viadev_post_send(c, v, subchannel);
    }
#else
    /* prepare descriptor and post */
    vbuf_init_send(v, sizeof(viadev_packet_rput_finish), subchannel);
    viadev_post_send(c, v, subchannel);
#endif
    D_PRINT("rput finish (for recv request %x), striping weight = %d\n",
            packet->rreq,
            packet->striping_weight);
    }

}


void viadev_post_rdmawrite(ibv_connection_t * c, vbuf * v
        , int subchannel)
{
    struct ibv_send_wr *bad_wr;


    /* set id to be the global_rank */
    v->grank = c->global_rank;
	v->subchannel = subchannel;
	v->desc.subchannel = subchannel;
    if (!c->send_wqes_avail[subchannel]) {
        viadev_ext_sendq_queue(c, v, subchannel);
        return;
    }
    c->send_wqes_avail[subchannel]--;

    /* Don't try to inline this data, since it is large */
    v->desc.u.sr.send_flags = IBV_SEND_SIGNALED;

    if(ibv_post_send(c->qp_hndl[subchannel], &(v->desc.u.sr), &bad_wr)) {
        error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
    }
}

void viadev_incoming_rput_finish(vbuf * v, ibv_connection_t * c,
                                 viadev_packet_rput_finish * h)
{

    MPIR_RHANDLE *rhandle;

    /* find the rhandle for this data */
    rhandle = (MPIR_RHANDLE *) ID_TO_REQ(h->rreq);
	assert(rhandle != NULL);
    rhandle->total_stripe_received +=  h->striping_weight;
    if ((int)rhandle->total_stripe_received < num_subchannels)
           return;
    rhandle->bytes_copied_to_user = rhandle->len;
    RECV_COMPLETE(rhandle);

}

int choose_sr_subchannel(ibv_connection_t *c)
{
    static int i = 0;
    if (sm_scheduling == ROUND_ROBIN){
        i = i - 1;
        if(i == -1) {
            i = num_subchannels - 1;
        }
        return i;
    }else if (sm_scheduling == ALT_ROUND_ROBIN){
        assert(num_subchannels >= 4);
        i = i - 1;
        if(i == -1) {
            i = num_subchannels - 1;
        }
        if (i == 1) 
            return 2;
        if (i == 2) 
            return 1;
        return i;
    } else if (PROCESS_BINDING == sm_scheduling) {
#ifdef _SMP_
        return (smpi.my_local_id % num_subchannels);
#else
        return (ibv_dev.me % num_subchannels);
#endif
    }
    return (num_subchannels - 1);
}


int choose_lm_subchannel(ibv_connection_t *c)
{
    static int i = 0;
    if (ROUND_ROBIN == lm_scheduling
            || STRIPE_BLOCKING == lm_scheduling){
        i = i - 1;
        if(i == -1) {
            i = num_subchannels - 1;
        }
        return i;
    } else if (PROCESS_BINDING == lm_scheduling) {
#ifdef _SMP_
        return (smpi.my_local_id % num_subchannels); 
#else
        return (ibv_dev.me % num_subchannels); 
#endif
    }
    return (num_subchannels - 1);
}


/* striping rdma messages across multiple subchannels */
void schedule_rdma_round_robin(ibv_connection_t *c, char *addr, int size,
    char** stripe_start_addr, int *stripe_size, int do_stripe)
{
    /* do even striping */
    int i, subchannel;
    subchannel = choose_lm_subchannel(c);

    /* use the first connection for small messages */

    stripe_start_addr[subchannel] = addr;
    stripe_size[subchannel] = size;
    for (i = 0; i < num_subchannels; i++) {
        if(i != subchannel){
                stripe_size[i] = 0;
        }
    }
}


/* striping rdma messages across multiple subchannels */
void schedule_rdma_stripe_blocking(ibv_connection_t *c, char *addr, int size,
    char** stripe_start_addr, int *stripe_size, int do_stripe)
{
	/* do even striping */
	int i, subchannel;
	char *next_addr;
	int inc;
	if ((size < striping_threshold) || !do_stripe) {
		subchannel = choose_lm_subchannel(c);
		/* use the first connection for small messages */
		stripe_start_addr[subchannel] = addr;
		stripe_size[subchannel] = size;
		for (i = 0; i < num_subchannels; i++)
			if(i != subchannel)
				stripe_size[i] = 0;
	} else {
		/* strip evenly across all connections */
		D_PRINT("striping triggered\n");
		inc = size / num_subchannels;
		next_addr = addr;
		for (i = 0; i < num_subchannels - 1; i++) {
			stripe_start_addr[i] = next_addr;
			stripe_size[i] = inc;
			next_addr += inc;
		}
		stripe_start_addr[num_subchannels - 1] = next_addr;
		stripe_size[num_subchannels - 1] = size - (num_subchannels - 1) * inc;
	}

}


void adjust_weights(ibv_connection_t *c, double start_time,
    double finish_time[MAX_SUBCHANNELS],
    double init_weight[MAX_SUBCHANNELS])
{
    int i;
    double bw[MAX_SUBCHANNELS]; 
    double bw_total = 0;
    int weight_assigned = 0;

    for (i = 0; i < num_subchannels; i++) {
        finish_time[i] -= start_time;
        finish_time[i] /= 1000;
        bw[i] = (init_weight[i]) / (double)(finish_time[i]);
        bw_total += bw[i];
    }

    for (i = 0; i < num_subchannels - 1; i++) {
        c->s_weight[i] = (int)
                (DYNAMIC_TOTAL_WEIGHT * bw[i] / bw_total);
        if (c->s_weight[i] == 0){
                c->s_weight[i]++;
        }
        weight_assigned += c->s_weight[i];
    }
    c->s_weight[num_subchannels - 1] =
            DYNAMIC_TOTAL_WEIGHT - weight_assigned;
}

void schedule_rdma_weighted_striping( ibv_connection_t *c,
                char *addr, int size, char** stripe_start_addr,
                int *stripe_size, int do_stripe)
{
    int i;
    char *next_addr;
    int inc;
    int s_total, subchannel;
    if (size < striping_threshold) {
            subchannel = choose_lm_subchannel(c);
            /* use the first connection for small messages */
            stripe_start_addr[subchannel] = addr;
            stripe_size[subchannel] = size;
            for (i = 0; i < num_subchannels; i++)
                if(i != subchannel)
                    stripe_size[i] = 0;
    } else {
    /* strip across all connections based on weights*/
        s_total = 0;
        for (i = 0; i < num_subchannels; i++) {
            s_total += c->s_weight[i];
        }
        next_addr = addr;
        for (i = 0; i < num_subchannels - 1; i++) {
            inc = c->s_weight[i] * size / s_total;
            stripe_start_addr[i] = next_addr;
            stripe_size[i] = inc;
            next_addr += inc;
        }
        stripe_start_addr[num_subchannels - 1] = next_addr;
        stripe_size[num_subchannels - 1] = addr
                            + size - next_addr;
    }

}


/* striping rdma messages across multiple subchannels */
void schedule_rdma_even_striping(ibv_connection_t *c, char *addr, int size,
	char** stripe_start_addr, int *stripe_size, int do_stripe)
{

	/* do even striping */
	int i, subchannel;
	char *next_addr;
	int inc;
	if (size < striping_threshold) {
		subchannel = choose_lm_subchannel(c);
		/* use the first connection for small messages */
		stripe_start_addr[subchannel] = addr;
		stripe_size[subchannel] = size;
		for (i = 0; i < num_subchannels; i++)
			if(i != subchannel)
				stripe_size[i] = 0;
	} else {
		/* strip evenly across all connections */
		D_PRINT("stripping triggered\n");
		inc = size / num_subchannels;
		next_addr = addr;
		for (i = 0; i < num_subchannels - 1; i++) {
			stripe_start_addr[i] = next_addr;
			stripe_size[i] = inc;
			next_addr += inc;
		}
		stripe_start_addr[num_subchannels - 1] = next_addr;
		stripe_size[num_subchannels - 1] = size - 
			(num_subchannels - 1) * inc;
	}

}

void schedule_rdma(ibv_connection_t *c, char *addr, int size,
	char** stripe_start_addr, int *stripe_size, int do_stripe)
{
    if(STRIPE_BLOCKING == lm_scheduling) {
        schedule_rdma_stripe_blocking(c, addr, size, 
					stripe_start_addr, stripe_size, do_stripe);
    } 
    if (EVEN_STRIPING == lm_scheduling){
        schedule_rdma_even_striping(c, addr, size,
                    stripe_start_addr, stripe_size, do_stripe);
    } 
    if (PROCESS_BINDING == lm_scheduling){
        schedule_rdma_round_robin(c, addr, size,
                    stripe_start_addr, stripe_size, do_stripe);
    } 
    if (ROUND_ROBIN == lm_scheduling) {
        schedule_rdma_round_robin(c, addr, size,
                    stripe_start_addr, stripe_size, do_stripe);
    } 
    if (USE_FIRST == lm_scheduling) {
        schedule_rdma_round_robin(c, addr, size,
                    stripe_start_addr, stripe_size, do_stripe);
    } 
    if (ADAPTIVE_STRIPING == lm_scheduling) {
        schedule_rdma_weighted_striping(c, addr, size,
                            stripe_start_addr, stripe_size, do_stripe);
    }
}



/* check ofo queues of all connections to see if
 * there is any packet (vbuf) that matches the current
 * expected sequence number
 * */

int check_ofo_queues(vbuf **vbuf_addr)
{
    int i;
    viadev_packet_header *h;
    vbuf *v;
    for (i = 0; i < ibv_dev.np; i++) {
        ibv_connection_t *c = &ibv_dev.connections[i];
        if (i == ibv_dev.me) {
			continue;
		}
        /* check the ofo queue */
        v = c->ofo_head;
        while (v) {
           h = (viadev_packet_header *)VBUF_BUFFER_START(v);
           /* in-order packet? */
           if (h->id == c->next_packet_expected) {
               remove_ofo_queue(v, c);
               *vbuf_addr = v;
               return 0;
           }
           v = (vbuf *)(v->desc.next);
        }
    }
    return 1;
}


/* add a packet (vbuf) to an ofo queue */
void enqueue_ofo_queue(vbuf *vbuf_addr, 
				ibv_connection_t *c)
{
    vbuf_addr->desc.next = c->ofo_head;
    c->ofo_head = vbuf_addr;
}


void remove_ofo_queue(vbuf *vbuf_addr, ibv_connection_t *c)
{
    vbuf *v, *v_next;

    v = c->ofo_head;
	if (v == NULL) {
    	fprintf(stderr, "remove_ofo_queue error, queue empty! vbuf = %p\n", 
						vbuf_addr);
        exit(-1);
    }

    if (v == vbuf_addr) {
        c->ofo_head = (vbuf *)(v->desc.next);
        return;
    }

    v_next = (vbuf *)(v->desc.next);
    while (v_next) {
	    if (v_next == vbuf_addr) {
    		v->desc.next = v_next->desc.next;
		    return;
    	}
	    v = v_next;
   		v_next = (vbuf *)(v_next->desc.next);
    }
    fprintf(stderr, "remove_ofo_queue error, not found! vbuf = %p\n", 
					vbuf_addr);
    exit(-1);
}


/* init all ofo queues to NULL */
void init_ofo_queues(void)
{
    int i;
    for (i = 0; i < ibv_dev.np; i++) {
        ibv_connection_t *c = &ibv_dev.connections[i];
        c->ofo_head = NULL;
    }
}

