/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "mpid.h"
#include "viadev.h"
#include "viapacket.h"
#include "queue.h"
#include "viaparam.h"
#include "viapriv.h"

int MPID_VIA_rendezvous_start(void *buf, int len, int src_lrank, int tag,
                              int context_id, int dest_grank,
                              MPIR_SHANDLE * shandle, int subchannel)
{

    /*
     * Construct a packet to start a rendezvous transfer. 
     * If dreg_entry is NULL, memory is not pinned, otherwise
     * it is. 
     */
    int i;
    dreg_entry *dreg_entry;
#if defined(RDMA_FAST_PATH)
    int rc = MPI_SUCCESS;
    ibv_connection_t *c = &ibv_dev.connections[dest_grank];
    vbuf *v;
    void *buffer;
    viadev_packet_rendezvous_start *packet;
    int proto = VIADEV_PROTOCOL_R3;
    int rdma_ok = fast_rdma_ok(c, 
            sizeof(viadev_packet_rendezvous_start), 0, subchannel);

    if (rdma_ok) {
        v = &(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]]);
        v->shandle = NULL;
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    } else {
        v = get_vbuf();
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    }


#else
    int rc = MPI_SUCCESS;

    vbuf *v = get_vbuf();
    void *buffer = VBUF_BUFFER_START(v);
    viadev_packet_rendezvous_start *packet = buffer;
    ibv_connection_t *c = &ibv_dev.connections[dest_grank];
    int proto = VIADEV_PROTOCOL_R3;
#endif

	assert(subchannel >= 0);
    /* if this fails, will return NULL, which is what we want. 
     * So no need to do anything other than try to register. 
     */
    dreg_entry = NULL;

    proto = VIADEV_PROTOCOL_RPUT;
    D_PRINT("rendzvous enter: set RPUT (proto=%d)\n", proto);

    if (0 == len) {
        proto = VIADEV_PROTOCOL_R3;
        D_PRINT("rendzvous enter: set R3 since len=0 (proto=%d)\n", proto);
    } else if ((buf != NULL) && (len > 0)) {
        dreg_entry = dreg_register(buf, len);
        if (NULL == dreg_entry) {
            proto = VIADEV_PROTOCOL_R3;
        }
    }

    /* fill in the packet */
    PACKET_SET_HEADER(packet, c, VIADEV_PACKET_RENDEZVOUS_START);
    PACKET_SET_ENVELOPE(packet, context_id, tag, len, src_lrank);
    packet->sreq = REQ_TO_ID(shandle);

    /* may be less than receive buffer size */
    packet->len = len;
    packet->buffer_address = buf;

    if (NULL != dreg_entry) {
        for(i = 0; i < num_hcas; i++)
            packet->memhandle[i] = *(struct ibv_mr*)dreg_entry->memhandle[i];
    }

    /* explicitly set protocol */
    shandle->protocol = proto;
    packet->protocol = proto;

    if (lm_scheduling == ADAPTIVE_STRIPING) {
        for (i = 0; i < num_subchannels; i++) {
                shandle->initial_weight[i] = c->s_weight[i];
                shandle->stripe_finish_time[i] = 0;
        }
        /* Need to find a uniform counter for all platforms
         * using MPI_Wtime for now */
        shandle->stripe_start_time = MPI_Wtime();
    }

#if defined(RDMA_FAST_PATH)
    if (rdma_ok) {
        post_fast_rdma_with_completion(c,
                                       sizeof
                                       (viadev_packet_rendezvous_start)
                                       , subchannel);
    } else {
        vbuf_init_send(v, sizeof(viadev_packet_rendezvous_start)
                , subchannel);
        viadev_post_send(c, v, subchannel);
    }
#else
    /* prepare descriptor and post */
    vbuf_init_send(v, sizeof(viadev_packet_rendezvous_start),
            subchannel);

    viadev_post_send(c, v, subchannel);
#endif

    /* we can post the vbuf before updating the send handle. 
     * If we go to a multithreaded implementation, in which 
     * another thread may process the completed vbuf, we will
     * need to update the shandle before posting the send
     * to avoid a race condition
     */

    shandle->local_address = buf;
    shandle->dreg_entry = dreg_entry;
    shandle->bytes_total = len;
    shandle->bytes_sent = 0;
    shandle->connection = c;
    shandle->is_complete = 0;

    /* fill in when we get the reply:
     * receive_id, remote_address, memory_handle
     */

    return rc;

}


int MPID_VIA_self_start(void *buf, int len, int src_lrank, int tag,
                        int context_id, MPIR_SHANDLE * shandle)
{
    MPIR_RHANDLE *rhandle;
    int rc = MPI_SUCCESS;
    int found;

    /* receive has been posted and is on the posted queue */
    MPID_Msg_arrived(src_lrank, tag, context_id, &rhandle, &found);

    if (found) {
        memcpy(rhandle->buf, buf, len);
        rhandle->s.MPI_TAG = tag;
        rhandle->s.MPI_SOURCE = src_lrank;
        RECV_COMPLETE(rhandle);

        rhandle->s.count = len;

        SEND_COMPLETE(shandle);
        /* need to call the finish routine for the receive? */
        return rc;
    }

    /* receive handle has been created 
     * and posted on unexpected queue */
    rhandle->connection = NULL;

    if (len < RENDEZVOUS_THRESHOLD) {
        /* copy the buffer and mark the send complete so there is a bit of
         * buffering for silly programs that send-to-self blocking.  Hide
         * info on the send in vbuf_{head,tail} which we know will not
         * be used.  */
        rhandle->send_id = 0;
        rhandle->vbuf_head = malloc(len);
        rhandle->vbuf_tail = (void *) ((aint_t) len);
        memcpy(rhandle->vbuf_head, buf, len);
        SEND_COMPLETE(shandle);
    } else {
        rhandle->send_id = REQ_TO_ID(shandle);
        shandle->local_address = buf;
        shandle->is_complete = 0;
        shandle->bytes_total = len;
    }

    return rc;
 
}

void MPID_VIA_self_finish(MPIR_RHANDLE * rhandle,
                          MPIR_RHANDLE * unexpected)
{
    MPIR_SHANDLE *shandle;

    void *send_ptr;
    int bytes_total;
    int truncate = 0;


    shandle = (MPIR_SHANDLE *) ID_TO_REQ(unexpected->send_id);

    if (shandle) {
        send_ptr = shandle->local_address;
        bytes_total = shandle->bytes_total;
    } else {
        send_ptr = unexpected->vbuf_head;
        bytes_total = (aint_t) unexpected->vbuf_tail;
    }

    if (bytes_total > rhandle->len)
        /* user error */
        truncate = 1;
    else
        /* fewer or same bytes were sent as were specified in the recv */
        rhandle->len = bytes_total;

    memcpy(rhandle->buf, send_ptr, rhandle->len);

    /* this is similate to copy_unexpected_handle_to_user_handle, but
     * we need many fewer fields 
     */
    rhandle->s.MPI_TAG = unexpected->s.MPI_TAG;
    rhandle->s.MPI_SOURCE = unexpected->s.MPI_SOURCE;
    rhandle->connection = NULL;
    RECV_COMPLETE(rhandle);

    if (truncate)
        rhandle->s.MPI_ERROR = MPI_ERR_TRUNCATE;

    if (shandle) {
        SEND_COMPLETE(shandle);
    } else {
        free(unexpected->vbuf_head);
    }

}

int MPID_VIA_eager_send(void *buf, int len, int src_lrank, int tag,
                        int context_id, int dest_grank, MPIR_SHANDLE * s,
                        int subchannel)
{

    int startmax = VBUF_DATA_SIZE(viadev_packet_eager_start);
    int nextmax = VBUF_DATA_SIZE(viadev_packet_eager_next);
    viadev_packet_eager_start *h;
    viadev_packet_eager_next *d;
    void *databuf;
    ibv_connection_t *c = &ibv_dev.connections[dest_grank];
    vbuf *v;

	D_PRINT("eager send\n");
#ifdef RDMA_FAST_PATH
    int rdma_ok;
#endif

    int bytes_sent;
	assert(subchannel >= 0);

    /*
     * we could fill in the send handle, but I think there is no
     * need, since we are sending all the data right here and now. 
     * However, is_complete should be set to zero because
     * the MPI send does not actually complete until the
     * VIA send completes. 
     */

    s->is_complete = 0;
    s->protocol = VIADEV_PROTOCOL_EAGER;

	s->dreg_entry = NULL;

    /* Set dreg_entry to NULL because SEND_COMPLETE will try to decrement
     * the reference count if it is non-null
     */
#if defined(RDMA_FAST_PATH)

    rdma_ok = fast_rdma_ok(c, 
            sizeof(viadev_packet_eager_start) + len, 1, subchannel);

#ifndef DISABLE_HEADER_CACHING
    if (rdma_ok && (len < 256)) {

        v = &(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]]);
		v->subchannel = subchannel;
		v->desc.subchannel = subchannel;
        h = (viadev_packet_eager_start *) v->buffer;
        /* set up the packet */
        PACKET_SET_ENVELOPE((&ibv_dev.match_hdr), context_id, tag, len, src_lrank);
        PACKET_SET_CREDITS(((viadev_packet_header *) &ibv_dev.match_hdr), c, subchannel);

        v->shandle = NULL;

        if (search_header_cache(c, 
                    (viadev_packet_eager_start *)&ibv_dev.match_hdr, subchannel)) {
            PACKET_SET_HEADER(h, c, FAST_EAGER_CACHED);
            h->header.fast_eager_size = (unsigned char) len;
            databuf = ((char *) h) + FAST_EAGER_HEADER_SIZE;
            memcpy(databuf, buf, len);
            /* figure out how much data to send in this packet */
            post_fast_rdma_with_completion(c,
                                           FAST_EAGER_HEADER_SIZE + len,
                                           subchannel);

        } else {
            PACKET_SET_ENVELOPE(h, context_id, tag, len, src_lrank);
            PACKET_SET_HEADER(h, c, VIADEV_PACKET_EAGER_START);

            /* Set some of the credit fields */
            h->header.vbuf_credit = ibv_dev.match_hdr.header.vbuf_credit;
            h->header.rdma_credit = ibv_dev.match_hdr.header.rdma_credit;
            h->header.remote_credit = ibv_dev.match_hdr.header.remote_credit;
            
            databuf = ((char *) h) + sizeof(viadev_packet_eager_start);
            memcpy(databuf, buf, len);
            h->bytes_in_this_packet = len;
            /* figure out how much data to send in this packet */
            post_fast_rdma_with_completion(c,
                                           sizeof
                                           (viadev_packet_eager_start) +
                                           len, subchannel);
        }

        /* mark the send as completed */
        SEND_COMPLETE(s);
        return 0;
    }
#endif                          /* of HEADER_CACHING */

    if (rdma_ok) {

        v = &(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]]);
		v->subchannel = subchannel;
		v->desc.subchannel = subchannel;
        h = (viadev_packet_eager_start *) v->buffer;
        databuf = ((char *) h) + sizeof(viadev_packet_eager_start);

        /* figure out how much data to send in this packet */
        h->bytes_in_this_packet = len;

        if (startmax < h->bytes_in_this_packet) {
            h->bytes_in_this_packet = startmax;
        }

        /* copy next chunk of data in to vbuf */
        if (h->bytes_in_this_packet != 0) {
            memcpy(databuf, buf, h->bytes_in_this_packet);
        }

        bytes_sent = h->bytes_in_this_packet;

        /* set up the packet */
        PACKET_SET_HEADER(h, c, VIADEV_PACKET_EAGER_START);
        PACKET_SET_ENVELOPE(h, context_id, tag, len, src_lrank);
        v->shandle = NULL;
		M_PRINT("h->header.id %d, subchannel %d\n",  h->header.id, subchannel);
        post_fast_rdma_with_completion(c,
                                       sizeof(viadev_packet_eager_start) +
                                       h->bytes_in_this_packet, subchannel);
        /* Multiple packet eager send over RDMA */
        while (bytes_sent < len) {

				/* Get the next vbuf */
            v = &(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]]);

            /* figure out where header and data are */
            d = (viadev_packet_eager_next *) VBUF_BUFFER_START(v);
            databuf = ((char *) d) + sizeof(viadev_packet_eager_next);

            /* set up the packet */
            PACKET_SET_HEADER(d, c, VIADEV_PACKET_EAGER_NEXT);

            /* figure out how much data to send in this packet */
            d->bytes_in_this_packet = len - bytes_sent;
            if (nextmax < d->bytes_in_this_packet) {
                d->bytes_in_this_packet = nextmax;
            }

            /* copy next chunk of data in to vbuf */
            memcpy(databuf, ((char *) buf) + bytes_sent,
                    d->bytes_in_this_packet);

            bytes_sent += d->bytes_in_this_packet;

            /* finally send it off */
            v->shandle = NULL;

            /* figure out how much data to send in this packet */
            post_fast_rdma_with_completion(c,
                    sizeof(viadev_packet_eager_next) +
                    d->bytes_in_this_packet, subchannel);
        }

        /* mark the send as completed */
        SEND_COMPLETE(s);

        return 0;
    }
#endif                          /* RDMA_FAST_PATH */

    /* If control reaches here, we have fallen back on 
     * send/recv channel for eager messages */

	D_PRINT("send message, subchannel[%d]\n",subchannel);
    v = get_vbuf();
    h = (viadev_packet_eager_start *) VBUF_BUFFER_START(v);
    databuf = ((char *) h) + sizeof(viadev_packet_eager_start);

	M_PRINT("h->header.id %d, subchannel %d\n",  h->header.id, subchannel);
	 
    /* set up the packet */
    PACKET_SET_HEADER(h, c, VIADEV_PACKET_EAGER_START);
    PACKET_SET_ENVELOPE(h, context_id, tag, len, src_lrank);

    /* figure out how much data to send in this packet */
    h->bytes_in_this_packet = len;

    if (startmax < h->bytes_in_this_packet) {
        h->bytes_in_this_packet = startmax;
    }

    /* copy next chunk of data in to vbuf */

    if (h->bytes_in_this_packet != 0) {
        memcpy(databuf, buf, h->bytes_in_this_packet);
    }

    bytes_sent = h->bytes_in_this_packet;

	assert(subchannel >= 0);

    /* finally send it off */
    vbuf_init_send(v, (sizeof(viadev_packet_eager_start) +
                       h->bytes_in_this_packet), subchannel);

    v->shandle = NULL;

    viadev_post_send(c, v, subchannel);

    while (bytes_sent < len) {

        v = get_vbuf();

        /* figure out where header and data are */
        d = (viadev_packet_eager_next *) VBUF_BUFFER_START(v);
        databuf = ((char *) d) + sizeof(viadev_packet_eager_next);

        /* set up the packet */
        PACKET_SET_HEADER(d, c, VIADEV_PACKET_EAGER_NEXT);

        /* figure out how much data to send in this packet */
        d->bytes_in_this_packet = len - bytes_sent;
        if (nextmax < d->bytes_in_this_packet) {
            d->bytes_in_this_packet = nextmax;
        }

        /* copy next chunk of data in to vbuf */
        memcpy(databuf, ((char *) buf) + bytes_sent,
               d->bytes_in_this_packet);

        bytes_sent += d->bytes_in_this_packet;

		assert(subchannel >= 0);

        /* finally send it off */
        vbuf_init_send(v, (sizeof(viadev_packet_eager_next) +
                           d->bytes_in_this_packet), subchannel);

        v->shandle = NULL;
        viadev_post_send(c, v, subchannel);
    }

    /* NOTE: even with the backlog queue, this assertion should
     * be true in a single threaded implementation.
     * The backlog queue only has something on it if remote_credit == 0.
     * We did a eager send because there were enough remote credits
     * to handle the transfer and these should have been the only
     * sends we posted since that check.
     */
    assert(c->remote_credit[subchannel] >= CREDIT_PRESERVE);

	assert(subchannel >= 0);

    /* mark the send as completed */
    SEND_COMPLETE(s);

    return 0;
}

int viadev_eager_ok(int len, ibv_connection_t * c, int subchannel)
{
    int nvbufs;

	assert(subchannel >= 0);

    nvbufs = viadev_calculate_vbufs_expected(len, VIADEV_PROTOCOL_EAGER);

    if ((int)c->remote_credit[subchannel] - (int)nvbufs >= CREDIT_PRESERVE) {
        return 1;
    }

    return 0;
}
