/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "collutils.h"
#include "viapriv.h"

#ifdef VIADEV_RPUT_SUPPORT

/*
 * Calculates logarithm to the base 2 of a number.
*/

int log_2(int np)
{
    int lgN, t;

    for (lgN = 0, t = 1; t < np; lgN++, t += t);

    return lgN;
}

/*
 * Calculates 2^x
 */
int power_two(int x)
{
    int pow = 1;

    while (x) {
        pow = pow * 2;
        x--;
    }

    return pow;
}

/* Register a block of memory */
void register_mem(void *buf, int len, VIP_MEM_HANDLE * hndl)
{
    DAT_REGION_DESCRIPTION region;
    DAT_VLEN reg_size;
    DAT_VADDR reg_addr;
    DAT_RETURN result;
    /* Register local memory */

    region.for_va = buf;
    result = dat_lmr_create (viadev.nic,
                         DAT_MEM_TYPE_VIRTUAL, region, len,
                         viadev.ptag, DAT_MEM_PRIV_ALL_FLAG,
                         &hndl->hndl, &hndl->lkey,
                         &hndl->rkey, &reg_size,
                         &reg_addr);

    if (result != DAT_SUCCESS) {
       udapl_error_abort(UDAPL_RETURN_ERR, "uDAPL error in register_mem: cannot create lmr");
    }
}

/* Deregister the memory pointed to by 'hndl' */

void deregister_mem(DAT_LMR_HANDLE * hndl)
{
    DAT_RETURN ret;

    ret = dat_lmr_free(*hndl);
    if (ret != DAT_SUCCESS) {
       udapl_error_abort(UDAPL_RETURN_ERR, "uDAPL error in deregister_mem: cannot create lmr");
    }
}

void copy_recv_addr(void *addr, struct Collbuf *coll,
                    VIP_MEM_HANDLE * hndl)
{
    aint_t recv_addr;
    void *dest_addr;

    /* Cast it to integer */
    recv_addr = (aint_t) addr;

    dest_addr = (void *) ((char *) coll->l_coll->buf + COLL_FLAG_SPACE);

    /* Copy the memory handle structure to this memory
     * location */

    *((DAT_RMR_CONTEXT *) dest_addr) = hndl->rkey;

    coll->atoa->buf_hndl_addr = dest_addr;
    coll->atoa->buf_hndl_addr_len = sizeof(DAT_RMR_CONTEXT);

    dest_addr = (void *) ((char *) dest_addr + sizeof(DAT_RMR_CONTEXT));

    /* Set the value to the recv addr */
    *((aint_t *) dest_addr) = recv_addr;

    coll->atoa->rdma_recv_addr = dest_addr;
    coll->atoa->rdma_recv_addr_len = sizeof(aint_t);

    dest_addr = (void *) ((char *) dest_addr + sizeof(aint_t));
    *((char *) dest_addr) = ALLTOALL_RDMA_FLAG;
}

/* A wrap around function for performing a
 * RDMA write for my collectives
 * Contiguous RDMA write
 */
void coll_rdma_write(int dest, void *local_addr,
                     int len, VIP_MEM_HANDLE * local_hndl,
                     void *remote_addr, VIP_MEM_HANDLE * remote_hndl)
{
    VIP_MEM_HANDLE l_hndl;
    VIP_MEM_HANDLE r_hndl;

    vbuf *v = NULL;
    viadev_connection_t *c = NULL;

    /* Grab a descriptor */
    v = get_vbuf();
    assert(v != NULL);

    /* Get connection from 'dest' global rank */
    c = &viadev.connections[dest];
    assert(c != NULL);

    l_hndl.lkey = local_hndl->lkey;
    l_hndl.hndl = local_hndl->hndl;
    l_hndl.rkey = local_hndl->rkey;

    r_hndl.rkey = remote_hndl->rkey;
    r_hndl.hndl = remote_hndl->hndl;
    r_hndl.lkey = remote_hndl->lkey;

    viadev_rput(c, v, local_addr, l_hndl, remote_addr, r_hndl, len);
}

/* Wait until all send descriptors on the connection are
 * available */

void coll_all_send_complete(int dest)
{
    viadev_connection_t *c = NULL;

    c = &viadev.connections[dest];

    while(c->send_wqes_avail != viadev_sq_size) {
        MPID_DeviceCheck(MPID_NOTBLOCKING);
    }
}


void Next_Power_of_Degree(int N, int degree, int *next_pow)
{
    int high = 65536;
    int low = 1;
    while ((high > N) && (low < N)) {
        high /= degree;
        low *= degree;
    }
    if (high <= N) {
        if (high == N)          /* no defect, power of 2! */
            (*next_pow) = N;
        else
            (*next_pow) = high << 1;
    } else {                    /* condition low >= N satisfied */
        if (low == N)           /* no defect, power of 2! */
            (*next_pow) = N;
        else
            (*next_pow) = low;
    }
}


int MPID_rdma_write(void *buf,
                    int len, int src_lrank, int dest_grank, int togle,
                    /*struct MPIR_COMMUNICATOR *comm_ptr */
                    void *remote_address, VIP_MEM_HANDLE * r_handle,
                    VIP_MEM_HANDLE * l_handle)
{
    viadev_connection_t *c = &viadev.connections[dest_grank];
    VIP_MEM_HANDLE remote_handle;
    DAT_RETURN ret4;
    UDAPL_DESCRIPTOR sr;

    remote_handle = *r_handle;

    sr.completion_flag = DAT_COMPLETION_UNSIGNALLED_FLAG;
    sr.op_code = UDAPL_RDMA_WRITE;
    sr.cookie.as_ptr = (void *)-1;
    sr.remote_iov.target_address = (DAT_VADDR) (unsigned long) remote_address;
    sr.remote_iov.rmr_context = r_handle->rkey;
    sr.remote_iov.segment_length = 1;
    sr.local_iov.lmr_context = l_handle->lkey;
    sr.local_iov.virtual_address = (DAT_VADDR) (unsigned long) buf;
    sr.local_iov.segment_length = 1;

    ret4 = dat_ep_post_rdma_write(c->vi, 1, &sr.local_iov,
                               sr.cookie, &sr.remote_iov, sr.completion_flag);
    if (ret4 != DAT_SUCCESS)
       udapl_error_abort(UDAPL_RETURN_ERR,
                            "Error in posting descriptor: MPID_rdma_write\n");


    return MPI_SUCCESS;
}

#endif                          /* VIADEV_RPUT_SUPPORT */
