/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */


/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#if defined(_SMP_)
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#ifdef _SMP_
#ifdef _AFFINITY_
#include <sched.h>
#endif /*_AFFINITY_*/
#endif

#ifdef MAC_OSX
#include <dlfcn.h>
#else
#include <sys/socket.h>
#endif
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "viadev.h"
#include "viaparam.h"
#include "viapriv.h"
#include "vbuf.h"
#include "dreg.h"
#include "process/pmgr_client.h"
#include "mpid_smpi.h"
#include <pthread.h>

#ifdef MAC_OSX
extern int mvapich_need_malloc_init;
extern void *vt_dylib;
extern void *(*vt_malloc_ptr) (size_t);
extern void (*vt_free_ptr) (void *);
#endif

extern viadev_info_t viadev;
DAT_IA_ATTR ia_attr;
DAT_EP_ATTR ep_attr;
DAT_EP_HANDLE temp_ep_handle;

#ifdef _SMP_
#ifdef _AFFINITY_
extern unsigned int viadev_enable_affinity;
#endif
#endif

static void udapl_conn_init (void);
static void udapl_conn_enable (void);
static void udapl_finalize (void);
static void get_address_table (conn_info_t * alladdrs, int len);
static void conn_server (int arg);
#ifdef USE_INLINE
int inline_size = 0;
#endif

#if defined(_SMP_)
static int get_host_id (char *myhostname, int hostname_len);
#endif

#if defined(LAZY_MEM_UNREGISTER) && !defined(MAC_OSX)
/* NOTE: Inclusion of malloc.h this makefile will cause redundant
 * declarations of malloc, free, etc.  If compiling under
 * gcc with -Wredundant-decls AND -Werror, the compilation
 * of this module will fail.  We have turned off
 * -Wredundant-decls for this module only.  If and when
 * this is fixed in glibc, the makefile should be repaired.
 */
#include <malloc.h>
#if defined(MALLOC_HOOK) && !defined(SOLARIS)
/* Initialize our malloc hook */
void (*__malloc_initialize_hook) (void) = mvapich_init_malloc_hook;
#endif

#endif


int
vapi_proc_info (int i, char **host, char **exename)
{
    int pid;
    pid = viadev.pids[i];
    *host = viadev.processes[i];
    *exename = viadev.execname;
    return (pid);
}

DAT_EVD_HANDLE async_evd_handle = DAT_HANDLE_NULL;

int
MPID_VIA_Init (int *argc, char ***argv, int *size, int *rank)
{
    int local_addr_len;
    conn_info_t *alladdrs = NULL;
    conn_info_t local_addr;
    DAT_RETURN ret;

#ifdef _SMP_
    int hostid;
#endif

    /*char debug_file_name[128]; */
    int mypid;

#if defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)
#ifdef LAZY_MEM_UNREGISTER
    /* Set glibc/stdlib malloc options to prevent handing
     * memory back to the system (brk) upon free.
     * Also, dont allow MMAP memory for large allocations.
     * This because we dont unregister memory when refcounts
     * go to zero.
     * Only need these options when RDMA operations are being
     * used.  Otherwise, MVICH layer owns all registered memory
     * and we will only free it if it is de-registered anyway.
     */
#if defined(MALLOC_HOOK) && !defined(SOLARIS)
    create_hash_table ();
#elif !defined(SOLARIS)
    mallopt (M_TRIM_THRESHOLD, -1);
    mallopt (M_MMAP_MAX, 0);
#endif /* MALLOC_HOOK */

#endif /* LAZY_MEM_UNREGISTER */
#endif /* VIADEV_RPUT_SUPPORT */

    /* init the VIA barrier counter */
    viadev.barrier_id = 0;

    /*
     * Find out from the process manager who we are. 
     */

    if (!pmgr_client_init (argc, argv, &viadev.np, &viadev.me,
                           &viadev.global_id, &viadev.processes))
      {
          udapl_error_abort (GEN_EXIT_ERR, "pmgr_client_init failed");
      }

    *size = viadev.np;
    *rank = viadev.me;
    strcpy (viadev.execname, ((const char **) (*argv))[0]);

    /* Now that we know our rank, we should also know
     * our name */
    viadev.my_name = (char *) malloc (sizeof (char) * HOSTNAME_LEN);

    if (NULL == viadev.my_name)
      {
          udapl_error_abort (GEN_EXIT_ERR, "malloc failed");
      }

    if (gethostname (viadev.my_name, sizeof (char) * HOSTNAME_LEN) < 0)
      {
          udapl_error_abort (GEN_EXIT_ERR, "gethostname failed");
      }

    D_PRINT ("size=%d, myrank=%d\n", *size, *rank);


    SET_ORIGINAL_MALLOC_HOOKS;

    /* For both SMP and no-SMP, we have the same address length */
    alladdrs = (conn_info_t *) malloc (viadev.np * sizeof (conn_info_t));

    viadev.remote_service_id =
        (DAT_CONN_QUAL *) malloc (viadev.np * sizeof (DAT_CONN_QUAL));
    viadev.remote_ia_addr =
        (DAT_SOCK_ADDR *) malloc (viadev.np * sizeof (DAT_SOCK_ADDR));
    viadev.pids = (int *) malloc (viadev.np * sizeof (int));
    if (viadev.remote_service_id == NULL || viadev.remote_ia_addr == NULL ||
        viadev.pids == NULL || alladdrs == NULL)
      {
          udapl_error_abort (GEN_EXIT_ERR, "malloc failed");
      }

    mypid = (int) getpid ();
    viadev.pids[viadev.me] = mypid;
    viadev.remote_service_id[viadev.me] = viadev.local_service_id =
        (DAT_CONN_QUAL) mypid + 1024;

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    /* set run-time parameters that have been passed in through
     * environment variables */
    viadev_init_parameters (viadev.np, viadev.me);
#ifdef VIADEV_HAVE_RDMA_LIMIT
    viadev.outstanding_rdmas = 0;
#endif
    /* initializations */
    ret = dat_ia_open (dapl_provider, VIADEV_DEFAULT_ASYNC_EVD_SIZE,
                       &async_evd_handle, &viadev.nic);
    if (ret != DAT_SUCCESS)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot open IA");
      }

    viadev.maxtransfersize = viadev_max_rdma_size;

    /* Get my local addresss by querying the IA */
    ret = dat_ia_query (viadev.nic, &async_evd_handle, DAT_IA_ALL, &ia_attr, 0, /* provider attr mask */
                        NULL);  /* provider attr */
    if (ret != DAT_SUCCESS)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot query IA");
      }

    /* Get my ia_address */
    memcpy (&viadev.local_ia_addr, ia_attr.ia_address_ptr,
            sizeof (DAT_SOCK_ADDR));

    /* we will create EPs first, exchange addresses and 
     * then connect them */

    udapl_conn_init ();

    local_addr.ia_addr = viadev.local_ia_addr;
    local_addr.service_id = viadev.local_service_id;

#ifdef _SMP_
    hostid = get_host_id (viadev.my_name, HOSTNAME_LEN);
    local_addr.host_id = hostid;

#else /* put a useless integer for no-SMP mode */
    local_addr.host_id = viadev.me;
#endif

    local_addr_len = sizeof (conn_info_t);
    pmgr_exchange_addresses (&local_addr, local_addr_len, alladdrs,
                             viadev.pids);

    get_address_table (alladdrs, local_addr_len);
    udapl_conn_enable ();

    free (alladdrs);

#if defined(VIADEV_RPUT_SUPPORT)
    dreg_init (viadev.nic, viadev.ptag);
#endif
    D_PRINT ("viainit complete\n");

#ifdef MAC_OSX
    mvapich_malloc_init ();
    mvapich_need_malloc_init = 0;
#endif
    return DAT_SUCCESS;
}

/*
 * Set up all communication and put VIs in the VI array in viadev
 * The algorithm is as follows (may change when M-VIA supports peer-to-peer
 * connections):
 *   . use client-server routines to establish connections
 *   . for each pair, the lower-numbered process acts as the server
 *   . we should be careful not to establish hotspots. E.g., if
 *     every process connects with others in order 0, 1, 2. We
 *     should also be careful not to serialize. 
 *     there is a nice fancy way to do this, with a communication pattern
 *     essentially the same as in an FFT. 
 *   . However, it is a bit tricky to get right so for now 
 *     do the stupid serialized version (xxx):
 *        for (i = 0; i < me; i++) connect(i)
 *        for (i = me+1; i <= np; i++) accept(i)
 *   
 *  Note that we do not set up a connection to ourselves 
 *
 */

void
MPID_Wtime (double *t)
{
    struct timeval tv;
    static int initialized = 0;
    static int sec_base;
    gettimeofday (&tv, NULL);
    if (!initialized)
      {
          sec_base = tv.tv_sec;
          initialized = 1;
      }
    *t = (double) (tv.tv_sec - sec_base) + (double) tv.tv_usec * 1.0e-6;
}

void
MPID_Wtick (double *t)
{
    *t = VIADEV_WTICK;
}

void
MPID_VIA_Finalize (void)
{

    D_PRINT ("entering MPID_VIA_Finalize ...\n");

    /* VAPI finalize called */
    udapl_finalize ();

    /* free connection array space */
    if (viadev.connections)
      {
          free (viadev.connections);
      }

    /* let the process manager reclaim data structures 
     * before termination */

    pmgr_finalize ();

    if (viadev.remote_service_id)
      {
          free (viadev.remote_service_id);
      }

    if (viadev.remote_ia_addr)
      {
          free (viadev.remote_ia_addr);
      }

    if (viadev.my_name)
      {
          free (viadev.my_name);
      }

    if (viadev.pids)
      {
          free (viadev.pids);
      }

    if (viadev.processes)
      {
          free (viadev.processes);
      }

#ifdef _SMP_
    if (smpi.local_nodes)
      {
          free (smpi.local_nodes);
      }
    if (smpi.l2g_rank) {
        free(smpi.l2g_rank);
    }

#endif

    D_PRINT ("exiting MPID_VIA_Finalize...\n");

}

static void
conn_server (int arg)
{
    DAT_RETURN status;
    DAT_EVENT event;
    DAT_COUNT count;
    DAT_CR_ARRIVAL_EVENT_DATA cr_stat;
    DAT_CR_PARAM cr_param;
    int i, num = 0;

    while (num < arg)
      {
          status =
              dat_evd_wait (viadev.creq_cq_hndl,
                            viadev_default_time_out, 1, &event, &count);
          if (status != DAT_SUCCESS)
            {
                continue;
            }
          if (event.event_number == DAT_CONNECTION_REQUEST_EVENT)
            {
                cr_stat = event.event_data.cr_arrival_event_data;
                if ((cr_stat.conn_qual != viadev.local_service_id) ||
                    (cr_stat.sp_handle.psp_handle != viadev.psp_hndl))
                  {
                      udapl_error_abort (GEN_EXIT_ERR,
                                         "error in conn_server");
                  }

                status =
                    dat_cr_query (cr_stat.cr_handle,
                                  DAT_CR_FIELD_PRIVATE_DATA, &cr_param);
                if (status != DAT_SUCCESS)
                  {
                      udapl_error_abort (GEN_EXIT_ERR,
                                         "Error: fail to get private data from the connection request");
                  }

                i = *(int *) cr_param.private_data;
                status =
                    dat_cr_accept (cr_stat.cr_handle, viadev.qp_hndl[i], 0,
                                   (DAT_PVOID) NULL);
                if (status != DAT_SUCCESS)
                  {
                      udapl_error_abort (GEN_EXIT_ERR,
                                         "error accepting request");
                  }
                num++;
            }                   /* if */
      }                         /* while */
}

/* internal functions used for initialization 
 * and finalization by uDAPL */
static void
udapl_conn_init ()
{
    DAT_EP_ATTR ep_attr;
    DAT_RETURN ret;
    int i;
    DAT_EP_PARAM param;

    /* allocate completion queues, protection handle before qp */
    ret = dat_pz_create (viadev.nic, &viadev.ptag);
    if (ret != DAT_SUCCESS)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot create pz");
      }

    ret = dat_evd_create (viadev.nic,
                          MIN (DEFAULT_MAX_CQ_SIZE,
                               ia_attr.max_evd_qlen), DAT_HANDLE_NULL,
                          DAT_EVD_DTO_FLAG, &viadev.cq_hndl);
    if (ret != DAT_SUCCESS)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot create evd");
      }

    ret = dat_evd_create (viadev.nic,
                          MIN (DEFAULT_MAX_CQ_SIZE,
                               ia_attr.max_evd_qlen), DAT_HANDLE_NULL,
                          DAT_EVD_CR_FLAG, &viadev.creq_cq_hndl);
    if (ret != DAT_SUCCESS)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot create evd");
      }

    ret = dat_evd_create (viadev.nic,
                          MIN (DEFAULT_MAX_CQ_SIZE,
                               ia_attr.max_evd_qlen), DAT_HANDLE_NULL,
                          DAT_EVD_CONNECTION_FLAG, &viadev.conn_cq_hndl);
    if (ret != DAT_SUCCESS)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot create evd");
      }

    while (1) {
          ret = dat_psp_create (viadev.nic,
                          viadev.local_service_id,
                          viadev.creq_cq_hndl,
                          DAT_PSP_CONSUMER_FLAG, &viadev.psp_hndl);
          if (ret == DAT_SUCCESS)
                  break;

          viadev.local_service_id += 100; 
          viadev.remote_service_id[viadev.me] = viadev.local_service_id;
    }

    SET_ORIGINAL_MALLOC_HOOKS;

    viadev.qp_hndl =
        (DAT_EP_HANDLE *) malloc (sizeof (DAT_EP_HANDLE) * viadev.np);
    if (viadev.qp_hndl == NULL)
      {
          udapl_error_abort (GEN_EXIT_ERR, "cannot allocate memory for ep");
      }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    if (strcmp (dapl_provider, "ccil") == 0)
      {
          ret = dat_ep_create (viadev.nic,
                               viadev.ptag,
                               viadev.cq_hndl,
                               viadev.cq_hndl,
                               viadev.conn_cq_hndl, NULL, &temp_ep_handle);
          ret =
              dat_ep_query (temp_ep_handle, DAT_EP_FIELD_EP_ATTR_ALL, &param);

          ep_attr = param.ep_attr;
          ep_attr.max_rdma_read_iov = 1;
          ep_attr.max_rdma_write_iov = 4;
          ep_attr.max_recv_dtos = viadev_max_wqe;
          ep_attr.max_request_dtos = viadev_max_wqe;
      }
    else if (strcmp (dapl_provider, "ib0") == 0)
      {
          ep_attr.service_type = DAT_SERVICE_TYPE_RC;
          ep_attr.max_mtu_size = DAPL_GEN2_MAX_MSG_SIZE;
          ep_attr.max_message_size = DAPL_GEN2_MAX_MSG_SIZE;
          ep_attr.max_rdma_size = 0;
          ep_attr.qos = DAT_QOS_BEST_EFFORT;
          ep_attr.recv_completion_flags = DAT_COMPLETION_DEFAULT_FLAG;
          ep_attr.request_completion_flags = DAT_COMPLETION_UNSIGNALLED_FLAG;
          ep_attr.max_recv_dtos = viadev_max_wqe;
          ep_attr.max_request_dtos = viadev_max_wqe;
          ep_attr.max_recv_iov = 4;
          ep_attr.max_request_iov = 4;
          ep_attr.max_rdma_read_in = DAPL_DEFAULT_MAX_RDMA_IN;
          ep_attr.max_rdma_read_out = DAPL_DEFAULT_MAX_RDMA_OUT;

          ep_attr.ep_transport_specific_count = 0;
          ep_attr.ep_transport_specific = NULL;
          ep_attr.ep_provider_specific_count = 0;
          ep_attr.ep_provider_specific = NULL;

          ep_attr.max_rdma_read_iov = 0;
          ep_attr.max_rdma_write_iov = 0;
          ep_attr.srq_soft_hw = 0;
      }
    else
      {
          ep_attr.service_type = DAT_SERVICE_TYPE_RC;
          ep_attr.max_mtu_size = viadev_default_mtu;
          ep_attr.max_rdma_size = ia_attr.max_rdma_size;
          ep_attr.qos = DAT_QOS_BEST_EFFORT;
          ep_attr.recv_completion_flags = DAT_COMPLETION_DEFAULT_FLAG;
          ep_attr.request_completion_flags = DAT_COMPLETION_UNSIGNALLED_FLAG;
          ep_attr.max_recv_dtos =
              MIN (viadev_max_wqe, ia_attr.max_dto_per_ep);
          ep_attr.max_request_dtos =
              MIN (viadev_max_wqe, ia_attr.max_dto_per_ep);
          ep_attr.max_recv_iov =
              MIN (viadev_default_max_sg_list,
                   ia_attr.max_iov_segments_per_dto);
          ep_attr.max_request_iov =
              MIN (viadev_default_max_sg_list,
                   ia_attr.max_iov_segments_per_dto);
          ep_attr.max_rdma_read_in = DAPL_DEFAULT_MAX_RDMA_IN;
          ep_attr.max_rdma_read_out = DAPL_DEFAULT_MAX_RDMA_OUT;

          ep_attr.ep_transport_specific_count = 0;
          ep_attr.ep_transport_specific = NULL;
          ep_attr.ep_provider_specific_count = 0;
          ep_attr.ep_provider_specific = NULL;

      }

    for (i = 0; i < viadev.np; i++)
      {

          if (i == viadev.me)
            {
                continue;
            }

          ret = dat_ep_create (viadev.nic,
                               viadev.ptag,
                               viadev.cq_hndl,
                               viadev.cq_hndl,
                               viadev.conn_cq_hndl,
                               &ep_attr, &(viadev.qp_hndl[i]));

          if (ret != DAT_SUCCESS)
            {
                udapl_error_abort (GEN_EXIT_ERR, "cannot create ep");
            }

          D_PRINT ("[%d] EP[%d] created\n", viadev.me, i);
      }

}

static void
udapl_conn_enable (void)
{
    int num_connected = 0;
    DAT_COUNT count;

#if defined(RDMA_FAST_PATH)
    VIP_MEM_HANDLE mem_handle;
    DAT_REGION_DESCRIPTION region;
    DAT_VLEN reg_size;
    DAT_VADDR reg_addr;
    DAT_RETURN result;
#ifdef SINGLE_RDMA_BUFFER
    int RDMA_buf_size = 0;
    char *RDMA_recv_buf = NULL;
    char *RDMA_send_buf = NULL;
    int pagesize;
    int rank_cnt;
#endif
#endif

    int i, j;
    DAT_RETURN ret;
    DAT_EVENT event;
    int step;

    for (step = 1; step < viadev.np; step++)
      {
          j = 0;
          if (viadev.me - step >= 0)
            {
                i = viadev.me - step;
                j++;
                ret = dat_ep_connect (viadev.qp_hndl[i], &viadev.remote_ia_addr[i], viadev.remote_service_id[i], DAT_TIMEOUT_INFINITE, sizeof (int),    /* private_data_size */
                                      (DAT_PVOID) & viadev.me,  /* private_data */
                                      DAT_QOS_BEST_EFFORT,      /* QoS */
                                      DAT_CONNECT_DEFAULT_FLAG);
                if (ret != DAT_SUCCESS)
                  {
                      udapl_error_abort (GEN_EXIT_ERR, "ep_connect error");
                  }
            }

          if (viadev.me + step < viadev.np)
            {
                conn_server (1);
                j++;
            }

          while (num_connected < j)
            {
                ret =
                    dat_evd_wait (viadev.conn_cq_hndl,
                                  DAT_TIMEOUT_INFINITE, 1, &event, &count);

                if (ret != DAT_SUCCESS)
                    continue;

                switch (event.event_number)
                  {
                  case DAT_CONNECTION_EVENT_TIMED_OUT:
                      break;

                  case DAT_CONNECTION_EVENT_ESTABLISHED:
                      num_connected++;
                      break;

                  default:
                      break;
                  }
            }
          num_connected = 0;
          pmgr_barrier (0);
      }

    /****************prepost receive descriptors ****************/

    D_PRINT ("viadev.nic=%d ptag=%d\n", viadev.nic, viadev.ptag);

    allocate_vbufs (viadev.nic, viadev.ptag, viadev_vbuf_pool_size);

    SET_ORIGINAL_MALLOC_HOOKS;

    /* initialize connection parameters */
    viadev.connections = (viadev_connection_t *)
        malloc (sizeof (viadev_connection_t) * viadev.np);

    if (NULL == viadev.connections)
      {
          udapl_error_abort (GEN_EXIT_ERR,
                             "Unable to malloc connection structures");
      }
#if defined(RDMA_FAST_PATH) && defined(SINGLE_RDMA_BUFFER)
    pagesize = getpagesize ();
    RDMA_buf_size =
        2 * (sizeof (struct vbuf) * viadev_num_rdma_buffer *
             (viadev.np - 1) + 2 * pagesize);
    viadev.RDMA_bufs = (char *) malloc (RDMA_buf_size);

    if (NULL == viadev.RDMA_bufs)
      {
          udapl_error_abort (GEN_EXIT_ERR,
                             "Unable to allocate Fast RDMA buffers");
      }
    memset (viadev.RDMA_bufs, 0, RDMA_buf_size);

    /* register RDMA buffers */
    region.for_va = (DAT_PVOID) viadev.RDMA_bufs;
    result = dat_lmr_create (viadev.nic,
                             DAT_MEM_TYPE_VIRTUAL, region,
                             RDMA_buf_size,
                             viadev.ptag, DAT_MEM_PRIV_ALL_FLAG,
                             &mem_handle.handle,
                             &mem_handle.lkey,
                             &mem_handle.rkey, &reg_size, &reg_addr);

    if (result != DAT_SUCCESS)
        udapl_error_abort (GEN_ASSERT_ERR, "cannot create lmr");

    viadev.RDMA_buf_hndl = mem_handle;

    /* page aligned buffers for RDMA send and recv */
    RDMA_send_buf =
        (char *) (((unsigned long) viadev.RDMA_bufs + pagesize - 1) /
                  pagesize * pagesize);
    RDMA_recv_buf =
        (char *) (RDMA_send_buf +
                  (RDMA_buf_size / 2 + pagesize - 1) / pagesize * pagesize);

    D_PRINT ("RDMA_send_buf:%p RDMA_recv_buf:%p, size:%d\n", RDMA_send_buf,
             RDMA_recv_buf, RDMA_buf_size);

    rank_cnt = 0;
#endif /* RDMA_FAST_PATH && SINGLE_RDMA_BUFFER */


    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    for (i = 0; i < viadev.np; i++)
      {

          viadev_connection_t *c = &viadev.connections[i];
          c->global_rank = i;
          c->next_packet_expected = 1;
          c->next_packet_tosend = 1;
          c->shandle_head = NULL;
          c->shandle_tail = NULL;
          c->rhandle = NULL;
          c->nextflow = NULL;
          c->inflow = 0;
          c->preposts = 0;
          c->send_wqes_avail = viadev_sq_size;
          c->ext_sendq_head = NULL;
          c->ext_sendq_tail = NULL;
          c->rendezvous_packets_expected = 0;
          viadev_init_backlog_queue (c);

#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
          /* init cached packet header */
          memset (&(c->cached_outgoing), 0,
                  sizeof (viadev_packet_eager_start));
          memset (&(c->cached_incoming), 0,
                  sizeof (viadev_packet_eager_start));
          c->cached_hit = c->cached_miss = 0;
          c->cached_incoming.header.type = c->cached_outgoing.header.type
              = FAST_EAGER_CACHED;
#endif
#endif

          if (i != viadev.me)
            {
                c->vi = viadev.qp_hndl[i];

                /* prepost receives */
                for (j = 0; j < viadev_initial_prepost_depth; j++)
                  {
                      PREPOST_VBUF_RECV (c);
                  }
                D_PRINT ("preposted %d receive desc\n",
                         viadev_initial_prepost_depth);

                c->remote_credit = viadev_initial_credits;
                c->remote_cc = viadev_initial_credits;
                c->local_credit = 0;
                c->preposts = viadev_initial_prepost_depth;
                c->initialized = (viadev_initial_prepost_depth
                                  ==
                                  viadev_prepost_depth +
                                  viadev_prepost_noop_extra);
            }
          else
            {
                /* these should be ignored for self-communication */
                c->vi = DAT_HANDLE_NULL;
                c->remote_credit = 0;
                c->remote_cc = 0;
                c->local_credit = 0;
            }

#if defined(RDMA_FAST_PATH)

          if (i != viadev.me)
            {

                /* initialize revelant fields */
                c->rdma_credit = 0;
                c->num_no_completion = 0;

                if (viadev_num_rdma_buffer)
                  {

#ifdef SINGLE_RDMA_BUFFER
                      c->RDMA_send_buf =
                          (struct vbuf *) ((unsigned long) RDMA_send_buf +
                                           rank_cnt * sizeof (struct vbuf) *
                                           viadev_num_rdma_buffer);
                      c->RDMA_recv_buf =
                          (struct vbuf *) ((unsigned long) RDMA_recv_buf +
                                           rank_cnt * sizeof (struct vbuf) *
                                           viadev_num_rdma_buffer);
                      rank_cnt++;
#else

                      SET_ORIGINAL_MALLOC_HOOKS;

#define RDMA_ALIGNMENT 64
#define RDMA_ALIGNMENT_OFFSET (64)
                      /* allocate RDMA buffers */
                      c->RDMA_send_buf_orig =
                          malloc (sizeof (struct vbuf) *
                                  viadev_num_rdma_buffer +
                                  2 * RDMA_ALIGNMENT);
                      c->RDMA_recv_buf_orig =
                          malloc (sizeof (struct vbuf) *
                                  viadev_num_rdma_buffer +
                                  2 * RDMA_ALIGNMENT);

                      /* align vbuf->buffer to 64 byte boundary */
                      c->RDMA_send_buf =
                          (struct vbuf
                           *) ((unsigned long) (c->RDMA_send_buf_orig) /
                               RDMA_ALIGNMENT * RDMA_ALIGNMENT +
                               RDMA_ALIGNMENT + RDMA_ALIGNMENT_OFFSET);
                      c->RDMA_recv_buf =
                          (struct vbuf
                           *) ((unsigned long) (c->RDMA_recv_buf_orig) /
                               RDMA_ALIGNMENT * RDMA_ALIGNMENT +
                               RDMA_ALIGNMENT + RDMA_ALIGNMENT_OFFSET);

                      /* zero buffers */
                      memset (c->RDMA_send_buf, 0,
                              sizeof (struct vbuf) * viadev_num_rdma_buffer);
                      memset (c->RDMA_recv_buf, 0,
                              sizeof (struct vbuf) * viadev_num_rdma_buffer);

                      SAVE_MALLOC_HOOKS;
                      SET_MVAPICH_MALLOC_HOOKS;
#endif

                      /* set pointers */
                      c->phead_RDMA_send = 0;
                      c->ptail_RDMA_send = viadev_num_rdma_buffer - 1;

                      c->p_RDMA_recv = 0;
                      c->p_RDMA_recv_tail = viadev_num_rdma_buffer - 1;

#ifndef SINGLE_RDMA_BUFFER
                      /* register RDMA buffers */
                      region.for_va = (DAT_PVOID) (c->RDMA_send_buf);
                      result = dat_lmr_create (viadev.nic,
                                               DAT_MEM_TYPE_VIRTUAL, region,
                                               sizeof (struct vbuf) *
                                               viadev_num_rdma_buffer,
                                               viadev.ptag,
                                               DAT_MEM_PRIV_ALL_FLAG,
                                               &mem_handle.hndl,
                                               &mem_handle.lkey,
                                               &mem_handle.rkey, &reg_size,
                                               &reg_addr);

                      if (result != DAT_SUCCESS)
                          udapl_error_abort (GEN_ASSERT_ERR,
                                             "cannot create lmr");

#endif

                      c->RDMA_send_buf_hndl = mem_handle;

#ifndef SINGLE_RDMA_BUFFER
                      region.for_va = (DAT_PVOID) (c->RDMA_recv_buf);
                      result = dat_lmr_create (viadev.nic,
                                               DAT_MEM_TYPE_VIRTUAL, region,
                                               sizeof (struct vbuf) *
                                               viadev_num_rdma_buffer,
                                               viadev.ptag,
                                               DAT_MEM_PRIV_ALL_FLAG,
                                               &mem_handle.hndl,
                                               &mem_handle.lkey,
                                               &mem_handle.rkey, &reg_size,
                                               &reg_addr);

                      if (result != DAT_SUCCESS)
                          udapl_error_abort (GEN_ASSERT_ERR,
                                             "cannot create lmr");

#endif

                      c->RDMA_recv_buf_hndl = mem_handle;
                  }
                else
                  {
                      c->phead_RDMA_send = 0;
                      c->ptail_RDMA_send = 0;
                      c->p_RDMA_recv = 0;
                      c->p_RDMA_recv_tail = 0;
                  }
                /* remote address and handle not available yet */
                c->remote_address_received = 0;

            }
#endif /* RDMA_FAST_PATH */

      }

    /* call barrier to make sure all preposting has been done */
    D_PRINT ("about to call pmgr_barrier...\n");

    /* call barrier to make sure all preposting has been done 
     * to gurantee all have preposted requests before real comm.
     */
    i = pmgr_barrier (1);

    D_PRINT ("process manager barrier returns %d\n", i);

    if (viadev_num_rdma_buffer == 0)
        return;

#if defined(RDMA_FAST_PATH)
    /* inform others about the receive 
     * RDMA buffer address and handle */

    for (i = 0; i < viadev.np; i++)
      {
          vbuf *v;
          viadev_packet_rdma_address *h;
          viadev_connection_t *c = &viadev.connections[i];

          if (i == viadev.me)
            {
                continue;
            }

          v = get_vbuf ();
          h = (viadev_packet_rdma_address *) VBUF_BUFFER_START (v);

          /* set up the packet */
          PACKET_SET_HEADER (h, c, VIADEV_RDMA_ADDRESS);
          h->envelope.src_lrank = i;

          /* set address and handle */
          h->RDMA_address = c->RDMA_recv_buf;
          h->RDMA_hndl = c->RDMA_recv_buf_hndl;

          /* finally send it off */
          vbuf_init_send (v, sizeof (viadev_packet_rdma_address));

          viadev_post_send (c, v);
      }

    /* get all the address */
    for (i = 0; i < viadev.np; i++)
      {
          viadev_connection_t *c = &viadev.connections[i];

          if (i == viadev.me)
            {
                continue;
            }

          while (!c->remote_address_received)
            {
                MPID_DeviceCheck (1);
            }
      }

#endif /* RDMA_FAST_PATH */

}

#if defined(RDMA_FAST_PATH)
void process_remote_rdma_address (viadev_connection_t * c,
                                  VIP_MEM_HANDLE remote_hndl,
                                  struct vbuf *remote_address);

void
process_remote_rdma_address (viadev_connection_t * c,
                             VIP_MEM_HANDLE remote_hndl,
                             struct vbuf *remote_address)
{
    int i;

    c->remote_address_received = 1;
    c->remote_RDMA_buf_hndl = remote_hndl;
    c->remote_RDMA_buf = remote_address;

    for (i = 0; i < viadev_num_rdma_buffer; i++)
      {

          /* set global_rank */
          c->RDMA_send_buf[i].grank = c->global_rank;
          c->RDMA_recv_buf[i].grank = c->global_rank;

          c->RDMA_recv_buf[i].desc.next = NULL;
          c->RDMA_recv_buf[i].desc.op_code = UDAPL_READ;

          /* fill out the descriptors */
          c->RDMA_send_buf[i].padding = FREE_FLAG;

          /* Mark initialization state of RDMA buffers to be busy */
          c->RDMA_recv_buf[i].padding = BUSY_FLAG;

          /* this field to be filled out later */
          c->RDMA_send_buf[i].desc.local_iov.segment_length = 0;
          c->RDMA_send_buf[i].desc.remote_iov.segment_length = 0;

          c->RDMA_send_buf[i].desc.local_iov.virtual_address =
              (DAT_VADDR) (unsigned long) (&(c->RDMA_send_buf[i]));
          c->RDMA_send_buf[i].desc.local_iov.lmr_context =
              c->RDMA_send_buf_hndl.lkey;

          c->RDMA_send_buf[i].desc.remote_iov.rmr_context =
              c->remote_RDMA_buf_hndl.rkey;
          c->RDMA_send_buf[i].desc.remote_iov.target_address =
              (DAT_VADDR) (unsigned long) (&(c->remote_RDMA_buf[i]));

          c->RDMA_send_buf[i].desc.completion_flag =
              DAT_COMPLETION_UNSIGNALLED_FLAG;
          c->RDMA_send_buf[i].desc.op_code = UDAPL_RDMA_WRITE;
          c->RDMA_send_buf[i].desc.cookie.as_ptr =
              (DAT_PVOID) & (c->RDMA_send_buf[i]);
      }
}

#endif

static void
udapl_finalize ()
{

    DAT_RETURN ret;
    int i;

    D_PRINT ("Entering udapl_finalize()\n");

#if defined(RDMA_FAST_PATH)

    if (viadev_num_rdma_buffer)
      {
          /* unpin rdma buffers */
#if defined(SINGLE_RDMA_BUFFER)
          ret = dat_lmr_free (viadev.RDMA_buf_hndl.hndl);
          if (DAT_SUCCESS != ret)
            {
                udapl_error_abort (UDAPL_RETURN_ERR,
                                   "could not unpin Fast RDMA buffer");
            }

          if (viadev.RDMA_bufs)
              free (viadev.RDMA_bufs);
          viadev.RDMA_bufs = NULL;
#else
          for (i = 0; i < viadev.np; i++)
            {
                viadev_connection_t *c = &viadev.connections[i];

                if (i == viadev.me)
                  {
                      continue;
                  }

                ret = dat_lmr_free (c->RDMA_send_buf_hndl.hndl);
                if (DAT_SUCCESS != ret)
                  {
                      udapl_error_abort (UDAPL_RETURN_ERR,
                                         "could not unpin send rdma buffer");
                  }

                ret = dat_lmr_free (c->RDMA_recv_buf_hndl.hndl);
                if (DAT_SUCCESS != ret)
                  {
                      udapl_error_abort (UDAPL_RETURN_ERR,
                                         "could not unpin recv rdma buffer");
                  }

                /* free the buffers */
                free (c->RDMA_send_buf_orig);
                free (c->RDMA_recv_buf_orig);
            }
#endif
      }
#endif

    /* Disconnect all the EPs now */
    for (i = 0; i < viadev.np; i++)
      {
          if (i == viadev.me)
              continue;
          ret =
              dat_ep_disconnect (viadev.qp_hndl[i], DAT_CLOSE_GRACEFUL_FLAG);
          if (DAT_SUCCESS != ret)
            {
                udapl_error_abort (UDAPL_RETURN_ERR, "fail to disconnect EP");
            }
      }

    /* Be the server or the client, there will be a disconnection event */
    {
        int num_disconnected = 0;
        DAT_EVENT event;
        DAT_COUNT count;
        while (num_disconnected < viadev.np - 1)
          {
              if ((DAT_SUCCESS ==
                   dat_evd_wait (viadev.conn_cq_hndl,
                                 DAT_TIMEOUT_INFINITE, 1, &event, &count))
                  && (event.event_number ==
                      DAT_CONNECTION_EVENT_DISCONNECTED))
                {
                    num_disconnected++;
                }
          }                     /* while */
    }                           /* disconnect event block */

    if (NULL == viadev.qp_hndl)
      {
          udapl_error_abort (GEN_EXIT_ERR, "Null queue pair handle");
      }

    if (strcmp (dapl_provider, "ccil") == 0)
      {
          dat_ep_free (temp_ep_handle);
      }

    for (i = 0; i < viadev.np; i++)
      {
          if (i == viadev.me)
              continue;

          ret = dat_ep_free (viadev.qp_hndl[i]);
          if (DAT_SUCCESS != ret)
            {
                udapl_error_abort (UDAPL_RETURN_ERR, "could not destroy EP");
            }
      }
    free (viadev.qp_hndl);

    ret = dat_psp_free (viadev.psp_hndl);
    if (DAT_SUCCESS != ret)
      {
          udapl_error_abort (UDAPL_RETURN_ERR, "could not destroy PSP");
      }

    deallocate_vbufs (viadev.nic);

    ret = dat_evd_free (viadev.cq_hndl);
    if (DAT_SUCCESS != ret)
      {
          udapl_error_abort (UDAPL_RETURN_ERR, "could not destroy DTO EVD");
      }

    ret = dat_evd_free (viadev.creq_cq_hndl);
    if (DAT_SUCCESS != ret)
      {
          udapl_error_abort (UDAPL_RETURN_ERR, "could not destroy creq EVD");
      }

    ret = dat_evd_free (viadev.conn_cq_hndl);
    if (DAT_SUCCESS != ret)
      {
          udapl_error_abort (UDAPL_RETURN_ERR, "could not destroy conn EVD");
      }

    /* unregister all user buffer registration */
    while (dreg_evict ());

    ret = dat_pz_free (viadev.ptag);

    /* to release all resources */
    ret = dat_ia_close (viadev.nic, DAT_CLOSE_GRACEFUL_FLAG);
    if (ret != DAT_SUCCESS)
      {
          ret = dat_ia_close (viadev.nic, DAT_CLOSE_ABRUPT_FLAG);
          if (ret != DAT_SUCCESS)
              udapl_error_abort (UDAPL_RETURN_ERR, "could not close IA");
      }

}

#if defined(_SMP_)
/* gethostname does return a NULL-terminated string if
 * the hostname length is less than "HOSTNAME_LEN". Otherwise,
 * it is unspecified.
 */
static int
get_host_id (char *myhostname, int hostname_len)
{
    int host_id = 0;
    struct hostent *hostent;

    hostent = gethostbyname (myhostname);
    if (hostent == NULL)
      {
          udapl_error_abort (GEN_EXIT_ERR,
                             "Null value returned by gethostbyname");
      }

    host_id = ((struct in_addr *) hostent->h_addr_list[0])->s_addr;

    return host_id;
}

#endif /* SMP */

static void
get_address_table (conn_info_t * alladdrs, int local_addr_len)
{
    int i;

#ifdef _SMP_
    int *hostnames_j = NULL;
    int *smpi_ptr = NULL;
    int j;
#ifdef _AFFINITY_
    long N_CPUs_online;
    unsigned long affinity_mask=1;
    unsigned long affinity_mask_len=sizeof(affinity_mask);

    /*Get number of CPU on machine*/
    if ( (N_CPUs_online=sysconf (_SC_NPROCESSORS_ONLN)) < 1 ) {
            perror("sysconf");
    }
#endif /*_AFFINITY_*/

    SET_ORIGINAL_MALLOC_HOOKS;

    hostnames_j = (int *) malloc (viadev.np * sizeof (int));
    smpi.local_nodes =
        (unsigned int *) malloc (viadev.np * sizeof (unsigned int));
    smpi_ptr = (unsigned int *) malloc(viadev.np * sizeof(int));


    if (hostnames_j == NULL || smpi.local_nodes == NULL || smpi_ptr == NULL) {
          udapl_error_abort (GEN_EXIT_ERR,
                             "malloc: in ib_rank_lid_table for SMP");
    }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;
#endif

    for (i = 0; i < viadev.np; i++)
      {
          viadev.remote_ia_addr[i] = alladdrs[i].ia_addr;
          viadev.remote_service_id[i] = alladdrs[i].service_id;
          D_PRINT ("service_id[%d]=%d\n", i,
                   (int) viadev.remote_service_id[i]);

#ifdef _SMP_
          hostnames_j[i] = alladdrs[i].host_id;
#endif
      }

#ifdef _SMP_

    smpi.only_one_device = 1;
    smpi.num_local_nodes = 0;
    for (j = 0; j < viadev.np; j++)
      {
          if (hostnames_j[viadev.me] == hostnames_j[j])
            {
                if (j == viadev.me)
                  {
                      smpi.my_local_id = smpi.num_local_nodes;
#ifdef _AFFINITY_
                      if ( viadev_enable_affinity > 0 ){
                              affinity_mask=1<<(smpi.my_local_id%N_CPUs_online);
                              if ( sched_setaffinity(0,affinity_mask_len,&affinity_mask)<0 ) {
                                     perror("sched_setaffinity");
                              }
                      }
#endif /* _AFFINITY_ */
                  }
                smpi.local_nodes[j] = smpi.num_local_nodes;
                smpi_ptr[smpi.num_local_nodes] = j;
                smpi.num_local_nodes++;
            }
          else
            {
                smpi.only_one_device = 0;
                smpi.local_nodes[j] = -1;
            }
      }
    smpi.l2g_rank = (unsigned int *)malloc(smpi.num_local_nodes * sizeof(int));
    if (smpi.l2g_rank == NULL) {
        udapl_error_abort(GEN_EXIT_ERR,
                            "malloc: in ib_rank_lid_table for SMP");
    }
    for (j = 0; j < smpi.num_local_nodes; j++) {
        smpi.l2g_rank[j] = smpi_ptr[j];
    }
    free(smpi_ptr);
    free (hostnames_j);
#endif
}

#ifdef PENTIUM_MHZ
double
get_us (void)
{
    unsigned long long now;
    __asm __volatile (".byte 0x0f, 0x31":"=A" (now));
    return (double) ((double) now / (PENTIUM_MHZ));
}
#else
double
get_us (void)
{
    struct timeval t;
    gettimeofday (&t, NULL);
    return (double) t.tv_sec * (double) 1e6 + (double) t.tv_usec;
}
#endif

#ifdef MAC_OSX
void
mvapich_malloc_init ()
{
    /* This dylib is Mac OS X specific. For other unix platforms
       replace libc.dylib with libc.so */

    if (vt_malloc_ptr != NULL)
        return;
    vt_dylib = dlopen ("/usr/lib/libc.dylib", RTLD_LAZY);

    if (vt_dylib == NULL)
      {
          printf ("Open error: Unable to open dynamic library ...\n");
          exit (1);
      }
#if defined(VIADEV_RPUT_SUPPORT) && defined(LAZY_MEM_UNREGISTER)

    vt_malloc_ptr = dlsym (vt_dylib, "malloc");

    if (vt_malloc_ptr == NULL)
      {
          printf
              ("Resolution error: Unable to locate symbol for malloc ...\n");
          exit (1);
      }

    vt_free_ptr = dlsym (vt_dylib, "free");

    if (vt_free_ptr == NULL)
      {
          printf ("Resolution error: Unable to locate symbol for free ...\n");
          exit (1);
      }
#endif
}
#endif
