/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "viaparam.h"
#include "vbuf.h"
#include "viapacket.h"
#include "udapl_const.h"

/*
 * ==============================================================
 * Initialize global parameter variables to default values
 * ==============================================================
 */

/* 
 *  Which HCA device is used?
 *  (1) If a user specifies "VIADEV_DEVICE" environmental variable,
 *      the specified HCA device will be used.
 *  (2) Otherwise, MVAPICH first gets a list of the device IDs 
 *      of the available HCA devices on the physical node.
 *  (3) Then, it chooses the first device by default.
 */

#ifdef VIADEV_PEER_TO_PEER
int viadev_use_peer_to_peer = 1;
#endif

/*
 *If we are limiting the number of outstanding RDMA operations,
 * Allow the limit to be set at run-time, but init variable
 * to default value.
 */
#ifdef VIADEV_HAVE_RDMA_LIMIT
int viadev_rdma_limit = VIADEV_RDMA_LIMIT;
#endif

/* Max number of entries on the Send Q of QPs per connection.
 * Should be about (prepost_depth + extra).
 * Must be within NIC MaxQpEntries limit.
 * Size will be adjusted below.
 */
unsigned long viadev_sq_size = 200;

/* Max number of entries on the RecvQ of QPs per connection.
 * computed to be:
 * prepost_depth + viadev_prepost_rendezvous_extra + viadev_prepost_noop_extra
 * Must be within NIC MaxQpEntries limit.
 */
unsigned long viadev_rq_size = 200;

/* Max number of entries on the completion queue at a time.
 * Needed when constructing the completion queue, but can
 * be grown dynamically if needed.
 * Should be about (viadev_sq_size + viadev_rq_size)*num_vi.
 * Must be within NIC MaxCQEntries limit.
 * Size will be adjusted below.
 */
unsigned long viadev_cq_size = 6000;

/* number of vbufs used in RDMA_FAST_PATH on each VI */
int viadev_num_rdma_buffer = VIADEV_NUM_RDMA_BUFFER;

/* number of vbufs to pre-post as receives on each VI */
int viadev_prepost_depth = VIADEV_PREPOST_DEPTH;
int viadev_prepost_threshold = 10;
int viadev_initial_prepost_depth = 5;
int viadev_initial_credits;

/* number of additional initial vbufs to allocate per VI
 * after initial allocation is exhausted.
 */
static int viadev_vbuf_extra = 10;

/* In the case of small numbers of VIs, allocate at least
 * this number of VBUFs whenever a new region is allocated.
 */

/* number of vbufs to allocate initially.
 * This will be re-defined after reading the parameters below
 * to scale to the number of VIs and other factors.
 */
int viadev_vbuf_pool_size = 5000;

/* number of vbufs to allocate in a secondary region if we should
 * run out of the initial allocation.  This is re-computed (below)
 * once other parameters are known.
 */
int viadev_vbuf_secondary_pool_size = 500;

/* max (total) number of vbufs to allocate, after which process 
 * terminates with a fatal error.
 * -1 means no limit.
 */
int viadev_vbuf_max = -1;

/* The number of "extra" vbufs that will be posted as receives
 * on a connection in anticipation of an R3 rendezvous message.
 * The TOTAL number of VBUFs posted on a receive queue at any
 * time is viadev_prepost_depth + viadev_prepost_rendezvous_extra
 * regardless of the number of outstanding R3 sends active on
 * a connection.
 */
int viadev_prepost_rendezvous_extra = VIADEV_PREPOST_RENDEZVOUS_EXTRA;

/* The message size (in bytes) at which we switch from an EAGER
 * protocol to the three-way RENDEZVOUS protocol.
 * In reality, smaller messages might be sent using R3
 * rather than EAGER if remote credits are low.
 */
int viadev_rendezvous_threshold = DEFAULT_RENDEZVOUS_THRESHOLD;

/* the number of times the MPID_deviceCheck will poll VipCQDone
 * before it reverts to blocking on VipCQWait.
 * This value could influence latency and should probably be
 * adjusted dynamically.
 */
int viadev_spin_count = 100;

/* allow some extra buffers for non-credited packets (eg. NOOP) */
int viadev_prepost_noop_extra = 10;

/* dont allow EAGER sends or R3 data packets to consume all
 * remote credits on a connection.  This parameter records
 * the number of credits per connection that will be preserved
 * for non-data, control packets.
 */
int viadev_credit_preserve = 10;

/* As receive descriptors (VBUFs) are consumed and reaped on
 * a connection, the receive side posts more receive VBUFs.
 * Each time a message is sent to the remote end of a connection,
 * a credit update value, representing the number of new receive
 * VBUFs posted since the last message was sent, is piggybacked
 * on the message.  The remote end can then update its remote
 * credit value, possibly unblocking stalled R3 data sends.
 * If the receiving side has no reason to send a message, the
 * credit update accumulate and do not get communicated to the
 * remote end.  If the credit update value exceeds this
 * viadev_credit_notify_threshold value, a NoOp message is
 * sent to the remote end with the sole purpose of communicating
 * the credit update and thereby preventing deadlock.
 *
 * There are two places where NoOp operations may be sent.
 * The first is after all the entries on the CQ are processed,
 * indicating that all incoming and outgoing messages have
 * completed.  If there are more than viadev_credit_notify_threshold
 * local credits we send a NoOp.
 */
int viadev_credit_notify_threshold = 5;

/* The second place is while we are processing receive descriptors.
 * If the local credit value gets reasonably large, we are not
 * sending messages to the remote side.  The remote end of this
 * connection may be stalled waiting for us to process all the
 * recieves so lets send a NoOp immediately so that they can
 * continue work.
 * This viadev_dynamic_credit_threshold should be a fraction
 * of the maximum receive prepost depth, but substantially
 * greater than the viadev_credit_notify_threshold.
 * It is re-computed below.
 */
int viadev_dynamic_credit_threshold = 10;

/*
 * Development verbose flag
 */
int viadev_verbose = 0;

int viadev_max_wqe = VIADEV_MAX_WQE;

#ifdef _SMP_
#ifdef _AFFINITY_
/* Affinity is enabled by default */
unsigned int viadev_enable_affinity=1;
#endif
#endif

/* 
 * Initialization parameters.
 */
unsigned int		viadev_ndreg_entries = 1000;
char                    dapl_provider[32] = VIADEV_DEFAULT_PROVIDER;
DAT_VLEN		viadev_default_mtu = VIADEV_DEFAULT_MTU;
unsigned long 		viadev_max_rdma_size = VIADEV_MAX_RDMA_SIZE;
DAT_TIMEOUT		viadev_default_time_out = VIADEV_DEFAULT_TIME_OUT;
int			viadev_max_fast_eager_size = VIADEV_MAX_FAST_EAGER_SIZE;
u_int32_t		viadev_default_max_sg_list = VIADEV_DEFAULT_MAX_SG_LIST;


void viadev_init_parameters(int num_proc, int me)
{
    char *value;

    /* ===============================================================
     * Examine the environment variables to see in any of these
     * were re-defined on the command line.
     * ===============================================================
     */

    if ((value = getenv("VIADEV_VERBOSE")) != NULL) {
        viadev_verbose = atoi(value);
    }

    if ((value = getenv("DAPL_PROVIDER")) != NULL) {
        strcpy(dapl_provider, value);
    }
    if (strcmp (dapl_provider, "ib0") == 0)
      {
          viadev_max_wqe = VIADEV_MAX_WQE_IBA;
          viadev_default_mtu = VIADEV_DEFAULT_MTU_SIZE_IBA;
      }
    else if (strcmp (dapl_provider, "ccil") == 0)
      {
          viadev_max_wqe = VIADEV_MAX_WQE_CCIL;
          viadev_default_mtu = VIADEV_DEFAULT_MTU_SIZE_CCIL;
          viadev_rendezvous_threshold = 65536;
          viadev_prepost_depth = 50;
      }
    else if (strcmp (dapl_provider, "gmg2") == 0)
      {
          viadev_max_wqe = VIADEV_MAX_WQE_GM;
          viadev_default_mtu = VIADEV_DEFAULT_MTU_SIZE_GM;
      }
   else if (strcmp (dapl_provider, "ibd0") == 0)
      {
          viadev_max_wqe = VIADEV_MAX_WQE_SOLARIS;
          viadev_default_mtu = VIADEV_DEFAULT_MTU_SIZE_SOLARIS;
      }

    
#ifdef VIADEV_HAVE_RDMA_LIMIT
    if ((value = getenv("VIADEV_RDMA_LIMIT")) != NULL) {
        viadev_rdma_limit = atoi(value);
    }
#endif

    if ((value = getenv("VIADEV_MAX_WQE")) != NULL) {
        viadev_max_wqe = (int)atoi(value);
    }

    if ((value = getenv("VIADEV_MAX_RDMA_SIZE")) != NULL) {
        viadev_max_rdma_size = (unsigned long)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MTU")) != NULL) {
            viadev_default_mtu = (DAT_VLEN)(int)atoi(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_TIME_OUT")) != NULL) {
        viadev_default_time_out = (DAT_TIMEOUT)atol(value);
    }
    if ((value = getenv("VIADEV_MAX_FAST_EAGER_SIZE")) != NULL) {
        viadev_max_fast_eager_size = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_SG_LIST")) != NULL) {
        viadev_default_max_sg_list = (u_int32_t)atol(value);
    }
    if ((value = getenv("VIADEV_NDREG_ENTRIES")) != NULL) {
        viadev_ndreg_entries = (unsigned int)atoi(value);
    }
    if ((value = getenv("VIADEV_SQ_SIZE")) != NULL) {
        viadev_sq_size = atol(value);
    }
    if ((value = getenv("VIADEV_CQ_SIZE")) != NULL) {
        viadev_cq_size = atol(value);
    }
    if ((value = getenv("VIADEV_NUM_RDMA_BUFFER")) != NULL) {
        viadev_num_rdma_buffer = atoi(value);
        if (viadev_num_rdma_buffer < 0)
            viadev_num_rdma_buffer = 0;
        if (viadev_num_rdma_buffer == 1)
            viadev_num_rdma_buffer = 2; /* need 1 extra */
    }
    if ((value = getenv("VIADEV_PREPOST_DEPTH")) != NULL) {
        viadev_prepost_depth = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_THRESHOLD")) != NULL) {
        viadev_prepost_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_POOL_SIZE")) != NULL) {
	viadev_vbuf_pool_size = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_SECONDARY_POOL_SIZE")) != NULL) {
	viadev_vbuf_secondary_pool_size = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_EXTRA")) != NULL) {
        viadev_vbuf_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_MAX")) != NULL) {
        viadev_vbuf_max = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_RENDEZVOUS_EXTRA")) != NULL) {
        viadev_prepost_rendezvous_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_RENDEZVOUS_THRESHOLD")) != NULL) {

        viadev_rendezvous_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_SPIN_COUNT")) != NULL) {
        viadev_spin_count = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_NOOP_EXTRA")) != NULL) {
        viadev_prepost_noop_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_INITIAL_PREPOST_DEPTH")) != NULL) {
        viadev_initial_prepost_depth = atoi(value);
    }
    /* TBD - temporary workaround, RDMA initialization mistakenly triggers
     * credit exchange and tries to use uninitialized RDMA remote addresses
     * to send Noops.
     */
    if (viadev_num_rdma_buffer)
        viadev_initial_prepost_depth =
            viadev_prepost_depth + viadev_prepost_noop_extra;
    if (viadev_initial_prepost_depth >
        viadev_prepost_depth + viadev_prepost_noop_extra)
        viadev_initial_prepost_depth =
            viadev_prepost_depth + viadev_prepost_noop_extra;
    if (viadev_initial_prepost_depth <= 0)
        viadev_initial_prepost_depth = 1;
    if (viadev_initial_prepost_depth <= viadev_prepost_noop_extra) {
        viadev_initial_credits = viadev_initial_prepost_depth;
    } else {
        viadev_initial_credits =
            viadev_initial_prepost_depth - viadev_prepost_noop_extra;
    }
    if ((value = getenv("VIADEV_CREDIT_PRESERVE")) != NULL) {
        viadev_credit_preserve = atoi(value);
    }
    if ((value = getenv("VIADEV_CREDIT_NOTIFY_THRESHOLD")) != NULL) {
        viadev_credit_notify_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_DYNAMIC_CREDIT_THRESHOLD")) != NULL) {
        viadev_dynamic_credit_threshold = atoi(value);
    }
    if ((viadev_prepost_depth - viadev_prepost_threshold) <=
        MAX(viadev_credit_notify_threshold,
            viadev_dynamic_credit_threshold)) {
        viadev_prepost_threshold =
            viadev_prepost_depth - 1 - MAX(viadev_credit_notify_threshold,
                                           viadev_dynamic_credit_threshold);
    }
    if (viadev_prepost_threshold < 0)
        viadev_prepost_threshold = 0;
#ifdef VIADEV_PEER_TO_PEER
    if ((value = getenv("VIADEV_USE_PEER_TO_PEER")) != NULL) {
        viadev_use_peer_to_peer = atoi(value);
    }
#endif

#ifdef _SMP_
    /* Should set in KBytes */
    if ( (value = getenv("SMP_EAGERSIZE")) != NULL ) {
        smp_eagersize = atoi(value);
    }
    /* Should set in MBytes */
    if ( (value = getenv("SMPI_LENGTH_QUEUE")) != NULL ) {
        smpi_length_queue = atoi(value);
    }
#ifdef _AFFINITY_
    if ((value = getenv("VIADEV_ENABLE_AFFINITY")) != NULL) {
            viadev_enable_affinity = atoi(value);
    }
#endif
#endif

    /* compute Receive Q size */
    viadev_rq_size =
        viadev_prepost_depth + viadev_prepost_rendezvous_extra +
        viadev_prepost_noop_extra;
}

void dump_param_values(FILE * fd, int me)
{
    char *int_fmt = "%3d: %32s = %d\n";
    char *str_fmt = "%3d: %32s = %s\n";

    /*
       char *hex_fmt = "%3d: %32s = %#x\n";
     */

    fprintf(fd, int_fmt, me, "VIADEV_VERBOSE", viadev_verbose);
    fprintf(fd, str_fmt, me, "VIADEV_DEVICE", dapl_provider);

    /* Print reliability Level for VIA, otherwise not implemented */
#ifdef VIADEV_PEER_TO_PEER
    fprintf(fd, int_fmt, me,
            "VIADEV_USE_PEER_TO_PEER", viadev_use_peer_to_peer);
#endif
#ifdef VIADEV_HAVE_RDMA_LIMIT
    fprintf(fd, int_fmt, me, "VIADEV_RDMA_LIMIT", viadev_rdma_limit);
#endif

    fprintf(fd, int_fmt, me, "VIADEV_MAX_RDMA_SIZE", (int)viadev_max_rdma_size);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MTU", (int)viadev_default_mtu);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_TIME_OUT", (int)viadev_default_time_out);
    fprintf(fd, int_fmt, me, "VIADEV_MAX_FAST_EAGER_SIZE", (int)viadev_max_fast_eager_size);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MAX_SG_LIST", (int)viadev_default_max_sg_list);
    fprintf(fd, int_fmt, me, "VIADEV_NDREG_ENTRIES", viadev_ndreg_entries);
    fprintf(fd, int_fmt, me, "VIADEV_SQ_SIZE", (int) viadev_sq_size);
    fprintf(fd, int_fmt, me, "VIADEV_CQ_SIZE", (int) viadev_cq_size);
    fprintf(fd, int_fmt, me, "VIADEV_NUM_RDMA_BUFFER",
            viadev_num_rdma_buffer);
    fprintf(fd, int_fmt, me, "VIADEV_PREPOST_DEPTH", viadev_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_THRESHOLD", viadev_prepost_threshold);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_POOL_SIZE", viadev_vbuf_pool_size);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_SECONDARY_POOL_SIZE",
            viadev_vbuf_secondary_pool_size);
    fprintf(fd, int_fmt, me, "VIADEV_VBUF_MAX", viadev_vbuf_max);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_RENDEZVOUS_EXTRA",
            viadev_prepost_rendezvous_extra);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_NOOP_EXTRA", viadev_prepost_noop_extra);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_PRESERVE",
            viadev_credit_preserve);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_NOTIFY_THRESHOLD",
            viadev_credit_notify_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_DYNAMIC_CREDIT_THRESHOLD",
            viadev_dynamic_credit_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_RENDEZVOUS_THRESHOLD",
            viadev_rendezvous_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_SPIN_COUNT", viadev_spin_count);
}


/* ================================================================
 * Adjust global parameters to conform with NIC attributes
 * ================================================================ */

/* This section only applicatble to VIA */
