/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include <stdio.h>
#include "mpid.h"
#include "udapl_header.h"
#include "viadev.h"
#include "viapriv.h"
#include "viutil.h"
#include "vbuf.h"
#include "viapacket.h"
#include "process/pmgr_client.h"
#include <unistd.h>

/* low level VIA information (hidden from ADI2) */
viadev_info_t viadev;

char nullsbuffer, nullrbuffer;

/* The vbuf backlog queue */
viadev_backlog_queue_t *backlog_queue = NULL;

void viadev_unimplemented(char *s)
{
   udapl_error_abort(GEN_EXIT_ERR, "Function %s not implemented", s);
}

/* to handle Send Q overflow, we maintain an extended send queue
 * above the HCA.  This permits use to have a virtually unlimited send Q depth
 * (limited by number of vbufs available for send)
 */
void viadev_ext_sendq_queue(viadev_connection_t * c, vbuf * v)
{
    v->desc.next = NULL;
    if (c->ext_sendq_head == NULL) {
        c->ext_sendq_head = v;
    } else {
        c->ext_sendq_tail->desc.next = v;
    }
    c->ext_sendq_tail = v;
}

/* dequeue and send as many as we can from the extended send queue
 * this is called in each function which may post send prior to it attempting
 * its send, hence ordering of sends is maintained
 */
void viadev_ext_sendq_send(viadev_connection_t * c)
{
    vbuf *v;

    while (c->send_wqes_avail && c->ext_sendq_head) {
        /* dequeue one vbuf */
        v = c->ext_sendq_head;
        c->ext_sendq_head = v->desc.next;
        if (v == c->ext_sendq_tail) {
            c->ext_sendq_tail = NULL;
        }
        v->desc.next = NULL;
        /* put the vbuf on the HCA's send Q */
        c->send_wqes_avail--;

        UDAPL_POST_SR(v, c, "VAPI_post_sr (viadev_ext_sendq_send)");
    }
}

/* init the backlog queue */
void viadev_init_backlog_queue(viadev_connection_t * c)
{
    c->backlog.len = 0;
    c->backlog.vbuf_head = NULL;
    c->backlog.vbuf_tail = NULL;
}

/* 
 * post a descriptor to the send queue
 * all outgoing packets go through this routine. 
 * takes a connection rather than a vi because we need to 
 * update flow control information on the connection. Also
 * it turns out it is always called in the context of a connection, 
 * i.e. it would be post_send(c->vi) otherwise. 
 */

#define BACKLOG_ENQUEUE(q,v) {                                      \
    v->desc.next = NULL;                                            \
        if (q->vbuf_tail == NULL) {                                 \
            q->vbuf_head = v;                                       \
        } else {                                                    \
            q->vbuf_tail->desc.next = v;                            \
        }                                                           \
    q->vbuf_tail = v;                                               \
        q->len++;                                                   \
}


#define BACKLOG_DEQUEUE(q,v)  {                                     \
    v = q->vbuf_head;                                               \
        q->vbuf_head = v->desc.next;                                \
        if (v == q->vbuf_tail) {                                    \
            q->vbuf_tail = NULL;                                    \
        }                                                           \
    q->len--;                                                       \
        v->desc.next = NULL;                                        \
}

/* Just got a credit update, send off as many backlogged
 * messages as possible
 */
void viadev_backlog_send(viadev_connection_t * c)
{
    viadev_backlog_queue_t *q = &c->backlog;

    while ((q->len > 0) && (c->remote_credit > 0)) {
        vbuf *v = NULL;
        viadev_packet_header *p;
        assert(q->vbuf_head != NULL);
        BACKLOG_DEQUEUE(q, v);

        /* Assumes packet header is at beginning of packet structure */
        p = (viadev_packet_header *) VBUF_BUFFER_START(v);

        PACKET_SET_CREDITS(p, c);
        c->remote_credit--;

        v->grank = c->global_rank;

        if (!c->send_wqes_avail) {
            viadev_ext_sendq_queue(c, v);
            continue;
        }
        c->send_wqes_avail--;

        UDAPL_POST_SR(v, c, "UDAPL_post_sr (viadev_backlog_push)");
    }
}

void viadev_post_send(viadev_connection_t * c, vbuf * v)
{
    viadev_packet_header *p =
        (viadev_packet_header *) VBUF_BUFFER_START(v);
    /* 
     * will need to grab lock in multithreaded 
     * will need to put sequence number in immediate data for unreliable
     * currently this will tell you if you're out of remote credit
     * but will not block or return error to caller.
     */

    if (c->remote_credit > 0 || p->type == VIADEV_PACKET_NOOP) {
        /* if we got here, the backlog queue better be  empty */
        assert(c->backlog.len == 0 || p->type == VIADEV_PACKET_NOOP);

        D_PRINT("post_send: SENT [vi %d][%d][vbuf %p] %s",
                c->global_rank, c->remote_credit, v,
                viadev_packet_to_string(c->vi, v));
        PACKET_SET_CREDITS(p, c);
        if (p->type != VIADEV_PACKET_NOOP)
            c->remote_credit--;
        v->grank = c->global_rank;

        D_PRINT("send message: len=%d", v->desc.sg_entry.len);
        D_PRINT("vbuf=%p\n", v->desc.sg_entry.addr);

        if (!c->send_wqes_avail) {
            viadev_ext_sendq_queue(c, v);
            return;
        }
        c->send_wqes_avail--;

        UDAPL_POST_SR(v, c, "UDAPL_post_sr (viadev_post_send)");

        /* do post receive after post_sr for last part of message
         * so latency hidden during post_sr
         */
        if (viadev_prepost_threshold && v->shandle != NULL
            && c->initialized) {
            int needed =
                viadev_prepost_depth + viadev_prepost_noop_extra +
                MIN(viadev_prepost_rendezvous_extra,
                    c->rendezvous_packets_expected);
            while (c->preposts < viadev_rq_size && c->preposts < needed) {
                PREPOST_VBUF_RECV(c);
            }
            viadev_send_noop_ifneeded(c);
        }
    } else {
        /* must delay the send until more credits arrive.
         * put on end of queue
         */
        viadev_backlog_queue_t *q = &c->backlog;
        BACKLOG_ENQUEUE(q, v);

        D_PRINT("post_send QUEUED: [vi %d][%d %d %d][vbuf %p] %s",
                c->global_rank, c->remote_credit, c->local_credit,
                c->remote_cc, v, viadev_packet_to_string(c->vi, v));
    }

}

#if defined(RDMA_FAST_PATH)

void post_fast_rdma_with_completion(viadev_connection_t * c, int len)
{

    int n = c->phead_RDMA_send;
    VBUF_FLAG_TYPE flag;
    viadev_packet_header *p = (viadev_packet_header *)
        (c->RDMA_send_buf[n].buffer);

    /* Mark RDMA buffer busy */
    c->RDMA_send_buf[n].padding = BUSY_FLAG;

    /* set credits info */
#ifndef DISABLE_HEADER_CACHING
    if ((p->type != FAST_EAGER_CACHED) &&
        (p->type != VIADEV_PACKET_EAGER_START))
        PACKET_SET_CREDITS(p, c);
    if ((p->type == VIADEV_PACKET_EAGER_START) &&
        (len >= viadev_max_fast_eager_size))
        PACKET_SET_CREDITS(p, c);
#else
    PACKET_SET_CREDITS(p, c);
#endif

    if (++(c->phead_RDMA_send) >= viadev_num_rdma_buffer)
        c->phead_RDMA_send = 0;

    /* set flags */
    if (*(VBUF_FLAG_TYPE *) (c->RDMA_send_buf[n].buffer + len)
        == len) {
        flag = (VBUF_FLAG_TYPE) (len + FAST_RDMA_ALT_TAG);
    } else {
        flag = (VBUF_FLAG_TYPE) len;
    }

    /* set head flag */
    c->RDMA_send_buf[n].head_flag = flag;
    /* set tail flag */
    *(VBUF_FLAG_TYPE *) (c->RDMA_send_buf[n].buffer + len) = flag;

    c->num_no_completion = 0;
    /* generate a completion */
    c->RDMA_send_buf[n].desc.completion_flag = DAT_COMPLETION_DEFAULT_FLAG;
    c->RDMA_send_buf[n].desc.cookie.as_ptr = (DAT_PVOID) & (c->RDMA_send_buf[n]);

#ifdef _IA64_
    len += 16;
    len -= (len & 7);
    if (len > VBUF_BUFFER_SIZE)
        len = VBUF_BUFFER_SIZE;
#endif

    c->RDMA_send_buf[n].desc.local_iov.segment_length =
        len + VBUF_FAST_RDMA_EXTRA_BYTES;
    c->RDMA_send_buf[n].desc.remote_iov.segment_length =
        len + VBUF_FAST_RDMA_EXTRA_BYTES;
    

    if (!c->send_wqes_avail) {
        viadev_ext_sendq_queue(c, &c->RDMA_send_buf[n]);
        return;
    }
    c->send_wqes_avail--;

    UDAPL_POST_SR((&c->RDMA_send_buf[n]), c, 
            "UDAPL_post_sr (post_fast_rdma)");
}

int fast_rdma_ok(viadev_connection_t * c, int data_size,
        int is_eager_packet)
{
    /* not OK if buffers are full */
    if (viadev_num_rdma_buffer == 0)
        return 0;
    if (c->phead_RDMA_send == c->ptail_RDMA_send) {
        return 0;
    }

    /* not OK if data size is too large */
    if (data_size > VBUF_BUFFER_SIZE) {
        return 0;
    }

    /* not OK if there are sends waiting in the backlog ??? */
    if (c->backlog.len > 0) {
        return 0;
    }

    /* See if head buffer is marked free */
    if(c->RDMA_send_buf[c->phead_RDMA_send].padding == BUSY_FLAG) {
        return 0;
    }

    /* otherwise OK */
    return 1;
}

void release_recv_rdma(viadev_connection_t * c, vbuf * v)
{
    vbuf *next_free;
    int next;
    int i;

    next = c->p_RDMA_recv_tail + 1;
    if (next >= viadev_num_rdma_buffer)
        next = 0;
    next_free = &(c->RDMA_recv_buf[next]);

    v->padding = FREE_FLAG;
    if (v != next_free) {
        return;
    }

    /* search all free buffers */
    for (i = next; i != c->p_RDMA_recv;) {

        if (c->RDMA_recv_buf[i].padding == FREE_FLAG) {
            c->rdma_credit++;
            if (++(c->p_RDMA_recv_tail) >= viadev_num_rdma_buffer)
                c->p_RDMA_recv_tail = 0;
            c->RDMA_recv_buf[i].padding = BUSY_FLAG;
        } else break;
        if (++i >= viadev_num_rdma_buffer)
            i = 0;
    }
}


#ifndef DISABLE_HEADER_CACHING

int search_header_cache(viadev_connection_t * c,
                        viadev_packet_eager_start * h)
{
    viadev_packet_eager_start *cached = &(c->cached_outgoing);

    if ((cached->header.vbuf_credit == h->header.vbuf_credit) &&
        (cached->header.remote_credit == h->header.remote_credit) &&
        (cached->header.rdma_credit == h->header.rdma_credit) &&
        (cached->envelope.context == h->envelope.context) &&
        (cached->envelope.tag == h->envelope.tag) &&
        (cached->envelope.src_lrank == h->envelope.src_lrank)) {
        c->cached_hit++;
        return 1;
    }

    c->cached_miss++;
    memcpy(cached, h, sizeof(viadev_packet_eager_start));
    return 0;
}

#endif


#endif

void viadev_post_rdmawrite(viadev_connection_t * c, vbuf * v)
{

    D_PRINT("post_rdmawrite: [vi %d] %s",
            c->global_rank, viadev_packet_to_string(c->vi, v));

    /* set id to be the global_rank */
    v->grank = c->global_rank;

    if (!c->send_wqes_avail) {
        viadev_ext_sendq_queue(c, v);
        return;
    }
    c->send_wqes_avail--;

    UDAPL_POST_SR(v, c, "viadev_post_rdmawrite");
}

/* 
 * post a descriptor to the receive queue
 */

void viadev_post_recv(viadev_connection_t * c, vbuf * v)
{
    DAT_RETURN ret;

    D_PRINT("viadev_post_recv: v = " AINT_FORMAT, (aint_t) v);
    /* set id to be the global_rank */
    v->grank = c->global_rank;
    v->desc.op_code = UDAPL_RECV;
    ret = dat_ep_post_recv(c->vi, 1, &v->desc.local_iov, v->desc.cookie, v->desc.completion_flag);   
    if(ret != DAT_SUCCESS){
        udapl_error_abort(UDAPL_RETURN_ERR,
                            "dat_ep_post_recv (viadev_post_recv)");
    }
}

char *viadev_packet_to_string(DAT_EVD_HANDLE vi, vbuf * v)
{
    return '\0';
}

int vi_to_grank(DAT_EVD_HANDLE vi)
{
    return 0;
}

void UDAPL_POST_SR(vbuf *v, viadev_connection_t *c, char *err_string)
    {
        DAT_RETURN ret;
        switch (v->desc.op_code){
        case UDAPL_RDMA_WRITE:
            {
               ret = dat_ep_post_rdma_write(c->vi, 1,
                                            &(v->desc.local_iov),
                                            v->desc.cookie,
                                            &(v->desc.remote_iov),
                                            v->desc.completion_flag);
               break;
            }
        case UDAPL_SEND:
            {
                ret = dat_ep_post_send (c->vi, 1, &(v->desc.local_iov), v->desc.cookie, v->desc.completion_flag);
                break;
            }
        case UDAPL_RECV:
            {
                ret = dat_ep_post_recv (c->vi, 1, &(v->desc.local_iov), v->desc.cookie, v->desc.completion_flag);
              break;
            }
        default:
              udapl_error_abort(UDAPL_RETURN_ERR, "unkown op_code");
        }
        if(ret != DAT_SUCCESS) {
            udapl_error_abort(UDAPL_RETURN_ERR, err_string);
        }
}

