/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include "mpid_bind.h"
#include "viadev.h"
#include "viapriv.h"
#include "queue.h"
#include "vbuf.h"
#include "dreg.h"
#include "viaparam.h"
#include "mpid_smpi.h"

void MPID_VIA_Irecv(void *buf, int len, int src_lrank, int tag,
                    int context_id, MPIR_RHANDLE * rhandle,
                    int *error_code)
{
    MPIR_RHANDLE *unexpected;

    /* assume success, change if necessary */
    *error_code = MPI_SUCCESS;

    /* error check. Both MPID_Recv and MPID_Irecv go through this code */
    if ((NULL == buf) && (len > 0)) {
        *error_code = MPI_ERR_BUFFER;
        return;
    }
    /* this is where we would register the buffer and get a dreg entry */
    /*rhandle->bytes_as_contig = len; changed for totalview */
    rhandle->len = len;
    /*rhandle->start = buf; changed for totalview support */
    rhandle->buf = buf;
#ifdef VIADEV_RPUT_SUPPORT
    rhandle->dreg_entry = 0;
#endif
    rhandle->is_complete = 0;
    rhandle->replied = 0;
    rhandle->bytes_copied_to_user = 0;
    rhandle->vbufs_received = 0;
    rhandle->protocol = VIADEV_PROTOCOL_RENDEZVOUS_UNSPECIFIED;
    rhandle->can_cancel = 1;

    /* At this point we don't fill in:
     * vbufs_expected, vbufs_received, bytes_received, remote_address, 
     * remote_memory_handle. connection, vbuf_head, vbuf_tail.
     */

    MPID_Search_unexpected_queue_and_post(src_lrank,
                                          tag, context_id, rhandle,
                                          &unexpected);

    if (unexpected) {
#ifdef _SMP_
        if ((VIADEV_PROTOCOL_SMP_SHORT == unexpected->protocol) && (NULL != unexpected->connection)) {
            MPID_SMP_Eagerb_unxrecv_start_short(rhandle, unexpected);
            return;
        }
#ifdef _SMP_RNDV_
        if ((VIADEV_PROTOCOL_SMP_RNDV == unexpected->protocol) && (NULL != unexpected->connection)){
            MPID_SMP_Rndvn_unxrecv_posted(rhandle, unexpected);
            return;
        }
#endif

#endif

        /* special case for self-communication */
        if (NULL == unexpected->connection) {
            MPID_VIA_self_finish(rhandle, unexpected);
            MPID_RecvFree(unexpected);
        } else {
            viadev_copy_unexpected_handle_to_user_handle(rhandle,
                                                         unexpected,
                                                         error_code);
            MPID_RecvFree(unexpected);

            if (*error_code != MPI_SUCCESS) {
                return;
            }

            /* packets of matching send have arrived, can no longer
             * cancel the recv */
            rhandle->can_cancel = 0;

            /* sender spedified a protocol, which one? */
            switch (rhandle->protocol) {
            case VIADEV_PROTOCOL_R3:
                {
                    viadev_recv_r3(rhandle);
                    break;
                }
            case VIADEV_PROTOCOL_RPUT:
                {
                    viadev_recv_rput(rhandle);
                    break;
                }
            case VIADEV_PROTOCOL_RGET:
                {
                    viadev_recv_rget(rhandle);
                    break;
                }
            case VIADEV_PROTOCOL_RENDEZVOUS_UNSPECIFIED:
                {
                    /* should always be specified on sender side */
                   udapl_error_abort(GEN_EXIT_ERR, "invalid protocol: RENDEZVOUS_UNSPECIFIED");
                }
            case VIADEV_PROTOCOL_EAGER:
                {
                    rhandle->vbufs_expected =
                        viadev_calculate_vbufs_expected
                        (rhandle->len, rhandle->protocol);
                    viadev_eager_pull(rhandle);
                    break;
                }
            default:
                {
                   udapl_error_abort(GEN_EXIT_ERR,
                                        "MPID_VIA_Irecv: unknown protocol %d\n",
                                        rhandle->protocol);
                }
            }

        }
    }
}

void viadev_recv_r3(MPIR_RHANDLE * rhandle)
{

    rhandle->vbufs_expected =
        viadev_calculate_vbufs_expected(rhandle->len,
                                        rhandle->protocol);
    viadev_prepost_for_rendezvous((viadev_connection_t *) rhandle->
                                  connection, rhandle->vbufs_expected);
    viadev_rendezvous_reply(rhandle);
}

void viadev_recv_rget(MPIR_RHANDLE * rhandle)
{
    /* sender has specified RDMA READ, can we match it? */

    /* first, if receive buffer is NULL, must be a zero length
     * transfer, just use a dummy buffer location so that we have
     * something to transfer */
    assert(0);
}

void viadev_recv_rput(MPIR_RHANDLE * rhandle)
{
#ifdef VIADEV_RPUT_SUPPORT
    /* only way buffer is NULL is if length is zero */
    if (NULL == rhandle->buf) {
        rhandle->buf = &nullrbuffer;
    }
    viadev_register_recvbuf_if_possible(rhandle);
    if (NULL == rhandle->dreg_entry) {
        /* failed to register memory, revert to R3 */
        rhandle->protocol = VIADEV_PROTOCOL_R3;
        viadev_recv_r3(rhandle);
        return;
    }
    /* ok, managed to register memory */
    rhandle->vbufs_expected =
        viadev_calculate_vbufs_expected(rhandle->len,
                                        rhandle->protocol);
    viadev_rendezvous_reply(rhandle);
#endif
}

#ifdef VIADEV_RPUT_SUPPORT
void viadev_register_recvbuf_if_possible(MPIR_RHANDLE * rhandle)
{
    /* note that dreg_register() will return NULL if it fails */
    rhandle->dreg_entry = dreg_register(rhandle->buf,
                                        rhandle->len, DREG_ACL_WRITE);
}
#endif

/* 
 * Copy a system-generated receive request from the unexpected queue
 * to a user request after finding a match.  
 *
 * After if search_unexpected_and_post finds a match, we now we have
 * two rhandles - the real one passed to us and and the placeholder
 * one on the unexpected queue, which has minimal information.
 *
 * This routine moves information from the placeholder handle into the
 * real handle, dequeues the queued handle and makes progress on the
 * receive if possible.  (for model, see chnrndv.c->Rndvn_unxrecv_start)
 *
 * xxx This is a lot of data to copy, but it all seems
 * necessary. 
 *      
 * xxx we could do sanity checks here. 
 *  
 */

void viadev_copy_unexpected_handle_to_user_handle(MPIR_RHANDLE * rhandle,
                                                  MPIR_RHANDLE *
                                                  unexpected,
                                                  int *error_code)
{

    rhandle->s.MPI_SOURCE = unexpected->s.MPI_SOURCE;
    rhandle->s.MPI_TAG = unexpected->s.MPI_TAG;
    rhandle->vbufs_received = unexpected->vbufs_received;
#ifdef VIADEV_RPUT_SUPPORT
    rhandle->bytes_sent = 0;
    rhandle->remote_address = unexpected->remote_address;
    rhandle->remote_memhandle = unexpected->remote_memhandle;
#endif
    rhandle->connection = unexpected->connection;
    rhandle->vbuf_head = unexpected->vbuf_head;
    rhandle->vbuf_tail = unexpected->vbuf_tail;
    rhandle->send_id = unexpected->send_id;
    rhandle->protocol = unexpected->protocol;

    /* check for truncation */
    if (unexpected->len > rhandle->len) {
        *error_code = MPI_ERR_COUNT;
        return;
        /*        error(ERR_FATAL, "message truncated. ax %d got %d", 
           rhandle->len, 
           unexpected->len);
         */
    } else {
        rhandle->s.count = unexpected->len;
        rhandle->len = unexpected->len;
    }

    /* if this is an eager send currently being transferred, 
     * connection handle may have a link to us
     */
    if (((viadev_connection_t *) (rhandle->connection))->rhandle
        == unexpected) {
        ((viadev_connection_t *) (rhandle->connection))->rhandle = rhandle;
    }
}

/*
 * calculate_vbufs_expected is called whenever we get
 * a match between a posted receive and an incoming packet. 
 * This happens either when a receive is posted and matches
 * an entry on the unexpected queue or when a packet
 * comes in that matches a pre-posted receive. 
 * 
 * In general, it is not until this match that we have information
 * from both sides that allows us to figure out how many packets
 * will be sent to complete the transfer. 
 * 
 * When the transfer is through a rendezvous protocol, the
 * specific type of rendezvous will have been determined before
 * this routine is called. 
 *
 */


int viadev_calculate_vbufs_expected(int nbytes, viadev_protocol_t protocol)
{
    int first_packet, next_packet;
    switch (protocol) {
    case VIADEV_PROTOCOL_EAGER:
        {
            first_packet = VBUF_BUFFER_SIZE -
                sizeof(viadev_packet_eager_start);
            next_packet = VBUF_BUFFER_SIZE -
                sizeof(viadev_packet_eager_next);
            break;
        }
    case VIADEV_PROTOCOL_R3:
        {
            first_packet = VBUF_BUFFER_SIZE -
                sizeof(viadev_packet_r3_data);
            next_packet = first_packet;
            break;
        }
#ifdef VIADEV_RPUT_SUPPORT
    case VIADEV_PROTOCOL_RPUT:
        {
            /* this is number of rputs by sender.
             *                  * receiver gets only one packet */
            first_packet = viadev.maxtransfersize;
            next_packet = viadev.maxtransfersize;
            break;
        }
#endif
    default:
        {
            /* return to MPI */
            first_packet = 1;
            next_packet = 1;
           udapl_error_abort(GEN_EXIT_ERR,
                                "calculate_vbufs_expected: unexpected protocol %d",
                                protocol);
        }
    }
    nbytes -= first_packet;
    if (nbytes > 0) {
        return (2 + (nbytes - 1) / next_packet);
    } else {
        return (1);
    }
}


/*
 * reply for rendezvous protocol 
 * This routine is called as soon as a match is made 
 */

void viadev_rendezvous_reply(MPIR_RHANDLE * rhandle)
{
    vbuf *v;
    viadev_packet_rendezvous_reply *packet;
#ifdef RDMA_FAST_PATH
    viadev_connection_t *c = (viadev_connection_t *) rhandle->connection;
    int rdma_ok;
#endif

    if (rhandle->replied) {
       udapl_error_abort(GEN_EXIT_ERR,
                            "Extra rendezvous reply attempted.");
        return;
    }
    if (rhandle->protocol != VIADEV_PROTOCOL_R3 &&
        rhandle->protocol != VIADEV_PROTOCOL_RPUT &&
        rhandle->protocol != VIADEV_PROTOCOL_RGET) {
       udapl_error_abort(GEN_EXIT_ERR,
                            "Inappropriate protocol %d for rendezvous reply\n",
                            rhandle->protocol);
    }
#if defined(RDMA_FAST_PATH)
    rdma_ok = fast_rdma_ok(c, sizeof(viadev_packet_rendezvous_reply), 0);

    if (rdma_ok) {
        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        packet = (viadev_packet_rendezvous_reply *) v->buffer;
    } else {
        v = get_vbuf();
        packet = (viadev_packet_rendezvous_reply *) VBUF_BUFFER_START(v);
    }
#else
    v = get_vbuf();
    packet = (viadev_packet_rendezvous_reply *) VBUF_BUFFER_START(v);
#endif

    PACKET_SET_HEADER(packet,
                      ((viadev_connection_t *) rhandle->connection),
                      VIADEV_PACKET_RENDEZVOUS_REPLY);

#ifdef VIADEV_RPUT_SUPPORT
    if (rhandle->dreg_entry != NULL) {
        packet->memhandle =
            ((dreg_entry *) rhandle->dreg_entry)->memhandle;
        packet->buffer_address = rhandle->buf;
    } else {
        packet->memhandle.hndl = 0;
        packet->buffer_address = NULL;
    }
#endif

    packet->sreq = rhandle->send_id;
    packet->rreq = REQ_TO_ID(rhandle);
    packet->protocol = rhandle->protocol;

    D_PRINT("VI %3d R3 OKTS len %d",
            ((viadev_connection_t *) rhandle->connection)->global_rank,
            rhandle->len);

#if defined(RDMA_FAST_PATH)
    if (rdma_ok) {
        post_fast_rdma_with_completion(c,
                                       sizeof
                                       (viadev_packet_rendezvous_reply));
    } else {
        vbuf_init_send(v, sizeof(viadev_packet_rendezvous_reply));
        viadev_post_send(((viadev_connection_t *) rhandle->connection), v);
    }
#else
    vbuf_init_send(v, sizeof(viadev_packet_rendezvous_reply));
    viadev_post_send(((viadev_connection_t *) rhandle->connection), v);
#endif

    rhandle->replied = 1;
}


/*
 * Make progress on a receive. This is invoked when we post
 * a receive that matches an eager entry in the unexpected queue. 
 * In this case we need to copy data that has come in so far. 
 * After this initial batch of vbufs, subsequent packets will be  
 * processed as they come in. 
 */

void viadev_eager_pull(MPIR_RHANDLE * rhandle)
{
    viadev_packet_eager_start *start;
    viadev_packet_eager_next *next;
    char *vbuf_data;
    viadev_connection_t *c = (viadev_connection_t *) rhandle->connection;
    vbuf *v;
    assert(rhandle->protocol == VIADEV_PROTOCOL_EAGER);

    /* we have just posted a receive that matches an in-progress
     * eager send. There is a chain of vbufs on the rhandle. First
     * vbuf is an EAGER_START, others are EAGER_NEXT
     */

    assert(rhandle->vbuf_head != NULL);

    v = (vbuf *) (rhandle->vbuf_head);
    start = (viadev_packet_eager_start *) VBUF_BUFFER_START(v);

#if !defined(DISABLE_HEADER_CACHING) && defined (RDMA_FAST_PATH)
    if (start->header.type == FAST_EAGER_CACHED) {
        vbuf_data = (char *) (start) + FAST_EAGER_HEADER_SIZE;
        c = &(viadev.connections[v->grank]);

        if (start->header.fast_eager_size != 0) {
            memcpy(rhandle->buf, vbuf_data,
                   start->header.fast_eager_size);
        }
        rhandle->bytes_copied_to_user = start->header.fast_eager_size;

        rhandle->vbuf_head = v->desc.next;
    } else {
        vbuf_data = (char *) (start) + sizeof(viadev_packet_eager_start);
        c = &(viadev.connections[v->grank]);

        if (start->bytes_in_this_packet != 0) {
            memcpy(rhandle->buf, vbuf_data, start->bytes_in_this_packet);
        }
        rhandle->bytes_copied_to_user = start->bytes_in_this_packet;

        rhandle->vbuf_head = v->desc.next;
    }

#else
    vbuf_data = (char *) (start) + sizeof(viadev_packet_eager_start);

    c = &(viadev.connections[v->grank]);

    if (start->bytes_in_this_packet != 0) {
        memcpy(rhandle->buf, vbuf_data, start->bytes_in_this_packet);
    }
    rhandle->bytes_copied_to_user = start->bytes_in_this_packet;
    rhandle->vbuf_head = v->desc.next;
#endif

#if defined(RDMA_FAST_PATH)
    if (v->padding == NORMAL_VBUF_FLAG)
        release_vbuf(v);
    else
        release_recv_rdma(c, v);
#else
    release_vbuf(v);
#endif

    while (rhandle->vbuf_head != NULL) {
        /* set up pointers */
        v = rhandle->vbuf_head;
        next = (viadev_packet_eager_next *) VBUF_BUFFER_START(v);
        vbuf_data = (char *) (next) + sizeof(viadev_packet_eager_next);

        assert(next->header.type == VIADEV_PACKET_EAGER_NEXT);

        memcpy((char *) (rhandle->buf) + rhandle->bytes_copied_to_user,
               vbuf_data, next->bytes_in_this_packet);

        rhandle->bytes_copied_to_user += next->bytes_in_this_packet;
        /* pop the just-copied vbuf off the list */

        rhandle->vbuf_head = v->desc.next;

#if defined(RDMA_FAST_PATH)
        if (v->padding == NORMAL_VBUF_FLAG)
            release_vbuf(v);
        else
            release_recv_rdma(c, v);
#else
        release_vbuf(v);
#endif
    }
    rhandle->vbuf_head = NULL;
    rhandle->vbuf_tail = NULL;

    if (rhandle->vbufs_expected == rhandle->vbufs_received) {
        assert(rhandle->bytes_copied_to_user == rhandle->len);
        RECV_COMPLETE(rhandle);
    }
}


/* When a rendezvous connection is in progress, we like to have
 * more than the default number of preposted vbufs available, so
 * that the sender can proceed without blocking 
 */

void viadev_prepost_for_rendezvous(viadev_connection_t * c,
                                   int vbufs_expected)
{

    int needed;
    /* extra preposts we want is c->rendezvous_packets_expected but
     * we are limited to VIADEV_PREPOST_RENDEZVOUS_EXTRA.
     * change RENDEZVOUS_EXTRA to a total PREPOST_MAX
     */

    /* record total number of vbufs expected for Rendzvous data receives */
    c->rendezvous_packets_expected += vbufs_expected;

    /* figure number of preposts needed to satisfy current demand, but
     * limit so that we don't prepost tons of vbufs.  */

    needed = c->rendezvous_packets_expected;
    if (needed > viadev_prepost_rendezvous_extra) {
        needed = viadev_prepost_rendezvous_extra;
    }
    needed += viadev_prepost_depth + viadev_prepost_noop_extra;

    assert(c->initialized);
    while (c->preposts < viadev_rq_size && needed > c->preposts) {
        PREPOST_VBUF_RECV(c);
    }
    viadev_send_noop_ifneeded(c);
}

