/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "mpid.h"
#include "viadev.h"
#include "viapriv.h"
#include "dreg.h"
#include "viapacket.h"
#include "queue.h"
#include "viaparam.h"

int MPID_VIA_rendezvous_start(void *buf, int len, int src_lrank, int tag,
                              int context_id, int dest_grank,
                              MPIR_SHANDLE * shandle)
{

    /*
     * Construct a packet to start a rendezvous transfer. 
     * If dreg_entry is NULL, memory is not pinned, otherwise
     * it is. 
     */
#if defined(VIADEV_RPUT_SUPPORT)||defined(VIADEV_RGET_SUPPORT)
    dreg_entry *dreg_entry;
#endif

#if defined(RDMA_FAST_PATH)
    int rc = MPI_SUCCESS;
    viadev_connection_t *c = &viadev.connections[dest_grank];
    vbuf *v;
    void *buffer;
    viadev_packet_rendezvous_start *packet;
    int proto = VIADEV_PROTOCOL_R3;
    int rdma_ok = fast_rdma_ok(c, 
            sizeof(viadev_packet_rendezvous_start), 0);

    if (rdma_ok) {
        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        v->shandle = NULL;
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    } else {
        v = get_vbuf();
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    }


#else
    int rc = MPI_SUCCESS;

    vbuf *v = get_vbuf();
    void *buffer = VBUF_BUFFER_START(v);
    viadev_packet_rendezvous_start *packet = buffer;
    viadev_connection_t *c = &viadev.connections[dest_grank];
    int proto = VIADEV_PROTOCOL_R3;
#endif

#if defined(VIADEV_RPUT_SUPPORT)||defined(VIADEV_RGET_SUPPORT)
    /* if this fails, will return NULL, which is what we want. 
     * So no need to do anything other than try to register. 
     */
    dreg_entry = NULL;

#ifdef VIADEV_RPUT_SUPPORT
    proto = VIADEV_PROTOCOL_RPUT;
    D_PRINT("rendzvous enter: set RPUT (proto=%d)\n", proto);
#else
    proto = VIADEV_PROTOCOL_RGET;
#endif

    if (0 == len) {
        proto = VIADEV_PROTOCOL_R3;
        D_PRINT("rendzvous enter: set R3 since len=0 (proto=%d)\n", proto);
    } else if ((buf != NULL) && (len > 0)) {

        dreg_entry = dreg_register(buf, len, DREG_ACL_READ);
        if (NULL == dreg_entry)
            proto = VIADEV_PROTOCOL_R3;

    }
#endif

    /* fill in the packet */
    PACKET_SET_HEADER(packet, c, VIADEV_PACKET_RENDEZVOUS_START);
    PACKET_SET_ENVELOPE(packet, context_id, tag, len, src_lrank);
    packet->sreq = REQ_TO_ID(shandle);

#if defined(VIADEV_RPUT_SUPPORT)||defined(VIADEV_RGET_SUPPORT)
    /* may be less than receive buffer size */
    packet->len = len;
    packet->buffer_address = buf;

    if (NULL != dreg_entry) {
        packet->memhandle = dreg_entry->memhandle;
    }
#endif

    /* explicitly set protocol */
    shandle->protocol = proto;
    packet->protocol = proto;

#if defined(RDMA_FAST_PATH)
    if (rdma_ok) {
        post_fast_rdma_with_completion(c,
                                       sizeof
                                       (viadev_packet_rendezvous_start));
    } else {
        vbuf_init_send(v, sizeof(viadev_packet_rendezvous_start));
        viadev_post_send(c, v);
    }
#else
    /* prepare descriptor and post */
    vbuf_init_send(v, sizeof(viadev_packet_rendezvous_start));

    viadev_post_send(c, v);
#endif

    /* we can post the vbuf before updating the send handle. 
     * If we go to a multithreaded implementation, in which 
     * another thread may process the completed vbuf, we will
     * need to update the shandle before posting the send
     * to avoid a race condition
     */

    shandle->local_address = buf;
#if defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)
    shandle->dreg_entry = dreg_entry;
#endif
    shandle->bytes_total = len;
    shandle->bytes_sent = 0;
    shandle->connection = c;
    shandle->is_complete = 0;

    /* fill in when we get the reply:
     * receive_id, remote_address, memory_handle
     */

    return rc;
}


int MPID_VIA_self_start(void *buf, int len, int src_lrank, int tag,
                        int context_id, MPIR_SHANDLE * shandle)
{

    MPIR_RHANDLE *rhandle;
    int rc = MPI_SUCCESS;
    int found;

    /* receive has been posted and is on the posted queue */
    MPID_Msg_arrived(src_lrank, tag, context_id, &rhandle, &found);

    if (found) {
        memcpy(rhandle->buf, buf, len);
        rhandle->s.MPI_TAG = tag;
        rhandle->s.MPI_SOURCE = src_lrank;
        RECV_COMPLETE(rhandle);

        rhandle->s.count = len;

        SEND_COMPLETE(shandle);
        /* need to call the finish routine for the receive? */
        return rc;
    }

    /* receive handle has been created 
     * and posted on unexpected queue */
    rhandle->connection = NULL;

    if (len < viadev_rendezvous_threshold) {
        /* copy the buffer and mark the send complete so there is a bit of
         * buffering for silly programs that send-to-self blocking.  Hide
         * info on the send in vbuf_{head,tail} which we know will not
         * be used.  */
        rhandle->send_id = 0;
        rhandle->vbuf_head = malloc(len);
        rhandle->vbuf_tail = (void *) ((aint_t) len);
        memcpy(rhandle->vbuf_head, buf, len);
        SEND_COMPLETE(shandle);
    } else {
        rhandle->send_id = REQ_TO_ID(shandle);
        shandle->local_address = buf;
        shandle->is_complete = 0;
        shandle->bytes_total = len;
    }

    return rc;
}

void MPID_VIA_self_finish(MPIR_RHANDLE * rhandle,
                          MPIR_RHANDLE * unexpected)
{
    MPIR_SHANDLE *shandle;

    void *send_ptr;
    int bytes_total;
    int truncate = 0;


    shandle = (MPIR_SHANDLE *) ID_TO_REQ(unexpected->send_id);

    if (shandle) {
        send_ptr = shandle->local_address;
        bytes_total = shandle->bytes_total;
    } else {
        send_ptr = unexpected->vbuf_head;
        bytes_total = (aint_t) unexpected->vbuf_tail;
    }

    if (bytes_total > rhandle->len)
        /* user error */
        truncate = 1;
    else
        /* fewer or same bytes were sent as were specified in the recv */
        rhandle->len = bytes_total;

    memcpy(rhandle->buf, send_ptr, rhandle->len);

    /* this is similate to copy_unexpected_handle_to_user_handle, but
     * we need many fewer fields 
     */
    rhandle->s.MPI_TAG = unexpected->s.MPI_TAG;
    rhandle->s.MPI_SOURCE = unexpected->s.MPI_SOURCE;
    rhandle->connection = NULL;
    RECV_COMPLETE(rhandle);

    if (truncate)
        rhandle->s.MPI_ERROR = MPI_ERR_TRUNCATE;

    if (shandle) {
        SEND_COMPLETE(shandle);
    } else {
        free(unexpected->vbuf_head);
    }
}


int viadev_eager_ok(int len, viadev_connection_t * c)
{
    int nvbufs;
    nvbufs = viadev_calculate_vbufs_expected(len, VIADEV_PROTOCOL_EAGER);

    if (c->remote_credit - nvbufs >= viadev_credit_preserve) {
        return 1;
    }

    return 0;
}

int MPID_VIA_eager_send(void *buf, int len, int src_lrank, int tag,
                        int context_id, int dest_grank, MPIR_SHANDLE * s)
{

    int startmax = VBUF_DATA_SIZE(viadev_packet_eager_start);
    int nextmax = VBUF_DATA_SIZE(viadev_packet_eager_next);
    viadev_packet_eager_start *h;
    viadev_packet_eager_next *d;
    void *databuf;
    viadev_connection_t *c = &viadev.connections[dest_grank];
    vbuf *v;

#ifdef RDMA_FAST_PATH
    int rdma_ok;
#endif

    int bytes_sent;

    D_PRINT("eager send \n");

    /*
     * we could fill in the send handle, but I think there is no
     * need, since we are sending all the data right here and now. 
     * However, is_complete should be set to zero because
     * the MPI send does not actually complete until the
     * VIA send completes. 
     */

    s->is_complete = 0;
    s->protocol = VIADEV_PROTOCOL_EAGER;


    /* Set dreg_entry to NULL because SEND_COMPLETE will try to decrement
     * the reference count if it is non-null
     */
#if defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)
    s->dreg_entry = NULL;
#endif

#if defined(RDMA_FAST_PATH)
    rdma_ok = fast_rdma_ok(c, 
            sizeof(viadev_packet_eager_start) + len, 1);


#ifndef DISABLE_HEADER_CACHING
    if (rdma_ok && (len < viadev_max_fast_eager_size)) {

        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        h = (viadev_packet_eager_start *) v->buffer;
        /* set up the packet */
        PACKET_SET_ENVELOPE((&viadev.match_hdr), context_id, tag, len, src_lrank);
        PACKET_SET_CREDITS(((viadev_packet_header *) &viadev.match_hdr), c);

#ifdef EARLY_SEND_COMPLETION
        v->shandle = NULL;
#else
        /* set the handle */
        v->shandle = s;
#endif

        if (search_header_cache(c, 
                    (viadev_packet_eager_start *)&viadev.match_hdr)) {
            PACKET_SET_HEADER(h, c, FAST_EAGER_CACHED);
            h->header.fast_eager_size = (unsigned char) len;
            databuf = ((char *) h) + FAST_EAGER_HEADER_SIZE;
            memcpy(databuf, buf, len);
            /* figure out how much data to send in this packet */
            post_fast_rdma_with_completion(c,
                                           FAST_EAGER_HEADER_SIZE + len);

        } else {
            PACKET_SET_ENVELOPE(h, context_id, tag, len, src_lrank);
            PACKET_SET_HEADER(h, c, VIADEV_PACKET_EAGER_START);

            /* Set some of the credit fields */
            h->header.vbuf_credit = viadev.match_hdr.header.vbuf_credit;
            h->header.rdma_credit = viadev.match_hdr.header.rdma_credit;
            h->header.remote_credit = viadev.match_hdr.header.remote_credit;
            
            databuf = ((char *) h) + sizeof(viadev_packet_eager_start);
            memcpy(databuf, buf, len);
            h->bytes_in_this_packet = len;
            /* figure out how much data to send in this packet */
            post_fast_rdma_with_completion(c,
                                           sizeof
                                           (viadev_packet_eager_start) +
                                           len);
        }

#ifdef EARLY_SEND_COMPLETION
        /* mark the send as completed */
        SEND_COMPLETE(s);
#endif
        return DAT_SUCCESS;

    }
#endif                          /* of HEADER_CACHING */

    if (rdma_ok) {

        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        h = (viadev_packet_eager_start *) v->buffer;
        databuf = ((char *) h) + sizeof(viadev_packet_eager_start);

        /* figure out how much data to send in this packet */
        h->bytes_in_this_packet = len;

        if (startmax < h->bytes_in_this_packet) {
            h->bytes_in_this_packet = startmax;
        }

        /* copy next chunk of data in to vbuf */
        if (h->bytes_in_this_packet != 0) {
            memcpy(databuf, buf, h->bytes_in_this_packet);
        }

        bytes_sent = h->bytes_in_this_packet;

        /* set up the packet */
        PACKET_SET_HEADER(h, c, VIADEV_PACKET_EAGER_START);
        PACKET_SET_ENVELOPE(h, context_id, tag, len, src_lrank);

#ifdef EARLY_SEND_COMPLETION
        v->shandle = NULL;
#else
        /*
         * This is a signal to the VIA send completion handler
         * that the MPI send should be marked complete.
         */
        if (bytes_sent == len) {
            v->shandle = s;
        } else {
            v->shandle = NULL;
        }
#endif
        post_fast_rdma_with_completion(c,
                                       sizeof(viadev_packet_eager_start) +
                                       h->bytes_in_this_packet);
        /* Multiple packet eager send over RDMA */
        while (bytes_sent < len) {

            /* Get the next vbuf */
            v = &(c->RDMA_send_buf[c->phead_RDMA_send]);

            /* figure out where header and data are */
            d = (viadev_packet_eager_next *) VBUF_BUFFER_START(v);
            databuf = ((char *) d) + sizeof(viadev_packet_eager_next);

            /* set up the packet */
            PACKET_SET_HEADER(d, c, VIADEV_PACKET_EAGER_NEXT);

            /* figure out how much data to send in this packet */
            d->bytes_in_this_packet = len - bytes_sent;
            if (nextmax < d->bytes_in_this_packet) {
                d->bytes_in_this_packet = nextmax;
            }

            /* copy next chunk of data in to vbuf */
            memcpy(databuf, ((char *) buf) + bytes_sent,
                    d->bytes_in_this_packet);

            bytes_sent += d->bytes_in_this_packet;

            /* finally send it off */
#ifdef EARLY_SEND_COMPLETION
            v->shandle = NULL;
#else
            /*
             * This is a signal to the VIA send completion handler
             * that the MPI send should be marked complete.
             */
            if (bytes_sent == len) {
                v->shandle = s;
            } else {
                v->shandle = NULL;
            }
#endif

            /* figure out how much data to send in this packet */
            post_fast_rdma_with_completion(c,
                    sizeof(viadev_packet_eager_next) +
                    d->bytes_in_this_packet);
        }

#ifdef EARLY_SEND_COMPLETION
        /* mark the send as completed */
        SEND_COMPLETE(s);
#endif

        return DAT_SUCCESS;
    }
#endif                          /* RDMA_FAST_PATH */

    /* If control reaches here, we have fallen back on 
     * send/recv channel for eager messages */

    v = get_vbuf();
    h = (viadev_packet_eager_start *) VBUF_BUFFER_START(v);
    databuf = ((char *) h) + sizeof(viadev_packet_eager_start);

    /* set up the packet */
    PACKET_SET_HEADER(h, c, VIADEV_PACKET_EAGER_START);
    PACKET_SET_ENVELOPE(h, context_id, tag, len, src_lrank);

    /* figure out how much data to send in this packet */
    h->bytes_in_this_packet = len;

    if (startmax < h->bytes_in_this_packet) {
        h->bytes_in_this_packet = startmax;
    }

    /* copy next chunk of data in to vbuf */

    if (h->bytes_in_this_packet != 0) {
        memcpy(databuf, buf, h->bytes_in_this_packet);
    }

    bytes_sent = h->bytes_in_this_packet;

    /* finally send it off */
    vbuf_init_send(v, (sizeof(viadev_packet_eager_start) +
                       h->bytes_in_this_packet));

#ifdef EARLY_SEND_COMPLETION
    v->shandle = NULL;
#else
    /*
     * This is a signal to the VIA send completion handler
     * that the MPI send should be marked complete.
     */
    if (bytes_sent == len) {
        v->shandle = s;
    } else {
        v->shandle = NULL;
    }
#endif

    viadev_post_send(c, v);

    while (bytes_sent < len) {

        v = get_vbuf();

        /* figure out where header and data are */
        d = (viadev_packet_eager_next *) VBUF_BUFFER_START(v);
        databuf = ((char *) d) + sizeof(viadev_packet_eager_next);

        /* set up the packet */
        PACKET_SET_HEADER(d, c, VIADEV_PACKET_EAGER_NEXT);

        /* figure out how much data to send in this packet */
        d->bytes_in_this_packet = len - bytes_sent;
        if (nextmax < d->bytes_in_this_packet) {
            d->bytes_in_this_packet = nextmax;
        }

        /* copy next chunk of data in to vbuf */
        memcpy(databuf, ((char *) buf) + bytes_sent,
               d->bytes_in_this_packet);

        bytes_sent += d->bytes_in_this_packet;

        /* finally send it off */
        vbuf_init_send(v, (sizeof(viadev_packet_eager_next) +
                           d->bytes_in_this_packet));

#ifdef EARLY_SEND_COMPLETION
        v->shandle = NULL;
#else
        /*
         * This is a signal to the VIA send completion handler
         * that the MPI send should be marked complete.
         */
        if (bytes_sent == len) {
            v->shandle = s;
        } else {
            v->shandle = NULL;
        }
#endif
        viadev_post_send(c, v);
    }

    /* NOTE: even with the backlog queue, this assertion should
     * be true in a single threaded implementation.
     * The backlog queue only has something on it if remote_credit == 0.
     * We did a eager send because there were enough remote credits
     * to handle the transfer and these should have been the only
     * sends we posted since that check.
     */
    assert(c->remote_credit >= viadev_credit_preserve);

#ifdef EARLY_SEND_COMPLETION
    /* mark the send as completed */
    SEND_COMPLETE(s);
#endif

    return DAT_SUCCESS;
}


#ifdef VIADEV_RPUT_SUPPORT
void viadev_rput(viadev_connection_t * c, vbuf * v, void *local_address,
                 VIP_MEM_HANDLE local_memhandle, void *remote_address,
                 VIP_MEM_HANDLE remote_memhandle, int nbytes)
{
#ifdef VIADEV_HAVE_RDMA_LIMIT
    while (viadev.outstanding_rdmas >= viadev_rdma_limit)
        MPID_DeviceCheck(MPID_BLOCKING);
    viadev.outstanding_rdmas++;
#endif
    D_PRINT("viadev_rput: RDMA write\n");
    vbuf_init_rput(v, local_address, local_memhandle,
                   remote_address, remote_memhandle, nbytes);
    viadev_post_rdmawrite(c, v);
}


#endif


