/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include "mpid_bind.h"
#include "viadev.h"
#include "viaparam.h"
#include "viapriv.h"
#include "mpid_smpi.h"
#include <unistd.h>

void MPID_SendContig(struct MPIR_COMMUNICATOR *comm_ptr,
                     void *buf,
                     int len,
                     int src_lrank,
                     int tag,
                     int context_id,
                     int dest_grank, MPID_Msgrep_t msgrep, int *error_code)
{



    MPIR_SHANDLE *shandle;

    MPID_SendAlloc(shandle);
    
    /* argument checking will be done by IsendContig */

    MPID_IsendContig(comm_ptr, buf, len, src_lrank, tag, context_id,
                     dest_grank, msgrep, (MPI_Request) shandle,
                     error_code);
    while (!shandle->is_complete)
        MPID_DeviceCheck(MPID_BLOCKING);

    MPID_SendFree(shandle);

}

void MPID_SsendContig(struct MPIR_COMMUNICATOR *comm_ptr,
                      void *buf,
                      int len,
                      int src_lrank,
                      int tag,
                      int context_id,
                      int dest_grank,
                      MPID_Msgrep_t msgrep, int *error_code)
{



    MPIR_SHANDLE *shandle;

    MPID_SendAlloc(shandle);

    /* argument checking will be done by IsendContig */

    MPID_IssendContig(comm_ptr, buf, len, src_lrank, tag, context_id,
                      dest_grank, msgrep, (MPI_Request) shandle,
                      error_code);

    while (!shandle->is_complete)
        MPID_DeviceCheck(MPID_BLOCKING);


    MPID_SendFree(shandle);

}




void MPID_IsendContig(struct MPIR_COMMUNICATOR *comm_ptr,
                      void *buf,
                      int len,
                      int src_lrank,
                      int tag,
                      int context_id,
                      int dest_grank,
                      MPID_Msgrep_t msgrep,
                      MPI_Request request, int *error_code)
{

    int rc;
    MPIR_SHANDLE *shandle = (MPIR_SHANDLE *) request;

    /*
     * This is the one argument check done in the model adi2 implementation 
     */

    if (buf == 0 && len > 0) {
        *error_code = MPI_ERR_BUFFER;
        return;
    }
#ifdef _SMP_
    if ((smpi.local_nodes[dest_grank] != -1)
        && (smpi.my_local_id != smpi.local_nodes[dest_grank])) {
        if (len < smp_eagersize){
           MPID_SMP_Eagerb_isend_short(buf, len, src_lrank,
                                      tag, context_id, dest_grank, msgrep,
                                      shandle);

           if(NULL != request->shandle.finish)
              request->shandle.finish(&(request->shandle));
           return;
        }

#ifdef _SMP_RNDV_
          MPID_SMP_Rndvn_isend(buf, len, src_lrank,
                               tag, context_id, dest_grank, msgrep,
                               shandle);
          return;
#endif
    }
#endif


    if (dest_grank == viadev.me) {
        rc = MPID_VIA_self_start(buf, len, src_lrank, tag,
                                 context_id, shandle);
    } else {

        if (len < viadev_rendezvous_threshold &&
            viadev_eager_ok(len, &viadev.connections[dest_grank])) {
            rc = MPID_VIA_eager_send(buf, len, src_lrank,
                                     tag, context_id, dest_grank, shandle);
        } else {
            if (NULL == buf || 0 == len) {
                buf = &nullsbuffer;
            }
            rc = MPID_VIA_rendezvous_start(buf, len, src_lrank,
                                           tag, context_id,
                                           dest_grank, shandle);
        }
    }
}



void MPID_IssendContig(struct MPIR_COMMUNICATOR *comm_ptr,
                       void *buf,
                       int len,
                       int src_lrank,
                       int tag,
                       int context_id,
                       int dest_grank,
                       MPID_Msgrep_t msgrep,
                       MPI_Request request, int *error_code)
{

    int rc;
    MPIR_SHANDLE *shandle = (MPIR_SHANDLE *) request;


    /*
     * This is the one argument check done in the model adi2 implementation 
     * Why isn't it done inside MPI? xxx
     */
    if (buf == 0 && len > 0) {
        *error_code = MPI_ERR_BUFFER;
        return;
    }

    if (dest_grank == viadev.me) {
        rc = MPID_VIA_self_start(buf, len, src_lrank, tag, context_id,
                                 shandle);
    } else {
        if (buf == NULL || len == 0) {
            buf = &nullsbuffer;
        }
#ifdef _SMP_
#ifdef _SMP_RNDV_
        if (smpi.local_nodes[dest_grank] != -1) {
           MPID_SMP_Rndvn_isend(buf, len, src_lrank, tag, context_id, dest_grank,
                                msgrep, shandle );
           return;
        }
#endif
#endif

        rc = MPID_VIA_rendezvous_start(buf, len, src_lrank,
                                       tag, context_id, dest_grank,
                                       shandle);
    }


}


void MPID_SendComplete(MPI_Request request, int *error_code)
{
    MPIR_SHANDLE *shandle = (MPIR_SHANDLE *) request;
    while (!shandle->is_complete)
        MPID_DeviceCheck(MPID_BLOCKING);
}

int MPID_SendIcomplete(MPI_Request request, int *error_code)
{
    MPIR_SHANDLE *shandle = (MPIR_SHANDLE *) request;
    MPID_DeviceCheck(MPID_NOTBLOCKING);
    if (shandle->is_complete)
        return 1;
    return 0;
}

void MPID_SendCancel(MPI_Request r, int *error_code)
{
#ifdef VIADEV_SEND_CANCEL
    MPIR_SHANDLE *shandle = (MPIR_SHANDLE *) r;

    if (shandle->is_complete) {
        /* too late to cancel, send is already complete */

    } else {
        /* Try to suspend the send operations on this end */
        shandle->send_cancel = MPID_SC_STALL;

        /* Send a cancel message to the remote process */

        /* Wait for message from remote process */
    }
#else
    vapidev_error_abort(GEN_EXIT_ERR, "MPID_SendCancel not implemented");
#endif
}

#ifdef MCST_SUPPORT

void MPID_RC_Send(void *, int, int, int, int, int, int *);
void MPID_RC_Send(void *buf,
                  int len,
                  int src_lrank,
                  int tag, int context_id, int dest_grank, int *error_code)
{

    int rc;
    MPIR_SHANDLE *shandle;

    MPID_SendAlloc(shandle);


    if (buf == 0 && len > 0) {
        *error_code = MPI_ERR_BUFFER;
        return;
    }

    if (dest_grank == viadev.me) {
        rc = MPID_VIA_self_start(buf, len, src_lrank, tag, context_id,
                                 shandle);
    } else {
        if (len < viadev_rendezvous_threshold &&
            viadev_eager_ok(len, &viadev.connections[dest_grank])) {
            rc = MPID_VIA_eager_send(buf, len, src_lrank,
                                     tag, context_id, dest_grank, shandle);
        } else {
            if (buf == NULL || len == 0) {
                buf = &nullsbuffer;
            }
            rc = MPID_VIA_rendezvous_start(buf, len, src_lrank,
                                           tag, context_id,
                                           dest_grank, shandle);
        }
    }
    co_print("Wait for device\n");
    co_print("The_shandle:%x\n", shandle);

    while (!shandle->is_complete)
        MPID_DeviceCheck(MPID_BLOCKING);
    co_print("Send complete\n");

    MPID_SendFree(shandle);
}

#endif
