/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#define _XOPEN_SOURCE 600
#include "vapi_header.h"

#include "vbuf.h"
#include "viutil.h"
#include <stdio.h>


#ifdef MALLOC_HOOK
#ifndef MAC_OSX
#include <malloc.h>             /* malloc */
#endif
#include <stdlib.h>
#else
#include <stdlib.h>
#endif                          /* MALLOC_HOOK */
#ifndef MAC_OSX
#include <malloc.h>
#endif
#include <stdlib.h>

#include <unistd.h>

#include "viapriv.h"            /* for viadevdebug */
#include "viaparam.h"
#include <string.h>

#include "dreg.h"


/*
 * vbufs
 * 
 * vbufs provide system buffers for VMPICH. They are analogous to mbufs
 * in BSD networking. 
 * The primary motivation for vbufs is that implementing MPI on VIA
 * seems to requiring pre-posting a number of fixed-sized buffers. 
 * These buffers must be registered (pinned). Life is easier if 
 * they are all registered at once so there is only one memory 
 * handle. We manage a fixed-size pool of vbufs that are
 * allocated and pinned when a progam starts up. We manage
 * the free vbuf list as a singly linked list. 
 * 
 *  Two different ways to manage the free list as a singly-linked list. 
 *  1. head and tail pointers. Add to tail, remove from head. 
 *  2. only head pointer, treat as a stack. 
 * 
 *  #1 Eliminates contention between adding to list and removing from list 
 *   Lock-free possible?
 * 
 *  #2 Has slightly less overhead when there is no contention, and is more 
 *  likely to produce a vbuf that is already in cache. 
 * 
 *  Currently anticipate that most access near-term will be single-threaded, 
 *  so go with head only.  (#2)
 */

/* head of list of allocated vbuf regions */
static vbuf_region *vbuf_region_head = NULL;

/* 
 * free_vbuf_head is the head of the free list
 */

static vbuf *free_vbuf_head = NULL;

/*
 * cache the nic handle, and ptag the first time a region is
 * allocated (at init time) for later additional vbur allocations
 */
static VAPI_hca_hndl_t nic_save = INVAL_HNDL;
static VAPI_pd_hndl_t ptag_save = INVAL_HNDL;

static int vbuf_n_allocated = 0;
static long num_free_vbuf = 0;
static long num_vbuf_get = 0;
static long num_vbuf_free = 0;


void dump_vbuf_region(vbuf_region * r)
{
}

void dump_vbuf_regions()
{
    vbuf_region *r = vbuf_region_head;

    while (r) {
        dump_vbuf_region(r);
        r = r->next;
    }
}
void deallocate_vbufs(VAPI_hca_hndl_t nic)
{
    vbuf_region *r = vbuf_region_head;
    VAPI_ret_t ret;

    while (r) {
        if (r->mem_handle.hndl != VAPI_INVAL_HNDL) {
            ret = VAPI_deregister_mr(nic, r->mem_handle.hndl);
            if (VAPI_OK != ret)
                vapidev_error_abort(VAPI_RETURN_ERR,
                                    "could not deregister MR (%s)",
                                    VAPI_strerror(ret));

            /* free vbufs add it later */
        }
        D_PRINT("deregister vbufs\n");
        r = r->next;
    }
}

static void allocate_vbuf_region(int nvbufs)
{
    struct vbuf_region *reg;
    void *mem;
    void *vbuf_dma_buffer;
    VIP_MEM_HANDLE mem_handle;

    int i;
    vbuf *cur;
    int alignment_vbuf = 64;
    int alignment_dma;

    alignment_dma = getpagesize();

    if (free_vbuf_head != NULL)
        vapidev_error_abort(GEN_ASSERT_ERR, "free_vbuf_head = NULL");
    /* are we limiting vbuf allocation?  If so, make sure
     * we dont alloc more than allowed
     */
    if (viadev_vbuf_max > 0) {
        nvbufs = MIN(nvbufs, viadev_vbuf_max - vbuf_n_allocated);
        if (nvbufs <= 0) {
            vapidev_error_abort(GEN_EXIT_ERR,
                                "VBUF alloc failure, limit exceeded");
        }
    }

    SET_ORIGINAL_MALLOC_HOOKS;

    reg = (struct vbuf_region *) malloc(sizeof(struct vbuf_region));
    if (NULL == reg) {
        vapidev_error_abort(GEN_EXIT_ERR,
                            "Unable to malloc a new struct vbuf_region");
    }

    if(posix_memalign((void **) &mem, alignment_vbuf, nvbufs * sizeof(vbuf))) {
        vapidev_error_abort(GEN_EXIT_ERR, 
                "unable to malloc vbuf struct");
    }

    /* ALLOCATE THE DMA BUFFER */

    if(posix_memalign((void **) &vbuf_dma_buffer, alignment_dma,
                nvbufs * viadev_vbuf_total_size)) {
        vapidev_error_abort(GEN_EXIT_ERR, 
                "unable to malloc vbufs DMA buffer");
    }

    SAVE_MALLOC_HOOKS;
    SET_ORIGINAL_MALLOC_HOOKS;

    memset(mem, 0, nvbufs * sizeof(vbuf));
    memset(vbuf_dma_buffer, 0, nvbufs * viadev_vbuf_total_size);

    vbuf_n_allocated += nvbufs;
    num_free_vbuf += nvbufs;
    reg->malloc_start = mem;

    reg->malloc_buf_start = vbuf_dma_buffer;
    reg->malloc_end = (void *) ((char *) mem + nvbufs * sizeof(vbuf));
    reg->malloc_buf_end = (void *) ((char *) vbuf_dma_buffer + 
            nvbufs * viadev_vbuf_total_size);
    
    reg->count = nvbufs;

    free_vbuf_head = (vbuf *) ((aint_t) mem);
    
    reg->vbuf_head = free_vbuf_head;

    D_PRINT("VBUF REGION ALLOCATION SZ %d TOT %d FREE %ld NF %ld NG %ld",
            nvbufs, vbuf_n_allocated, num_free_vbuf,
            num_vbuf_free, num_vbuf_get);

    /* REGISTER DMA BUFFER */
    REGISTER_VBUF_DMA(ptag_save, nic_save, mem_handle, 
            nvbufs * viadev_vbuf_total_size, vbuf_dma_buffer);

    reg->mem_handle = mem_handle;


    /* init the free list */
    for (i = 0; i < nvbufs - 1; i++) {
        cur = free_vbuf_head + i;

        cur->desc.next = free_vbuf_head + i + 1;
        cur->region = reg;

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        cur->head_flag = (VBUF_FLAG_TYPE *) ((char *)(vbuf_dma_buffer) + 
                (i * viadev_vbuf_total_size));
        cur->buffer = (char *) ((char *)(cur->head_flag) + 
                sizeof(VBUF_FLAG_TYPE));
#else
        cur->buffer = (char *) ((char *)(vbuf_dma_buffer) +
                (i * viadev_vbuf_total_size));
#endif
    }
    /* last one needs to be set to NULL */
    cur = free_vbuf_head + nvbufs - 1;

    cur->desc.next = NULL;

    cur->region = reg;

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    cur->head_flag = (VBUF_FLAG_TYPE *) ((char *)vbuf_dma_buffer + 
            ((nvbufs - 1) * viadev_vbuf_total_size));

    cur->buffer = (char *) ((char *)(cur->head_flag) + 
            sizeof(VBUF_FLAG_TYPE));
#else
    cur->buffer = (char *) ((char *)vbuf_dma_buffer + 
            ((nvbufs - 1) * viadev_vbuf_total_size));

#endif

    /* thread region list */
    reg->next = vbuf_region_head;
    vbuf_region_head = reg;
}
void allocate_vbufs(VAPI_hca_hndl_t nic, VAPI_pd_hndl_t ptag, int nvbufs)
{
    /* this function is only called by the init routines.
     * cache the nic handle and ptag for later vbuf_region allocations
     */
    nic_save = nic;
    ptag_save = ptag;

    /* now allocate the first vbuf region */
    allocate_vbuf_region(nvbufs);
}


/* 
 * Get a vbuf off the free list 
 */

vbuf *get_vbuf(void)
{
    vbuf *v;
    /* 
     * It will often be possible for higher layers to recover
     * when no vbuf is available, but waiting for more descriptors
     * to complete. For now, just abort. 
     */
    if (NULL == free_vbuf_head) {
        D_PRINT("Allocating new vbuf region\n");
        allocate_vbuf_region(viadev_vbuf_secondary_pool_size);
        if (NULL == free_vbuf_head) {
            vapidev_error_abort(GEN_EXIT_ERR,
                                "No free vbufs. Pool size %d",
                                vbuf_n_allocated);
        }
    }
    v = free_vbuf_head;
    num_free_vbuf--;
    num_vbuf_get++;

    /* this correctly handles removing from single entry free list */
    free_vbuf_head = free_vbuf_head->desc.next;
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    /* need to change this to RPUT_VBUF_FLAG or RGET_VBUF_FLAG later
     * if we are doing rput */
    v->padding = NORMAL_VBUF_FLAG;
#endif

    /* this is probably not the right place to initialize shandle to NULL.
     * Do it here for now because it will make sure it is always initialized.
     * Otherwise we would need to very carefully add the initialization in
     * a dozen other places, and probably miss one. 
     */
    v->shandle = NULL;

    return (v);
}

/*
 * Put a vbuf back on the free list 
 */

void release_vbuf(vbuf * v)
{
    /* note this correctly handles appending to empty free list */

    D_PRINT("release_vbuf: releasing %p previous head = %p",
            v, free_vbuf_head);

    assert(v != free_vbuf_head);

    v->desc.next = free_vbuf_head;

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    if ((v->padding != NORMAL_VBUF_FLAG)
        && (v->padding != RPUT_VBUF_FLAG)
        && (v->padding != RGET_VBUF_FLAG)) {
        vapidev_error_abort(GEN_EXIT_ERR, "vbuf %p not correct!!! %d %d %d %d\n", 
                v, v->padding, NORMAL_VBUF_FLAG, RPUT_VBUF_FLAG, RGET_VBUF_FLAG);
    }
#endif


    free_vbuf_head = v;
    num_free_vbuf++;
    num_vbuf_free++;
}


/*
 * fill in vbuf descriptor with all necessary info 
 */

#ifdef MCST_SUPPORT
void vbuf_init_msend(vbuf *, unsigned long);

void vbuf_init_msend(vbuf * v, unsigned long len)
{
    v->desc.sr.comp_type = VAPI_SIGNALED;
    v->desc.sr.opcode = VAPI_SEND;
    v->desc.sr.remote_qkey = 0;
    v->desc.sr.remote_qp = IB_MULTICAST_QP;
    v->desc.sr.remote_ah = viadev.av_hndl;      /* Check this out .. */
    v->desc.sr.id = (aint_t) v;
    v->desc.sr.sg_lst_len = 1;
    v->desc.sr.sg_lst_p = &(v->desc.sg_entry);
    v->desc.sg_entry.len = len;
    v->desc.sg_entry.lkey = v->region->mem_handle.lkey;
    v->desc.sg_entry.addr = (VAPI_virt_addr_t) (virt_addr_t) (v->buffer);
    v->desc.sr.set_se = FALSE;
    v->desc.sr.fence = FALSE;
}
#endif


void vbuf_init_send(vbuf * v, unsigned long len)
{
    v->desc.sr.comp_type = VAPI_SIGNALED;
    v->desc.sr.opcode = VAPI_SEND;
    v->desc.sr.remote_qkey = 0;
    v->desc.sr.remote_qp = 0;
    v->desc.sr.id = (aint_t) v;
    v->desc.sr.sg_lst_len = 1;
    v->desc.sr.sg_lst_p = &(v->desc.sg_entry);
    v->desc.sg_entry.len = len;
    v->desc.sg_entry.lkey = v->region->mem_handle.lkey;
    v->desc.sg_entry.addr = (VAPI_virt_addr_t) (virt_addr_t) (v->buffer);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    v->padding = NORMAL_VBUF_FLAG;
#endif
}

void vbuf_init_recv(vbuf * v, unsigned long len)
{
    v->desc.rr.comp_type = VAPI_SIGNALED;
    v->desc.rr.opcode = VAPI_RECEIVE;
    v->desc.rr.id = (aint_t) v;
    v->desc.rr.sg_lst_len = 1;
    v->desc.rr.sg_lst_p = &(v->desc.sg_entry);
    v->desc.sg_entry.len = len;
    v->desc.sg_entry.lkey = v->region->mem_handle.lkey;
    v->desc.sg_entry.addr = (VAPI_virt_addr_t) (virt_addr_t) (v->buffer);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    v->padding = NORMAL_VBUF_FLAG;
#endif
}
void vbuf_init_sendrecv(vbuf * v, unsigned long len)
{
    vapidev_error_abort(GEN_EXIT_ERR, "don't use vbuf_init_sendrecv, "
                        "use vbuf_init_send or vbuf_init_recv\n");
}

void vbuf_init_sendrecv_immediate(vbuf * v, unsigned long len,
                                  int immediate)
{
    vapidev_error_abort(GEN_EXIT_ERR, "vbuf_init_sendrecv_immediate: "
                        "not implemented yet!\n");
}

void vbuf_init_rput(vbuf * v, void *local_address,
                    VIP_MEM_HANDLE local_memhandle, void *remote_address,
                    VIP_MEM_HANDLE remote_memhandle, int len)
{
    v->desc.sr.comp_type = VAPI_SIGNALED;
    v->desc.sr.opcode = VAPI_RDMA_WRITE;
    v->desc.sr.remote_qkey = 0;
    v->desc.sr.remote_qp = 0;
    v->desc.sr.id = (aint_t) v;
    v->desc.sr.sg_lst_len = 1;
    v->desc.sr.remote_addr =
        (aint_t) (VAPI_virt_addr_t) (virt_addr_t) (remote_address);
    v->desc.sr.r_key = remote_memhandle.rkey;
    v->desc.sr.sg_lst_p = &(v->desc.sg_entry);
    v->desc.sg_entry.len = len;
    v->desc.sg_entry.lkey = local_memhandle.lkey;
    v->desc.sg_entry.addr =
        (VAPI_virt_addr_t) (virt_addr_t) (local_address);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    v->padding = RPUT_VBUF_FLAG;
#endif

    D_PRINT("RDMA write\n");
}


#ifdef VIADEV_RGET_SUPPORT

void vbuf_init_rget(vbuf * v,
                    void *local_address,
                    VIP_MEM_HANDLE local_memhandle,
                    void *remote_address,
                    VIP_MEM_HANDLE remote_memhandle, int len)
{
    v->desc.sr.comp_type = VAPI_SIGNALED;
    v->desc.sr.opcode = VAPI_RDMA_READ;
    v->desc.sr.remote_qkey = 0;
    v->desc.sr.remote_qp = 0;
    v->desc.sr.id = (aint_t) v;
    v->desc.sr.sg_lst_len = 1;
    v->desc.sr.remote_addr =
        (aint_t) (VAPI_virt_addr_t) (virt_addr_t) (remote_address);
    v->desc.sr.r_key = remote_memhandle.rkey;
    v->desc.sr.sg_lst_p = &(v->desc.sg_entry);
    v->desc.sg_entry.len = len;
    v->desc.sg_entry.lkey = local_memhandle.lkey;
    v->desc.sg_entry.addr =
        (VAPI_virt_addr_t) (virt_addr_t) (local_address);

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    v->padding = RGET_VBUF_FLAG;
#endif

}
#endif

/*
 * print out vbuf contents for debugging 
 */

void dump_vbuf(char *msg, vbuf * v)
{
    int i, len;
    viadev_packet_header *header = NULL;
    header = (viadev_packet_header *) VBUF_BUFFER_START(v);
    D_PRINT("%s: dump of vbuf %p, seq#=%d, type = %d\n",
            msg, v, header->id, header->type);
    len = 100;
#if defined(RDMA_FAST_PATH)
    D_PRINT("flag = 0x%x, total_size = %u\n", v->head_flag,
            (v->head_flag & FAST_RDMA_SIZE_MASK));
#endif
    for (i = 0; i < len; i++) {
        if (0 == i % 16)
            D_PRINT("\n  ");
        D_PRINT("%2x  ", (unsigned int) v->buffer[i]);
    }
    D_PRINT("\n");
    D_PRINT("  END OF VBUF DUMP\n");
}

#ifdef ADAPTIVE_RDMA_FAST_PATH
void vbuf_fast_rdma_alloc (viadev_connection_t * c, int dir) 
{
    vbuf * v;
    int vbuf_alignment = 64;
    int pagesize = getpagesize();
    int i;
    VIP_MEM_HANDLE mem_handle;

    vbuf * vbuf_ctrl_buf;
    void * vbuf_rdma_buf;

    /* initialize revelant fields */
    c->rdma_credit = 0;
    c->num_no_completion = 0;

    if (viadev_num_rdma_buffer) {

        SET_ORIGINAL_MALLOC_HOOKS;

        /* allocate vbuf struct buffers */
        if(posix_memalign((void **)&vbuf_ctrl_buf, vbuf_alignment,
                    sizeof(struct vbuf) * viadev_num_rdma_buffer)) {
            vapidev_error_abort(GEN_EXIT_ERR,
                            "malloc: vbuf in vbuf_fast_rdma_alloc");
        }

        memset(vbuf_ctrl_buf, 0,
                sizeof(struct vbuf) * viadev_num_rdma_buffer);

        /* allocate vbuf RDMA buffers */
        if(posix_memalign((void **)&vbuf_rdma_buf, pagesize,
                    viadev_vbuf_total_size * viadev_num_rdma_buffer)) {
            vapidev_error_abort(GEN_EXIT_ERR,
                            "malloc: vbuf DMA in vbuf_fast_rdma_alloc");
        }

        memset(vbuf_rdma_buf, 0,
                viadev_vbuf_total_size * viadev_num_rdma_buffer);

        /* REGISTER RDMA SEND BUFFERS */
        REGISTER_VBUF_DMA(viadev.ptag, viadev.nic, mem_handle, 
                viadev_vbuf_total_size * viadev_num_rdma_buffer, vbuf_rdma_buf);

        /* Connect the DMA buffer to the vbufs */
        for (i = 0; i < viadev_num_rdma_buffer; i++) {
            v = &(vbuf_ctrl_buf[i]);
            v->head_flag = (VBUF_FLAG_TYPE *) ( (char *)(vbuf_rdma_buf) + (i * viadev_vbuf_total_size));
            v->buffer = (char *) ((char *)(v->head_flag) + sizeof(VBUF_FLAG_TYPE));
        }

        /* Some vbuf initialization */
        for (i = 0; i < viadev_num_rdma_buffer; i++) {

            /* set global_rank */
            vbuf_ctrl_buf[i].grank = c->global_rank;

            if (dir==0) {
                vbuf_ctrl_buf[i].desc.next = NULL;
                vbuf_ctrl_buf[i].padding = FREE_FLAG;
            } else {
                vbuf_ctrl_buf[i].padding = BUSY_FLAG;
            }

        }

        if (dir==0) {
            c->RDMA_send_buf      = vbuf_ctrl_buf;
            c->RDMA_send_buf_DMA  = vbuf_rdma_buf;
            c->RDMA_send_buf_hndl = mem_handle;
            /* set pointers */
            c->phead_RDMA_send = 0;
            c->ptail_RDMA_send = viadev_num_rdma_buffer - 1;
        } else {
            c->RDMA_recv_buf      = vbuf_ctrl_buf;
            c->RDMA_recv_buf_DMA  = vbuf_rdma_buf;
            c->RDMA_recv_buf_hndl = mem_handle;
            /* set pointers */
            c->p_RDMA_recv = 0;
            c->p_RDMA_recv_tail = viadev_num_rdma_buffer - 1;

            /* Add the connection to the RDMA polling list */
            viadev.RDMA_polling_group[viadev.RDMA_polling_group_size] = c->global_rank;
            viadev.RDMA_polling_group_size++;
        }

        SAVE_MALLOC_HOOKS;
        SET_MVAPICH_MALLOC_HOOKS;


    }
}

void vbuf_rdma_address_send (viadev_connection_t * c)
{
    /* inform others about the receive 
     * RDMA buffer address and handle */

    vbuf *v;
    viadev_packet_rdma_address *h;

    v = get_vbuf();
    h = (viadev_packet_rdma_address *) VBUF_BUFFER_START(v);

    /* set up the packet */
    PACKET_SET_HEADER(h, c, VIADEV_RDMA_ADDRESS);
    h->envelope.src_lrank = c->global_rank;

    /* set address and handle */
    h->RDMA_address = c->RDMA_recv_buf_DMA;
    h->RDMA_hndl = c->RDMA_recv_buf_hndl;

#ifdef MCST_SUPPORT
    if (is_mcast_enabled) {
        /* set address and handle for UD Acks, this is for now remove later */
        h->RDMA_ACK_address = (viadev.bcast_info.ack_buffer);
        h->RDMA_ACK_hndl.rkey = viadev.bcast_info.ack_mem_hndl.rkey;
    }
#endif                          /* MCST_SUPPORT */

    /* finally send it off */
    vbuf_init_send(v, sizeof(viadev_packet_rdma_address));

    viadev_post_send(c, v);

}

#endif
