/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "mpid_bind.h"
#include <vapi_header.h>
#include "viutil.h"
#include "viapriv.h"
#include "viapacket.h"
#include "queue.h"
#include "viaparam.h"
#include "mpid_smpi.h"
#include "vbuf.h"

#ifdef MCST_SUPPORT
extern int ud_prepost_count;
extern int ud_prepost_threshold;
extern int period;
extern int retransmission;
void viadev_process_msend(void *);
void viadev_process_mrecv(void *);
void viadev_rendezvous_cancel(MPIR_RHANDLE *);;
void process_acks(void);
void slide_window(void);
void check_time_out(void);
void viadev_post_ud_recv(vbuf *);
extern int is_mcast_enabled;
#endif

static void viutil_spinandwaitcq(int spin_count, VAPI_wc_desc_t * sc,
                                 VAPI_wc_desc_t * rc, int *inspin);
/*
 * MPID_DeviceCheck processes ALL completed descriptors. 
 * This means it handles ALL incoming packets as well (as 
 * takes care of marking send requests complete when data is
 * transferred -- still not sure if it actually does something here 
 * 
 * Each time it is
 * called, it processes ALL descriptors that have completed since
 * the last time it was called, even if it was called nonblocking. 
 * 
 * If the blocking form is called, and there are no packets waiting, 
 * it spins for a short time and then waits on the connection queue 
 * 
 * Most of the work of the VIA device is done in this routine. 
 *
 * As this routine is on the critical path, will have to work hard
 * to optimize it. Reducing function calls will be important. 
 */


/* 
 * The flowlist is a list of connections that have received new flow 
 * credits and may be able to make progress on pending rendezvous sends.
 * The flowlist is a singly-linked list of connections, linked by 
 * the "nextflow" field in the connection. In order to prevent loops, 
 * there is also a connection field "inflow" which is set to one
 * if it is in the flowlist and zero otherwise. It must be 
 * initialized to 0 and it only changed by the flow control 
 * logic in MPID_CheckDevice
 * 
 * We add a connection to the flowlist when either
 *  1) it gets new credits and there is a rendezvous in progress
 *  2) a rendezvous ack is received
 * 
 * Note: flowlist fields are touched only routines in this file. 
 */



static viadev_connection_t *flowlist;
#define PUSH_FLOWLIST(c) {                                          \
    if (0 == c->inflow) {                                           \
        c->inflow = 1;                                              \
        c->nextflow = flowlist;                                     \
        flowlist = c;                                               \
    }                                                               \
}

#define POP_FLOWLIST() {                                            \
    if (flowlist != NULL) {                                         \
        viadev_connection_t *_c;                                    \
        _c = flowlist;                                              \
        flowlist = _c->nextflow;                                    \
        _c->inflow = 0;                                             \
        _c->nextflow = NULL;                                        \
    }                                                               \
}

#if defined(_IA64_)
           #define RMB()  asm volatile ("mf" ::: "memory");
#elif defined(_IA32_)
           #define RMB()  asm volatile ("lfence" ::: "memory");
#elif defined(_X86_64_)
           #define RMB()  asm volatile ("lfence" ::: "memory");
#elif defined(_EM64T_)
           #define RMB()  asm volatile ("lfence" ::: "memory");
#else
           #define RMB()
#endif /* defined(_IA64_ ... */

static void process_flowlist(void);

/*
 * Attached to each connection is a list of send handles that
 * represent rendezvous sends that have been started and acked but not
 * finished. When the ack is received, the send is placed on the list;
 * when the send is complete, it is removed from the list.  The list
 * is an "in progress sends" queue for this connection.  We need it to
 * remember what sends are pending and to remember the order sends
 * were acked so that we complete them in that order. This
 * prevents a situation where we receive an ack for message 1, block
 * because of flow control, receive an ack for message 2, and are able
 * to complete message 2 based on the new piggybacked credits.
 * 
 * The list head and tail are given by the shandle_head and
 * shandle_tail entries on viadev_connection_t and the list is linked
 * through the nexthandle entry on a send handle.  
 *
 * The queue is FIFO because we must preserve order, so we maintain
 * both a head and a tail. 
 *
 */

#define RENDEZVOUS_IN_PROGRESS(c, s) {                              \
    if (NULL == c->shandle_tail) {                                  \
        c->shandle_head = s;                                        \
    } else {                                                        \
        c->shandle_tail->nexthandle = s;                            \
    }                                                               \
    c->shandle_tail = s;                                            \
    s->nexthandle = NULL;                                           \
}

#define RENDEZVOUS_DONE(c) {                                        \
    c->shandle_head = c->shandle_head->nexthandle;                  \
        if (NULL == c->shandle_head) {                              \
            c->shandle_tail = NULL;                                 \
        }                                                           \
}

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))

int poll_rdma_buffer(void **vbuf_addr, int *out_of_order)
{
    int i, max_conn;
    vbuf *v;
    viadev_packet_header *h;
    VBUF_FLAG_TYPE size;
    volatile VBUF_FLAG_TYPE *tail;
    volatile VBUF_FLAG_TYPE *head;
    *vbuf_addr = NULL;
    *out_of_order = 0;

    if (viadev_num_rdma_buffer == 0)
        return VAPI_CQ_EMPTY;

#if defined(RDMA_FAST_PATH)
    max_conn = viadev.np;

#elif defined(ADAPTIVE_RDMA_FAST_PATH)
    max_conn= viadev.RDMA_polling_group_size;
#endif

    for (i = 0; i < max_conn; i++) {

#ifdef RDMA_FAST_PATH
        viadev_connection_t *c = &viadev.connections[i];
        if (i == viadev.me)
            continue;
#elif ADAPTIVE_RDMA_FAST_PATH
        viadev_connection_t *c = &viadev.connections[viadev.RDMA_polling_group[i]];
#endif
        /* get the current polling buffer */
        v = &(c->RDMA_recv_buf[c->p_RDMA_recv]);
        head = v->head_flag;
        /* check the flag */
        if (*head) {

            size = (*head & FAST_RDMA_SIZE_MASK);
            tail = (VBUF_FLAG_TYPE *) ((aint_t) (v->buffer) + size);

            /* check tail */

            /* If tail has not come go ahead and poll other
             * connections. Busy waiting for a message being
             * DMA'd to user memory causes unnecessary loss of
             * CPU cycles */
            if(*tail != *head) {
                continue;
            }

             /* Read memory barrier - make sure the header is read
              * only after the head and tails are read and compared
              * disabled by default for PGI Compiler*/
#ifndef DISABLE_RMB 
            RMB();
#endif
            /* out of order message? */
            h = (viadev_packet_header *) (v->buffer);

            if (h->type != VIADEV_PACKET_NOOP
                && h->id != c->next_packet_expected) {
                continue;
            }

            *vbuf_addr = v;
            *out_of_order = 0;

            /* clear head flag */
            *head = 0;

            /* advance receive pointer */
            if (++(c->p_RDMA_recv) >= viadev_num_rdma_buffer)
                c->p_RDMA_recv = 0;
            return VAPI_OK;

        } else
            continue;

    }

    return VAPI_CQ_EMPTY;

}

#endif  /* RDMA_FAST_PATH || ADAPTIVE_RDMA_FAST_PATH */


int MPID_DeviceCheck(MPID_BLOCKING_TYPE blocking)
{
    VAPI_ret_t ret1, ret2;
    VAPI_wc_desc_t rc, sc;
    int inspin, gotone = 0;
    void *vbuf_addr;

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    int ofo;
#endif

#ifdef MCST_SUPPORT
    VAPI_ret_t ret3;
    ret3 = VAPI_OK;
#endif


    D_PRINT("Enter DeviceCheck Blocking = %s",
            (blocking == MPID_NOTBLOCKING ? "FALSE" : "TRUE"));

  checkstart:
    flowlist = NULL;

    /* process all waiting packets */
    ret1 = ret2 = VAPI_OK;


    /* Check the fastest channel */
#ifdef _SMP_
    if (SMP_INIT) {
        if (MPI_SUCCESS == MPID_SMP_Check_incoming())
            gotone = 1;
    
    }
#endif

#ifdef MCST_SUPPORT
    if ((viadev.bcast_info.bcast_called == 1) && (is_mcast_enabled)) {

        if (retransmission == 0) {

            /* Calling this if there are no retransmissions else where */
            process_acks();

            slide_window();

            check_time_out();

        }
    }
#endif

#ifdef MCST_SUPPORT
    while ((ret1 == VAPI_OK) ||
           ((ret2 == VAPI_OK) && (viadev.bcast_info.bcast_called == 1)
            && (is_mcast_enabled)) ||
           ((ret3 == VAPI_OK) && (viadev.bcast_info.bcast_called == 1)
            && (is_mcast_enabled))) {
#else
    while (ret1 == VAPI_OK) {   /* at least success on one */
#endif

#ifdef MCST_SUPPORT
        if ((viadev.bcast_info.bcast_called == 1) && (is_mcast_enabled)) {
            ret2 = VAPI_poll_cq(viadev.nic, viadev.ud_scq_hndl, &sc);
            if (ret2 == VAPI_OK) {
                vbuf_addr = (void *) ((aint_t) sc.id);
                if (sc.status != VAPI_SUCCESS) {
                    vapidev_error_abort(VAPI_STATUS_ERR,
                                        "[%s:%d] Got completion with error, "
                                        "code=%s, vendor code=%x, dest rank=%d\n",
                                        viadev.my_name, viadev.me,
                                        VAPI_wc_status_sym(sc.status),
                                        sc.vendor_err_syndrome,
                                        ((vbuf *) vbuf_addr)->grank);
                }
                viadev_process_msend(vbuf_addr);
            }

            ret3 = VAPI_poll_cq(viadev.nic, viadev.ud_rcq_hndl, &sc);
            if (ret3 == VAPI_OK) {
                vbuf_addr = (void *) ((aint_t) sc.id);
                if (sc.status != VAPI_SUCCESS) {
                    vapidev_error_abort(VAPI_STATUS_ERR,
                                        "[%s:%d] Got completion with error, "
                                        "code=%s, vendor code=%x, dest rank=%d\n",
                                        viadev.my_name, viadev.me,
                                        VAPI_wc_status_sym(sc.status),
                                        sc.vendor_err_syndrome,
                                        ((vbuf *) vbuf_addr)->grank);
                }
                co_print("ud_pkt recvd \n");
                viadev_process_mrecv(vbuf_addr);
            }
        }
#endif

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        ret1 = poll_rdma_buffer(&vbuf_addr, &ofo);
        if (ret1 == VAPI_OK) {
            gotone = 1;
            viadev_process_recv(vbuf_addr);
        }
#endif
        ret1 = VAPI_poll_cq(viadev.nic, viadev.cq_hndl, &sc);
        if (ret1 == VAPI_OK) {

#ifdef MCST_SUPPORT
            if (sc.id == -1) {
                viadev.bcast_info.sbuf_tail =
                    (++viadev.bcast_info.sbuf_tail) % MAX_ACKS;
                viadev.bcast_info.ack_full = 0;
            } else {
#endif

                gotone = 1;
                vbuf_addr = (void *) ((aint_t) sc.id);

                /* Need to check if it is a completion with error */
                if (sc.status != VAPI_SUCCESS) {
                    vapidev_error_abort(VAPI_STATUS_ERR,
                                        "[%s:%d] Got completion with error, "
                                        "code=%s, vendor code=%x, dest rank=%d\n",
                                        viadev.my_name, viadev.me,
                                        VAPI_wc_status_sym(sc.status),
                                        sc.vendor_err_syndrome,
                                        ((vbuf *) vbuf_addr)->grank);
                }

                if (sc.opcode == VAPI_CQE_SQ_SEND_DATA ||
                    sc.opcode == VAPI_CQE_SQ_RDMA_WRITE ||
                    sc.opcode == VAPI_CQE_SQ_RDMA_READ) {

                    viadev_process_send(vbuf_addr);
                } else {
                    viadev_process_recv(vbuf_addr);
                }

#ifdef MCST_SUPPORT
            }
#endif
        }
    }

    /* make progress on pending rendezvous transfers */
    if (flowlist) {
        process_flowlist();
        /* In case some new credits have come in while processing
         * go check again. Little overhead to do this.
         */
        goto checkstart;
    }

    if (gotone || blocking == MPID_NOTBLOCKING) {
        D_PRINT("Leaving DeviceCheck gotone = %d", gotone);
        return gotone ? MPI_SUCCESS : -1;       /* nonblock must return -1 */
    }

    rc.id = 0;                  /* for sanity */
    sc.id = 0;
    viutil_spinandwaitcq(viadev_spin_count, &sc, &rc, &inspin);

#ifdef MCST_SUPPORT
    if (sc.id == -1) {
        viadev.bcast_info.sbuf_tail =
            (++viadev.bcast_info.sbuf_tail) % MAX_ACKS;
        viadev.bcast_info.ack_full = 0;
    } else {
#endif

        if (sc.id) {            /* both in sc */
            vbuf_addr = (void *) ((aint_t) sc.id);

            if (sc.opcode == VAPI_CQE_SQ_SEND_DATA ||
                sc.opcode == VAPI_CQE_SQ_RDMA_WRITE ||
                sc.opcode == VAPI_CQE_SQ_RDMA_READ) {

                viadev_process_send(vbuf_addr);
            } else {
                if (sc.opcode != VAPI_CQE_RQ_SEND_DATA &&
                    sc.opcode != VAPI_CQE_RQ_RDMA_WITH_IMM) {

                    vapidev_error_abort(VAPI_RETURN_ERR,
                                        "(after spinandwait) invalid "
                                        "receive opcode(%d)\n", sc.opcode);
                }
                viadev_process_recv(vbuf_addr);
            }
        }
#ifdef MCST_SUPPORT
    }
#endif


    if (flowlist) {
        process_flowlist();
    }
    D_PRINT("Leaving DeviceCheck after wait");

    return MPI_SUCCESS;
}

#ifdef MCST_SUPPORT

void viadev_process_msend(void *vbuf_addr)
{
    vbuf *v;
    v = (vbuf *) (vbuf_addr);
    release_vbuf(v);
}

void viadev_process_mrecv(void *vbuf_addr)
{
    vbuf *v = NULL;
    viadev_packet_eager_start *header = NULL;
    MPIR_RHANDLE *rhandle;
    int Bcnt, root, found;

    co_print("Entering process mrecv\n");

    --ud_prepost_count;
    co_print("ud_prepost_count:%d, ud_prepost_threshold:%d\n",
             ud_prepost_count, ud_prepost_threshold);

    if (ud_prepost_count < ud_prepost_threshold) {
        co_print("Preposting ud descriptors\n");
        while (ud_prepost_count < VIADEV_UD_PREPOST_DEPTH) {
            PREPOST_UD_VBUF_RECV();
            ud_prepost_count++;
        }
    }

    v = (vbuf *) vbuf_addr;

    co_print("[%d] received message: vbuf=%p\n", viadev.me, v);
    header = (viadev_packet_eager_start *) (VBUF_BUFFER_START(v) + 40);

    root = (header->envelope).src_lrank;
    Bcnt = (header->envelope.tag);
    co_print("ud msg:%d arrived\n", Bcnt);

    /* Drop packet received out of window ..Need to do even for rc */
    if ((Bcnt < viadev.bcast_info.Bcnt[root]) || (root == viadev.me)) {
        if (root == viadev.me)
            co_print("Dropping self pkt\n");
        else
            co_print("Dropping UD duplicate pkt\n");
        release_vbuf(v);
        return;
    }

    /* Search the unexpected Q and drop if an already existing pkt. is 
     * found */
    MPID_Search_unexpected_queue((header->envelope).src_lrank,
                                 (header->envelope).tag, -1, 0, &rhandle);
    if (rhandle) {
        co_print("Searched Q, dropping UD duplicate\n");
        release_vbuf(v);
        return;
    }


    MPID_Msg_arrived((header->envelope).src_lrank, (header->envelope).tag,
                     -1, &rhandle, &found);


    rhandle->s.count = header->envelope.data_length;
    rhandle->protocol = VIADEV_PROTOCOL_UD_MCST;
    co_print("Qing into unexpected Q in mrecv"
             ",protocol:%d\n", VIADEV_PROTOCOL_UD_MCST);
    rhandle->vbufs_received = 1;


    if (found) {
        char *vbuf_data =
            ((char *) header) + sizeof(viadev_packet_eager_start);

        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            printf("message truncated. ask %d got %d",
                   rhandle->len, header->envelope.data_length);
        }

        rhandle->len = header->envelope.data_length;
        memcpy((char *) (rhandle->buf), vbuf_data,
               header->bytes_in_this_packet);
        rhandle->bytes_copied_to_user = header->bytes_in_this_packet;

        assert(rhandle->bytes_copied_to_user == rhandle->len);
        RECV_COMPLETE(rhandle);
        release_vbuf(v);
        /*incrementing the count so that duplicates can be detected */
        ++viadev.bcast_info.Bcnt[root];

    } else {                    /*if not found */
        rhandle->len = header->envelope.data_length;
        rhandle->vbuf_head = v;
        rhandle->vbuf_tail = v;
        rhandle->bytes_copied_to_user = 0;

        /* add later */
        v->desc.next = NULL;
    }


}

#endif

/* Input: address of vbuf, which contains the send or recv decscriptor */
void viadev_process_send(void *vbuf_addr)
{
    vbuf *v;
    viadev_packet_header *header;
    viadev_connection_t *c = NULL;
    VAPI_sr_desc_t *desc = NULL;
    VAPI_sg_lst_entry_t *sg_lst_entry = NULL;

    MPIR_SHANDLE *s;

    /* process a completed descriptor on the send queue.  
     * dispatch to an appropriate routine based on the type 
     * in the header field of the vbuf. xxx Later (with RDMA) we will 
     * have to handle packets that have immediate data 
     * but no packet header in the vbuf.  */

    v = (vbuf *) (vbuf_addr);
    desc = &v->desc.sr;
    sg_lst_entry = &(v->desc.sg_entry);

    /* since we have a completion, one more WQE is available
     * first process any existing WQEs on extended send q
     * if we implement send coallesing for RDMA fast Path (MAX_NO_COMPLETION)
     * we will need to handle >1 additions to send_wqes_avail
     */
    c = &(viadev.connections[v->grank]);
    c->send_wqes_avail++;
    if (c->ext_sendq_head) {
        viadev_ext_sendq_send(c);
    }

    /* RDMA read is being treated as a send completion.
     * The rationale is that it is tied to to the send work queue.
     * perhaps its not ok, to be treated in viadev_process_send,
     * rather viadev_process_recv. But anyways, it doesn't matter */

#if (defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT))

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    if (v->padding == RPUT_VBUF_FLAG) {
        release_vbuf(v);
        return;
    }
#if defined(VIADEV_RGET_SUPPORT)
    if (v->padding == RGET_VBUF_FLAG) {
        process_rdma_read_completion(v);
        return;
    }
#endif /* VIADEV_RGET_SUPPORT */

#else
    if (v->desc.sr.opcode == VAPI_RDMA_WRITE) {
        release_vbuf(v);
        return;
    }

#if defined(VIADEV_RGET_SUPPORT)
    if (v->desc.sr.opcode == VAPI_RDMA_READ) {
        process_rdma_read_completion(v);
        return;
    }
#endif  /* VIADEV_RGET_SUPPORT */

#endif  /* RDMA_FAST_PATH || ADAPTIVE_RDMA_FAST_PATH */

#endif  /* VIADEV_RPUT_SUPPORT || VIADEV_RGET_SUPPORT */

    header = (viadev_packet_header *) VBUF_BUFFER_START(v);

    switch (header->type) {
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    case FAST_EAGER_CACHED:
#endif
#endif
    case VIADEV_PACKET_EAGER_START:
    case VIADEV_PACKET_EAGER_NEXT:
    case VIADEV_PACKET_R3_DATA:
        s = (MPIR_SHANDLE *) (v->shandle);
        if (s != NULL)
            SEND_COMPLETE(s);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;

#ifdef VIADEV_RGET_SUPPORT
    case VIADEV_PACKET_RGET_FINISH:
        {
            MPIR_RHANDLE *r = NULL;
            r = (MPIR_RHANDLE *)v->shandle;
            if(r != NULL) {
                RECV_COMPLETE(r);
            }
        }
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;
#endif  /* VIADEV_RGET_SUPPORT */

    case VIADEV_PACKET_RPUT_FINISH:
        s = (MPIR_SHANDLE *) (v->shandle);
        if (s == NULL) {
            vapidev_error_abort(GEN_ASSERT_ERR,
                                "s == NULL, s is the send "
                                "handler of the rput finish");
        }
        SEND_COMPLETE(s);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;

#if (defined(RDMA_FAST_PATH) || defined(MCST_SUPPORT) || \
        defined(ADAPTIVE_RDMA_FAST_PATH))
    case VIADEV_RDMA_ADDRESS:
#endif
    case VIADEV_PACKET_RENDEZVOUS_START:
    case VIADEV_PACKET_RENDEZVOUS_REPLY:
#ifdef MCST_SUPPORT
    case VIADEV_PACKET_RENDEZVOUS_CANCEL:
#endif
    case VIADEV_PACKET_NOOP:
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_BARRIER:
        D_PRINT("[%d] reaped barrier [%d]",
                viadev.me, ((viadev_packet_barrier *) header)->barrier_id);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            v->padding = FREE_FLAG;
        }
#else
        release_vbuf(v);
#endif
        break;
    default:
        vapidev_error_abort(VAPI_STATUS_ERR,
                            "Unknown packet type %d in "
                            "viadev_process_send", header->type);
    }
}

/*
 * The workhorse of CheckDevice
 */
void viadev_process_recv(void *vbuf_addr)
{
    vbuf *v;
    viadev_packet_header *header;
    viadev_connection_t *c;
    int needed;
    VAPI_rr_desc_t *desc;
    VAPI_sg_lst_entry_t *sg_lst_entry;
    int barrier_id, count;


    /* *All* packets come IN through this routine and go OUT
     * through viadev_post_send. This is asymmetrical, so 
     * it doesn't feel right. But perhaps it's ok?
     */

    /* Process a completed descriptor on the recv queue.  dispatch to
     * an appropriate routine based on the type in the header field of
     * the vbuf. Flow control credits are always accumulated before
     * the packet is processed, and we check the sequence number.
     * When done, we release the vbuf and repost if necessary. 
     */

    v = (vbuf *) vbuf_addr;
    desc = &v->desc.rr;
    sg_lst_entry = &v->desc.sg_entry;
    c = &(viadev.connections[v->grank]);

    header = (viadev_packet_header *) VBUF_BUFFER_START(v);
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    if (header->type == FAST_EAGER_CACHED) {
        c->cached_incoming.header.id = header->id;
        c->cached_incoming.header.type = FAST_EAGER_CACHED;
        c->cached_incoming.envelope.data_length = header->fast_eager_size;
        c->cached_incoming.header.fast_eager_size = 
            header->fast_eager_size;
        header = &(c->cached_incoming.header);
    }
#endif
#endif

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    /* handle out of order messages for send/receive channel */
    /* RDMA channel don't return out of order message !!! */
    if (v->padding == NORMAL_VBUF_FLAG) {
        vbuf *v1;
        int ofo;
        int r1;
        while (header->type != VIADEV_PACKET_NOOP &&
               header->id != c->next_packet_expected) {
            /* switch to the rdma channel */
#ifdef _SMP_
            if(SMP_INIT) {
                smpi_net_lookup();
            }
#endif
            r1 = poll_rdma_buffer((void *) &v1, &ofo);
            if (r1 == VAPI_OK) {
                viadev_process_recv(v1);
            }
        }
    }
#endif

    if (header->type != VIADEV_PACKET_NOOP) {
        if (header->id != c->next_packet_expected) {
            vapidev_error_abort(GEN_ASSERT_ERR,
                                "header->id != c->next_packet_expected");
        }
        c->next_packet_expected++;
    }
    D_PRINT("[%d]header->type=%d\n", viadev.me, header->type);

    /* just consumed a vbuf from a VI receive queue, decrement prepost
     * count and check if we should re-post.  Want to do this here
     * since we may be sending a packet in response to this receive
     * and want it to include the credit update asap.
     */
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    if (v->padding == NORMAL_VBUF_FLAG)
        c->preposts--;
#else
    c->preposts--;
#endif
    needed = viadev_prepost_depth + viadev_prepost_noop_extra
        + MIN(viadev_prepost_rendezvous_extra,
              c->rendezvous_packets_expected);

    if (c->initialized) {
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (header->type == VIADEV_PACKET_NOOP
            && v->padding == NORMAL_VBUF_FLAG) {
#else
        if (header->type == VIADEV_PACKET_NOOP) {
#endif
            PREPOST_VBUF_RECV(c);
            /* noops don't count for credits */
            c->local_credit--;
        } else if (c->preposts < (int) viadev_rq_size &&
                   c->preposts + viadev_prepost_threshold < needed) {
            count = 0;

            /* Collect all the vbufs into an array */
            do {
                collect_vbuf_for_recv(count, c);
                count++;
            } while (c->preposts < (int) viadev_rq_size && c->preposts < needed);

            if(0 != count) {
                viadev_post_recv_list(c, count);
            }
        }
    } else {
        int i;

        /* viadev.initialized means that MPI_Init has finished.
         *
         * Packets following that should "initialize" the
         * connection.
         * 
         * Even if MPI_Init has not yet finished, R3 packets
         * may have arrived, either as part of MPI_Init, or
         * rendezvous MPI sends of 0 bytes (Ssend). Incoming
         * rendezvous messages may have been discovered during
         * Device Check poll after the MPI_Init messages, but
         * before viadev.initialized is set by upper layer
         * in src/env/initutil.c 
         * In such case, we have to post buffers for the
         * connection, and move it to "initialized" state*/

        if((!viadev.initialized &&
                (header->type == VIADEV_PACKET_RENDEZVOUS_START) &&
                (((viadev_packet_rendezvous_start *)header)->protocol
                 == VIADEV_PROTOCOL_R3)) ||
                viadev.initialized) {

            for (i = 0;
                    i <
                    viadev_prepost_noop_extra - (viadev_initial_prepost_depth -
                        viadev_initial_credits); ++i) {
                PREPOST_VBUF_RECV(c);
                c->local_credit--;
            }


            /* fill out buffers to make up for small initial preposts */
            while (c->preposts < (int) viadev_rq_size && c->preposts < needed) {
                PREPOST_VBUF_RECV(c);
            }

            c->initialized = TRUE;
        }
    }                           /* c->initialized */

    if (header->type == VIADEV_PACKET_R3_DATA) {
        D_PRINT("[%d]header->type=%d is R3 DATA\n", viadev.me,
                header->type);
        /* got an R3 Data packet, do we need to pre-post additional
         * receives?  This next call will ensure that we keep the
         * the number of additional pre-posts for R3 data sends
         * up to the maximum allowed.
         */
        if (c->rendezvous_packets_expected > 0) {
            viadev_prepost_for_rendezvous(c, 0);
        }
    }

    D_PRINT("VI %3d PKT RECV LC %3d PP %3d RC %3d CU %3d ",
            c->global_rank, c->local_credit,
            c->preposts, c->remote_credit, header->vbuf_credit);

    /* update our knowledge of remote sides credit count */
    c->remote_cc = header->remote_credit;

    /* accumulate vbuf credits and remember */
    if (header->vbuf_credit > 0) {
        c->remote_credit += header->vbuf_credit;
        if (header->type == VIADEV_PACKET_NOOP) {
            D_PRINT("VI %3d RECV NOOP with credit %d",
                    c->global_rank, header->vbuf_credit);
        }
        if (c->backlog.len > 0) {
            /* We have pending messages on the backlog queue.
             * Send them out immediately.
             */
            viadev_backlog_send(c);
        }
        /* if any credits remain, schedule rendezvous progress */
        if ((c->remote_credit > 0) && (c->shandle_head != NULL)) {
            PUSH_FLOWLIST(c);
        }
    }
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    /* process rdma credit */
    if (header->rdma_credit > 0) {
        c->ptail_RDMA_send += header->rdma_credit;
        if (c->ptail_RDMA_send >= viadev_num_rdma_buffer)
            c->ptail_RDMA_send -= viadev_num_rdma_buffer;
    }
#endif


    switch (header->type) {
#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    case FAST_EAGER_CACHED:
#endif
#endif
    case VIADEV_PACKET_EAGER_START:

#ifdef ADAPTIVE_RDMA_FAST_PATH
        if ((c->RDMA_recv_buf == NULL) && (c->initialized) &&
                viadev_num_rdma_buffer) {
            if (viadev.RDMA_polling_group_size < viadev_rdma_eager_limit) {
                c->eager_start_cnt++;
                if (viadev_rdma_eager_threshold < c->eager_start_cnt) {
                    vbuf_fast_rdma_alloc (c, 1);
                    vbuf_rdma_address_send(c);
                }
            }
        }
#endif
        
        viadev_incoming_eager_start(v, c,
                                    (viadev_packet_eager_start *) header);
        /* dont release vbuf may have been unexpected. 
         * eager packet processing
         * will release vbufs when possible.  */
        break;
    case VIADEV_PACKET_EAGER_NEXT:
        viadev_incoming_eager_next(v, c,
                                   (viadev_packet_eager_next *) header);
        /* dont release vbuf may have been unexpected. 
         * eager packet processing
         * will release vbufs when possible. 
         */
        break;
    case VIADEV_PACKET_RENDEZVOUS_START:
        viadev_incoming_rendezvous_start(v, c,
                                         (viadev_packet_rendezvous_start *)
                                         header);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_RENDEZVOUS_REPLY:
#ifdef MCST_SUPPORT
    case VIADEV_PACKET_RENDEZVOUS_CANCEL:
#endif
        viadev_incoming_rendezvous_reply(v, c,
                                         (viadev_packet_rendezvous_reply *)
                                         header);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_R3_DATA:
        viadev_incoming_r3_data(v, c, (viadev_packet_r3_data *) header);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;

#if (defined(RDMA_FAST_PATH) || defined(MCST_SUPPORT) || \
        defined(ADAPTIVE_RDMA_FAST_PATH))
    case VIADEV_RDMA_ADDRESS:

#ifdef MCST_SUPPORT
        if (is_mcast_enabled) {
            viadev.bcast_info.remote_add[c->global_rank] =
                (((viadev_packet_rdma_address *) header)->
                 RDMA_ACK_address);
            viadev.bcast_info.remote_rkey[c->global_rank] =
                (((viadev_packet_rdma_address *) header)->RDMA_ACK_hndl).
                rkey;
#ifndef RDMA_FAST_PATH
            c->remote_address_received = 1;
#endif
        }
#endif


#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        viadev_incoming_rdma_address(v, c, (viadev_packet_rdma_address *)
                                     header);

        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
#endif

#ifdef VIADEV_RPUT_SUPPORT
    case VIADEV_PACKET_RPUT_FINISH:
        viadev_incoming_rput_finish(v, c,
                                    (viadev_packet_rput_finish *) header);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
#endif
    case VIADEV_PACKET_NOOP:
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;
    case VIADEV_PACKET_BARRIER:
        barrier_id = ((viadev_packet_barrier *) header)->barrier_id;
        D_PRINT("[%d] got barrier [%d] from [%d]",
                viadev.me, barrier_id, c->global_rank);
        c->barrier_id = barrier_id;
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif
        break;

#ifdef VIADEV_RGET_SUPPORT
    case VIADEV_PACKET_RGET_FINISH :
        viadev_incoming_rget_finish(v, c,
                (viadev_packet_rget_finish *) header);
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG) {
            release_vbuf(v);
        } else {
            release_recv_rdma(c, v);
        }
#else
        release_vbuf(v);
#endif  /* RDMA_FAST_PATH */
        break;
#endif  /* VIADEV_RGET_SUPPORT */

    default:
        vapidev_error_abort(VAPI_STATUS_ERR,
                            "Unknown packet type %d in "
                            "viadev_process_recv", header->type);
    }

    viadev_send_noop_ifneeded(c);
}

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
void process_remote_rdma_address(viadev_connection_t * c,
                                 VIP_MEM_HANDLE remote_hndl,
                                 vbuf * remote_address);

void viadev_incoming_rdma_address(vbuf * v,
                                  viadev_connection_t * c,
                                  viadev_packet_rdma_address * header)
{

    /* address also received */
    if (c->remote_address_received != 0) {
        vapidev_error_abort(GEN_EXIT_ERR,
                            "global rank for this connection: %d",
                            c->global_rank);
    }
    process_remote_rdma_address(c, header->RDMA_hndl,
                                header->RDMA_address);

}
#endif


void viadev_incoming_rendezvous_start(vbuf * v, viadev_connection_t * c,
                                      viadev_packet_rendezvous_start *
                                      header)
{

    MPIR_RHANDLE *rhandle;
    int found;

#ifdef MCST_SUPPORT

    MPIR_RHANDLE *rhandle1;
    int root, Bcnt;
    co_print("Entered incoming rendezvous start\n");
    root = header->envelope.src_lrank;

    if (header->envelope.context == -1) {
        Bcnt = (header->envelope.tag);

        /* Drop packet received out of window ..Need to do even for rc */
        if (Bcnt < viadev.bcast_info.Bcnt[root]) {
            co_print("Bcnt:%d, viadev.bcast_info.Bcnt:%d\n",
                     Bcnt, viadev.bcast_info.Bcnt[root]);
            co_print("Dropping duplicate pkt in rendezvous start\n");
            co_print("Cancelling rendezvous\n");
            MPID_RecvAlloc(rhandle1);
            rhandle1->connection = c;
            rhandle1->send_id = header->sreq;
            viadev_rendezvous_cancel(rhandle1);
            return;
        }
        /* Search the unexpected Q and drop if an already existing pkt. is 
         * found */
        MPID_Search_unexpected_queue((header->envelope).src_lrank,
                                     (header->envelope).tag,
                                     -1, 0, &rhandle);
        if (rhandle) {

            co_print("Searched Q, dropping duplicate in rendezvous\n");
            co_print("Cancelling rendezvous\n");
            MPID_RecvAlloc(rhandle1);
            rhandle1->connection = c;
            rhandle1->send_id = header->sreq;
            viadev_rendezvous_cancel(rhandle1);
            return;
        }
    }

    co_print("in rend. start:src:%d,tag:%d,context:%d\n",
             header->envelope.src_lrank, header->envelope.tag,
             header->envelope.context);
#endif

    MPID_Msg_arrived(header->envelope.src_lrank,
                     header->envelope.tag,
                     header->envelope.context, &rhandle, &found);

    /* This information needs to be filled in whether a preposted 
     * receive was found or not.  */

    rhandle->connection = c;
    rhandle->send_id = header->sreq;
    rhandle->s.count = header->envelope.data_length;
    rhandle->vbufs_received = 0;
    rhandle->vbuf_head = NULL;
    rhandle->vbuf_tail = NULL;
    rhandle->replied = 0;
    rhandle->protocol = header->protocol;

#if defined(VIADEV_RGET_SUPPORT)
    rhandle->remote_address = header->buffer_address;
    rhandle->remote_memhandle = header->memhandle;
#endif


    if (!found) {
        /* recv not yet posted, set size of incoming data in
         * unexpected queue rhandle */
        rhandle->len = header->envelope.data_length;
    } else {
        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            vapidev_error_abort(VAPI_STATUS_ERR,
                                "message truncated. ask %d got %d",
                                rhandle->len,
                                header->envelope.data_length);
        }

        rhandle->len = header->envelope.data_length;

        /* packet now arrived, can no-longer cancel recv */
        rhandle->can_cancel = 0;

        /* can we match sender's protocol? */
        switch (header->protocol) {
        case VIADEV_PROTOCOL_R3:
            viadev_recv_r3(rhandle);
            break;
        case VIADEV_PROTOCOL_RGET:
            viadev_recv_rget(rhandle);
            break;
        case VIADEV_PROTOCOL_RPUT:
            viadev_recv_rput(rhandle);
            break;
        case VIADEV_PROTOCOL_RENDEZVOUS_UNSPECIFIED:
            vapidev_error_abort(VAPI_STATUS_ERR,
                                "VIRS: unspecified protocol %d\n",
                                rhandle->protocol);
            break;
        default:
            vapidev_error_abort(VAPI_STATUS_ERR,
                                "VIRS: invalid protocol %d\n",
                                rhandle->protocol);
        }
    }
}

void viadev_incoming_eager_start(vbuf * v, viadev_connection_t * c,
                                 viadev_packet_eager_start * header)
{

    MPIR_RHANDLE *rhandle;
    int found;

#ifdef MCST_SUPPORT
    int root, Bcnt;

    root = 0;

    co_print("Entering incoming eager start \n");
    if (header->envelope.context == -1) {

        root = header->envelope.src_lrank;
        Bcnt = (header->envelope.tag);
        co_print("Bcnt:%d,bcast_info.Bcnt:%d\n", Bcnt,
                 viadev.bcast_info.Bcnt[root]);

        /* Drop packet received out of window ..Need to do even for rc */
        if (Bcnt < viadev.bcast_info.Bcnt[root]) {
            co_print("Dropping duplicate pkt in eager start\n");
#if defined(RDMA_FAST_PATH)
            if (v->padding == NORMAL_VBUF_FLAG)
                release_vbuf(v);
            else
                release_recv_rdma(c, v);
#else
            release_vbuf(v);
#endif
            return;
        }

        /* Search the unexpected Q and drop if an already existing pkt. is 
         * found */
        MPID_Search_unexpected_queue((header->envelope).src_lrank,
                                     (header->envelope).tag,
                                     -1, 0, &rhandle);
        if (rhandle) {
            co_print("Searched Q, dropping duplicate\n");
#if defined(RDMA_FAST_PATH)
            if (v->padding == NORMAL_VBUF_FLAG)
                release_vbuf(v);
            else
                release_recv_rdma(c, v);
#else
            release_vbuf(v);
#endif
            return;
        }
    }
#endif

    MPID_Msg_arrived(header->envelope.src_lrank,
                     header->envelope.tag,
                     header->envelope.context, &rhandle, &found);

#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    /* caching header */
    if ((header->header.type == VIADEV_PACKET_EAGER_START) &&
        (header->envelope.data_length < viadev_max_fast_eager_size) &&
        (v->padding != NORMAL_VBUF_FLAG)) {
        memcpy(&(c->cached_incoming), header,
               sizeof(viadev_packet_eager_start));
    }

#endif
#endif

    /* This information needs to be filled in whether a preposted 
     * receive was found or not. 
     */

    rhandle->connection = c;
    rhandle->s.count = header->envelope.data_length;

    rhandle->protocol = VIADEV_PROTOCOL_EAGER;
    rhandle->vbufs_received = 1;

    if (found) {

        int truncated = 0;
        int copy_bytes;
#if (!defined(DISABLE_HEADER_CACHING) && defined (RDMA_FAST_PATH))
        char *vbuf_data;
#else
        char *vbuf_data = ((char *) header) +
            sizeof(viadev_packet_eager_start);
#endif
        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            truncated = 1;
        } else {
            rhandle->len = header->envelope.data_length;
        }

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))

#if (!defined(DISABLE_HEADER_CACHING) && defined(RDMA_FAST_PATH))
        if (header->header.type == FAST_EAGER_CACHED) {
            vbuf_data =
                ((char *) (VBUF_BUFFER_START(v))) + FAST_EAGER_HEADER_SIZE;
	    header->bytes_in_this_packet = header->header.fast_eager_size;
	    copy_bytes = rhandle->len;
	} else {
            vbuf_data = ((char *) header) +
                sizeof(viadev_packet_eager_start);
            copy_bytes = MIN(rhandle->len, 
                    header->bytes_in_this_packet);
        }
#else
        copy_bytes = MIN(rhandle->len, 
                header->bytes_in_this_packet);
#endif

#else
        copy_bytes = MIN(rhandle->len, 
                header->bytes_in_this_packet);
#endif

        /* matching packet is here, can no longer cancel recv */
        rhandle->can_cancel = 0;

        rhandle->vbufs_expected =
            viadev_calculate_vbufs_expected(rhandle->len,
                                            rhandle->protocol);

         if(copy_bytes > 0) {
             memcpy((char *) (rhandle->buf), vbuf_data, copy_bytes);
         }

        rhandle->bytes_copied_to_user = copy_bytes;
        if (rhandle->vbufs_expected == 1) {
            RECV_COMPLETE(rhandle);
#ifdef MCST_SUPPORT
            /* Incrementing the count so that duplicates can be found */
            if (header->envelope.context == -1) {
                ++viadev.bcast_info.Bcnt[root];
            }
#endif
            if (truncated) {
                rhandle->s.MPI_ERROR = MPI_ERR_TRUNCATE;
            }

            D_PRINT("VI %3d EAGER START RECV COMPLETE, total %d",
                    c->global_rank, rhandle->s.count);
        } else {
            c->rhandle = rhandle;
            rhandle->s.MPI_ERROR =
                truncated ? MPI_ERR_TRUNCATE : MPI_SUCCESS;
        }
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG)
            release_vbuf(v);
        else
            release_recv_rdma(c, v);
#else
        release_vbuf(v);
#endif
    } else {
        /* Recv not yet posted, set size of incoming data in
         * unexpected queue rhandle
         */
        rhandle->len = header->envelope.data_length;
        rhandle->vbuf_head = v;
        rhandle->vbuf_tail = v;
        rhandle->bytes_copied_to_user = 0;
        v->desc.next = NULL;
        c->rhandle = rhandle;
    }
}

void viadev_incoming_eager_next(vbuf * v, viadev_connection_t * c,
                                viadev_packet_eager_next * header)
{
    MPIR_RHANDLE *rhandle = c->rhandle;
    int copy_bytes = 0;

    if (rhandle == NULL) {
        vapidev_error_abort(GEN_ASSERT_ERR, "rhandle is NULL");
    }

    rhandle->vbufs_received++;

    if (rhandle->bytes_copied_to_user > 0) {
        /* this is a user receive handle, and has already been matched */
        char *vbuf_data = ((char *) header) +
            sizeof(viadev_packet_eager_next);

        copy_bytes = MIN(header->bytes_in_this_packet,
                (rhandle->len - rhandle->bytes_copied_to_user));

	if (copy_bytes) {
            memcpy((char *) (rhandle->buf) +
                   rhandle->bytes_copied_to_user,
                   vbuf_data, copy_bytes);
            rhandle->bytes_copied_to_user += copy_bytes;
        }

        if (rhandle->vbufs_expected == rhandle->vbufs_received) {
            int err = rhandle->s.MPI_ERROR;
            RECV_COMPLETE(rhandle);
            rhandle->s.MPI_ERROR = err;
            D_PRINT("VI %3d EAGER NEXT RECV COMPLETE, total %d",
                    c->global_rank, rhandle->s.count);
        }
#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        if (v->padding == NORMAL_VBUF_FLAG)
            release_vbuf(v);
        else
            release_recv_rdma(c, v);
#else
         release_vbuf(v);
#endif
    } else {
        /* haven't posted the message yet. accumulate vbufs on a chain */
        ((vbuf *) rhandle->vbuf_tail)->desc.next = v;
        v->desc.next = NULL;
        rhandle->vbuf_tail = v;
    }
}

/* 
 * We previously initiated a rendezvous send. 
 * Receiver has posted a matching receive and is 
 * telling us about it. 
 * The fact that they are telling us about it means
 * it is not an rget. It is either an r3 or an rput
 * depending on whether we are sending data directly 
 * into the user buffer on the other side
 * For the first implementation, only r3 is supported, 
 * copied through vbufs on both sides. 
 * 
 * Since messages can be very large and vbuf resources
 * are limited, we have to handle the case where there
 * are not enough preposted vbufs on the other side
 * to handle our data (this is true even for rput, 
 * which still consumes vbufs). 
 * 
 * For a non-pathological receiver (philosophical
 * debates about the MPI progress rule being of little
 * relevance to the real world -- and for the record we 
 * state that in the initial single-threaded imlementation 
 * we do not obey the strongest version of the rule)
 * we will receive credit renewals that allow us
 * to proceed. However, to get those credits, we have
 * to process some incoming packets. To handle this, we
 * put send handles for started-but-not-completed
 * rendezvous transfers on a queue of send
 * handles attached to the connection. MPID_DeviceCheck
 * makes progress on these through "viadev_rendezvous_push_transfer"
 * when it has run out of packets to process. 
 * 
 * All we do in this routine is set up the send handle and 
 * defer actually pushing data until all packets are processed. 
 *
 */

void viadev_incoming_rendezvous_reply(vbuf * v, viadev_connection_t * c,
                                      viadev_packet_rendezvous_reply *
                                      header)
{
    MPIR_SHANDLE *s;
    /* first make sure we have a match */
    s = (MPIR_SHANDLE *) (ID_TO_REQ(header->sreq));
    if (s == NULL) {
        vapidev_error_abort(GEN_ASSERT_ERR,
                            "s == NULL, send handler at viadev_incoming_rendezvous_reply");
    }
#ifdef MCST_SUPPORT

    if (header->header.type == VIADEV_PACKET_RENDEZVOUS_CANCEL) {
        SEND_COMPLETE(s);
        return;
    }
#endif

    /* fill in some fields */
    s->bytes_sent = 0;
    s->receive_id = header->rreq;


    /* we are the sender and initially set a protocol.  The
     * receiver may not be able to support our protocol and
     * may have reverted to R3.  If so, we have to revert as
     * well, and unregister any memory */

    switch (s->protocol) {
    case VIADEV_PROTOCOL_RPUT:
        if (header->protocol == VIADEV_PROTOCOL_R3) {
#ifdef VIADEV_RPUT_SUPPORT
            if (s->dreg_entry != NULL) {
                dreg_unregister(s->dreg_entry);
                s->dreg_entry = NULL;
            }
            s->remote_address = NULL;
            s->remote_memhandle.hndl = VAPI_INVAL_HNDL;
#endif
            s->protocol = VIADEV_PROTOCOL_R3;
        } else {
            assert(header->protocol == VIADEV_PROTOCOL_RPUT);

#ifdef VIADEV_RPUT_SUPPORT
            s->remote_address = header->buffer_address;
            s->remote_memhandle = header->memhandle;
#endif
        }
        break;
    case VIADEV_PROTOCOL_RGET:
        if (header->protocol == VIADEV_PROTOCOL_R3) {
            s->protocol = VIADEV_PROTOCOL_R3;
        } else {
            /* should never have gotten here if both sides
             * agreed on RGET */
            vapidev_error_abort(VAPI_STATUS_ERR,
                                "VIRR: RGET protocol failure");
        }
        break;

    case VIADEV_PROTOCOL_R3:
        assert(header->protocol == VIADEV_PROTOCOL_R3);
        break;

    default:
        vapidev_error_abort(VAPI_STATUS_ERR,
                            "VIRR: invalid sender protocol [%d]",
                            s->protocol);
    }

    /* 
     * even if we have an rput, which we know we could finish, 
     * put it on the list so that it is processed in order. 
     * Need to satisfy MPI ordering requirements. 
     */

    D_PRINT("VI %3d R3 OKTS RECVED len %d",
            c->global_rank, s->bytes_total);

    RENDEZVOUS_IN_PROGRESS(c, s);

    /*
     * this is where all rendezvous transfers are started,
     * so it is the only place we need to set this kludgy
     * field
     */

    s->nearly_complete = 0;

    PUSH_FLOWLIST(c);

}


static void process_flowlist(void)
{
    MPIR_SHANDLE *s;
    while (flowlist) {

        /* Push on the the first ongoing receive with 
         * viadev_rendezvous_push. If the receive
         * finishes, it will advance the shandle_head
         * pointer on the connection. 
         * 
         * xxx the side effect of viadev_rendezvous_push is
         * bad practice. Find a way to do this so the logic
         * is obvious. 
         */

        s = flowlist->shandle_head;
        while (s != NULL) {
            viadev_rendezvous_push(s);
            if (!s->nearly_complete) {
                break;
            }
            RENDEZVOUS_DONE(flowlist);
            s = flowlist->shandle_head;
        }
        /* now move on to the next connection */
        POP_FLOWLIST();
    }
}


void viadev_rendezvous_push(MPIR_SHANDLE * s)
{
    vbuf *v;
    viadev_packet_r3_data *h;
    char *databuf;
    int datamax = VBUF_DATA_SIZE(viadev_packet_r3_data);

    viadev_connection_t *c = (viadev_connection_t *) s->connection;


#ifdef VIADEV_RPUT_SUPPORT
    if (s->protocol == VIADEV_PROTOCOL_RPUT) {
        int nbytes;
        if (s->bytes_sent != 0) {
            vapidev_error_abort(GEN_ASSERT_ERR,
                                "s->bytes_sent != 0 Rendezvous Push");
        }

        if (s->bytes_total > 0) {
            assert(s->dreg_entry != NULL);
            assert(s->remote_address != NULL);
        }
        while (s->bytes_sent < s->bytes_total) {
            v = get_vbuf();
            assert(v != NULL);
            nbytes = s->bytes_total - s->bytes_sent;
            if (nbytes > (int) viadev.maxtransfersize) {
                nbytes = viadev.maxtransfersize;
            }
            viadev_rput(c, v,
                        (char *) (s->local_address) + s->bytes_sent,
                        ((dreg_entry *) s->dreg_entry)->memhandle,
                        (char *) (s->remote_address) + s->bytes_sent,
                        s->remote_memhandle, nbytes);
            s->bytes_sent += nbytes;
        }
        assert(s->bytes_sent == s->bytes_total);

        viadev_rput_finish(s);
        s->nearly_complete = 1;

        return;
    }
#endif


    /* if we get here, then protocol is R3 */

    while (c->remote_credit >= viadev_credit_preserve &&
           s->nearly_complete == 0) {
        v = get_vbuf();

        /* figure out where header and data are */
        h = (viadev_packet_r3_data *) VBUF_BUFFER_START(v);
        databuf = ((char *) h) + sizeof(viadev_packet_r3_data);

        /* set up the packet */
        PACKET_SET_HEADER(h, c, VIADEV_PACKET_R3_DATA);
        h->rreq = s->receive_id;

        /* figure out how much data to send in this packet */
        h->bytes_in_this_packet = s->bytes_total - s->bytes_sent;
        if (datamax < h->bytes_in_this_packet) {
            h->bytes_in_this_packet = datamax;
        }

        /* copy next chunk of data in to vbuf */
        memcpy(databuf, ((char *) s->local_address) +
               s->bytes_sent, h->bytes_in_this_packet);

        s->bytes_sent += h->bytes_in_this_packet;

        /* finally send it off */
        vbuf_init_send(v, sizeof(viadev_packet_r3_data) +
                       h->bytes_in_this_packet);
        /* If this is the last packet, mark the vbuf so that the MPI
         * send is marked complete when the VIA send completes. 
         * Do not remove from the shandle list on the connection. Whoever
         * called us will do that. 
         */
        if (s->bytes_sent == s->bytes_total) {
            /* mark MPI send complete when VIA send completes */
            v->shandle = s;
            s->nearly_complete = 1;
        }
        viadev_post_send(c, v);

    }
    if (s->nearly_complete) {
        D_PRINT("VI %3d R3 PUSH COMPLETE, total %d",
                c->global_rank, s->bytes_total);
    } else {
        D_PRINT("VI %3d R3 PUSH STALLED, total %d, sent %d",
                c->global_rank, s->bytes_total, s->bytes_sent);
    }
}



void viadev_incoming_r3_data(vbuf * v, viadev_connection_t * c,
                             viadev_packet_r3_data * h)
{

    MPIR_RHANDLE *rhandle;
    void *databuf;
    int bytes_to_copy;

    /* find the rhandle for this data */
    rhandle = (MPIR_RHANDLE *) ID_TO_REQ(h->rreq);

    databuf = ((char *) h) + sizeof(viadev_packet_r3_data);
    bytes_to_copy = h->bytes_in_this_packet;

    rhandle->vbufs_received++;

    memcpy(((char *) rhandle->buf) +
           rhandle->bytes_copied_to_user, databuf, bytes_to_copy);

    rhandle->bytes_copied_to_user += bytes_to_copy;

    c->rendezvous_packets_expected--;

    /* sanity check */
    assert(rhandle->bytes_copied_to_user <= rhandle->len);

    if (rhandle->bytes_copied_to_user == rhandle->len) {
        assert(rhandle->vbufs_received == rhandle->vbufs_expected);
        RECV_COMPLETE(rhandle);
        D_PRINT("VI %3d R3 RECV COMPLETE total %d",
                c->global_rank, rhandle->len);
    }
}

#ifdef VIADEV_RPUT_SUPPORT

void viadev_incoming_rput_finish(vbuf * v, viadev_connection_t * c,
                                 viadev_packet_rput_finish * h)
{

    MPIR_RHANDLE *rhandle;

    /* find the rhandle for this data */
    rhandle = (MPIR_RHANDLE *) ID_TO_REQ(h->rreq);

    rhandle->bytes_copied_to_user = rhandle->len;
    RECV_COMPLETE(rhandle);

}

#endif

#ifdef VIADEV_RGET_SUPPORT
void viadev_incoming_rget_finish(vbuf * v, viadev_connection_t * c,
        viadev_packet_rget_finish * h)
{
    MPIR_SHANDLE *shandle;

    /* find the rhandle for this data */
    shandle = (MPIR_SHANDLE *) ID_TO_REQ(h->sreq);
    SEND_COMPLETE(shandle);
}

#endif /* VIADEV_RGET_SUPPORT */

void viadev_send_noop_ifneeded(viadev_connection_t * c)
{
    /* Dont want remote end to stall too long waiting for a credit
     * update.  If the local credits have built up above the dynamic
     * threshold, send a no-op immediately
     * c->remote_cc should accurately reflect the remote sides credit situation.
     * If its low and we have built-up local credits, send a noop
     * for the purposes of a credit refresh and to prevent a deadlock.
     */
    if (c->local_credit >= viadev_dynamic_credit_threshold
        || (c->remote_cc <= viadev_credit_preserve &&
            c->local_credit >= viadev_credit_notify_threshold)
        ) {
        viadev_send_noop(c);
    }
}

void viadev_send_noop(viadev_connection_t * c)
{
    /* always send a noop when it is needed even if there is a backlog.
     * noops do not consume credits.
     * this is necessary to avoid credit deadlock.
     * RNR NAK will protect us if receiver is low on buffers.
     * by doing this we can force a noop ahead of any other queued packets.
     */

#if defined(RDMA_FAST_PATH)
    vbuf *v;
    viadev_packet_noop *p;

    if (fast_rdma_ok(c, sizeof(viadev_packet_noop), 0)) {
        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        v->shandle = NULL;
        p = (viadev_packet_noop *) v->buffer;
        PACKET_SET_HEADER(p, c, VIADEV_PACKET_NOOP);
        post_fast_rdma_with_completion(c, sizeof(viadev_packet_noop));
    } else {
        v = get_vbuf();
        p = (viadev_packet_noop *) VBUF_BUFFER_START(v);
        PACKET_SET_HEADER(p, c, VIADEV_PACKET_NOOP);
        vbuf_init_send(v, sizeof(viadev_packet_header));
        viadev_post_send(c, v);
    }

#else
    vbuf *v = get_vbuf();
    viadev_packet_noop *p = (viadev_packet_noop *) VBUF_BUFFER_START(v);
    D_PRINT("VI %3d SEND NOOP with %d credits", c->global_rank,
            c->local_credit);
    PACKET_SET_HEADER(p, c, VIADEV_PACKET_NOOP);
    vbuf_init_send(v, sizeof(viadev_packet_header));
    viadev_post_send(c, v);
#endif
}


#ifdef VIADEV_RPUT_SUPPORT
void viadev_rput_finish(MPIR_SHANDLE * s)
{
#if defined(RDMA_FAST_PATH)
    viadev_connection_t *c = (viadev_connection_t *) (s->connection);
    vbuf *v;
    void *buffer;
    viadev_packet_rput_finish *packet;
    int rdma_ok = fast_rdma_ok(c, sizeof(viadev_packet_rput_finish), 0);

    if (rdma_ok) {
        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        v->shandle = NULL;
        packet = (viadev_packet_rput_finish *) v->buffer;
    } else {
        v = get_vbuf();
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    }
#else
    vbuf *v = get_vbuf();
    void *buffer = VBUF_BUFFER_START(v);
    viadev_packet_rput_finish *packet = buffer;
    viadev_connection_t *c = (viadev_connection_t *) (s->connection);
#endif

    /* fill in the packet */
    PACKET_SET_HEADER(packet, c, VIADEV_PACKET_RPUT_FINISH);
    packet->rreq = REQ_TO_ID(s->receive_id);

    /* mark MPI send complete when VIA send completes */
    v->shandle = s;

#if defined(RDMA_FAST_PATH)
    if (rdma_ok) {
        post_fast_rdma_with_completion(c,
                                       sizeof(viadev_packet_rput_finish));
    } else {
        vbuf_init_send(v, sizeof(viadev_packet_rput_finish));
        viadev_post_send(c, v);
    }
#else
    /* prepare descriptor and post */
    vbuf_init_send(v, sizeof(viadev_packet_rput_finish));
    viadev_post_send(c, v);
#endif


}
#endif


#ifdef VIADEV_RGET_SUPPORT
void viadev_rget_finish(MPIR_RHANDLE * r)
{
#if defined(RDMA_FAST_PATH)
    viadev_connection_t *c = (viadev_connection_t *) (r->connection);
    vbuf *v;
    void *buffer;
    viadev_packet_rget_finish *packet;
    int rdma_ok = fast_rdma_ok(c, sizeof(viadev_packet_rget_finish), 0);

    if (rdma_ok) {
        v = &(c->RDMA_send_buf[c->phead_RDMA_send]);
        v->shandle = NULL;
        packet = (viadev_packet_rget_finish *) v->buffer;
    } else {
        v = get_vbuf();
        buffer = VBUF_BUFFER_START(v);
        packet = buffer;
    }
#else
    vbuf *v = get_vbuf();
    void *buffer = VBUF_BUFFER_START(v);
    viadev_packet_rget_finish *packet = buffer;
    viadev_connection_t *c = (viadev_connection_t *) (r->connection);
#endif

    /* fill in the packet */
    PACKET_SET_HEADER(packet, c, VIADEV_PACKET_RGET_FINISH);
    packet->sreq = REQ_TO_ID(r->send_id);

    v->shandle = r;

#if defined(RDMA_FAST_PATH)
    if (rdma_ok) {
        post_fast_rdma_with_completion(c,
                                       sizeof(viadev_packet_rget_finish));
    } else {
        vbuf_init_send(v, sizeof(viadev_packet_rget_finish));
        viadev_post_send(c, v);
    }
#else
    /* prepare descriptor and post */
    vbuf_init_send(v, sizeof(viadev_packet_rget_finish));
    viadev_post_send(c, v);
#endif
}
#endif /* VIADEV_RGET_SUPPORT */

/* add a function here directly instead of putting it in viutil.c */
static void viutil_spinandwaitcq(int spincount, VAPI_wc_desc_t * sc,
                                 VAPI_wc_desc_t * rc, int *inspin)
{

    VAPI_ret_t ret1;

#ifdef BLOCKING_SUPPORT
    unsigned long long spin_count = 0;
#endif

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    vbuf *v;
    int ofo;
#endif

    sc->id = rc->id = 0;
    *inspin = 1;

    while (1) {

#ifdef MCST_SUPPORT
        if ((viadev.bcast_info.bcast_called == 1) && (is_mcast_enabled)) {
            if (retransmission == 0) {
                process_acks();

                slide_window();

                check_time_out();

            }
        }
#endif

#ifdef _SMP_
        if (SMP_INIT) {
            if (MPID_SMP_Check_incoming() == MPI_SUCCESS) {
                sc->id = rc->id = 0;
                ret1 = VAPI_OK;
                break;
            }
        }
#endif

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        ret1 = poll_rdma_buffer((void *) &v, &ofo);
        if (ret1 == VAPI_OK) {
            sc->id = (aint_t) v;
            sc->opcode = VAPI_CQE_RQ_SEND_DATA;
            break;
        }
#endif

#ifdef BLOCKING_SUPPORT

        while(spin_count < viadev_max_spin_count) {
            ret1 = VAPI_poll_cq(viadev.nic, viadev.cq_hndl, sc);

            if(ret1 == VAPI_CQ_EMPTY) {
                spin_count++;
            } else {
                break;
            }
        }

        if(spin_count == viadev_max_spin_count) {
            /* Okay ... spinned long enough, now time to go to sleep! */
            ret1 = EVAPI_poll_cq_block(viadev.nic, viadev.cq_hndl, 0, sc);
        }
#else
        ret1 = VAPI_poll_cq(viadev.nic, viadev.cq_hndl, sc);

#endif
        if (ret1 == VAPI_OK) {
            if (sc->status != VAPI_SUCCESS) {
                vapidev_error_abort(VAPI_STATUS_ERR,
                                    "[%s:%d] Got completion with error, "
                                    "code=%s, vendor code=%x\n",
                                    viadev.my_name, viadev.me,
                                    VAPI_wc_status_sym(sc->status),
                                    sc->vendor_err_syndrome);
            }

            break;
        }

        if (ret1 == VAPI_CQ_EMPTY)
            continue;
        /* error when comes here */
        vapidev_error_abort(GEN_ASSERT_ERR,
                            "[%s:%d] Error in (%s:%d): ret1=%d\n",
                            viadev.my_name, viadev.me, __FILE__, __LINE__,
                            ret1);
        break;
    }
}

#if defined(VIADEV_RGET_SUPPORT)
void process_rdma_read_completion(vbuf *v)
{
    MPIR_RHANDLE *rhandle;
    viadev_connection_t *c = NULL;

    viadev_packet_rget *packet = 
        (viadev_packet_rget *) VBUF_BUFFER_START(v);

    c = &(viadev.connections[v->grank]);

    /* Since we have a RDMA read completion,
     * we have to make progress on the ones which
     * were queued due to lack of outstanding reads
     */
    c->rdma_reads_avail++;
    if (c->ext_rdma_read_head) {
        viadev_ext_rdma_read_start(c);
    }

    rhandle = (MPIR_RHANDLE *) packet->rreq;

    /* Process the RDMA read completion */
    viadev_rget_finish(rhandle);

    /* Reset some fields for
     * sanity check */
    v->shandle = NULL;
    release_vbuf(v);
}
#endif /* VIADEV_RGET_SUPPORT */
