/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "viaparam.h"
#include "vapi_header.h"
#include "vbuf.h"
#include "vapi_const.h"
#include "viapacket.h"

/*
 * ==============================================================
 * Initialize global parameter variables to default values
 * ==============================================================
 */


/*
 *  Which HCA device is used?
 *  (1) If a user specifies "VIADEV_DEVICE" environmental variable,
 *      the specified HCA device will be used.
 *  (2) Otherwise, MVAPICH first gets a list of the device IDs
 *      of the available HCA devices on the physical node.
 *  (3) Then, it chooses the first device by default.
 */

#define DEFAULT_HCA_ID      (0)

char *viadev_device[MAX_NUM_HCAS]; 


#ifdef VIADEV_PEER_TO_PEER
int viadev_use_peer_to_peer = 1;
#endif

/*
 *If we are limiting the number of outstanding RDMA operations,
 * Allow the limit to be set at run-time, but init variable
 * to default value.
 */
#ifdef VIADEV_HAVE_RDMA_LIMIT
int viadev_rdma_limit = VIADEV_RDMA_LIMIT;
#endif


/* Max number of entries on the Send Q of QPs per connection.
 * Should be about (prepost_depth + extra).
 * Must be within NIC MaxQpEntries limit.
 * Size will be adjusted below.
 */

unsigned long viadev_sq_size = 200;

/* Max number of entries on the RecvQ of QPs per connection.
 * computed to be:
 * prepost_depth + viadev_prepost_rendezvous_extra + viadev_prepost_noop_extra
 * Must be within NIC MaxQpEntries limit.
 */

unsigned long viadev_rq_size;
/* Max number of entries on the completion queue at a time.
 * Needed when constructing the completion queue, but can
 * be grown dynamically if needed.
 * Should be about (prepost_depth + extra)*num_vi*2.
 * Must be within NIC MaxCQEntries limit.
 * Size will be adjusted below.
 */
unsigned long viadev_cq_size = 40000;

/* number of vbufs used in RDMA_FAST_PATH on each VI */
int viadev_num_rdma_buffer = VIADEV_NUM_RDMA_BUFFER;

/* number of vbufs to pre-post as receives on each VI */
int viadev_prepost_depth = VIADEV_PREPOST_DEPTH;
int viadev_prepost_threshold = 10;
int viadev_initial_prepost_depth = 5;
int viadev_initial_credits;


/* number of additional initial vbufs to allocate per VI
 * after initial allocation is exhausted.
 */
static int viadev_vbuf_extra = 10;

/* In the case of small numbers of VIs, allocate at least
 * this number of VBUFs whenever a new region is allocated.
 */
#if 0
#define VIADEV_VBUF_MIN_POOL_SIZE 100
#endif
/* number of vbufs to allocate initially.
 * This will be re-defined after reading the parameters below
 * to scale to the number of VIs and other factors.
 */
int viadev_vbuf_pool_size = 5000;

/* number of vbufs to allocate in a secondary region if we should
 * run out of the initial allocation.  This is re-computed (below)
 * once other parameters are known.
 */
int viadev_vbuf_secondary_pool_size = 500;

/* max (total) number of vbufs to allocate, after which process 
 * terminates with a fatal error.
 * -1 means no limit.
 */
int viadev_vbuf_max = -1;

/* The number of "extra" vbufs that will be posted as receives
 * on a connection in anticipation of an R3 rendezvous message.
 * The TOTAL number of VBUFs posted on a receive queue at any
 * time is viadev_prepost_depth + viadev_prepost_rendezvous_extra
 * regardless of the number of outstanding R3 sends active on
 * a connection.
 */
#if 0
#define VIADEV_PREPOST_RENDEZVOUS_EXTRA VIADEV_PREPOST_DEPTH 
#endif
int viadev_prepost_rendezvous_extra =  VIADEV_PREPOST_RENDEZVOUS_EXTRA;

/* The message size (in bytes) at which we switch from an EAGER
 * protocol to the three-way RENDEZVOUS protocol.
 * In reality, smaller messages might be sent using R3
 * rather than EAGER if remote credits are low.
 */
int viadev_rendezvous_threshold = (VBUF_BUFFER_SIZE -
        sizeof(viadev_packet_eager_start));


/* the number of times the MPID_deviceCheck will poll VipCQDone
 * before it reverts to blocking on VipCQWait.
 * This value could influence latency and should probably be
 * adjusted dynamically.
 */
int viadev_spin_count = 100;

int viadev_prepost_noop_extra = 10;
/* dont allow EAGER sends or R3 data packets to consume all
 * remote credits on a connection.  This parameter records
 * the number of credits per connection that will be preserved
 * for non-data, control packets.
 */
int viadev_credit_preserve = 10;

/* As receive descriptors (VBUFs) are consumed and reaped on
 * a connection, the receive side posts more receive VBUFs.
 * Each time a message is sent to the remote end of a connection,
 * a credit update value, representing the number of new receive
 * VBUFs posted since the last message was sent, is piggybacked
 * on the message.  The remote end can then update its remote
 * credit value, possibly unblocking stalled R3 data sends.
 * If the receiving side has no reason to send a message, the
 * credit update accumulate and do not get communicated to the
 * remote end.  If the credit update value exceeds this
 * viadev_credit_notify_threshold value, a NoOp message is
 * sent to the remote end with the sole purpose of communicating
 * the credit update and thereby preventing deadlock.
 *

 * There are two places where NoOp operations may be sent.
 * The first is after all the entries on the CQ are processed,
 * indicating that all incoming and outgoing messages have
 * completed.  If there are more than viadev_credit_notify_threshold
 * local credits we send a NoOp.
 */
int viadev_credit_notify_threshold = 5;

/* The second place is while we are processing receive descriptors.
 * If the local credit value gets reasonably large, we are not
 * sending messages to the remote side.  The remote end of this
 * connection may be stalled waiting for us to process all the
 * recieves so lets send a NoOp immediately so that they can
 * continue work.
 * This viadev_dynamic_credit_threshold should be a fraction
 * of the maximum receive prepost depth, but substantially
 * greater than the viadev_credit_notify_threshold.
 * It is re-computed below.
 */
int viadev_dynamic_credit_threshold = 10;
/*
 * Development verbose flag
 */
int viadev_verbose = 0;

#ifdef _SMP_
#ifdef _AFFINITY_
/* Affinity is enabled by default */
unsigned int viadev_enable_affinity=1;
#endif
#endif

/*
 * Initialization parameters.
 */
unsigned int        viadev_ndreg_entries = 1000;

IB_mtu_t        viadev_default_mtu = VIADEV_DEFAULT_MTU;
unsigned long       viadev_max_rdma_size = VIADEV_MAX_RDMA_SIZE;
int             viadev_default_port = VIADEV_DEFAULT_PORT;
u_int8_t        viadev_default_qp_ous_rd_atom = VIADEV_DEFAULT_QP_OUS_RD_ATOM;
VAPI_psn_t      viadev_default_psn = VIADEV_DEFAULT_PSN;
VAPI_pkey_ix_t      viadev_default_pkey_ix = VIADEV_DEFAULT_PKEY_IX;
u_int32_t       viadev_default_p_key = VIADEV_DEFAULT_P_KEY;
IB_rnr_nak_timer_code_t viadev_default_min_rnr_timer = VIADEV_DEFAULT_MIN_RNR_TIMER;
IB_sl_t         viadev_default_service_level = VIADEV_DEFAULT_SERVICE_LEVEL;
VAPI_timeout_t      viadev_default_time_out = VIADEV_DEFAULT_TIME_OUT;
IB_static_rate_t    viadev_default_static_rate = VIADEV_DEFAULT_STATIC_RATE;
u_int8_t        viadev_default_src_path_bits = VIADEV_DEFAULT_SRC_PATH_BITS;
int         viadev_max_fast_eager_size = VIADEV_MAX_FAST_EAGER_SIZE;
VAPI_retry_count_t  viadev_default_retry_count = VIADEV_DEFAULT_RETRY_COUNT;
VAPI_retry_count_t  viadev_default_rnr_retry = VIADEV_DEFAULT_RNR_RETRY;
/*
   VAPI_rkey_t     viadev_default_r_key = VIADEV_DEFAULT_R_KEY;
   VAPI_lkey_t     viadev_default_l_key = VIADEV_DEFAULT_L_KEY;
   unsigned int        viadev_max_num_dhandle = VIADEV_MAX_NUM_DHANDLE;
   */
u_int32_t       viadev_default_max_sg_list = VIADEV_DEFAULT_MAX_SG_LIST;
u_int8_t        viadev_default_max_rdma_dst_ops = VIADEV_DEFAULT_MAX_RDMA_DST_OPS;

void viadev_init_parameters(int num_proc, int me)
{
    char *value;
    unsigned int number_hcas;
    int m, n;
    VAPI_hca_id_t* viadev_device_list;
    VAPI_ret_t ret;
    int j;
    /* ===============================================================
     * Examine the environment variables to see in any of these
     * were re-defined on the command line.
     * ===============================================================
     */
    if ( (value = getenv("VIADEV_VERBOSE")) != NULL) {
        viadev_verbose = atoi(value);
    }

    ret = EVAPI_list_hcas(0, &number_hcas, NULL);

    for(n = 0; n < number_hcas ; n++)
        viadev_device[n] = malloc(sizeof(VAPI_hca_id_t)); 
    /* Set the HCA device */
    if ( (value = getenv("VIADEV_DEVICE")) != NULL ) {
        strncpy(viadev_device[0], value, sizeof(VAPI_hca_id_t));
    }else{
        /* First, to get the number of hcas currently available */
        ret = EVAPI_list_hcas(0, &number_hcas, NULL);

        if ( VAPI_OK == ret ){
            vapidev_error_abort(GEN_EXIT_ERR,
                    "Error: No HCA Available and found by EVAPI_list_hcas.\n");
        }
        if ( VAPI_EAGAIN != ret ){
            vapidev_error_abort(VAPI_RETURN_ERR,
                    "Error: EVAPI_list_hcas: %s\n",
                    VAPI_strerror(ret));
        }

        /* Second,  to get the list of all devices currently available */

        viadev_device_list = malloc(number_hcas * sizeof(VAPI_hca_id_t));
        if ( NULL == viadev_device_list ){
            vapidev_error_abort(GEN_EXIT_ERR,
                    "Error:malloc for viadev_device_list.\n");
        }

        ret = EVAPI_list_hcas(number_hcas, &number_hcas, viadev_device_list);
        if ( VAPI_OK != ret ){
            vapidev_error_abort(VAPI_RETURN_ERR,
                    "Error: EVAPI_list_hcas : %s\n",
                    VAPI_strerror(ret));
        }
        /* Third,  choose the first one currently available.
         * This is decided by "DEFAULT_HCA_ID" by default.
         */
        for(m = 0; m < number_hcas; m++) {
            strncpy(viadev_device[m], (char*)&viadev_device_list[m],
                    sizeof(VAPI_hca_id_t));
        }
    }


    /* striping threshold */
    if ((value = getenv("STRIPING_THRESHOLD")) != NULL) {
        striping_threshold = atoi(value);
    } else {
        striping_threshold = STRIPING_THRESHOLD;
    }
#ifdef VIADEV_HAVE_RDMA_LIMIT
    if ((value = getenv("VIADEV_RDMA_LIMIT")) != NULL) {
        viadev_rdma_limit = atoi(value);
    }
#endif

    /* number of HCAs and ports supported */
    if ((value = getenv("NUM_HCAS")) != NULL) {
        num_hcas = atoi(value);
    }

    if ((value = getenv("NUM_PORTS")) != NULL) {
        num_ports = atoi(value);
    }

    num_subchannels = num_ports * num_hcas;

    if ( (value = getenv("SM_SCHEDULING")) != NULL) {
        if (!strcmp(value, "USE_FIRST")) {
            sm_scheduling = USE_FIRST;
        } else if (!strcmp(value, "ROUND_ROBIN")) {
            sm_scheduling = ROUND_ROBIN;
        } else if (!strcmp(value, "PROCESS_BINDING")) {
            sm_scheduling = PROCESS_BINDING;
        } else {
            fprintf(stderr, "invalid small message scheduling\n");
        }
    } else {
        sm_scheduling = ROUND_ROBIN;
    }

    if ( (value = getenv("LM_SCHEDULING")) != NULL) {
        if (!strcmp(value, "ROUND_ROBIN")) {
            lm_scheduling = ROUND_ROBIN;
        } else if (!strcmp(value, "WEIGHTED_STRIPING")) {
            lm_scheduling = WEIGHTED_STRIPING;
        } else if (!strcmp(value, "EVEN_STRIPING")) {
            lm_scheduling = EVEN_STRIPING;
        } else if (!strcmp(value, "STRIPE_BLOCKING")) {
            lm_scheduling = STRIPE_BLOCKING;
        } else if (!strcmp(value, "ADAPTIVE_STRIPING")) {
            lm_scheduling = ADAPTIVE_STRIPING;
        } else if (!strcmp(value, "PROCESS_BINDING")) {
            lm_scheduling = PROCESS_BINDING;
        } else if (!strcmp(value, "USE_FIRST")) {
            lm_scheduling = USE_FIRST;
        } else { 
            fprintf(stderr, "invalid large message scheduling\n");
        }
    } else {
        lm_scheduling = ADAPTIVE_STRIPING;
    }

    /* static link rates */

    for (j = 0; j < MAX_SUBCHANNELS; j++)
        link_speed[j] = DEFAULT_STATIC_RATE;  

    rdma_scheduling = RDMA_EVEN_STRIPING;
    sr_scheduling = SR_USE_FIRST;

    if ((value = getenv("VIADEV_MAX_RDMA_SIZE")) != NULL) {
        viadev_max_rdma_size = (unsigned long)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MTU")) != NULL) {
        if (strcmp(value,"MTU256")==0)
            viadev_default_mtu = MTU256;
        else if (strcmp(value,"MTU512")==0)
            viadev_default_mtu = MTU512;
        else if (strcmp(value,"MTU1024")==0)
            viadev_default_mtu = MTU1024;
        else if (strcmp(value,"MTU2048")==0)
            viadev_default_mtu = MTU2048;
        else if (strcmp(value,"MTU4096")==0)
            viadev_default_mtu = MTU4096;
        else
            viadev_default_mtu = MTU1024;
    }
    if ((value = getenv("VIADEV_DEFAULT_PORT")) != NULL) {
        viadev_default_port = (int)atoi(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_QP_OUS_RD_ATOM")) != NULL) {
        viadev_default_qp_ous_rd_atom = (u_int8_t)atoi(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_PSN")) != NULL) {
        viadev_default_psn = (VAPI_psn_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_PKEY_IX")) != NULL) {
        viadev_default_pkey_ix = (VAPI_pkey_ix_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_P_KEY")) != NULL) {
        viadev_default_p_key = (u_int32_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MIN_RNR_TIMER")) != NULL) {
        viadev_default_min_rnr_timer = (IB_rnr_nak_timer_code_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_SERVICE_LEVEL")) != NULL) {
        viadev_default_service_level = (IB_sl_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_TIME_OUT")) != NULL) {
        viadev_default_time_out = (VAPI_timeout_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_STATIC_RATE")) != NULL) {
        viadev_default_static_rate = (IB_static_rate_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_SRC_PATH_BITS")) != NULL) {
        viadev_default_src_path_bits = (u_int8_t)atoi(value);
    }
    if ((value = getenv("VIADEV_MAX_FAST_EAGER_SIZE")) != NULL) {
        viadev_max_fast_eager_size = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_RETRY_COUNT")) != NULL) {
        viadev_default_retry_count = (VAPI_retry_count_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_RNR_RETRY")) != NULL) {
        viadev_default_rnr_retry = (VAPI_retry_count_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_SG_LIST")) != NULL) {
        viadev_default_max_sg_list = (u_int32_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_RDMA_DST_OPS")) != NULL) {
        viadev_default_max_rdma_dst_ops = (u_int8_t)atol(value);
    }
    if ((value = getenv("VIADEV_NDREG_ENTRIES")) != NULL) {
        viadev_ndreg_entries = (unsigned int)atoi(value);
    }
    if ((value = getenv("VIADEV_SQ_SIZE")) != NULL) {
        viadev_sq_size = atol(value);
    }

    if ( (value = getenv("VIADEV_CQ_SIZE")) != NULL) {
        viadev_cq_size = atol(value);
    }

    if ((value = getenv("VIADEV_NUM_RDMA_BUFFER")) != NULL) {
        viadev_num_rdma_buffer = atoi(value);
        if (viadev_num_rdma_buffer < 0)
            viadev_num_rdma_buffer = 0;
        if (viadev_num_rdma_buffer == 1)
            viadev_num_rdma_buffer = 2; /* need 1 extra */
    }

    if ( (value = getenv("VIADEV_PREPOST_DEPTH")) != NULL) {
        viadev_prepost_depth = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_THRESHOLD")) != NULL) {
        viadev_prepost_threshold = atoi(value);
    }

    if ( (value = getenv("VIADEV_VBUF_EXTRA")) != NULL) {
        viadev_vbuf_extra = atoi(value);
    }
    if ( (value = getenv("VIADEV_VBUF_MAX")) != NULL) {
        viadev_vbuf_max = atoi(value);
    }
    if ( (value = getenv("VIADEV_PREPOST_RENDEZVOUS_EXTRA")) != NULL) {
        viadev_prepost_rendezvous_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_RENDEZVOUS_THRESHOLD")) != NULL) {
        if(atoi(value) > viadev_rendezvous_threshold) {

            if(viadev.me == 0) {
                vapidev_error_abort(GEN_EXIT_ERR, "You have tried to"
                        " set the rendezvous threshold to greater than"
                        " vbuf size.\n Currently, it is not possible to do so at"
                        " runtime.\n Please change the VBUF_TOTAL_SIZE in file"
                        " $PREFIX/mpid/vapi/vbuf.h.\n\n");
            }
            exit(-1);
        } else {
            viadev_rendezvous_threshold = atoi(value);
        }
    }
    if ( (value = getenv("STRIPING_THRESHOLD")) == NULL){
        striping_threshold = viadev_rendezvous_threshold;
    }

    if ( (value = getenv("VIADEV_SPIN_COUNT")) != NULL) {
        viadev_spin_count = atoi(value);
    }
    if ( (value = getenv("VIADEV_CREDIT_PRESERVE")) != NULL) {
        viadev_credit_preserve = atoi(value);
    }
    if ( (value = getenv("VIADEV_CREDIT_NOTIFY_THRESHOLD")) != NULL) {
        viadev_credit_notify_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_NOOP_EXTRA")) != NULL) {
        viadev_prepost_noop_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_INITIAL_PREPOST_DEPTH")) != NULL) {
        viadev_initial_prepost_depth = atoi(value);
    }
    /* TBD - temporary workaround, RDMA initialization mistakenly triggers
     * credit exchange and tries to use uninitialized RDMA remote addresses
     * to send Noops.
     */

#ifdef _SMP_
    /* Should set in KBytes */
    if ( (value = getenv("SMP_EAGERSIZE")) != NULL ) {
        smp_eagersize = atoi(value);
    }
    /* Should set in MBytes */
    if ( (value = getenv("SMPI_LENGTH_QUEUE")) != NULL ) {
        smpi_length_queue = atoi(value);
    }
#ifdef _AFFINITY_
    if ((value = getenv("VIADEV_ENABLE_AFFINITY")) != NULL) {
        viadev_enable_affinity = atoi(value);
    }
#endif
#endif

    if (viadev_num_rdma_buffer)
        viadev_initial_prepost_depth =
            viadev_prepost_depth + viadev_prepost_noop_extra;
    if (viadev_initial_prepost_depth >
            viadev_prepost_depth + viadev_prepost_noop_extra)
        viadev_initial_prepost_depth =
            viadev_prepost_depth + viadev_prepost_noop_extra;
    if (viadev_initial_prepost_depth <= 0)
        viadev_initial_prepost_depth = 1;
    if (viadev_initial_prepost_depth <= viadev_prepost_noop_extra) {
        viadev_initial_credits = viadev_initial_prepost_depth;
    } else {
        viadev_initial_credits =
            viadev_initial_prepost_depth - viadev_prepost_noop_extra;
    }
    if ((value = getenv("VIADEV_CREDIT_PRESERVE")) != NULL) {
        viadev_credit_preserve = atoi(value);
    }
    if ((value = getenv("VIADEV_CREDIT_NOTIFY_THRESHOLD")) != NULL) {
        viadev_credit_notify_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_DYNAMIC_CREDIT_THRESHOLD")) != NULL) {
        viadev_dynamic_credit_threshold = atoi(value);
    }
    if ((viadev_prepost_depth - viadev_prepost_threshold) <=
            MAX(viadev_credit_notify_threshold,
                viadev_dynamic_credit_threshold)) {
        viadev_prepost_threshold =
            viadev_prepost_depth - 1 - MAX(viadev_credit_notify_threshold,
                    viadev_dynamic_credit_threshold);
    }
    if (viadev_prepost_threshold < 0)
        viadev_prepost_threshold = 0;

#ifdef VIADEV_PEER_TO_PEER
    if ( (value = getenv("VIADEV_USE_PEER_TO_PEER")) != NULL) {
        viadev_use_peer_to_peer = atoi(value);
    }
#endif    
    viadev_rq_size =
        viadev_prepost_depth + viadev_prepost_rendezvous_extra +
        viadev_prepost_noop_extra;

}

void dump_param_values(FILE * fd, int me)
{
    char *int_fmt = "%3d: %32s = %d\n";
    char *str_fmt = "%3d: %32s = %s\n";

    /*
       char *hex_fmt = "%3d: %32s = %#x\n";
       */

    fprintf(fd, int_fmt, me, "VIADEV_VERBOSE", viadev_verbose);
    fprintf(fd, str_fmt, me, "VIADEV_DEVICE", viadev_device);

    /* Print reliability Level for VIA, otherwise not implemented */
#ifdef VIADEV_PEER_TO_PEER
    fprintf(fd, int_fmt, me,
            "VIADEV_USE_PEER_TO_PEER", viadev_use_peer_to_peer);
#endif
#ifdef VIADEV_HAVE_RDMA_LIMIT
    fprintf(fd, int_fmt, me, "VIADEV_RDMA_LIMIT", viadev_rdma_limit);
#endif
    fprintf(fd, int_fmt, me, "VIADEV_SQ_SIZE", (int) viadev_sq_size);
    fprintf(fd, int_fmt, me, "VIADEV_CQ_SIZE", (int) viadev_cq_size);
    fprintf(fd, int_fmt, me, "VIADEV_NUM_RDMA_BUFFER",
            viadev_num_rdma_buffer);
    fprintf(fd, int_fmt, me, "VIADEV_PREPOST_DEPTH", viadev_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_THRESHOLD", viadev_prepost_threshold);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_POOL_SIZE", viadev_vbuf_pool_size);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_SECONDARY_POOL_SIZE",
            viadev_vbuf_secondary_pool_size);
    fprintf(fd, int_fmt, me, "VIADEV_VBUF_MAX", viadev_vbuf_max);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_RENDEZVOUS_EXTRA",
            viadev_prepost_rendezvous_extra);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_NOOP_EXTRA", viadev_prepost_noop_extra);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_PRESERVE",
            viadev_credit_preserve);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_NOTIFY_THRESHOLD",
            viadev_credit_notify_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_DYNAMIC_CREDIT_THRESHOLD",
            viadev_dynamic_credit_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_RENDEZVOUS_THRESHOLD",
            viadev_rendezvous_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_SPIN_COUNT", viadev_spin_count);
}




/* ================================================================
 * Adjust global parameters to conform with NIC attributes
 * ================================================================ */

/* This section only applicatble to VIA */
